
{-----------------------------------------------------------------------------
 Unit Name: FDrawParamSurf
 Project Name : ImmersionR3R6
 Author:    Angel Montesinos-Amilibia
 Date: 2009
 Purpose: Drawing of a parameterized surface; the surface is illuminated by a
          pointlike focus that casts shadows
-----------------------------------------------------------------------------}
{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

{Some code is desactivated. It is to be used when the user clips the
surface. Currently this option is not supported.}

unit FDrawParamSurf;
interface
uses
  Windows, Graphics, Types, Classes,
  VectorsAndUtil;

type
  threeBytes = array[1..3] of Byte;
  
  pointData = record
    pos, nor, proj, projF: Vector3;
    c3: Double;
    key: array[0..1] of Word;
    inside: Boolean;
  end;

  triang = array[1..3] of pointData;
  pointDataGrid = array of array of PointData;

  trFill = procedure(bm: TBitmap; vertices: triang) of object;

  mapR2R3= function (const p: Vector2): Vector3 of object;

  TDrawParamSurf = class
    Hdifs, DD: array of Vector2;
    val, initialVal, rowInitialVal: array of Double;
    wdth, wdthMinus1, gr: Integer;
    rct, rRed: TRect;
    theF: mapR2R3;
    numberOfColors: Byte;
    c0, c1, c2, c3, cx, cy: Double;
    vert: pointDataGrid;
    min3DRed, max3DRed, ctr3DRed: Vector3; {the cube containing the Veronese, in
      the affine frame centered at the tip of the mean curvature vector, and
      basis 'MapR3R6.nn'}
    cf0, cf1, cf2, cf3, cfx, cfy, span3D, dist: Double;
    fN, fH, fV, vH, vV, vN, viewPoint, focus: Vector3;
    procedure AssignCol(var pd: PointData);
    function ClipSegmentToRect3D(r: TRect; var x, y: Vector3; yDirector: Boolean): Boolean;
    procedure ComputeFBuffer;
    procedure ComputeGridData;
    function ComputeSurface: Boolean;
    constructor Create;
    function DefinedUnitVector(place: TPoint; var theVector: Vector3): Boolean;
    destructor Destroy;override;
    procedure DrawSegment(can: TCanvas; vec1, vec2: Vector3;
             maxPart: Integer;  baseColor: TColor);
    procedure DrawSurface;
    function DrawSurfaceWithoutClips: Boolean;
    function Face(a,b,c: Vector3): Byte;
    procedure FillFBufferTriangle(triangleF: triang);
    procedure FillPointData(var pd: PointData);
    procedure FillTriangle(bm: TBitmap; vertices: triang);
    procedure FillTriangleAsymptness(bm: TBitmap; vertices: triang);
    function FindRay(pix: TPoint): Vector3;
    procedure FrameFromFocus(w: Integer);
    procedure FrameFromViewPoint(w: Integer);
    procedure GetCorrectViewPoint(w: Integer);
    procedure InitializeBuffers;
    procedure InitializeCameras(w: Integer);
    procedure MoveSurface(shift: TShiftState; prevPt, curPt: TPoint);
    procedure MoveViewPoint(prevPt,curPt: TPoint; Shift: TShiftState);
    function ProjectToFocusPlaneRealAndDist(x: Vector3): Vector3;
    function ProjectToPaneRealAndInv(x: Vector3): Vector3;
    procedure ReceiveBasicData(min3D, max3D: Vector3; grid, w: Integer);
    procedure ScaleCoefficients(sc: Double);
  end;

  var
    dps: TDrawParamSurf;

implementation
uses
 SysUtils, Math, Forms,
 ResourceStrings, {ManageClips,} FGammaPicker, FSupplWindow, FMapR3R6;

const

  kBigFiller: Double = MaxInt;
  cFarthness = 3.5;
  gId: threeBytes =(1,2,3);
  
type

  FillTri = procedure(bm: TBitmap; vertices: triang);



procedure TDrawParamSurf.AssignCol(var pd: PointData);
{compute illumination index of a point in the surface, in its ambient and
 focus components}
const
  csm = 1.0e-6;  {A conventional little number}
  dullness: double = 0.3; {A conventional measure of the surface dullness}
  OneMinusDullness: double = 1 - 0.3 {1 - dullness};
  intFocus: double = 0.6; {Relative importance of the focus illumination}
  intAmb: double = 0.4; {Relative importance of the ambient illumination.
                         Must be 1 - intFocus}
var
  k: Integer;
  ray, fRay: Vector3;
  rayNorm, fRayNorm, sigma, lambda, omega, fComp: Double;
begin
  if (pd.nor[1]< 2) then begin
    for k:= 1 to 3 do begin
      ray[k] := viewPoint[k] - pd.pos[k];
      fRay[k] := focus[k] - pd.Pos[k];
    end;
    rayNorm := EuclNorm(ray);
    fRayNorm := EuclNorm(fRay);
    for k := 1 to 3 do begin
      ray[k] := ray[k] / rayNorm; {'ray' becomes unit vector from 'x' to eye}
      fRay[k] := fRay[k] / fRayNorm; {'fRay' becomes unit vector from 'x' to focus}
    end;
    sigma := EuclInnerProd(ray,pd.nor);
    lambda := EuclInnerProd(fRay,pd.nor);
    omega := 2 * lambda * sigma; {the sign of 'omega' says that we are looking
    to the illuminated (by the focus) side of the surface}

    {Compute and store ambient component to 'pd.clKeys[0]}
    pd.key[0] := round(cGammaN*intAmb * (1 + abs(sigma)) / 2);

    if omega > 0 then begin
      {We are looking to the illuminated side}
      omega := (1 + omega - EuclInnerProd(fRay,ray)) / 2;
      fComp := abs(lambda) * (OneMinusDullness * sqr(omega) + dullness * (1 +
        abs(sigma)) / 2);

      {Compute and store focus component to 'pd.clKeys[1]}
      pd.key[1] := Min(cGammaN - pd.key[0], round(intFocus *cGammaN*fComp));
    end
    else
      {don't leave this field without initializing. Put it the background color}
      pd.key[1]:= 0;
  end
  else begin
    pd.key[0]:= 0;
    pd.key[1]:= 0;
  end;
end;

function TDrawParamSurf.ClipSegmentToRect3D(r: TRect; var x, y: Vector3; yDirector: Boolean): Boolean;
{If yDirector = True, clips the line  t -> x + t*y inside 'r' by giving on
exit on 'x' and 'y' the points where that line cuts the boundary of 'r', on the
understanding that one of them may be at the infinity, what we will substitute
by a very far point in the line.
If director = False, clips inside 'r' the segment from 'x' to 'y', giving on
exit, on 'x' and 'y', the first and last points of that segment that lie in 'r'.
Result is true if the clip is nonempty.
See the paper entitled "ClippingASegmentOrLine.pdf" that accompanies the source
of this program.}
const
  cPInf= 1.0e6;   {This is plus infinity for our practical purposes}
  cMInf= -1.0e6;   {And this is minus infinity}
var
  a,b, pa, pb: Vector3;
  minC, maxC: Vector2;
  minT, maxT, t, t0, t1, k0, k1, k0n, k1n, aux: Double;
  i, sgn: Integer;
  prox: Boolean;

  procedure ReduceInterval; {A frequently used task}
  begin
    if t0 > t1 then begin
      t:= t0;
      t0:= t1;
      t1:= t;
    end;
    minT:= max(minT, t0);
    maxT:= min(maxT, t1);
    result:= ( maxT >= minT );
  end;

begin
  minC[1]:= r.Left;
  minC[2]:= r.Top;
  maxC[1]:= r.Right - 1;
  maxC[2]:= r.Bottom - 1;

  a:= x;
  if yDirector then
    for i:= 1 to 3 do
      b[i]:= x[i] + y[i]
  else
    b:= y;
  pa:= ProjectToPaneRealAndInv(a);
  pb:= ProjectToPaneRealAndInv(b);
  if not yDirector then begin
    result:= PtInRect(r,Point(round(pa[1]),round(pa[2]))) and
           PtInRect(r,Point(round(pb[1]),round(pb[2])));
    if result then
      Exit;
  end
  else
    result:= False;
  {These are the false infinite ends that we use as initializations when
   'director' is true}
  if yDirector then begin
    minT:= cMInf;
    maxT:= cPInf;
  end
  else begin  {the initial interval is the initial limit}
    minT:= 0;
    maxT:= 1;
  end;

  prox:= False;

  {We first study the special cases}
  if abs(pa[3]-pb[3]) < 1.0e-12 then begin
    prox:= True;
    if abs(pa[1]-pb[1]) > 1.0e-12 then begin
      t0:= (pa[1]-minC[1])/(pa[1]-pb[1]);
      t1:= (pa[1]-maxC[1])/(pa[1]-pb[1]);
      ReduceInterval;
      if  result and (abs(pa[2]-pb[2]) > 1.0e-12) then begin
        t0:= (pa[2]-minC[2])/(pa[2]-pb[2]);
        t1:= (pa[2]-maxC[2])/(pa[2]-pb[2]);
        ReduceInterval;
      end;
    end;
  end
  {Now the usual cases. First for the horizontal coordinate}
  else begin
    sgn:= 1- 2* ord( pa[3]*pb[3]/(pa[3]-pb[3]) < 0);
    aux:= pa[1]-pb[1];
    if (abs(aux) < 1.0e-10) then begin
      {the line has constant horizontal projection. Hence}
      if (pa[1] < minC[1]) or (pa[1] > maxC[1]) then
        Exit;
    end
    else begin
      k0n:= minC[1]*(pa[3]-pb[3])+ pa[1]*pb[3]-pb[1]*pa[3];
      k0:= k0n/aux;
      k1n:=maxC[1]*(pa[3]-pb[3])+ pa[1]*pb[3]-pb[1]*pa[3];
      k1:= k1n/aux;
      if (k0 <= 0) and (k1 <= 0) then begin
        result:= False;
        Exit;
      end;
      if (k0 > 0) and (k1 > 0) then begin
        t0:= (pa[1]-minC[1])*pb[3]/k0n;
        t1:= (pa[1]-maxC[1])*pb[3]/k1n;
        ReduceInterval;
      end
      else if (k0 > 0) and (k1 < 0) then begin
        aux:= (pa[1]-minC[1])*pb[3]/k0n;
        if sgn = 1 then begin
          t0:= aux;
          t1:= cPInf;
        end
        else begin
          t0:= cMInf;
          t1:= aux;
        end;
        ReduceInterval;
      end
      else if (k0 < 0) and (k1 > 0) then begin
        aux:= (pa[1]-maxC[1])*pb[3]/k1n;
        if sgn = 1 then begin
          t0:= aux;
          t1:= cPInf;
        end
        else begin
          t0:= cMInf;
          t1:= aux;
        end;
        ReduceInterval;
      end;

      {And now for the vertical coordinate}

      if result then begin
        aux:= pa[2]-pb[2];
        if (abs(aux) < 1.0e-10) then begin
          {the line has constant vertical projection. Hence}
          if (pa[2] < minC[2]) or (pa[2] > maxC[2]) then
            Exit;
        end
        else if result then begin
          k0n:= minC[2]*(pa[3]-pb[3])+ pa[2]*pb[3]-pb[2]*pa[3];
          k0:= k0n/aux;
          k1n:=maxC[2]*(pa[3]-pb[3])+ pa[2]*pb[3]-pb[2]*pa[3];
          k1:= k1n/aux;
          if (k0 <= 0) and (k1 <= 0) then begin
            result:= False;
            Exit;
          end;
          if (k0 > 0) and (k1 > 0) then begin
            t0:= (pa[2]-minC[2])*pb[3]/k0n;
            t1:= (pa[2]-maxC[2])*pb[3]/k1n;
            ReduceInterval;
          end
          else if (k0 > 0) and (k1 < 0) then begin
            aux:= (pa[2]-minC[2])*pb[3]/k0n;
            if sgn = 1 then begin
              t0:= aux;
              t1:= cPInf;
            end
            else begin
              t0:= cMInf;
              t1:= aux;
            end;
            ReduceInterval;
          end
          else if (k0 < 0) and (k1 > 0) then begin
            aux:= (pa[2]-maxC[2])*pb[3]/k1n;
            if sgn = 1 then begin
              t0:= aux;
              t1:= cPInf;
            end
            else begin
              t0:= cMInf;
              t1:= aux;
            end;
            ReduceInterval;
          end;
        end;
      end;
    end;
  end;
  if result then begin
    for i:= 1 to 3 do begin
      x[i]:= (1-minT)*a[i]+ minT*b[i];
      y[i]:= (1-maxT)*a[i]+ maxT*b[i];
    end;
    if prox then begin
      pa:= ProjectToPaneRealAndInv(x);
      pb:= ProjectToPaneRealAndInv(y);
      result:= PtInRect(r,Point(round(pa[1]),round(pa[2]))) and
           PtInRect(r,Point(round(pb[1]),round(pb[2])));
    end
  end
end;

procedure TDrawParamSurf.ComputeFBuffer;
{Compute the z-buffer of the surface as seen from the illuminating focus.
 No clips are active}
var
  i, j: integer;
  triangle: triang;
begin
  for i := 1 to gr do
    for j := 1 to gr do begin
      triangle[1] := vert[i, j - 1];
      triangle[2] := vert[i, j];
      triangle[3] := vert[i - 1, j - 1];
      FillFBufferTriangle(triangle);
      triangle[1] := vert[i - 1, j];
      triangle[2] := vert[i - 1, j - 1];
      triangle[3] := vert[i, j];
      FillFBufferTriangle(triangle);
    end;
end;

procedure TDrawParamSurf.ComputeGridData;
{Store data for the points of a grid gr x gr of the parametr domain}
var
  i, j: Integer;
begin
  for i := 0 to gr do
    for j := 0 to gr do
      FillPointData(vert[i,j]);
end;

function TDrawParamSurf.ComputeSurface: Boolean;
begin
  GetCorrectViewPoint(rct.right);
  FrameFromFocus(rct.right);
  MapR3R6.ComputeVeroneseVertexes(vert);
  ComputeGridData;
  DrawSurface;
  result := True;
end;

constructor TDrawParamSurf.Create;
var
  i: Integer;
begin
  inherited;
  SetLength(Hdifs,10);
  SetLength(DD,10);
  SetLength(val,10);
  SetLength(initialVal,10);
  SetLength(rowInitialVal,10);
  numberOfColors:= 0;
  for i:= 1 to 3 do begin
    min3DRed[i]:= -1;
    max3DRed[i]:= 1;
    ctr3DRed[i]:= 0;
  end;
  InitializeCameras(kWidthMainBox);
end;

function TDrawParamSurf.DefinedUnitVector(place: TPoint; var theVector: Vector3): Boolean;
{We use an inverse stereographic projection, with respect to the point
 'ctr3DRed + r*vN' to the sphere of center 'ctr3DRed' and radius 'r'.
 All calculations are made as if the origin of R^3 was at 'ViewPoint'}
const
  cReductor = 0.3; {The lesser, the more sensitive will be the movement of
                    the surface to the movement of the mouse}
var
  b, a, k, bMa: Vector3;
  t,r: double;
  i: Integer;
begin
  r:= cReductor*span3D/2;
  k:= FindRay(place); {vector whose direction is determined by the mouse
                       position 'place'. }
  t:= (dist - r)/EuclInnerProd(k, vN); {So, the line 'viewpoint + t*k'
               cuts the plane orthogonal to 'vN' passing by
               'ctr3DRed + r*vN' when 't' has this value}
  {Now, we calculate the inverse stereographic projection of 'viewpoint + t*k'
   to the sphere}
  for i:= 1 to 3 do begin
    b[i]:= t*k[i];
    a[i]:= (dist + r)*vN[i];
    bMa[i]:= b[i]-a[i];
  end;
  t:= -2*r*EuclInnerProd(vN,bMa)/EuclInnerProd(bMa, bMa);
  for i:= 1 to 3 do
    theVector[i]:= -dist*vN[i]+(1-t)*a[i] + t*b[i]; {This is the inverse
    stereographic projection relative to the sphere center. We now normalize
    'theVector'}
  r:= EuclNorm(theVector);
  result:= r > 1.0e-12;
  if result then for i:= 1 to 3 do
    theVector[i]:= theVector[i]/r;
end;

destructor TDrawParamSurf.Destroy;
var
  i: Integer;
begin
  Hdifs:= nil;
  DD:= nil;
  val:= nil;
  initialVal:= nil;
  rowinitialVal:= nil;
  if high(vert) >= 0 then begin
    for i:= 0 to high(vert) do
      vert[i]:= nil;
    vert:= nil;
  end;
  inherited;
end;

procedure TDrawParamSurf.DrawSegment(can: TCanvas; vec1, vec2: Vector3;
     maxPart: Integer;  baseColor: TColor);
{draw a segment using the zbuffer to check its visibility}
const
  cTolZBuffer = 0.999;
var
  i, num: Integer;
  d, b,span, lambda: Double;
  pixR1, pixR2, pixDif, pix: Vector3;
  pt1, pt2: Tpoint;
  vis1, vis2: boolean;
  x1, x2: Vector3;
begin
  x1:= vec1;
  x2:= vec2;
  can.Pen.Color:= baseColor;
  if not ClipSegmentToRect3D(rRed, x1, x2,False) then
    Exit;
  pixR1 := ProjectToPaneRealAndInv(x1);
  pixR2 := ProjectToPaneRealAndInv(x2);
  span:= hypot(pixR1[1]-pixR2[1], pixR1[2]-pixR2[2]);
  num:= 1 + trunc(span/max(2,maxPart));
  for i:= 1 to 3 do
    pixDif[i]:= pixR2[i]- pixR1[i];

  pt2.x:= round(pixR1[1]);
  pt2.y:= round(pixR1[2]);
  b:= zBuffer[pt2.x,pt2.y]+(pixR1[1]-pt2.x)*(zBuffer[pt2.x+1,pt2.y]-
          zBuffer[pt2.x,pt2.y])+ (pixR1[2]-pt2.y)*(zBuffer[pt2.x,pt2.y+1]-
          zBuffer[pt2.x,pt2.y]);
  vis2 := (pixR1[3] > cTolZBuffer*b);
  for i:= 1 to num do begin
    pt1:= pt2;
    vis1:= vis2;
    lambda:= i/num;
    d:= pixR1[3] + lambda*pixDif[3];
    pix[1]:= pixR1[1]+ lambda*pixDif[1];
    pix[2]:= pixR1[2]+ lambda*pixDif[2];
    pt2.x:= round(pix[1]);
    pt2.y:= round(pix[2]);
    b:= zBuffer[pt2.x,pt2.y]+(pix[1]-pt2.x)*(zBuffer[pt2.x+1,pt2.y]-
          zBuffer[pt2.x,pt2.y])+ (pix[2]-pt2.y)*(zBuffer[pt2.x,pt2.y+1]-
          zBuffer[pt2.x,pt2.y]);
    vis2 := (d > cTolZBuffer*b);
    if vis1 and vis2 then begin
      can.moveto(pt1.x,pt1.y);
      can.lineto(pt2.x,pt2.y);
    end;
  end;
end;

procedure TDrawParamSurf.DrawSurface;
{currently it is supposed that there are no clips}
  var
    c: TColor;
begin
  {Store the background color}
  c:= gammas[0,0];
  bmp.Canvas.Brush.Color:= (c and $00FF0000) shr 16+
   (c and $0000FF00)+ (c and $000000FF) shl 16;

  {Draw the background}
  ClearBitmap(bmp, bmp.Width, c);

{Draw the surface}
 {if gClips <> nil then
    DrawSurfaceWithClips(bm)
  else}
  DrawSurfaceWithoutClips;
end;


(*function DrawSurfaceWithClips(var bm: TBitmap): Boolean;
var
  i, j: Integer;
  vertices: triang;
  FillT, FillTC: FillTri;
begin
  Result := False;

  {Reserve romm for 'vert'}
  if (vert = nil) or (high(vert[0]) <> gGridV) or
    (high(vert) <> gGridU) then begin
    for i:= 0 to high(vert) do
      vert[i]:=nil;
    vert := nil;
    SetLength(vert, gGridU + 1);
    for i := 0 to gGridU do
      SetLength(vert[i], gGridV + 1);
  end;
  gClipTop := Length(gClips) - 1;

  InitializeZBuffer;
  InitializeParamBuffer;
  InitializeFBuffer;
  ComputeGridData;
  if ord(gCm)< 1 then begin
    {Draw surface with color per illumination}
    FillT:= FillTriangle;
    FillTC:= FillTriangleWithClips;

    {Reserve room for some variables according to the length of the clip list}
    if High(Hdifs) < 8+gClipTop then begin
      SetLength(Hdifs, 9+gClipTop);
      SetLength(DD, 9+gClipTop);
      SetLength(val,  9+gClipTop);
      SetLength(initialVal, 9+gClipTop);
      SetLength(rowInitialVal, 9+gClipTop);
    end;
  end
  else begin
    {Draw surface with color per the curvature chosen by user}
    FillT:= FillTriangleCurv;
    FillTC:= FillTriangleCurvWithClips;

    {Reserve room for some variables according to the length of the clip list}
    if High(Hdifs) < 9+gClipTop then begin
      SetLength(Hdifs, 10+gClipTop);
      SetLength(DD, 10+gClipTop);
      SetLength(val,  10+gClipTop);
      SetLength(initialVal, 10+gClipTop);
      SetLength(rowInitialVal, 10+gClipTop);
    end;
  end;

  {A previous pass to gather information about visiblity from focus}
  ComputeFBufferWithClips;

  for i := 1 to gGridU do begin
    for j := 1 to gGridV do begin
      {Each rectangle in the grid of the parameter domain is divided in
      two triangles. The first is passed now to 'FillT'}
      vertices[1] := vert[i, j - 1];
      vertices[2] := vert[i, j];
      vertices[3] := vert[i - 1, j - 1];
      if vertices[1].inside and vertices[2].inside and vertices[3].inside then
        {The triangle is inside clips: clips do not affecf it}
        FillT(bm, vertices)
      else
        {The triangle is cut by a clip plane}
        FillTC(bm, vertices);

      {Now the second triangle is passed to FillT or to FillTC. Note that
       the vertexes are ordered so as to maintain the orientation. That is,
       the orientation of this second triangle is the same as that of the
       former}
      vertices[1] := vertices[2];
      vertices[2] := vertices[3];
      vertices[3] := vertices[1];
      vertices[1] := vert[i - 1, j];
      if vertices[1].inside and vertices[2].inside and vertices[3].inside then
        FillT(bm, vertices)
      else
        FillTC(bm, vertices);
    end;
  end;
  Result := True;
end; *)



function TDrawParamSurf.DrawSurfaceWithoutClips: Boolean;
{See documentation on function DrawSurfaceWithClips}
var
  i, j: Integer;
  x, y: Vector3;
  ft: trFill;
  vertices: triang;
begin
  InitializeBuffers;
  ComputeGridData;
  ComputeFBuffer;
  if numberOfColors < 4 then
    ft:= FillTriangle
  else
    ft:= FillTriangleAsymptness;
  for i := 1 to gr do begin
    for j := 1 to gr do begin
      vertices[1] := vert[i, j - 1];
      vertices[2] := vert[i, j];
      vertices[3] := vert[i - 1, j - 1];
      ft(bmp, vertices);
      vertices[1] := vertices[2];
      vertices[2] := vertices[3];
      vertices[3] := vertices[1];
      vertices[1] := vert[i - 1, j];
      ft(bmp, vertices);
    end;
  end;
  x[1]:= 0;
  x[2]:= 0;
  x[3]:= 0;
  for i:= 1 to 3 do
    y[i]:= - MapR3R6.meanCurv[i];
  DrawSegment(bmp.canvas, x, y, 5, clLime);
  Result := True;
end;

function TDrawParamSurf.Face(a,b,c: Vector3): Byte;
{'a,b,c' are the vertexes of an oriented triangle in space. We compute the
 inner product of the vector from 'a' to 'viewPoint' with the oriented
 normal to the triangle 'a,b,c'. The result is 1 if the inner product is
 negative, 0 otherwise}
  var
    h: double;
    va,ba,ca: Vector3;
    i: Integer;
begin
  for i:= 1 to 3 do begin
    va[i]:= a[i] - viewpoint[i];
    ba[i] := b[i] - a[i];
    ca[i] := c[i] - a[i];
  end;
  h:= va[1]*(ba[2]*ca[3]-ba[3]*ca[2])+va[2]*(ba[3]*ca[1]-ba[1]*ca[3])+va[3]*(ba[1]*ca[2]-ba[2]*ca[1]);
  result:= ord(h < 0);
end;

procedure TDrawParamSurf.FillFBufferTriangle(triangleF: triang);
{Fills the values of the z-buffer of the illuminating focus on the points
 of a triangle of 'pointData'. The surface is not clipped, that is there aren't
 active clips.}
{See documentation of FillTriangleCurvWithClips  in the source of the program
 Superficies, unit DrawParamSurf.pas}
var
  permu: threeBytes;
  A, B, C: PointData;
  BA, CA: array[1..2] of single;
  uu: array[1..2] of Vector2Single;
  Hdifs, DD: array[1..2] of single;
  val, initialVal, rowInitialVal: single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, fromHere, toThere: Integer;
  auxab, auxac, auxbc: Single;
  CAtLeftOfB: Boolean;
begin
  permu := gId;
  if triangleF[2].projF[2] < triangleF[1].projF[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if triangleF[3].projF[2] < triangleF[permu[1]].projF[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if triangleF[3].projF[2] < triangleF[permu[2]].projF[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := triangleF[permu[1]];
  B := triangleF[permu[2]];
  C := triangleF[permu[3]];

  for i:= 1 to 2 do begin
    BA[i] := B.projF[i] - A.projF[i];
    CA[i] := C.projF[i] - A.projF[i];
  end;
  det := BA[1] * CA[2] - BA[2] * CA[1];
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  uu[1,1]:= det*CA[2];
  uu[1,2]:= - det*BA[2];
  uu[2,1]:= - det*CA[1];
  uu[2,2]:= det* BA[1];

  initialVal := A.ProjF[3];
  Hdifs[1]:= B.projF[3] - initialVal;
  Hdifs[2]:= C.projF[3] - initialVal;

  for i:= 1 to 2 do
    DD[i]:= Hdifs[1]* uu[i,1] + Hdifs[2]*uu[i,2];

  coefAC := CA[1] / CA[2];
  if B.projF[2] > A.projF[2] then begin
    coefAB := BA[1] / BA[2];
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.projF[2]));
    jSup := min(wdthMinus1, floor(B.projF[2]));
    if jSup >= jInf then begin
      for j := jInf to jSup do begin
        difj:= j-A.projF[2];
        rowInitialVal := initialVal+ difj*DD[2] ;
        auxac := A.projF[1] + difj * coefAC;
        auxab := A.projF[1] + difj * coefAB;
        if CAtLeftOfB then begin
          fromHere := Max(0, ceil(auxac));
          toThere := Min(floor(auxab), wdthMinus1);
        end
        else begin
          fromHere := Max(0, ceil(auxab));
          toThere := Min(floor(auxac), wdthMinus1);
        end;
        if fromHere <= toThere then begin
          for i := fromHere to toThere do begin
            difi:= i-A.projF[1];
            val:= rowInitialVal + difi*DD[1];
            if (val < fBuffer[i, j]) then
              fBuffer[i, j] := val;
          end;
        end;
      end;
    end;
  end;

  if B.projF[2]< C.projF[2] then begin
    coefBC := (B.projF[1]-C.projF[1]) / (B.projF[2]-C.projF[2]);
    jInf := max(0, ceil(B.projF[2]));
    jSup := min(wdthMinus1, floor(C.projF[2]));
    if jSup < jInf then
      Exit;
    for j := jInf to jSup do begin
      difj:= j - A.projF[2];
      difb:= j - B.projF[2];
      rowInitialVal := initialVal+ difj*DD[2] ;
      auxac := A.projF[1] + difj * coefAC;
      auxbc := B.projF[1] + difb * coefBC;
      if auxac < auxbc then begin
        fromHere := Max(0, ceil(auxac));
        toThere := Min(floor(auxbc), wdthMinus1);
      end
      else begin
        fromHere := Max(0, ceil(auxbc));
        toThere := Min(floor(auxac), wdthMinus1);
      end;
      if fromHere <= toThere then begin
        for i := fromHere to toThere do begin
          difi:= i - A.projF[1];
          val:= rowInitialVal + difi*DD[1];
          if (val < fBuffer[i, j]) then
            fBuffer[i, j] := val;
        end;
      end;
    end;
  end;
end;

procedure TDrawParamSurf.FillPointData(var pd: PointData);
{On call, 'pd.pos, pd.nor' are already been filled. This procedure
 fills the remaining fields. The field 'inside' is used only when there
 are clipping planes}
begin
  pd.proj := ProjectToPaneRealAndInv(pd.pos);
  {pd.inside := True;
  if gClips <> nil then
    for i := 0 to High(gClips) do
      if HeightOverClip(gClips[i], pd.pos) > 0 then begin
        pd.inside := True;
        break;
      end;}
  pd.projF := ProjectToFocusPlaneRealAndDist(pd.pos);
  AssignCol(pd);
end;

procedure TDrawParamSurf.FillTriangle(bm: TBitmap; vertices: triang);
{Fills several data of the pixels where some surface triangle projects.
  'gzBuffer[i,j]' :  z-buffer of pixel [i,j] as seen from 'gViewPoint'
  'gParamBuffer[i,j]' :  pixel [i,j] is the projection of a point surface
      given as the image of the parameter vector 'gParamBuffer[i,j]'
  'gBasicMap^[i, j]' :  code for the value of the color of pixel [i,j]
      in the palette, and it is from the obverse or the reverse side
      of the surface
In top of this, the pixel [i,j] of the bitmap 'bm' is given its corresponding
color.
The surface is not clipped, that is there aren't active clips.}

{See mote documentation of FillTriangleCurvWithClips in the source of the program
 Superficies, unit DrawParamSurf.pas}
var
  permu: threeBytes;
  reverse: Byte;
  virt: TPoint;
  A, B, C: PointData;
  BA, CA: array[1..2] of single;
  uu: array[1..2] of Vector2Single;
  Hdifs, DD: array[0..5,1..2] of single;
  val, initialVal, rowInitialVal: array[0..5] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC, fb1, fb2: single;
  k, j, jInf, jSup, i, fromHere, toThere, pm, cc: Integer;
  auxab, auxac, auxbc, aux, light: Single;
  line: pTColorArray;
  pt: TPoint;
  CAtLeftOfB, noFContour: Boolean;


  procedure ComputeDerivatives;
    var
      BAL, CAL: Single;
      i,j: Integer;
  begin
    uu[1,1]:= det*CA[2];
    uu[1,2]:= - det*BA[2];
    uu[2,1]:= - det*CA[1];
    uu[2,2]:= det* BA[1];

    initialVal[0]:= A.Proj[3];
    initialVal[1] := A.key[0] * A.proj[3];
    initialVal[2] := A.key[1] * A.Proj[3];
    initialVal[3] := A.projF[1] * A.Proj[3];
    initialVal[4] := A.projF[2] * A.Proj[3];
    initialVal[5] := A.projF[3] * A.Proj[3];

    Hdifs[0,1]:= B.proj[3] - A.proj[3];
    Hdifs[0,2]:= C.proj[3] - A.proj[3];
    Hdifs[1,1]:= B.proj[3] * B.key[0] - initialVal[1];
    Hdifs[1,2]:= C.proj[3] * C.key[0] - initialVal[1];
    Hdifs[2,1]:= B.proj[3] * B.key[1] - initialVal[2];
    Hdifs[2,2]:= C.proj[3] * C.key[1] - initialVal[2];
    for i:= 3 to 5 do begin
      Hdifs[i,1]:= B.proj[3] * B.projF[i-2] - initialVal[i];
      Hdifs[i,2]:= C.proj[3] * C.projF[i-2] - initialVal[i];
    end;

    for i:= 1 to 2 do
      for j:= 0 to 5 do
        DD[j,i]:= Hdifs[j,1]* uu[i,1] + Hdifs[j,2]*uu[i,2];

    BAL:= B.projF[3]-A.projF[3];
    CAL:= C.projF[3]-A.projF[3];

    det:= (B.projF[1]-A.projF[1])*(C.projF[2]-A.projF[2])-
        (B.projF[2]-A.projF[2])*(C.projF[1]-A.projF[1]);
    noFContour:= abs(det) > 1.0e-6;
    if noFContour then begin
      fb1:= (BAL*(C.projF[2]-A.projF[2])- CAL*(B.projF[2]-A.projF[2]))/det;
      fb2:= (CAL*(B.projF[1]-A.projF[1])-BAL*(C.projF[1]-A.projF[1]))/det;
    end;
  end;

begin
  reverse := numberOfColors*Face(vertices[1].pos, vertices[2].pos, vertices[3].pos);
  permu := gId;
  if vertices[2].proj[2] < vertices[1].proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].proj[2] < vertices[permu[1]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].proj[2] < vertices[permu[2]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];
  if A.proj[2] >= C.proj[2] then
    Exit;

  for i:= 1 to 2 do begin
    BA[i] := B.proj[i] - A.proj[i];
    CA[i] := C.proj[i] - A.proj[i];
  end;
  det := BA[1] * CA[2] - BA[2] * CA[1];
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ComputeDerivatives;

  coefAC := CA[1] / CA[2];

  if B.proj[2] > A.proj[2] then begin
    coefAB := BA[1] / BA[2];
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.proj[2]));
    jSup := min(wdthMinus1, floor(B.proj[2]));
    if jSup >= jInf then begin
      for j := jInf to jSup do begin
        difj:= j-A.proj[2];
        for k:= 0 to 5 do
          rowInitialVal[k] := initialVal[k]+ difj*DD[k,2];
        auxac := A.proj[1] + difj * coefAC;
        auxab := A.proj[1] + difj * coefAB;
        line := bm.ScanLine[j];
        if CAtLeftOfB then begin
          fromHere := Max(0, ceil(auxac));
          toThere := Min(floor(auxab), wdthMinus1);
        end
        else begin
          fromHere := Max(0, ceil(auxab));
          toThere := Min(floor(auxac), wdthMinus1);
        end;
        if fromHere <= toThere then begin
          for i := fromHere to toThere do begin
            difi:= i-A.proj[1];
            val[0]:= rowInitialVal[0] + difi*DD[0,1];
            if (val[0] > zBuffer[i, j]) then begin
              zBuffer[i, j] := val[0];
              aux:= 1/val[0];
              for k:= 1 to 5 do
                val[k]:= (rowInitialVal[k] + difi*DD[k,1])*aux;
              pt.x := Floor(val[3]);
              pt.y := Floor(val[4]);
              light := val[1];
              if PtInRect(rct, pt) and noFContour then begin
                aux:=val[5] + fb1*(pt.x - val[3])+ fb2*(pt.y - val[4]);
                if ( aux < 1.001*fBuffer[pt.x, pt.y]) then
                  light := light + val[2];
              end;
              pm := round(light);
              map[i, j] := pm + reverse shl 15;
              line^[i] := gammas[reverse,pm];
            end
          end;
        end;
      end;
    end;
  end;

  if B.proj[2]< C.proj[2] then begin
    coefBC := (B.proj[1]-C.proj[1]) / (B.proj[2]-C.proj[2]);
    jInf := max(0, ceil(B.proj[2]));
    jSup := min(wdthMinus1, floor(C.proj[2]));
    if jSup < jInf then
      Exit;
    for j := jInf to jSup do begin
      difj:= j - A.proj[2];
      difb:= j - B.proj[2];
      for k:= 0 to 5 do
        rowInitialVal[k] := initialVal[k]+ difj*DD[k,2] ;
      line := bm.ScanLine[j];
      auxac := A.proj[1] + difj * coefAC;
      auxbc := B.proj[1] + difb * coefBC;
      if auxac < auxbc then begin
        fromHere := Max(0, ceil(auxac));
        toThere := Min(floor(auxbc), wdthMinus1);
      end
      else begin
        fromHere := Max(0, ceil(auxbc));
        toThere := Min(floor(auxac), wdthMinus1);
      end;
      if fromHere <= toThere then begin
        for i := fromHere to toThere do begin
          difi:= i-A.proj[1];
          val[0]:= rowInitialVal[0] + difi*DD[0,1];
          if (val[0] > zBuffer[i, j]) then begin
            zBuffer[i, j] := val[0];
            aux:= 1/val[0];
            for k:= 1 to 5 do
              val[k]:= (rowInitialVal[k] + difi*DD[k,1])*aux;
            pt.x := Floor(val[3]);
            pt.y := Floor(val[4]);
            light := val[1];
            if PtInRect(rct, pt) and noFContour then begin
              aux:=val[5] + fb1*(pt.x - val[3])+ fb2*(pt.y - val[4]);
              if ( aux < 1.001*fBuffer[pt.x, pt.y]) then
                light := light + val[2];
            end;
            pm := round(light);
            map[i, j] := pm + reverse shl 15;
            line^[i] := gammas[reverse,pm];
          end
        end;
      end;
    end;
  end;
end;

procedure TDrawParamSurf.FillTriangleAsymptness(bm: TBitmap; vertices: triang);
{Fills several data of the pixels where some surface triangle projects.
  'gzBuffer[i,j]' :  z-buffer of pixel [i,j] as seen from 'gViewPoint'
  'gParamBuffer[i,j]' :  pixel [i,j] is the projection of a point surface
      given as the image of the parameter vector 'gParamBuffer[i,j]'
  'gBasicMap^[i, j]' :  code for the value of the color of pixel [i,j]
      in the palette, and it is from the obverse or the reverse side
      of the surface
In top of this, the pixel [i,j] of the bitmap 'bm' is given its corresponding
color.
The surface is not clipped, that is there aren't active clips.}

{See mote documentation of FillTriangleCurvWithClips in the source of the program
 Superficies, unit DrawParamSurf.pas}
var
  permu: threeBytes;
  virt: TPoint;
  A, B, C: PointData;
  BA, CA: array[1..2] of single;
  uu: array[1..2] of Vector2Single;
  Hdifs, DD: array[0..6,1..2] of single;
  val, initialVal, rowInitialVal: array[0..6] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC, fb1, fb2: single;
  k, j, jInf, jSup, i, fromHere, toThere, pm, cc: Integer;
  auxab, auxac, auxbc, aux, light: Single;
  line: pTColorArray;
  pt: TPoint;
  CAtLeftOfB, noFContour: Boolean;


  procedure ComputeDerivatives;
    var
      BAL, CAL: Single;
      i,j: Integer;
  begin
    uu[1,1]:= det*CA[2];
    uu[1,2]:= - det*BA[2];
    uu[2,1]:= - det*CA[1];
    uu[2,2]:= det* BA[1];

    initialVal[0]:= A.Proj[3];
    initialVal[1] := A.key[0] * A.proj[3];
    initialVal[2] := A.key[1] * A.Proj[3];
    initialVal[3] := A.projF[1] * A.Proj[3];
    initialVal[4] := A.projF[2] * A.Proj[3];
    initialVal[5] := A.projF[3] * A.Proj[3];
    initialVal[6] := A.c3 * A.Proj[3];

    Hdifs[0,1]:= B.proj[3] - A.proj[3];
    Hdifs[0,2]:= C.proj[3] - A.proj[3];
    Hdifs[1,1]:= B.proj[3] * B.key[0] - initialVal[1];
    Hdifs[1,2]:= C.proj[3] * C.key[0] - initialVal[1];
    Hdifs[2,1]:= B.proj[3] * B.key[1] - initialVal[2];
    Hdifs[2,2]:= C.proj[3] * C.key[1] - initialVal[2];
    for i:= 3 to 5 do begin
      Hdifs[i,1]:= B.proj[3] * B.projF[i-2] - initialVal[i];
      Hdifs[i,2]:= C.proj[3] * C.projF[i-2] - initialVal[i];
    end;
    Hdifs[6,1]:= B.proj[3] * B.c3 - initialVal[6];
    Hdifs[6,2]:= C.proj[3] * C.c3 - initialVal[6];

    for i:= 1 to 2 do
      for j:= 0 to 6 do
        DD[j,i]:= Hdifs[j,1]* uu[i,1] + Hdifs[j,2]*uu[i,2];

    BAL:= B.projF[3]-A.projF[3];
    CAL:= C.projF[3]-A.projF[3];

    det:= (B.projF[1]-A.projF[1])*(C.projF[2]-A.projF[2])-
        (B.projF[2]-A.projF[2])*(C.projF[1]-A.projF[1]);
    noFContour:= abs(det) > 1.0e-6;
    if noFContour then begin
      fb1:= (BAL*(C.projF[2]-A.projF[2])- CAL*(B.projF[2]-A.projF[2]))/det;
      fb2:= (CAL*(B.projF[1]-A.projF[1])-BAL*(C.projF[1]-A.projF[1]))/det;
    end;
  end;

begin
  permu := gId;
  if vertices[2].proj[2] < vertices[1].proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].proj[2] < vertices[permu[1]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].proj[2] < vertices[permu[2]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];
  if A.proj[2] >= C.proj[2] then
    Exit;

  for i:= 1 to 2 do begin
    BA[i] := B.proj[i] - A.proj[i];
    CA[i] := C.proj[i] - A.proj[i];
  end;
  det := BA[1] * CA[2] - BA[2] * CA[1];
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ComputeDerivatives;

  coefAC := CA[1] / CA[2];

  if B.proj[2] > A.proj[2] then begin
    coefAB := BA[1] / BA[2];
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.proj[2]));
    jSup := min(wdthMinus1, floor(B.proj[2]));
    if jSup >= jInf then begin
      for j := jInf to jSup do begin
        difj:= j-A.proj[2];
        for k:= 0 to 6 do
          rowInitialVal[k] := initialVal[k]+ difj*DD[k,2];
        auxac := A.proj[1] + difj * coefAC;
        auxab := A.proj[1] + difj * coefAB;
        line := bm.ScanLine[j];
        if CAtLeftOfB then begin
          fromHere := Max(0, ceil(auxac));
          toThere := Min(floor(auxab), wdthMinus1);
        end
        else begin
          fromHere := Max(0, ceil(auxab));
          toThere := Min(floor(auxac), wdthMinus1);
        end;
        if fromHere <= toThere then begin
          for i := fromHere to toThere do begin
            difi:= i-A.proj[1];
            val[0]:= rowInitialVal[0] + difi*DD[0,1];
            if (val[0] > zBuffer[i, j]) then begin
              zBuffer[i, j] := val[0];
              aux:= 1/val[0];
              for k:= 1 to 6 do
                val[k]:= (rowInitialVal[k] + difi*DD[k,1])*aux;
              pt.x := Floor(val[3]);
              pt.y := Floor(val[4]);
              light := val[1];
              if PtInRect(rct, pt) and noFContour then begin
                aux:=val[5] + fb1*(pt.x - val[3])+ fb2*(pt.y - val[4]);
                if ( aux < 1.001*fBuffer[pt.x, pt.y]) then
                  light := light + val[2];
              end;
              pm := round(light);
              map[i, j] := pm;
              cc:= 2+ ord(val[6] > 0);
              line^[i] := gammas[cc,pm];
            end
          end;
        end;
      end;
    end;
  end;

  if B.proj[2]< C.proj[2] then begin
    coefBC := (B.proj[1]-C.proj[1]) / (B.proj[2]-C.proj[2]);
    jInf := max(0, ceil(B.proj[2]));
    jSup := min(wdthMinus1, floor(C.proj[2]));
    if jSup < jInf then
      Exit;
    for j := jInf to jSup do begin
      difj:= j - A.proj[2];
      difb:= j - B.proj[2];
      for k:= 0 to 6 do
        rowInitialVal[k] := initialVal[k]+ difj*DD[k,2] ;
      line := bm.ScanLine[j];
      auxac := A.proj[1] + difj * coefAC;
      auxbc := B.proj[1] + difb * coefBC;
      if auxac < auxbc then begin
        fromHere := Max(0, ceil(auxac));
        toThere := Min(floor(auxbc), wdthMinus1);
      end
      else begin
        fromHere := Max(0, ceil(auxbc));
        toThere := Min(floor(auxac), wdthMinus1);
      end;
      if fromHere <= toThere then begin
        for i := fromHere to toThere do begin
          difi:= i-A.proj[1];
          val[0]:= rowInitialVal[0] + difi*DD[0,1];
          if (val[0] > zBuffer[i, j]) then begin
            zBuffer[i, j] := val[0];
            aux:= 1/val[0];
            for k:= 1 to 6 do
              val[k]:= (rowInitialVal[k] + difi*DD[k,1])*aux;
            pt.x := Floor(val[3]);
            pt.y := Floor(val[4]);
            light := val[1];
            if PtInRect(rct, pt) and noFContour then begin
              aux:=val[5] + fb1*(pt.x - val[3])+ fb2*(pt.y - val[4]);
              if ( aux < 1.001*fBuffer[pt.x, pt.y]) then
                light := light + val[2];
            end;
            pm := round(light);
            map[i, j] := pm;
            cc:= 2+ ord(val[6] > 0);
            line^[i] := gammas[cc,pm];
          end
        end;
      end;
    end;
  end;
end;

function TDrawParamSurf.FindRay(pix: TPoint): Vector3;
{Returns a vector 'result' such that all points of the line given by
 't  --> viewpoint + t* result' projects upon the pixel 'pix' of
 the 3D panel}
var
  k: Integer;
  ah, av: Double;
begin
  ah := pix.x - cx;
  av := pix.y - cy;
  for k := 1 to 3 do
    result[k] := ah * vH[k] + av * vV[k] + c0 * vN[k];
end;

procedure TDrawParamSurf.FrameFromFocus(w: Integer);
{For cast shadows computation, we project the scene on a virtual
 panel in the same manner as for the projection of the scene on the
 drawing panel. Now we compute the projection parameters to that plane.}
var
  r: Double;
  i, j, k, h: Integer;
  maximum, minimum: Vector2;
  pos, proj: Vector3;
begin
  r := 0.3 * EuclDistance(ctr3DRed, viewpoint);
  for i := 1 to 3 do
    focus[i] := viewpoint[i] - r * (vH[i] + vV[i]);
  r := EuclDistance(ctr3DRed, focus);
  for i := 1 to 3 do
    fN[i] := (ctr3DRed[i] - focus[i]) / r;
  r := hypot(fN[1], fN[2]);
  if (r > 1.0E-6) then begin
    fH[1] := fN[2] / r;
    fH[2] := -fN[1] / r;
    fH[3] := 0.0;
    fV[1] := -fH[2] * fN[3];
    fV[2] := fH[1] * fN[3];
    fV[3] := -r;
  end
  else begin
    fH[1] := 1.0;
    fH[2] := 0.0;
    fH[3] := 0.0;
    fV[1] := 0.0;
    fV[2] := 1.0;
    fV[3] := 0.0;
  end;
  cf0 := 1;
  cf1 := EuclInnerProd(focus, fH);
  cf2 := EuclInnerProd(focus, fV);
  cf3 := EuclInnerProd(focus, fN);
  cfx := w div 2;
  cfy := w div 2;

  for i:= 1 to 2 do begin
    maximum[i]:= -1.0e20;
    minimum[i]:=  1.0e20;
  end;
  for i:= 0 to 1 do
    for j:= 0 to 1 do
      for k:= 0 to 1 do begin
         pos[1]:= i*min3DRed[1] + (1-i)*max3DRed[1];
         pos[2]:= j*min3DRed[2] + (1-j)*max3DRed[2];
         pos[3]:= k*min3DRed[3] + (1-k)*max3DRed[3];
         proj:= ProjectToFocusPlaneRealAndDist(pos);
         for h:= 1 to 2 do begin
           maximum[h]:= max(maximum[h], proj[h]);
           minimum[h]:= min(minimum[h], proj[h]);
         end;
      end;
  r:= min(w / (maximum[1] - minimum[1]+1.1e-6), w /
    (maximum[2] - minimum[2] + 1.1e-6));
  cfx:= round(w - r*(maximum[1]-cfx)/cf0);
  cfy:= round(w - r*(maximum[2]-cfy)/cf0);
  cf0:= r;
end;

procedure TDrawParamSurf.FrameFromViewPoint(w: Integer);
{We compute the parameters used in the conic projection with center
 in 'viewpoint' on a plane orthogonal to 'vN'}
var
  maxDim: Double;
  i, j, k, h: Integer;
  maximum, minimum: Vector2;
  pos, proj: Vector3;
begin
  span3D:= EuclDistance(min3DRed, max3DRed);
  dist:= cFarthness*span3D;
  viewPoint[1]:= dist + ctr3DRed[1];
  viewPoint[2]:= ctr3DRed[2];
  viewPoint[3]:= ctr3DRed[3];

  vN[1]:= -1;
  vN[2]:= 0;
  vN[3]:= 0;

  vH[1]:= 0;
  vH[2]:= 1;
  vH[3]:= 0;

  vV[1]:= 0;
  vV[2]:= 0;
  vV[3]:= -1;

  {provisional projections coefficientes. 'gC0' is a scale factor}
  c0 := 1;
  c1 := EuclInnerProd(viewpoint, vH);
  c2 := EuclInnerProd(viewpoint, vV);
  c3 := EuclInnerProd(viewpoint, vN);

  {displacement of the origin of the plane so that the
  center of the cube is projected into the center of the plane}
  cx := w div 2;
  cy := w div 2;

  for i:= 1 to 2 do begin
    maximum[i]:= -1.0e20;
    minimum[i]:=  1.0e20;
  end;
  for i:= 0 to 1 do
    for j:= 0 to 1 do
      for k:= 0 to 1 do begin
         pos[1]:= i*min3DRed[1] + (1-i)*max3DRed[1];
         pos[2]:= j*min3DRed[2] + (1-j)*max3DRed[2];
         pos[3]:= k*min3DRed[3] + (1-k)*max3DRed[3];
         proj:= ProjectToPaneRealAndInv(pos);
         for h:= 1 to 2 do begin
           maximum[h]:= max(maximum[h], proj[h]);
           minimum[h]:= min(minimum[h], proj[h]);
         end;
      end;
  maxDim:= 1.0e-6 + max(maximum[1]-minimum[1], maximum[2]-minimum[2]);
  if maxDim > 1.0e-6 then {give 5% more room than strictly needed}
    c0:= 0.95*w/maxDim;
end;

procedure TDrawParamSurf.GetCorrectViewPoint(w: Integer);
{place the viewpoint at a given relative distance from the cube,
and adjust the scale of the projection}
var
  aux: Double;
  i, j,k, h: Integer;
  maximum, minimum: Vector2;
  pos, proj: Vector3;
begin
  span3D:= EuclDistance(min3DRed, max3DRed);
  dist:= span3D*cFarthness;
  for i:= 1 to 3 do {distance from the cube center to the viewpoint is
  made equal to 'cFarthness' times the diagonal of the cube comprising
  the surface}
    viewPoint[i]:= ctr3DRed[i] - dist*vN[i];

  c3 := EuclInnerProd(viewpoint, vN);
  c0:= 1;
  for i:= 1 to 2 do begin
    maximum[i]:= -1.0e20;
    minimum[i]:=  1.0e20;
  end;
  for i:= 0 to 1 do
    for j:= 0 to 1 do
      for k:= 0 to 1 do begin
         pos[1]:= i*min3DRed[1] + (1-i)*max3DRed[1];
         pos[2]:= j*min3DRed[2] + (1-j)*max3DRed[2];
         pos[3]:= k*min3DRed[3] + (1-k)*max3DRed[3];
         proj:= ProjectToPaneRealAndInv(pos);
         for h:= 1 to 2 do begin
           maximum[h]:= max(maximum[h], proj[h]);
           minimum[h]:= min(minimum[h], proj[h]);
         end;
      end;
  aux:= 1.0e-6 + max(maximum[1]-minimum[1], maximum[2]-minimum[2]);
  if aux > 1.0e-6 then {adjust the scale of the conic projection}
    c0:= 0.95*w/aux;
end;

procedure TDrawParamSurf.InitializeBuffers;
  var
    i, j: Integer;
begin
  for i:= 0 to wdthMinus1 do
    for j:= 0 to wdthMinus1 do begin
      fBuffer[i,j]:= kBigFiller;
      zBuffer[i,j]:= 0;
      map[i,j]:= 0;
    end;
end;

procedure TDrawParamSurf.InitializeCameras(w: Integer);
begin
  FrameFromViewPoint(w);
  FrameFromFocus(w);
end;

procedure TDrawParamSurf.MoveSurface(shift: TShiftState; prevPt, curPt: TPoint);
{Change the viewpoint to get the effect that it is the surface that is moving.
 Then redraw the surface}
begin
  MoveViewPoint(prevPt,curPt,Shift);
  {Redraw the surface}
  DrawSurface;
end;

procedure TDrawParamSurf.MoveViewPoint(prevPt,curPt: TPoint; Shift: TShiftState);
{Uses Dick Palais algorithm for rotating. The rotation is defined by
two unit vectors 'e1,e2' as if we had a rigid ball that could move around
its fixed center and one moved the point of the ball at position 'e1'
to the position 'e2' dragging so all the ball}
var
  e1, e2, b: Vector3;
  norm, twoVe2, e2b, vb, scale, c0abs: double;
  i, mode: Integer;
  good: Boolean;
begin
  c0abs := abs(c0);
  good := DefinedUnitVector(prevPt, e1); {Uses stereographic projection}

  if ssCtrl in Shift then
    mode := 1  {Translate surface}
  else if ssShift in Shift then
    mode := 2  {Zoom surface}
  else
    mode := 0; {Rotate surface}

  case mode of
    1: begin   {add the mouse translation to all points of the surface}
        cx := cx - prevPt.x + curPt.x;
        cy := cy - prevPt.y + curPt.y;
      end;
    2: begin {modify the scale coefficient 'c0' of the projection}
        scale := (1 + (prevPt.y - curPt.y) / 500);
        if ((c0abs > 1) or (scale > 1)) and ((c0abs < 10000) or (scale <
          1)) then
          c0 := c0 * scale;
      end;
  else begin  {apply the Dick Palais rotation algorithm to all of the
               vectors that appear in the projection formula and to the
               viewpoint}
      good := good and DefinedUnitVector(curPt, e2);
      if good then begin
        for i := 1 to 3 do
          b[i] := e1[i] + e2[i];
        norm := EuclNorm(b);
        for i := 1 to 3 do
          b[i] := b[i] / norm;
        e2b := EuclInnerProd(e2, b);
        twoVe2 := 2 * EuclInnerProd(e2, vH);
        vb := EuclInnerProd(b, vH);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          vH[i] := vb * b[i] - twoVe2 * e2[i] + vH[i];
        twoVe2 := 2 * EuclInnerProd(e2, vV);
        vb := EuclInnerProd(b, vV);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          vV[i] := vb * b[i] - twoVe2 * e2[i] + vV[i];
        twoVe2 := 2 * EuclInnerProd(e2, vN);
        vb := EuclInnerProd(b, vN);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          vN[i] := vb * b[i] - twoVe2 * e2[i] + vN[i];
        c3 := EuclInnerProd(ctr3DRed, vN) - dist;
        c1 := EuclInnerProd(ctr3DRed, vH);
        c2 := EuclInnerProd(ctr3DRed, vV);
        for i := 1 to 3 do
          viewpoint[i] := c3 * vN[i] + c1 * vH[i] + c2 * vV[i];
        FrameFromFocus(wdth);
      end;
    end;
  end;
end;

function TDrawParamSurf.ProjectToFocusPlaneRealAndDist(x: Vector3): Vector3;
begin
  result[3] :=fN[1] * x[1] + fN[2] * x[2] + fN[3] * x[3] - cf3;
  result[1] := (fH[1] * x[1] + fH[2] * x[2] + fH[3] * x[3] - cf1) * cf0 +
     cfx;
  result[2] := (fV[1] * x[1] + fV[2] * x[2] + fV[3] * x[3] - cf2) * cf0 +
     cfy;
end;

function TDrawParamSurf.ProjectToPaneRealAndInv(x: Vector3): Vector3;
{Gives in 'result[1], result[2]' the coni projection of 'x' in the bitmap, Also,
 gives the inverse, 'result[3]', of the height of 'x' with respect to the plane passing by 'gViewpoint' that is
 orthogonal to the line from 'viewpoint' to 'ctr3DRed'}
begin
  result[3] := 1/(vN[1] * x[1] + vN[2] * x[2] + vN[3] * x[3] - c3);
  result[1] := (vH[1] * x[1] + vH[2] * x[2] + vH[3] * x[3] - c1) * c0 * result[3] +
    cx;
  result[2] := (vV[1] * x[1] + vV[2] * x[2] + vV[3] * x[3] - c2) * c0 * result[3] +
    cy;
end;

procedure TDrawParamSurf.ReceiveBasicData(min3D, max3D: Vector3; grid, w: Integer);
{the cube containing the surface is computed elsewhere and copied here. Also,
 the width of the bitmap is passed in 'w' and in 'grid' the density of rectangles
 in each coordinate direction}
var
  i: Integer;
begin
  min3DRed:= min3D;
  max3Dred:= max3D;
  for i:= 1 to 3 do
    ctr3DRed[i]:= (min3DRed[i]+ max3DRed[i])/2;
  gr:= grid;
  wdth:= w;
  rct:= Rect(0,0,wdth,wdth);
  wdthMinus1:= wdth-1;
  rRed:= Rect(0,0,wdthMinus1,wdthMinus1);
  SetBufferSizes(wdth, True);
  if high(vert) <> gr then begin
    SetLength(vert,gr+1);
    for i:= 0 to gr do
      SetLength(vert[i], gr+1);
  end;
  if bmp.width <> wdth then begin
    bmp.Width:= wdth;
    bmp.Height:= wdth;
  end;
end;

procedure TDrawParamSurf.ScaleCoefficients(sc: Double);
{used when the drawing is to be made for the printer}
begin
  c0:= sc*c0;
  cx:= sc*cx;
  cy:= sc*cy;
end;


procedure Initialize;
begin
  dps:= TDrawParamSurf.Create;
End;

procedure Finalize;
begin
  FreeAndNil(dps);
end;


initialization
   Initialize;
finalization
   Finalize;

end.
