
{-----------------------------------------------------------------------------
 Unit Name: FEvaluator_32
 Author:    Angel Montesinos Amilibia
 October 15, 2011 -  February 24, 2011
 Purpose: Class TEvaluator_32. It is ancillary to TFunctionParser_32.
 From an instance of class TFunctionParser_32, and using its RPN
 equivalent, it creates evaluators of the function and its derivatives first and
 second. Provides also some useful applications of these evaluators

 For source code you may email me to:
     montesin@uv.es
-----------------------------------------------------------------------------}
{

Unit Name: FEvaluator_32
Evaluator of functions of several variables built from an instance of
TFunctionParser_32

Copyright (C) 2011 Angel Montesinos Amilibia

This unit is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This unit is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can obtain more details by emailing me to:
     montesin@uv.es

Or writing me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andres Estelles, 1
     46100 Burjasot (Valencia)
     Spain
}

{See OCFunctDDDoubleSafeIntegrals for documentation}

unit FEvaluator_32;

interface
uses
  FunctionParser_32, Globals;

type

FirstDeriv= record
  fa0, dfa0: double;
end;

SecondDeriv= record
  fa0, dfa0, ddfa0: double;
end;


TEvaluator_32 = class(TObject)
  PRIVATE
    data : array of Double; {stores the value of the
      parameters and constants that appear in the formula}
    u, w: array of Double;
    emptyRegisters: Byte;
    minCorner, jump1, jump2, invJump1, invJump2: array of double;
    intList1: IntegralDataList1;
    intList2: IntegralDataList2;
    lastIntegral1, lastIntegral2: Integer;
    fC, dC, ddC: AnsiString;
    theArg: array of Double;
    currentJ: Integer; {Used for the derivatives to know where to read the
                       current position of 'u' and 'w'; Don't use for any else }
    argNum, parNum, lengthActorList : Integer; {argNum=Length[allowedVars]}
    actorList : AnsiString; {the RPN string that represents the function}
    {Gives the value of the funcion at the point (X[0],..., X[argNum-1])}
    ffBlock: PByte;
    vvBlock: PByte;
    aaBlock: PByte;
    procedure AddAddressToCode(p: PAnsiChar; var cod: AnsiString);
    procedure BuildFirstDerivative;
    procedure BuildSecondDerivative;
    procedure CodifyFunction;
    function EvaluateBSplinesFunction(index: Integer): Double;
    function EvaluateBSplinesFunctionAndFirstDer(index: Integer): Double;
    function EvaluateBSplinesFunctionFirstAndSecondDer(index: Integer): Double;
    procedure LeaveRoomInFPUStack(neededPlaces: Integer; var cod: AnsiSTring);
    function LoadAd(i: Integer): AnsiString;
    function LoadVd(i: Integer): AnsiString;
    function WriteDdP(const a): AnsiString;
    procedure AddIntToCode(n: Integer; var cod: AnsiString);
    procedure PopFromAppStack(var cod: AnsiSTring);
    procedure PutCodeInExeBlock(const cod: AnsiString; var exeBlock: PByte);
    procedure SetUWLength;
  PUBLIC
    ff: function(const X: array of Double): Double;
    {Gives the value of the funcion at the point X}
    vv: function(const XV: array of Double): Double;
    {Gives the value of the funcion at the point (XV[0],..., XV[argNum-1])
     and leaves in u[1] the value of the directional derivative of the
     function with respect to the vector (XV[argNum],..., XV[2*argNum-1])}
    aa: function(const XVA: array of Double): Double;
    {Gives the value of the function at the point (XVA[0],..., XVA[argNum-1])
     and leaves in u[1] the value of the directional derivative of the
     function with respect to the vector (XVA[argNum],..., XVA[2*argNum-1]),
     and leaves in w[1] the value of the second derivative of the
     function with respect to the velocity (XVA[argNum],..., XVA[2*argNum-1]),
     and acceleration (XVA[2*argNum],..., XVA[3*argNum-1])}

    usedVarsAndParams: AnsiString;
    procedure Build(theF: TFunctionParser_32);
    destructor Destroy; OVERRIDE;
    procedure DiffAndHessian(x: Vector3;  var df: Vector3; var hess: matrix3); overload;
       {Use this when ArgNum = 3}
    procedure DiffAndHessian(x: Vector2; var f: Double; var df: Vector2; var hess: matrix2); overload;
       {Use this when ArgNum = 2}
    function FirstD(const XV: array of Double): FirstDeriv;
    function gradF(const x: Vector3): Vector3; overload;
       {Use this when ArgNum = 3}
    function gradF(const x: Vector2): Vector2; overload;
       {Use this when ArgNum = 2}
    function SecondD(const XV: array of Double): SecondDeriv;
   // function TheF(const x: Vector3): Double;
    procedure UpdateParameters(parameters: array of Double);
  end;

  Evaluator_32List= array of TEvaluator_32;

  var
    gTheF1, gTheF2, gTheF3, gTheF4: TEvaluator_32;
    gTheF1Clones: Evaluator_32List; {for drawing an impl surf, clones of 'gTheF1',
       one for each thread}
    gImplParts: Evaluator_32List; {used for drawing geometric curves on the
       implicit surface}
    gLastImplPart: Integer;

implementation

uses
  SysUtils, Math, Windows, 
  Stutz_Get_CPU_data, Utilities, ResStrings;

const
  halfPi: double = 1.5707963267948966192313;
  twoPi: double = 6.283185307179586476925;
  eMaxNum = 1.0E30;
  eMinNum = 1.0E-30;
  eMaxExp = 100;

  kLastChar = #$FF;
  {FPU registers number}

  c_stackDisplacDouble = 8;
  c_stackDisplacDoubleHStr= #$08; {hexadecimal string of c_stackDisplacDouble}
  c_LastReg = 7; {last register in the FPU}

  kZeroPointerStr32 = #$00#$00#$00#$00;

  {Code constants}
  c_fstp0ESPd = #$DD#$9C#$24#$00#$00#$00#$00; //fstp st(0) to [ESP] as Double
  c_fldESPd = #$DD#$84#$24#$00#$00#$00#$00;  //fld from [ESP] as a Double.

  c_LoadPi = #$D9#$EB;
  c_Load1 = #$D9#$E8;
  c_Load0 = #$D9#$EE;
  c_zero_comp = #$D9#$EE#$DF#$F1;
    {4 bytes. ST(cLastReg) must be empty. Compara 0 con x=ST(0). Despues, JA salta
     si 0>x; JNA salta si 0 <= x; JB, si 0 < x; JNB, si 0 >= x; JNE si 0<>x}
  c_1_comp = #$D9#$E8#$DF#$F1;
    {4 bytes.  ST(cLastReg) must be empty. Compara 1 con x=ST(0). Despues, JA salta
     si 1>x; JNA salta si 1 <= x; JB, si 1 < x; JNB, si 1 >= x; JNE si 1<>x}
  c_mod_pi = #$D9#$EB#$D9#$C9#$D9#$F8#$DD#$D9;
    {8 bytes. ST(cLastReg) must be empty}
  {c_mod_2pi = #$D9#$EB#$D8#$C0#$D9#$C9#$D9#$F8#$DD#$D9;}
    {10 bytes. ST(cLastReg) must be empty}
 // c_push_eax = #$50;
 // c_pop_eax = #$58;
  c_ln = #$D9#$E8#$D9#$C9#$D9#$F1#$D9#$ED#$DE#$C9;
    {10 bytes. Requires ST(0) > 0 and ST(cLastReg) empty}
  c_2toX = #$D9#$C0#$D9#$FC#$D9#$C9#$D8#$E1#$D9#$F0#$D9#$E8#$DE#$C1#$D9#$C9#$D9
    +
    #$E8#$D9#$FD#$DD#$D9#$DE#$C9;
    {24 bytes. Requires ST6) and ST(cLastReg) empty}
  c_div2 = c_Load1 + #$D9#$E0#$D9#$C9#$D9#$FD#$DD#$D9; 
      {10 bytes. Requires ST(cLastReg) empty}
var
  uSpecialValues: array[0..7] of double;
  c_mod_2Pi, c_div_S, c_ln_S, c_2toX_S: AnsiString;

function EvaluateBSplineOnCurve(order, k1: Integer; a1: Double; dd: PArrayNodes1): Double;
var
  b0, a: Double;
  i,r: Integer;
  p: array[0..4,0..4] of Double; //4 is the expected maximal order
begin
  for i:= 0 to order do
    p[0,i]:= dd^[i+k1-order];
  for r:= 1 to order do begin
    b0:= 1/(order+1-r);
    for i:= r to order do begin
      a:= b0*(a1 - i - k1 + order);
      p[r,i]:= a*p[r-1,i]+(1-a)*p[r-1,i-1];
    end;
  end;
  result:= p[order,order];  
end; 

function EvaluateBSplineOnSurface(order1, order2, k1,k2: Integer; a1, a2: Double; dd: PArrayArrayNodes2): Double;
var
  b0, a: Double;
  i,j, r: Integer;
  d: array[0..4] of double;  //4 is the expected maximal order
  p: array[0..4, 0..4] of Double; //4 is the expected maximal order
begin
  for j:= k2-order2 to k2 do begin
    for i:= 0 to order1 do
      p[0,i]:=dd^[i+k1-order1,j];
    for r:= 1 to order1 do begin
      b0:= 1/(order1+1-r);
      for i:= r to order1 do begin
        a:= b0*(a1 - i - k1 + order1);
        p[r,i]:= a*p[r-1,i]+(1-a)*p[r-1,i-1];
      end;
    end;
    d[j-k2+order2]:= p[order1,order1];
  end;
  for i:= 0 to order2 do
    p[0,i]:= d[i];
  for r:=1 to order2 do begin
    b0:= 1/(order2+1-r);
    for i:= r to order2 do begin
      a:= b0*(a2 - i - k2 + order2);
      p[r,i]:= a*p[r-1,i]+(1-a)*p[r-1,i-1];
    end;
  end;
  result:= p[order2,order2];  
end;
  

function sn(n: Pointer): AnsiString;
{Converts a pointer to a string in the format that the CPU needs to
 read correctly a memory address}
var
  c: PAnsiChar;
  i: Integer;
begin
  c:= @n;
  Result := '';
  for i := 0 to SizeOf(Pointer) - 1 do
    Result := Result + AnsiChar(c[i]);
end;

procedure TEvaluator_32.AddIntToCode(n: Integer; var
   cod: AnsiString);
var
  c: PAnsiChar;
  s: AnsiString;
  i: Integer;
begin
  c := @n;
  s := '';
  for i := 0 to SizeOf(Integer) - 1 do
    s := s + AnsiChar(c[i]);
  cod := cod + s;
end;


function LoadXd(i: Integer): AnsiString;
{Reads from the function stack the independent variable 'i' of Double
type and stores it in ST(0)}
begin
  if i = 1 then
    result:= #$DD#$00
  else if i > 1 then
    result:= #$DD#$40 + chr(SizeOf(Double) * (i-1));
end;

procedure TEvaluator_32.AddAddressToCode(p: PAnsiChar; var
   cod: AnsiString);
{Adds to 'cod' the address written in 'p', in the format that is
required by the CPU}
var
  s: AnsiString;
  i: Integer;
begin
  s := '';
  for i := 0 to SizeOf(Pointer) - 1 do
    s := s + p[i];
  cod := cod + s;
end;

procedure TEvaluator_32.DiffAndHessian(x: Vector3; var df: Vector3; var hess: matrix3);
begin
  aa([x[1],x[2],x[3],1,0,0,0,0,0]);
  df[1]:= u[1];
  hess[1,1]:= w[1];
  aa([x[1],x[2],x[3],0,1,0,0,0,0]);
  df[2]:= u[1];
  hess[2,2]:= w[1];
  aa([x[1],x[2],x[3],0,0,1,0,0,0]);
  df[3]:= u[1];
  hess[3,3]:= w[1];
  aa([x[1],x[2],x[3],1,1,0,0,0,0]);
  hess[1,2]:= (w[1] -  hess[1,1]- hess[2,2])/2;
  hess[2,1]:= hess[1,2];
  aa([x[1],x[2],x[3],0,1,1,0,0,0]);
  hess[2,3]:= (w[1] - hess[2,2]- hess[3,3])/2;
  hess[3,2]:= hess[2,3];
  aa([x[1],x[2],x[3],1,0,1,0,0,0]);
  hess[1,3]:= (w[1] - hess[3,3]- hess[1,1])/2;
  hess[3,1]:= hess[1,3];
end;

procedure TEvaluator_32.DiffAndHessian(x: Vector2; var f: Double; var df: Vector2; var hess: matrix2);
begin
  f:= aa([x[1],x[2],1,0,0,0]);
  df[1]:= u[1];
  hess[1,1]:= w[1];
  aa([x[1],x[2],0,1,0,0]);
  df[2]:= u[1];
  hess[2,2]:= w[1];
  aa([x[1],x[2],1,1,0,0]);
  hess[1,2]:= (w[1] -  hess[1,1]- hess[2,2])/2;
  hess[2,1]:= hess[1,2];
end;

function TEvaluator_32.FirstD(const XV: array of Double): FirstDeriv;
begin
  result.fa0:= vv(XV);
  result.dfa0:= u[1];
end;

function TEvaluator_32.LoadVd(i: Integer): AnsiString;
{Reads from the function stack the velocity 'i' of Double
type and stores it in ST(0)}
begin
  result:= #$DD#$40 + chr(SizeOf(Double )* (argNum + i-1));
end;

function TEvaluator_32.LoadAd(i: Integer): AnsiString;
{Reads from the function stack the velocity 'i' of Double
type and stores it in ST(0)}
begin
  result:= #$DD#$40 + chr(SizeOf(Double )* (2*argNum + i-1));
end;

function LoadDd(const a): AnsiString;
{Reads from the address of the var 'a' as a Double and stores it in ST(0)}
begin
  result:= #$DD#$05 + sn(@a);
end;

function TEvaluator_32.WriteDdP(const a): AnsiString;
{Writes ST(0), as an Double, into the memory location @a, and pops ST(0)}
begin
  result:= #$DD#$1D + sn(@a);
end;

function WriteDd(const a): AnsiString;
{Writes ST(0), as an Double, into the memory location @a, but not pops}
begin
  result:= #$DD#$15 + sn(@a);
end;

function TEvaluator_32.EvaluateBSplinesFunction(index: Integer): Double;
  var
    u0, v0, a1, a2: Double;
    dim, k1, k2, var1, var2: Integer;
begin
  if index <= $0F then begin{dimension = 1}
    dec(index);
    var1:= intList1[index].var1;
    u0:= minCorner[var1]-(cQ+2)*Jump1[var1];
    a1:= (theArg[var1]-u0)*invJump1[var1];
    k1:= Floor(a1);
    k1:= Min(k1, cN1+cQ+2);
    k1:= max(k1,cQ+2);
    result:= EvaluateBSplineOnCurve(4,k1,a1,@intList1[index].Data[0]);
  end
  else begin
    index:= index shr 4 - 1;
    var1:= intList2[index].var1;
    var2:= intList2[index].var2;
    u0:= minCorner[var1]-(cQ+2)*Jump2[var1];
    v0:= minCorner[var2]-(cQ+2)*Jump2[var2];
    a1:= (theArg[var1]-u0)*invJump2[var1];
    a2:= (theArg[var2]-v0)*invJump2[var2];
    k1:= Floor(a1);
    k1:= Min(k1, cN2+cQ+2);
    k1:= max(k1,cQ+2);
    k2:= Floor(a2);
    k2:= Min(k2, cN2+cQ+2);
    k2:= max(k2,cQ+2);
    result:= EvaluateBSplineOnSurface(4,4,k1,k2,a1,a2,@intList2[index].Data[0,0]);
  end
end;

function TEvaluator_32.EvaluateBSplinesFunctionAndFirstDer(index: Integer):Double;
  var
    u0, v0, a1, a2, aux: Double;
    k1, k2, var1, var2: Integer;
begin
  if index <= $0F then begin{dimension = 1}
    dec(index);
    var1:= intList1[index].var1;
    u0:= minCorner[var1]-(cQ+2)*Jump1[var1];
    a1:= (theArg[var1]-u0)*invJump1[var1];
    k1:= Floor(a1);
    k1:= Min(k1, cN1+cQ+2);
    k1:= max(k1,cQ+2);
    result:= EvaluateBSplineOnCurve(4,k1,a1,@intList1[index].Data[0]);
    u[currentJ]:= theArg[var1+ argNum]*EvaluateBSplineOnCurve(3,k1,a1,@intList1[index].Data[1]);
  end
  else begin
    index:= index shr 4 - 1;
    var1:= intList2[index].var1;
    var2:= intList2[index].var2;
    u0:= minCorner[var1]-(cQ+2)*Jump2[var1];
    v0:= minCorner[var2]-(cQ+2)*Jump2[var2];
    a1:= (theArg[var1]-u0)*invJump2[var1];
    a2:= (theArg[var2]-v0)*invJump2[var2];
    k1:= Floor(a1);
    k1:= Min(k1, cN2+cQ+2);
    k1:= max(k1,cQ+2);
    k2:= Floor(a2);
    k2:= Min(k2, cN2+cQ+2);
    k2:= max(k2,cQ+2);
    result:= EvaluateBSplineOnSurface(4,4,k1,k2,a1,a2,@intList2[index].Data[0,0]);
    u[currentJ]:= 0;
    aux:= theArg[var1+ argNum];
    if aux <>0 then
      u[currentJ]:= aux*EvaluateBSplineOnSurface(3,4,k1,k2,a1,a2,@intList2[index].Data[1,0]);
    aux:= theArg[var2+ argNum];
    if aux <> 0 then
      u[currentJ]:= u[currentJ] + aux*EvaluateBSplineOnSurface(3,4,k1,k2,a1,a2,@intList2[index].Data[0,1]);
  end
end;

function TEvaluator_32.EvaluateBSplinesFunctionFirstAndSecondDer(index: Integer):Double;
var
    u0, v0, a1, a2, d1,d2,d3: Double;
    k1, k2, var1, var2: Integer;
begin
  if index <= $0F then begin{dimension = 1}
    dec(index);
    var1:= intList1[index].var1;
    u0:= minCorner[var1]-(cQ+2)*Jump1[var1];
    a1:= (theArg[var1]-u0)*invJump1[var1];
    k1:= Floor(a1);
    k1:= Min(k1, cN1+cQ+2);
    k1:= max(k1,cQ+2);
    result:= EvaluateBSplineOnCurve(4,k1,a1,@intList1[index].Data[0]);
    d1:= EvaluateBSplineOnCurve(3,k1,a1,@intList1[index].Data[1]);
    u[currentJ]:= d1 * theArg[var1+ argNum];
    w[currentJ]:= d1 * theArg[var1 + 2*argNum];
    w[currentJ]:= w[currentJ] +
       sqr(theArg[var1+ argNum])*EvaluateBSplineOnCurve(2,k1,a1,@intList1[index].Data[2]);
  end
  else begin
    index:= index shr 4 - 1;
    var1:= intList2[index].var1;
    var2:= intList2[index].var2;
    u0:= minCorner[var1]-(cQ+2)*Jump2[var1];
    v0:= minCorner[var2]-(cQ+2)*Jump2[var2];
    a1:= (theArg[var1]-u0)*invJump2[var1];
    a2:= (theArg[var2]-v0)*invJump2[var2];
    k1:= Floor(a1);
    k1:= Min(k1, cN2+cQ+2);
    k1:= max(k1,cQ+2);
    k2:= Floor(a2);
    k2:= Min(k2, cN2+cQ+2);
    k2:= max(k2,cQ+2);
    result:= EvaluateBSplineOnSurface(4,4,k1,k2,a1,a2,@intList2[index].Data[0,0]);
    d1:= EvaluateBSplineOnSurface(3,4,k1,k2,a1,a2,@intList2[index].Data[1,0]);
    d2:= EvaluateBSplineOnSurface(4,3,k1,k2,a1,a2,@intList2[index].Data[0,1]);
    u[currentJ]:= d1*theArg[var1+ argNum]+ d2*theArg[var2 + argNum];
    w[currentJ]:= d1*theArg[var1 + 2*argNum]+ d2*theArg[var2 + 2*argNum];
    d1:= EvaluateBSplineOnSurface(2,4,k1,k2,a1,a2,@intList2[index].Data[2,0]);
    d2:= EvaluateBSplineOnSurface(4,2,k1,k2,a1,a2,@intList2[index].Data[0,2]);
    d3:= EvaluateBSplineOnSurface(3,3,k1,k2,a1,a2,@intList2[index].Data[1,1]);
    w[currentJ]:= w[currentJ] + d1*sqr(theArg[var1+ argNum])+ d2*sqr(theArg[var2 + argNum])+
             2*d3*theArg[var1 + argNum]*theArg[var2 + argNum] ;

  end
end;

procedure TEvaluator_32.CodifyFunction;
var
  i, j: Integer;
  k: Int64;
  prot: Word;
  ch: AnsiChar;
  c: PAnsiChar;
  escape: Boolean;
  n: array of Integer;
  lstn: Integer;
begin
  j:= $2000;
  if FFBlock = nil then begin
    prot:= PAGE_EXECUTE_READWRITE;
    FFBlock:= VirtualAlloc(nil, j, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(FFBlock, j, PAGE_EXECUTE_READWRITE, @prot);
  end;
  lstn:= -1;
  fC := '';
  if (lastIntegral1 >= 0) or (lastIntegral2 >= 0) then
    for i:= 1 to argNum do
      fC:= fC + LoadXd(i) + WriteDdP(theArg[i-1]);
  emptyRegisters := c_LastReg + 1;
  i := 0;
  while i < lengthActorList do begin
    inc(i);
      ch := actorList[i];
      escape := (ch = kLastChar);
      if escape or (ch = 'p') or (ch = '1') then
        begin
          if escape then inc(i);
          ch := actorList[i];
          LeaveRoomInFPUStack(1, fC);
          if escape then
            begin
              if (ord(ch) > argNum) then
                fC := fC + LoadDd(data[ord(ch) - argNum])
              else if (ord(ch)<= argNum) then
                fC := fC + LoadXd(ord(ch));
            end
          else if (ch = 'p') then
            fC := fC + c_LoadPi
          else
            fC := fC + c_Load1;
          dec(emptyRegisters);
        end
      else
        begin
          if (ch in ['+', '-', '*', '/', '^']) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack(fC);
              case ch of
                '+' :
                  fC := fC + #$DE#$C1;
                '-' :
                  fC := fC + #$DE#$E9;
                '*' :
                  fC := fC + #$DE#$C9;
                '/' :
                  begin
                    LeaveRoomInFPUStack(1, fC);
                    fC := fC + c_div_S;
                  end;
                '^' :
                  begin
                    LeaveRoomInFPUStack(1, fC);
                    fC := fC + c_zero_comp +
                      #$72#$08#$DD#$D8#$DD#$D8#$D9#$EE#$EB#$42#$D9#$F1 +
                      c_2toX_S;
                  end;
              end;
              inc(emptyRegisters);
            end
      else {opCode of a function}  begin
        if (ch in ['A', 'B', 'C', 'E', 'F', 'G', 'M']) then
          LeaveRoomInFPUStack(1, fC)
        else if (ch in ['H', 'I', 'J', 'K', 'L', 'O', 'P', 'Q', 'R']) then
          LeaveRoomInFPUStack(2, fC)
        else if (ch ='S') then begin
          LeaveRoomInFPUStack(c_LastReg+1, fC);
        end;
        case ch of
           '#' :
                  fC := fC + #$D9#$E0;
                'A' :
                  fC := fC + c_mod_2pi + #$D9#$FF;
                'B' :
                  fC := fC + c_mod_2pi + #$D9#$FE;
                'C' :
                  fC := fC + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S;
                'D' :
                  fC := fC + #$D8#$C8;
                'E' :
                  fC := fC + c_zero_comp +
                    #$72#$06 +
                    #$DD#$D8#$D9#$EE#$EB#$02#$D9#$FA;
                'F' :
                  fC := fC + c_ln_S;
                'G' :
                  fC := fC + c_mod_pi + #$D9#$F2#$DD#$D8;

                'H' :
                  fC := fC + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D9#$C0#$D9#$E8#$DE +
                    #$F1#$DE#$E9+ c_div2;
                'I' :
                  fC := fC + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D9#$C0#$D9#$E8#$DE +
                    #$F1#$DE#$C1+ c_div2;
                'J' :
                  fC := fC + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D8#$C8#$D9#$E8#$D8#$E9#$D9#$C9 +
                    #$D9#$E8#$DE#$C1#$DE#$F9;
                'K' :
                  fC := fC + #$D9#$C0#$D9#$E1 +
                    c_1_comp +
                    #$DD#$D8#$72#$0E#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1#$D9#$FA#$D9#$F3 +
                    #$EB#$10 + c_zero_comp + #$DD#$D8 +
                    LoadDd(uSpecialValues[0]) +
                    #$72#$02#$D9#$E0;
                'L' :
                  fC := fC + #$D9#$C0#$D9#$E1 +
                    c_1_comp +
                    #$DD#$D8#$72#$10#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1#$D9#$FA#$D9#$C9 +
                    #$D9#$F3#$EB#$0E + c_zero_comp +
                    #$DD#$D8#$72#$04#$D9#$EB#$EB#$02 +
                    #$D9#$EE;
                'M' : fC := fC + #$D9#$E8#$D9#$F3;
                'N' : fC := fC + #$D9#$E1;
                'O' :
                  fC := fC + c_mod_pi +
                    #$D9#$FB#$D9#$C9 + c_div_S;
                'P' :
                  fC := fC + c_1_comp +
                    #$77#$18#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E9#$D9#$FA#$DE#$C1 +
                    c_ln + #$EB#$04#$DD#$D8#$D9#$EE;
                'Q' :
                  fC := fC +
                    #$D9#$C0#$D9#$E1 + c_1_comp + #$DD#$D8#$76#$24
                    +
                    #$D9#$C0#$D9#$E8#$DE#$C1#$D9#$C9#$D9#$E8#$DE#$E1#$DE#$F9 +
                    c_ln + c_div2 + #$EB#$16
                    + c_zero_comp +
                    #$DD#$D8#$77#$08 +LoadDd(uSpecialValues[2]) +
                    #$EB#$06 + LoadDd(uSpecialValues[3]);
                'R' :
                  fC := fC + c_zero_comp +
                    #$D9#$E1#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$C1
                    +
                    #$D9#$FA#$DE#$C1 + c_ln +
                      #$76#$02#$D9#$E0;
          'S': begin
              fC := fC + #$BA; {#$BA mov edx imm32}
              i:= i+2;
              ch := actorList[i];
              AddIntToCode(Integer(ord(ch)), fC);
              fC:= fC+ #$50#$B8; {#$50 push eax; #$B8 mov eax imm32}
              AddAddressToCode(@self, fC);
              inc(lstn);
              if high(n) < lstn then
                 SetLength(n, lstn+2);
              n[lstn]:= Length(fC)+1;
              fC := fC +  #$E8 + kzeroPointerStr32 + #$58;
                    {#$E8 call} {#$58 pop eax}
              emptyRegisters:= 7;
            end;
        else
          ;
        end;
      end;
    end;
  end; 
  fC := fC + #$C3;
  PutCodeInExeBlock(fC, ffBlock);
  @ff := @ffBlock[0];
  if ((lastIntegral1 >= 0) or (lastIntegral2 >= 0)) and (lstn >= 0) then
    for i:= 0 to lstn do begin
      k:= Int64(@TEvaluator_32.EvaluateBSplinesFunction)-(Integer(@ff) + n[i] + SizeOf(pointer) );
      c:= @k;
      for j := 0 to SizeOf(Pointer) - 1 do
        ffBlock[n[i]+j] := Byte(c[j]);
    end;
  prot:= PAGE_EXECUTE_READWRITE;
  VirtualProtect(ffBlock, SizeOf(ffBlock), PAGE_EXECUTE_READ, @prot);
  fc:= '';
  n:= nil;
end;


procedure TEvaluator_32.Build(theF: TFunctionParser_32);
var
  i, j, k, a, b: Integer;
begin
  argNum:= theF.argNum;
  parNum:= theF.parNum;
  lastIntegral1:= theF.lastIntegral1;
  lastIntegral2:= theF.lastIntegral2;
  SetLength(theArg,3*argNum);

  k:= High(theF.data);
  if k >= 0 then begin
    SetLength(data,k+1);
    for j:= 0 to k do
      data[j]:= theF.data[j];
  end;
  
  lengthActorList:= Length(theF.actorList);
  actorList:= Copy(theF.actorList, 1, lengthActorList);
  
  k:= High(theF.minCorner);
  if k >= 0 then begin
    SetLength(minCorner,k+1);
    SetLength(jump1,k+1);
    SetLength(invJump1,k+1);
    SetLength(jump2,k+1);
    SetLength(invJump2,k+1);
    for i:= 0 to k do begin
      minCorner[i]:= theF.minCorner[i];
      jump1[i]:= theF.jump1[i];
      invJump1[i]:= theF.invJump1[i];
      jump2[i]:= theF.jump2[i];
      invJump2[i]:= theF.invJump2[i];
    end;
  end;

  if lastIntegral1 >= 0 then begin
    if high(intList1) < lastIntegral1 then
      SetLength(intList1, lastIntegral1+1);
    for i:= 0 to lastIntegral1 do begin
      intList1[i].var1:= theF.intList1[i].var1;
      for j:= 0 to 2 do
        for k:= 0 to cNodes1 do
         intList1[i].data[j,k]:= theF.intList1[i].data[j,k];
    end;
  end;
  if lastIntegral2 >= 0 then begin
    if high(intList2) < lastIntegral2 then
      SetLength(intList2, lastIntegral2+1);
    for i:= 0 to lastIntegral2 do begin
      intList2[i].var1:= theF.intList2[i].var1;
      intList2[i].var2:= theF.intList2[i].var2;
      for j:= 0 to 2 do
        for k:= 0 to 2 do
          for a:= 0 to cNodes2 do
            for b:= 0 to cNodes2 do
              IntList2[i].data[j,k,a,b]:= theF.IntList2[i].data[j,k,a,b];
    end;
  end;
  SetUWLength;
  CodifyFunction;
  BuildFirstDerivative;
  BuildSecondDerivative;
end;

procedure TEvaluator_32.BuildFirstDerivative;
var
  i, j: Integer;
  k: Int64;
  prot: Word;
  ch : AnsiChar;
  c: PAnsiChar;
  escape : Boolean;
  n: array of Integer;
  lstn: Integer;
begin
  j:= $4000;
  if vvBlock = nil then begin
    prot:= PAGE_EXECUTE_READWRITE;
    vvBlock:= VirtualAlloc(nil, j, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(vvBlock, j, PAGE_EXECUTE_READWRITE, @prot);
  end;
  lstn:= -1;
  dC := ''; {Initializes the object code of the derivative}
  if (lastIntegral1 >= 0) or (lastIntegral2 >= 0) then
    for i:= 1 to 2*argNum do
      dC:= dC + LoadXd(i) + WriteDdP(theArg[i-1]);
      {this is where the integrals will found the values of the independent
      variables}

  emptyRegisters := c_LastReg + 1;
  i := 0;
  j:= 0;
  while i < lengthActorList do
    begin
      inc(i);
      ch := actorList[i];
      escape := (ch = kLastChar);
      if escape or (ch = 'p') or (ch = '1') then
        begin
          if escape then inc(i); {the next byte in'actorList' will
              identify the constant, parameter or variable that must be
              loaded to ST(0)}

          ch := actorList[i];
          LeaveRoomInFPUStack(1, dC); {only last register in RFPU must be clear}
          j:= j+1; {point where things happen in the array 'u'}
          if escape then
            begin
              if (ord(ch) > argNum) then
                begin  {Load a constant or parameter}
                  dC:= dC + c_Load0;
                  dC:= dC + WriteDdP(u[j]);
                  dC := dC + LoadDd(data[ord(ch) - argNum]);
                end
              else if (ord(ch)<= argNum) then
                begin {Load an independent variable}
                  dC := dC + LoadXd(ord(ch));
                  dC:= dC + LoadVd(ord(ch));
                  dC:= dC + WriteDdP(u[j]);
                end
            end
          else if (ch = 'p') then begin  {Load "pi"}
            dC:= dC + c_Load0;
            dC:= dC + WriteDdP(u[j]);
            dC := dC + c_LoadPi;
          end
          else begin; {Load 1}
            dC:= dC + c_Load0;
            dC:= dC + WriteDdP(u[j]);
            dC := dC + c_Load1
          end;
          dec(emptyRegisters); {One more register has become occupied}
        end
      else
        begin
          if (ch in ['+', '-', '*', '/', '^']) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack(dC); {nly ST(0) is occupied, but any of the
                following binary operators require two arguments. Hence, the
                missing argument has been temporally sent to the Application
                stack: reclaim it}

              if ch = '^' then
                LeaveRoomInFPUStack(3, dC)
              else if ch in ['*','/'] then
                LeaveRoomInFPUStack(2, dC)
              else
                LeaveRoomInFPUStack(1, dC);
              case ch of
                '+' : begin
                    dC:= dC + #$DE#$C1;
                    dC:= dC + LoadDd(u[j-1]);
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C1;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
                '-' : begin
                    dC:= dC + #$DE#$E9;
                    dC:= dC + LoadDd(u[j-1]);
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$E9;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
                '*' : begin
                    dC:= dC + #$D9#$C0#$D8#$CA#$D9#$CA;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + LoadDd(u[j-1]);
                    dC:= dC + #$DE#$CA#$DE#$C1;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
                '/' :
                  begin
                    dC:= dC + c_zero_comp;
                    dC:= dC + #$75#$14#$DD#$D8;
                    dC:= dC + LoadDd(uSpecialValues[2]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + c_Load0;
                    dC:= dC + WriteDdP(u[j-1]);
                    dC:= dC + #$EB#$28;
                    dC:= dC + c_Load1;
                    dC:= dC + #$DE#$F1#$D9#$C0#$D8#$CA#$D9#$CA;
                    dC:= dC + LoadDd(u[j-1]);
	                  dC:= dC + #$D8#$CA#$D9#$CA#$D8#$C8#$DE#$C9;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9;
	                  dC:= dC + #$DE#$E9;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
                '^' :
                  begin
                    dC:= dC + c_zero_comp;
                    dC:= dC + #$72#$0A#$DD#$D8#$DD#$D8;
                    dC:= dC + c_Load0 + c_Load0;
                    dC:= dC + #$EB#$70#$D9#$C0#$D8#$FA;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9#$D9#$CA#$D9#$C9;
                    dC:= dC + c_Load1;
	                  dC:= dC + #$D9#$C9#$D9#$F1#$D9#$C0#$D9#$CA#$DE#$C9;
                    dC:= dC + c_2toX_S + #$D9#$C9;
                    dC:= dC + LoadDd(u[j-1]);
	                  dC:= dC + #$DE#$C9#$D9#$ED#$DE#$C9#$DE#$C2#$DC#$C9#$D9#$C9;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
              end;
              inc(emptyRegisters);
              j:= j-1;
            end
          else {opCode of a function}
            begin  {Unary operators}
              if ch='#' then
                LeaveRoomInFPUStack(1, dC)
              else if ch= 'S' then
                LeaveRoomInFPUStack(c_LastReg+1,dC)
              else
                LeaveRoomInFPUStack(2, dC);
              case ch of
                '#' : begin
                    dC:= dC + #$D9#$E0;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$D9#$E0;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'A' : begin
                    dC:= dC + c_mod_2pi;
                    dC:= dC + #$D9#$FB#$D9#$C9#$D9#$E0;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'B' : begin
                    dC:= dC + c_mod_2pi; 
                    dC:= dC + #$D9#$FB;
	                  dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'C' : begin
                    dC:= dC + #$D9#$EA#$DE#$C9;
                    dC:= dC + c_2ToX_S;
	                  dC:= dC + #$D9#$C0;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'D' : begin
                    dC:= dC + #$D9#$C0#$D8#$C8#$D9#$C9#$D8#$C0;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'E' : begin
                    dC:= dC +c_zero_comp + #$72#$08;
	                  dC:= dC + #$DD#$D8 + c_Load0;
                    dC:= dC + c_Load0;
	                  dC:= dC + #$EB#$0E#$D9#$FA#$D9#$C0#$D8#$C0;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$F1;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'F' : begin
                    dC:= dC +c_zero_comp + #$72#$0E;
                    dC:= dC + #$DD#$D8;
	                  dC:= dC + LoadDd(uSpecialValues[6]);
                    dC:= dC + #$D9#$E0;
	                  dC:= dC + c_Load0;
                    dC:= dC + #$EB#$16#$D9#$C0;
                    dC:= dC + c_ln;
                    dC:= dC + #$D9#$C9;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$F1;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'G' : begin
                    dC:= dC + c_mod_pi + #$D9#$F2#$D9#$C1#$D8#$C8#$DE#$C1;
                    dC:= dC + LoadDd(u[j]) + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'H' : begin
                    dC:= dC + #$D9#$EA#$DE#$C9;
	                  dC:= dC + c_2ToX_S;
	                  dC:= dC + c_Load1;
	                  dC:= dC + #$D9#$E0#$D9#$C1#$D8#$F9#$D9#$FD#$D9#$CA#$D9#$FD;
                    dC:= dC + #$DD#$D9#$D9#$C0#$D8#$C2#$D9#$CA#$DE#$E9;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'I' : begin
                    dC:= dC + #$D9#$EA#$DE#$C9;
	                  dC:= dC + c_2ToX_S;
	                  dC:= dC + c_Load1;
                    dC:= dC + #$D9#$E0#$D9#$C1#$D8#$F9#$D9#$FD#$D9#$CA#$D9#$FD;
                    dC:= dC + #$DD#$D9#$D9#$C0#$D8#$C2#$D9#$C9#$DE#$E2;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'J' : begin
                    dC:= dC + #$D9#$EA#$DE#$C9;
                    dC:= dC + c_2ToX_S;
                    dC:= dC + #$D8#$C8;
                    dC:= dC + c_Load1;
                    dC:= dC + #$D8#$E9#$D9#$C9;
                    dC:= dC + c_Load1;
                    dC:= dC + #$DE#$C1#$DE#$F9#$D9#$C0#$D8#$C8;
                    dC:= dC + c_Load1;
                    dC:= dC + #$DE#$E1;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'K' : begin
                    dC:= dC + #$D9#$C0#$D9#$E1 + c_1_comp;
                    dC:= dC + #$DD#$D8#$76#$18#$D9#$C0#$D8#$C8;
                    dC:= dC + c_Load1 + #$DE#$E1#$D9#$FA;
                    dC:= dC + c_Load1 + #$D8#$F1#$D9#$CA#$D9#$C9#$D9#$F3#$D9#$C9;
                    dC:= dC + #$EB#$16;
                    dC:= dC + c_zero_comp;
                    dC:= dC + #$DD#$D8 + c_Load1 + #$D9#$E0 + c_LoadPi + #$D9#$FD;
                    dC:= dC + #$DD#$D9#$72#$02#$D9#$E0 + c_Load0;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'L' : begin
                    dC:= dC + #$D9#$C0#$D9#$E1;
                    dC:= dC + c_1_comp;
                    dC:= dC + #$DD#$D8#$76#$18#$D9#$C0#$D8#$C8;
                    dC:= dC + c_Load1 + #$DE#$E1#$D9#$FA;
                    dC:= dC + c_Load1 + #$D8#$F1#$D9#$E0#$D9#$CA#$D9#$F3#$D9#$C9#$EB#$10;
                    dC:= dC + c_zero_comp + #$DD#$D8#$72#$04 + c_LoadPi + #$EB#$02;
                    dC:= dC + c_Load0 + c_Load0 + LoadDd(u[j]) + #$DE#$C9 +
                    WriteDdP(u[j]);
                  end;
                'M' : begin
                    dC:= dC + #$D9#$C0 + c_Load1 + #$D9#$F3#$D9#$C9#$D8#$C8;
                    dC:= dC + c_Load1 + #$DC#$C1#$DE#$F1 + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9 + WriteDdP(u[j]);
                  end;
                'N' : begin
                    dC:= dC + c_zero_comp + #$76#$10#$D9#$E0 + LoadDd(u[j]);
                    dC:= dC + #$D9#$E0 + WriteDdP(u[j]) ;
                  end;
                'O' : begin
                    dC:= dC + c_mod_pi + c_zero_comp + #$74#$12#$D9#$F2#$DC#$F1;
                    dC:= dC + #$D9#$C9#$D9#$C0#$D8#$C8#$DE#$C2#$D9#$C9#$D9#$E0;
                    dC:= dC + #$EB#$06#$DD#$D8 + c_Load0 + c_Load0;
                    dC:= dC + LoadDd(u[j]) + #$DE#$C9 + WriteDdP(u[j]);
                  end;
                'P' : begin
                    dC:= dC + c_1_comp + #$73#$22#$D9#$C0#$D8#$C8 + c_Load1 + #$DE#$E9;
                    dC:= dC + #$D9#$FA#$D9#$C0#$DE#$C2#$D9#$C9 + c_ln;
                    dC:= dC + #$D9#$C9 + c_Load1 + #$DE#$F1#$EB#$06#$DD#$D8;
                    dC:= dC + c_Load0 + c_Load0 + LoadDd(u[j]) + #$DE#$C9 + WriteDdP(u[j]);
                  end;
                'Q' : begin
                    dC:= dC + #$D9#$C0#$D9#$E1 + c_1_comp + #$DD#$D8#$76#$2E;
                    dC:= dC + #$D9#$C0 + c_Load1 + #$DE#$C1#$D9#$C9 + c_Load1;
                    dC:= dC + #$DE#$E1#$D9#$C0#$D8#$CA#$D9#$CA#$DE#$F1;
                    dC:= dC + c_ln + c_Load1 + #$D9#$E0#$D9#$C9#$D9#$FD#$D9#$CA;
                    dC:= dC + #$DE#$F9#$D9#$E0#$EB#$18 + c_zero_comp;
                    dC:= dC + #$DD#$D8#$77#$08 + LoadDd(uSpecialValues[2]);
                    dC:= dC + #$EB#$06 + LoadDd(uSpecialValues[3]) + c_Load0;
                    dC:= dC;
                    dC:= dC + LoadDd(u[j]) + #$DE#$C9 + WriteDdP(u[j]);
                  end;
                'R' : begin
                    dC:= dC + #$D9#$C0#$D8#$C8 + c_Load1 + #$DE#$C1#$D9#$FA;
                    dC:= dC + #$D9#$C0#$D9#$CA#$DE#$C1 + c_ln + #$D9#$C9;
                    dC:= dC + c_Load1 + #$DE#$F1 + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9 + WriteDdP(u[j]);
                  end;
                'S' : begin {This is special, because the 'S' means that the
                      value of an integral has to be calculated and pushed into
                      the RFPU}
                    j:= j+1; {the value of the integral must be pushed, it not
                              substitutes the previous value of ST(0)}

                    {the four lines that follow will be interpreted by the CPU as:
                     load in variable 'currentJ' the value of  'j'}
                    dC:= dC+ #$50#$B8; {#$50 push eax; #$B8 mov eax imm32}
                    AddIntToCode(j, dC);
                    dC:= dC+ #$89#$05     {MOV r/m32,r32, mov m32 eax}
                           + sn(@currentJ);
                    dC:= dC + #$58; {pop eax}

                    i:= i+2; {the character 'S' will be followed by kLastChar and
                      the position, as a byte, of that integral in the list
                      of integrals that are involved in the function.
                      In the following call to the function
                      EvaluateBSplinesFunctionAndFirstDer, this value will be
                      read as the value of the first argument of that function,
                      that is as the index of the integral in the list of
                      integrals, as desired}
                    ch := actorList[i]; {Hence, in 'ch' is that byte. Thus the
                    next two lines instruct the CPU to store that number in the
                    general register edx}
                    dC := dC + #$BA; {#$BA mov edx imm32}
                    AddIntToCode(Integer(ord(ch)), dC);

                    {The following three lines will be interpreted as Load in the
                     register eax the address of the present instance of
                     TOCFunctDDDoubleSafeIntegrals.}
                    dC:= dC+ #$50#$B8; {#$50 push eax; #$B8 mov eax imm32}

                    AddAddressToCode(@self, dC);

                    {The following line will be interpreted as Call the function
                     code that starts at $00000000. At the end of this long "while"
                     that number will be changed into the jump that the CPU must
                     do to reach the start of the function,
                     TOCFunctDDDoubleSafeIntegrals.EvaluateBSplinesFunctionAndFirstDer}
                    inc(lstn);
                    if high(n) < lstn then
                      SetLength(n, lstn+2);
                    n[lstn]:= Length(dC)+1;
                    dC := dC + #$E8 + kzeroPointerStr32 + #$58;
                              {#$E8 call} {#$58 pop eax}
                    emptyRegisters:= c_LastReg;
                  end;
              else
                 ;
              end;
            end;
        end;
    end; 
  dC := dC + #$C3;
  PutCodeInExeBlock(dC, vvBlock);
  @VV := @vvBlock[0];
  if ((lastIntegral1 >= 0) or (lastIntegral2 >= 0)) and (lstn >= 0) then
    for i:= 0 to lstn do begin
      k:= Int64(@TEvaluator_32.EvaluateBSplinesFunctionAndFirstDer)-(Integer(@VV) + n[i] + SizeOf(pointer) );
      c:= @k;
      for j := 0 to SizeOf(Pointer) - 1 do
        vvBlock[n[i]+j] := Byte(c[j]);
    end;
  prot:= PAGE_EXECUTE_READWRITE;
  VirtualProtect(vvBlock, SizeOf(vvBlock), PAGE_EXECUTE_READ, @prot);
  dC:= '';
  n:= nil;
end;


procedure TEvaluator_32.BuildSecondDerivative;
var
  i, j : Integer;
  k: Int64;
  prot: Word;
  ch : AnsiChar;
  c: PAnsiChar;
  escape : Boolean;
  n: array of Integer;
  lstn: Integer;
begin
  j:= $8000;
  if aaBlock = nil then begin
    prot:= PAGE_EXECUTE_READWRITE;
    aaBlock:= VirtualAlloc(nil, j, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(aaBlock, j, PAGE_EXECUTE_READWRITE, @prot);
  end;
  lstn:= -1;
  ddC := '';
  if (lastIntegral1 >= 0) or (lastIntegral2 >= 0) then
    for i:= 1 to 3*argNum do
      ddC:= ddC + LoadXd(i) + WriteDdP(theArg[i-1]);
  emptyRegisters := c_LastReg + 1;
  i := 0;
  j:= 0;
  while i < lengthActorList do begin
    inc(i);
    ch := actorList[i];
    escape := (ch = kLastChar);
    if escape or (ch = 'p') or (ch = '1') then begin
      if escape then inc(i);
      ch := actorList[i];
      LeaveRoomInFPUStack(1, ddC);
      j:= j+1;
      if escape then begin
        if (ord(ch) > argNum) then
          ddC := ddC + 
              c_Load0 +  //PUSH  0 in ST(0)
              WriteDdP(u[j]) + // save ST(0) in u[j], and pop
              c_Load0 +  //PUSH  0 in ST(0)
              WriteDdP(w[j]) +  // save ST(0) in u[j], and pop
              LoadDd(data[ord(ch) - argNum])
        else if (ord(ch)<= argNum) then
          ddC := ddC  +
              LoadVd(ord(ch))+
              WriteDdP(u[j]) +
              LoadAd(ord(ch)) +
              WriteDdP(w[j]) +
              LoadXd(ord(ch))
      end
      else if (ch = 'p') then
        ddC := ddC + 
            c_Load0 +  //PUSH  0 in ST(0)
            WriteDdP(u[j]) + // save ST(0) in u[j], and pop
            c_Load0 +  //PUSH  0 in ST(0)
            WriteDdP(w[j])+  // save ST(0) in u[j], and pop
            c_LoadPi 
      else
        ddC := ddC + 
            c_Load0 +  //PUSH  0 in ST(0)
            WriteDdP(u[j]) + // save ST(0) in u[j], and pop
            c_Load0 +  //PUSH  0 in ST(0)
            WriteDdP(w[j]) + // save ST(0) in u[j], and pop
            c_Load1;
      dec(emptyRegisters);
    end
    else begin {opcode of a binary operator}
      if (ch in ['+', '-', '*', '/', '^']) then begin
        if (emptyRegisters = c_LastReg) then
          PopFromAppStack(ddC);
        if ch in ['+', '-']  then
          LeaveRoomInFPUStack(1, ddC)
        else
          LeaveRoomInFPUStack(3, ddC);
        case ch of
          '+': begin
             ddC:= ddC + #$DE#$C1 + // FADDP r_j + r_{j-1} in ST(0)
                LoadDd(u[j]) +
                LoadDd(u[j-1]) +
                #$DE#$C1 + // FADDP. v_j + v_{j-1} in ST(0), r_j + r_{j-1} in ST(1)
                WriteDdP(u[j-1]) +
                LoadDd(w[j]) +
                LoadDd(w[j-1]) +
                #$DE#$C1 + // FADDP. a_j + a_{j-1} in ST(0), r_j + r_{j-1} in ST(1)
                WriteDdP(w[j-1]);
               end;
          '-': begin
             ddC:= ddC + #$DE#$E9 + // FSUBP r_{j-1}-r_j in ST(0)
                LoadDd(u[j-1]) +
                LoadDd(u[j]) + // v_{j-1} in ST(0), v_j in ST(1), r_{j-1}-r_j in ST(2)
                #$DE#$E9 + // FSUBP. v_{j-1} - v_j in ST(0), r_{j-1}-r_j in ST(1)
                WriteDdP(u[j-1]) +
                LoadDd(w[j-1]) +
                LoadDd(w[j]) + // a_{j-1} in ST(0), a_j in ST(1), r_{j-1}-r_j in ST(2)
                #$DE#$E9 + // FSUBP. a_{j-1} - a_j in ST(0), r_{j-1}-r_j in ST(1)
                WriteDdP(w[j-1]);
               end;
          '*': begin
             ddC:= ddC + LoadDd(u[j])+ //v_j in ST(0),  r_j  in ST(1), r_{j-1} in ST(2)
                LoadDd(u[j-1]) + //u_{j-1}in ST(0), v_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                #$D9#$C0 + // FLD ST(0). u_{j-1}in ST(0) and ST(1), v_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D8#$CA + // FMUL ST(0) ST(2). u_{j_1}v_j in ST(0), v_{j-1} in ST(1), v_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D8#$C0 + // FADD ST(0) ST(0). 2u_{j_1}v_j in ST(0), v_{j-1} in ST(1), v_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D9#$CA + // FXCH ST(2).v_j in ST(0), v_{j-1} in ST(1), 2u_{j_1}v_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D8#$CC + // FMUL ST(0) ST(4). r_{j-1}v_j in ST(0), v_{j-1} in ST(1), 2u_{j_1}v_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D9#$C9 + // FXCH. v_{j-1} in ST(1), r_{j-1}v_j in ST(1),  2u_{j_1}v_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D8#$CB + // FMUL ST(0) ST(3). r_ju_{j-1} in ST(0), r_{j-1}v_j in ST(1),  2u_{j_1}v_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$DE#$C1 + // FADDP. r_ju_{j-1} +r_{j-1}v_j in ST(0), 2u_{j_1}v_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                WriteDdP(u[j-1])+ // 2u_{j_1}v_j in ST(0),  r_j  in ST(1), r_{j-1} in ST(2)
                LoadDd(w[j]) + // a_j in ST(0), 2u_{j_1}v_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                #$D8#$CB + // FMUL ST(0) ST(3).  a_j r_{j-1} in ST(0), 2u_{j_1}v_j in ST(1), r_j  in ST(2), r_{j-1} in ST(3)
                #$DE#$C1 + // FADDP. a_j r_{j-1} + 2u_{j_1}v_j in ST(0),  r_j  in ST(1), r_{j-1} in ST(2)
                LoadDd(w[j-1]) + // a_{j-1} in ST(0), a_j r_{j-1} + 2u_{j_1}v_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                #$D8#$CA + // FMUL ST(0) ST(2).  a_{j-1} r_j in ST(0), a_j r_{j-1} + 2u_{j_1}v_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                #$DE#$C1 + // FADDP. a_j r_{j-1} + 2u_{j_1}v_j +a_{j-1} r_j in ST(0),  r_j  in ST(1), r_{j-1} in ST(2)
                WriteDdP(w[j-1])+ // FADDP. r_j in ST(0), r_{j-1} in ST(1)
                #$DE#$C9; // FMULP. r_j  r_{j-1} in ST(0)
               end;
          '/': begin
             ddC:= ddC + c_zero_comp +
                #$75#$1C +  //JNE, if r_j <> 0 jumps 28 bytes
                #$DD#$D8 + // POP ST(0). r_{j-1} (numerador) in ST(0)
                LoadDd(uSpecialValues[2]) +  //eMaxNum in ST(0), r_{j-1} in ST(1)
                #$DE#$C9 + //FMULP. r_{j-1*eMaxNum  in ST(0)
                c_Load0 +
                WriteDdP(u[j-1])+
                c_Load0 +
                WriteDdP(w[j-1])+
                #$EB#$48+ //JMP. jumps 72 bytes
                c_Load1 + //1 in ST(0), r_j in ST(1), r_{j-1} in ST(2)
                #$DE#$F1 + //FDIVRP.1/r_j in ST(0), r_{j-1} in ST(1)
                #$DC#$C9 + //FMUL ST(1) ST(0). 1/ r_j in ST(0), r_{j-1}/r_j in ST(1)
                LoadDd(u[j-1])+
                #$D8#$C9 + //FMUL ST(0) ST(1) v_{j-1}/r_j in ST(0), 1/r_j in ST(1), 
                           //r_{j-1}/r_j in ST(2)
                LoadDd(u[j])+
                #$D8#$CA + //FMUL ST(0) ST(2) v_j/r_j in ST(0), v_{j-1}/r_j in 
                           //ST(1), 1/r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D9#$C0 + //FLD ST(0). v_j/r_j in ST(0) and ST(1), v_{j-1}/r_j in 
                           //ST(2), 1/ r_j in ST(3), r_{j-1}/r_j in ST(4)
                #$D8#$CC + //FMUL ST(0) ST(4). u_jr_{j-1}/r_j^2 in ST(0), v_j/r_j
                           //ST(1), v_{j-1}/r_j in ST(2), 1/ r_j in ST(3), r_{j-1}/r_j in ST(4)
                #$D8#$EA + //FSUBR ST(0) ST(2). v_{j-1}/r_j - u_jr_{j-1}/r_j^2 in
                           //ST(0), v_j/r_j  ST(1), v_{j-1}/r_j in ST(2), 1/ r_j in ST(3), r_{j-1}/r_j in ST(4)
                WriteDdP(u[j-1])+ //v_j/r_j  ST(0), v_{j-1}/r_j in ST(1), 1/ r_j in
                           //ST(2), r_{j-1}/r_j in ST(3)
                #$DC#$C9 + //FMUL ST(1) ST(0). v_j/r_j  in ST(0), u_{j-1}v_j/r_j ^2
                           //in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D8#$C8 + //FMUL ST(0) ST(0). v_j^2/r_j^2  in ST(0), u_{j-1}v_j/r_j
                           //^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D8#$CB + //FMUL ST(0) ST(3). v_j^2r_{j-1}/r_j^3 in ST(0),
                           //u_{j-1}v_j/r_j ^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$DE#$E1 + //FSUBRP. v_j^2r_{j-1}/r_j^3 -u_{j-1}v_j/r_j ^2 in
                           //ST(0), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D8#$C0 + //FADD ST(0) ST(0). 2u_j^2r_{j-1}/r_j^3 -2u_{j-1}v_j/r_j
                           //^2 in  ST(0), 1/ r_j in ST(1), r_{j-1}/r_j in ST(2)
                LoadDd(w[j-1])+
                #$D8#$CA + //FMUL ST(0) ST(2). a_{j-1}/r_j in ST(0),
                           //2u_j^2r_{j-1}/r_j^3 -2u_{j-1}v_j/r_j ^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$DE#$C1 + //FADDP. a_{j-1}/r_j +2u_j^2r_{j-1}/r_j^3
                           //-2u_{j-1}v_j/r_j ^2 in  ST(0), 1/ r_j in ST(1), r_{j-1}/r_j in ST(2)
                LoadDd(w[j])+ //a_j in ST(0), a_{j-1}/r_j +2u_j^2r_{j-1}/r_j^3
                              //-2u_{j-1}v_j/r_j ^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D8#$CB + //FMUL ST(0) ST(3). a_j r_{j-1}/r_j in ST(0), a_{j-1}/r_j
                           //+2u_j^2r_{j-1}/r_j^3 -2u_{j-1}v_j/r_j ^2 in ST(1),   1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$DE#$CA + //FMULP ST(2) ST(0). a_{j-1}/r_j +2u_j^2r_{j-1}/r_j^3
                           //-2u_{j-1}v_j/r_j ^2 in  ST(0), a_j r_{j-1}/r_j^2 in ST(1), r_{j-1}/r_j in ST(2)
                #$DE#$E1 + //FSUBRP. a_{j-1}/r_j-a_j r_{j-1}/r_j^2
                           //+2u_j^2r_{j-1}/r_j^3 -2u_{j-1}v_j/r_j ^2 in ST(0), r_{j-1}/r_j in ST(1)
                WriteDdP(w[j-1]);
               end;
          else
            ;
        end;
        inc(emptyRegisters);
        j:= j-1;
      end
      else begin {opCode of a function}
        if ch in ['#','N'] then
          LeaveRoomInFPUStack(1, ddC)
        else if ch= 'C' then
          LeaveRoomInFPUStack(2, ddC)
        else if ch= 'S' then
          LeaveRoomInFPUStack(c_LastReg+1,ddC)
        else
          LeaveRoomInFPUStack(3, ddC);
        case ch of
          '#': begin
                 ddC:= ddC + #$D9#$E0 + //FCHS, - r_j in  ST(0)
                   LoadDd(u[j]) + // u[j] in ST(0), -r_j in ST(1)
                   #$D9#$E0 + //FCHS,  -u[j] in ST(0), -r_j in ST(1)
                   WriteDdP(u[j]) +
                   LoadDd(w[j]) + // a_j in ST(0), -r_j in ST(1)
                   #$D9#$E0 + //FCHS,  -a_j in ST(0), -r_j in ST(1)
                   WriteDdP(w[j]);
               end;
          'A': begin
                 ddC:= ddC + c_mod_2pi +
                   #$D9#$FB + //FSINCOS. cos(r_j) in ST(0), sen(r_j) in ST(1)
                   #$D9#$C9 + //FXCH ,intercambiar ST(0) and ST(1).
                   #$D9#$E0 + //FCHS. -sen(r_j) in ST(0), cos(r_j) in ST(1)
                   LoadDd(u[j]) + // v_j in ST(0), -sen(r_j) in ST(1), cos(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0).v_j in ST(0) and ST(1), -sen(r_j) in ST(2), cos(r_j) in ST(3)
                   #$D8#$C8 + // FMUL ST(0) ST(0).  v_j^2 in ST(0), v_j in ST(1), -sen(r_j) in ST(2), cos(r_j) in ST(3)
                   #$D8#$CB + // FMUL ST(0) ST(3).  v_j^2 cos(r_j) in ST(0), v_j in ST(1), -sen(r_j) in ST(2), cos(r_j) in ST(3)
                   #$D9#$C9 + //FXCH. v_j in ST(0), v_j^2 cos(r_j) in ST(1), -sen(r_j) in ST(2), cos(r_j) in ST(3)
                   #$D8#$CA + // FMUL ST(0) ST(2). -u_jsen(r_j) in ST(0), v_j^2 cos(r_j) in ST(1), -sen(r_j) in ST(2), cos(r_j) in ST(3)
                   WriteDdP(u[j])+ //v_j^2 cos(r_j) in ST(0), -sen(r_j) in ST(1), cos(r_j) in ST(2)
                   LoadDd(w[j]) + //a_j in ST(0), v_j^2 cos(r_j) in ST(1), -sen(r_j) in ST(2), cos(r_j) in ST(3)
                   #$DE#$CA + //FMULP ST(2) ST(0). v_j^2 cos(r_j) in ST(0), -a_j sen(r_j) in ST(1), cos(r_j) in ST(2)
                   #$DE#$E9 + //FSUBP ST(1) ST(0).  -v_j^2 cos(r_j) -a_j sen(r_j) in ST(0), cos(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'B': begin
                 ddC:= ddC + c_mod_2pi +
                   #$D9#$FB + //FSINCOS. cos(r_j) in ST(0), sen(r_j) in ST(1)
                   LoadDd(u[j]) + // v_j in ST(0), cos(r_j) in ST(1), sen(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0).v_j in ST(0) and ST(1), cos(r_j) in ST(2), sen(r_j) in ST(3)
                   #$D8#$C8 + // FMUL ST(0) ST(0).  v_j^2 in ST(0), v_j in ST(1), cos(r_j) in ST(2), sen(r_j) in ST(3)
                   #$D8#$CB + // FMUL ST(0) ST(3).  v_j^2 sen(r_j) in ST(0), v_j in ST(1), cos(r_j) in ST(2), sen(r_j) in ST(3)
                   #$D9#$C9 + //FXCH. v_j in ST(0), v_j^2 sen(r_j) in ST(1), cos(r_j) in ST(2), sen(r_j) in ST(3)
                   #$D8#$CA + // FMUL ST(0) ST(2). u_jcos(r_j) in ST(0), v_j^2 sen(r_j) in ST(1), cos(r_j) in ST(2), sen(r_j) in ST(3)
                   WriteDdP(u[j])+ //v_j^2 sen(r_j) in ST(0), cos(r_j) in ST(1), sen(r_j) in ST(2)
                   LoadDd(w[j]) + //a_j in ST(0), v_j^2 sen(r_j) in ST(1), cos(r_j) in ST(2), sen(r_j) in ST(3)
                   #$DE#$CA + //FMULP ST(2) ST(0). v_j^2 sen(r_j) in ST(0), a_j cos(r_j) in ST(1), sen(r_j) in ST(2)
                   #$DE#$E9 + //FSUBP ST(1) ST(0).  a_j sen(r_j) -v_j^2 cos(r_j) in ST(0), senj) in ST(1)
                   WriteDdP(w[j]);
               end;
          'C': begin
                 ddC:= ddC + #$D9#$EA + //FLDL2E, Cargar log in base 2 de e
                   #$DE#$C9 + //FMULP. r_j*log_2e  in ST(0)
                   c_2ToX_S + //exp(r_j) in ST(0)
                   LoadDd(u[j])+ //v_j in ST(0), exp(r_j) in ST(1)
                   #$D9#$C0 + // FLD ST(0). v_j in ST(0) and ST(1), exp(r_j) in ST(2)
                   #$D8#$CA + // FMUL ST(0) ST(2).  v_j exp(r_j) in ST(0), v_j in ST(1), exp(r_j) in ST(2)
                   WriteDdP(u[j]) + //  v_j in ST(0), exp(r_j) in ST(1)
                   #$D8#$C8 + // FMUL ST(0) ST(0). v_j^2 in ST(0), exp(r_j) in ST(1)
                   LoadDd(w[j]) + //a_j in ST(0), v_j^2 in ST(1), exp(r_j) in ST(2)
                   #$DE#$C1 + // ADDP. a_j +v_j^2 in ST(0), exp(r_j) in ST(1)
                   #$D8#$C9 + //FMULP. ( a_j +v_j^2 )exp(r_j) in ST(0), exp(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'D': begin
                 ddC:= ddC + #$D9#$C0 +  //FLD ST(0). r_j in ST(0)  and  in  ST(1)
                   #$D8#$C8 + //FMUL ST(0), ST(0). r_j^2 in ST(0), r_j in ST(1)
                   #$D9#$C9 + //FXCH. r_j in ST(0), r_j ^2en ST(1)
                   #$D8#$C0 + //ADD ST(0) ST(0). 2r_j in ST(0), r_j^2 in ST(1)
                   LoadDd(w[j])+ //a_j in ST(0), 2r_j in ST(1), r_j ^2en ST(2)
                   #$D8#$C9 + // FMUL ST(0) ST(1). 2w_jr_j in ST(0), 2r_j in ST(1), r_j^2 in ST(2)
                   LoadDd(u[j])+ //v_j in ST(0), 2w_jr_j in ST(1), 2r_j in ST(2), r_j^2 in ST(3)
                   #$DC#$CA + // FMUL ST(2) ST(0) v_j in ST(0), 2w_jr_j in ST(1), 2r_ju_j in ST(2), r_j^2 in ST(3)
                   #$D9#$CA + // FXCH ST(2). 2r_ju_j in ST(0), 2w_jr_j in ST(1), v_j in ST(2), r_j^2 in ST(3)
                   WriteDdP(u[j]) + //2w_jr_j in ST(0), v_j in ST(1), r_j^2 in ST(2)
                   #$D9#$C9 + //FXCH. v_j in ST(0), 2w_jr_j in ST(1),  r_j^2 in ST(2)
                   #$D8#$C8 + // FMUL ST(0) ST(0). v_j^2 in ST(0), 2w_jr_j in ST(1),  r_j^2 in ST(2)
                   #$D8#$C0 + //ADD ST(0) ST(0). 2u_j^2 in ST(0), 2w_jr_j in ST(1),  r_j^2 in ST(2)
                   #$DE#$C1 + //ADDP. 2u_j^2 +2w_jr_j in ST(0),  r_j^2 in ST(1)
                   WriteDdP(w[j]);
               end;
          'E': begin
                 ddC:= ddC +c_zero_comp + #$72#$16 + //if ST(0)>0, jumps 22 bytes)
                   #$DD#$D8 + c_Load0 + //FSTP ST(0) and carga 0. Tira
                                       //ST(0) and sustityelo por 0
		               c_Load0 +  //0 in ST(0) and in ST(1)
                   WriteDdP(u[j]) +
		               c_Load0 +  //0 in ST(0) and in ST(1)
                   WriteDdP(w[j]) +
                   #$EB#$32 +  //jumps los 50 bytes siguientes
                   #$D9#$FA + // FSQRT. sqrt{r_j} in ST(0)
		               #$D9#$C0 + // FLD ST(0).  sqrt{r_j} in ST(0) and  ST(1)
                   #$D8#$C0 + // FADD ST(0) ST(0). 2sqrt{r_j} in ST(0) ,
                              //sqrt(r_j) in ST(1)
                   c_Load1 + // 1 in ST(0), 2sqrt{r_j} in ST(1) , sqrt(r_j) in ST(2)
                   #$DE#$F1 + // FDIVRP. 1/(2sqrt{r_j} ) in ST(0), sqrt(r_j) in ST(1)
                   #$D9#$C0 + // FLD ST(0). 1/(2sqrt{r_j} ) in ST(0) and ST(1),
                              //sqrt(r_j) in ST(2)
                   LoadDd(u[j])+ //v_j in ST(0), 1/(2sqrt{r_j} ) in ST(1) and
                                 //ST(2), sqrt(r_j) in ST(3)
                   #$DE#$C9 + //FMULP. v_j/(2sqrt{r_j} ) in ST(0) , 1/(2sqrt{r_j}) in ST(1),
                              //sqrt(r_j) in ST(2)
                   #$D9#$C0 + //FLD ST(0). v_j/(2sqrt{r_j} ) in ST(0) and ST(1),
                              //1/(2sqrt{r_j} ) in ST(2) , sqrt(r_j) in ST(3)
                   WriteDdP(u[j]) + // v_j/(2sqrt{r_j} ) in ST(0),
                              //1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   #$D8#$C8 + // FMUL ST(0) ST(0). v_j^2/(4 r_j) in ST(0),
                              //1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   #$D8#$C9 + // FMUL ST(0) ST(2). v_j^2/(8 r_jsqrt(r_j)) in
                              //ST(0),  1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   #$D8#$C0 + // FADD ST(0) ST(0).  v_j^2/(4 r_jsqrt(r_j)) in
                              //ST(0),  1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   LoadDd(w[j])+ //a_j in ST(0), v_j^2/(4 r_jsqrt(r_j) ) in
                                 //ST(1), 1/(2sqrt{r_j}) in ST(2) , sqrt(r_j) in ST(3)
  		             #$DE#$CA + //FMULP ST(2) ST(0). v_j^2/(4 r_jsqrt(r_j) ) in
                              //ST(0), a_j/(2sqrt{r_j}) in ST(1), sqrt(r_j) in ST(2)
		               #$DE#$E9 + //FSUBP.  a_j/(2sqrt{r_j})  - v_j^2/(4r_jsqrt(r_j))
                              //in ST(0), sqrt(r_j) in ST(1)
		               WriteDdP(w[j]) ;
               end;
          'F': begin
                 ddC:= ddC +c_zero_comp + #$72#$1C + //if ST(0)>0, jumps 28 bytes
                   #$DD#$D8 + // FSTP ST(0) :pop. expulsa  r_j
                   LoadDd(uSpecialValues[6])+ // carga eMaxExp in ST(0)
                   #$D9#$E0 + // FCHS, -eMaxExp in ST(0)
                   c_Load0 + // 0 in ST(0), -eMaxExp in ST(1)
                   WriteDdP(u[j])+
                   c_Load0 +  //0 in ST(0) and in ST(1)
                   WriteDdP(w[j]) +
                   #$EB#$34 +  //jumps los 52 bytes siguientes
                   #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
                   c_ln + // ln(r_j) in ST(0), r_j in ST(1)
                   #$D9#$C9 + // FXCH. r_j in ST(0), ln(r_j) in ST(1)
                   c_Load1 +  // 1 in ST(0), r_j in ST(1), ln(r_j) in ST(2)
                   #$DE#$F1 +  //FDIVRP 1/r_j in ST(0), ln(r_j) in ST(1)
	  	             LoadDd(u[j]) + //v_j in ST(0), 1/r_j in ST(1), ln(r_j) in ST(2)
                   #$D8#$C9 +  //FMUL ST(0) ST(1). v_j/r_j in ST(0), 1/r_j in
                               //ST(1), ln(r_j) in ST(2)
		               #$D9#$C0 + // FLD ST(0). v_j/r_j in ST(0) and in ST(1), 1/r_j in
                              //ST(2), ln(r_j) in ST(3)
                   WriteDdP(u[j]) + // v_j/r_j in ST(0), 1/r_j in ST(1), ln(r_j)
                              //in ST(2)
                   #$D8#$C8 + // FMUL ST(0) ST(0). v_j^2/r_j^2 in ST(0), 1/r_j in
                              //ST(1), ln(r_j) in ST(2)
                   LoadDd(w[j])+ //a_j in ST(0), v_j^2/r_j^2 in ST(1), 1/r_j in
                              //ST(2), ln(r_j) in ST(3)
		               #$DE#$CA + // FMULP ST(2) ST(0) . v_j^2/r_j^2 in ST(0),
                              //a_j/r_j in ST(1), ln(r_j) in ST(2)
		               #$DE#$E9 + //FSUBP. a_j/r_j -v_j^2/r_j^2  in ST(0), ln(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'G': begin
                 ddC:= ddC + c_mod_pi + #$D9#$F2 + // FPTAN.  1 in ST(0),
                              //tan(r_j) in ST(1)
		               #$D9#$C1 + // FLD ST(1). tan(r_j) in ST(0), 1 in ST(1),
                              //tan(r_j) in ST(2)
                   #$D8#$C8 + //FMUL ST(0), ST(0).  tan(r_j)^2 in ST(0), 1 in
                              //ST(1), tan(r_j) in ST(2)
                   #$DE#$C1 + //FADDP.  1+tan(r_j)^2 in ST(0),  tan(r_j) in ST(1)
                   LoadDd(u[j])+ // v_j in ST(0), 1+tan(r_j)^2 in ST(1),
                              //tan(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0). v_j in ST(0) and ST(1),  1+tan(r_j)^2
                              //in ST(2),  tan(r_j) in ST(3)
 		               #$D8#$C8 + //FMUL ST(0) ST(0).  v_j^2 in ST(0), v_j in ST(1),
                              //1+tan(r_j)^2 in ST(2),  tan(r_j) in ST(3)
                   #$D9#$C9 + // FXCH. v_j in ST(0), v_j^2 in ST(1), 1+tan(r_j)^2
                              //in ST(2),  tan(r_j) in ST(3)
                   #$D8#$CA + //FMUL ST(0) ST(2). v_j(1+tan^2) enST(0), v_j^2 in
                              //ST(1), 1+tan(r_j)^2 in ST(2),  tan(r_j) in ST(3)
		               WriteDdP(u[j])+ // v_j^2 in ST(0), 1+tan(r_j)^2 in ST(1),
                              //tan(r_j) in ST(2)
                   #$D8#$C0 + //FADD ST(0) ST(0) 2u_j^2 in ST(0), 1+tan(r_j)^2
                              //in ST(1),  tan(r_j) in ST(2)
                   #$D8#$CA + //FMUL ST(0) ST(2). 2u_j^2 tan(r_j)  in ST(0),
                              //1+tan(r_j)^2 in ST(1),  tan(r_j) in ST(2)
                   #$D8#$C9 + //FMUL ST(0) ST(1). 2u_j^2 tan(r_j)(1+tan(r_j)^2 )
                              //in ST(0),   1+tan(r_j)^2 in ST(1),  tan(r_j) in ST(2)
		               LoadDd(w[j])+
		               #$DE#$CA + // FMULP ST(2) ST(0). 2u_j^2 tan(r_j)(1+tan(r_j)^2)
                              //in ST(0),  a_j( 1+tan(r_j)^2) in ST(1),  tan(r_j) in ST(2)
                   #$DE#$C1 + //FADDP.  2u_j^2 tan(r_j)(1+tan(r_j)^2 ) + a_j(1+tan(r_j)^2)
                              //in ST(0),  tan(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'H': begin
                 ddC:= ddC + #$D9#$EA#$DE#$C9 + //FLDL2E ,FMULP. r_j *ln_2e in ST(0)
                   c_2ToX_S + // e^{r_j} in ST(0)
		               c_Load1 +  // 1 in ST(0),  exp (r_j) in ST(1)
		               #$D9#$E0 + // FCHS. -1 in ST(0), exp (r_j) in ST(1)
                   #$D9#$C1 + // FLF ST(1). exp(r_j) in ST(0), -1 in ST(1), exp
                            //(r_j) in ST(2)
                   #$D8#$F9 + // FDIVR ST(0) ST(1). -exp (-r_j) in ST(0), -1 in
                            //ST(1), exp (r_j) in ST(2)
                   #$D9#$FD + // FSCALE. -exp (-r_j)/2 in ST(0), -1 in ST(1), exp
                            //(r_j) in ST(2)
                   #$D9#$CA + // FXCH ST(2). exp (r_j) in ST(0), -1 in ST(1) ,
                            //-exp (-r_j)/2 in ST(2)
                   #$D9#$FD + // FSCALE. exp (r_j)/2 in ST(0), -1 in ST(1), -exp(-r_j)/2 in ST(2)
		               #$DD#$D9 + // FSTP  ST(1). exp (r_j)/2 in ST(0) , -exp(-r_j)/2 in ST(1)
		               #$D9#$C0 + // FLD ST(0).  exp (r_j)/2 in ST(0) and in ST(1), -exp (-r_j)/2 in ST(2)
                   #$D8#$C2 + // FADD ST(0) ST(2). sinh(r_j) in ST(0), exp(r_j)/2 in ST(1), -exp (-r_j)/2 in ST(2)
                   #$D9#$CA + //FXCH ST(2). -exp (-r_j)/2 in ST(0), exp (r_j)/2 in ST(1),  sinh(r_j) in ST(2)
                   #$DE#$E9 + // FSUBP ST(1) ST(0). cosh(r_j) in ST(0), sinh(r_j) in ST(1)
                   LoadDd(u[j])+ // v_j in ST(0), cosh(r_j) in ST(1), sinh(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0). v_j in ST(0) and ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
		               #$D8#$C8 +  //FMUL ST(0) ST(0).  v_j^2 in ST(0), v_j in ST(1),
                            //cosh(r_j) in ST(2), sinh(r_j) in ST(3)
  		             #$D8#$CB + //FMUL ST(0) ST(3). v_j^2 sinh(r_j) in ST(0), v_j
                            //in ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
                   #$D9#$C9 + // FXCH. v_j in ST(0), v_j^2 sinh(r_j) in ST(1),
                            //cosh(r_j) in ST(2), sinh(r_j) in ST(3)
		               #$D8#$CA + //FMUL ST(0) ST(2). v_j ch(r_j) in ST(0), v_j^2
                            //sh(r_j) in ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
                   WriteDdP(u[j])+ //v_j^2 sh(r_j) in ST(0), cosh(r_j) in ST(1),
                            //sinh(r_j) in ST(2)
                   LoadDd(w[j])+ //a_j in ST(0), v_j^2 sh(r_j) in ST(1),
                            //cosh(r_j) in ST(2), sinh(r_j) in ST(3)
                   #$DE#$CA + //FMULP ST(2) ST(0).  v_j^2 sh(r_j) in ST(0), a_j
                            //cosh(r_j) in ST(1), sinh(r_j) in ST(2)
		               #$DE#$C1 + //FADDP. v_j^2 sh(r_j) + a_j cosh(r_j) in ST(0),
                            //sinh(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'I': begin
                 ddC:= ddC + #$D9#$EA#$DE#$C9 + //FLDL2E ,FMULP. r_j *ln_2e in ST(0)
		               c_2ToX_S + // e^{r_j} in ST(0)
		               c_Load1 +  // 1 in ST(0),  exp (r_j) in ST(1)
		               #$D9#$E0 + // FCHS. -1 in ST(0), exp (r_j) in ST(1)
		               #$D9#$C1 + // FLF ST(1). exp(r_j) in ST(0), -1 in ST(1), exp
                              //(r_j) in ST(2)
                   #$D8#$F9 + // FDIVR ST(0) ST(1). -exp (-r_j) in ST(0), -1 in
                              //ST(1), exp (r_j) in ST(2)
                   #$D9#$FD + // FSCALE. -exp (-r_j)/2 in ST(0), -1 in ST(1), exp
                              //(r_j) in ST(2)
                   #$D9#$CA + // FXCH ST(2). exp (r_j) in ST(0), -1 in ST(1) ,
                              //-exp (-r_j)/2 in ST(2)
                   #$D9#$FD + // FSCALE. exp (r_j)/2 in ST(0), -1 in ST(1), -exp
                              //(-r_j)/2 in ST(2)
                   #$DD#$D9 + // FSTP  ST(1). exp (r_j)/2 in ST(0) , -exp
                              //(-r_j)/2 in ST(1)
                   #$D9#$C0 + // FLD ST(0).  exp (r_j)/2 in ST(0) and in ST(1),
                              //-exp (-r_j)/2 in ST(2)
                   #$D8#$C2 + // FADD ST(0) ST(2). sinh(r_j) in ST(0), exp
                              //(r_j)/2 in ST(1), -exp (-r_j)/2 in ST(2)
                   #$D9#$CA + //FXCH ST(2). -exp (-r_j)/2 in ST(0), exp (r_j)/2
                              //in ST(1),  sinh(r_j) in ST(2)
		               #$DE#$E9 + // FSUBP ST(1) ST(0). cosh(r_j) in ST(0), sinh(r_j)
                              //in ST(1)
		               #$D9#$C9 + // FXCH. sinh(r_j) in ST(0), cosh(r_j) in ST(1)
                   LoadDd(u[j])+ // v_j in ST(0), sinh(r_j) in ST(1), cosh(r_j) in ST(2)
		               #$D9#$C0 + // FLD ST(0). v_j in ST(0) and ST(1), sinh(r_j) in
                              //ST(2), cosh(r_j) in ST(3)
		               #$D8#$C8 + //FMUL ST(0) ST(0).  v_j^2 in ST(0), v_j in ST(1),
                              //sinh(r_j) in ST(2), cosh(r_j) in ST(3)
                   #$D8#$CB + //FMUL ST(0) ST(3). v_j^2 cosh(r_j) in ST(0), v_j
                              //in ST(1), sinh(r_j) in ST(2), cosh(r_j) in ST(3)
                   #$D9#$C9 + // FXCH. v_j in ST(0), v_j^2 cosh(r_j) in ST(1),
                              //sinh(r_j) in ST(2), cosh(r_j) in ST(3)
		               #$D8#$CA + //FMUL ST(0) ST(2). v_j sh(r_j) in ST(0), v_j^2
                              //ch(r_j) in ST(1), sinh(r_j) in ST(2), cosh(r_j) in ST(3)
		               WriteDdP(u[j])+ //v_j^2 ch(r_j) in ST(0), csinh(r_j) in ST(1),
                              //cosh(r_j) in ST(2)
                   LoadDd(w[j])+ //a_j in ST(0), v_j^2 ch(r_j) in ST(1), sh(r_j)
                              //in ST(2), ch(r_j) in ST(3)
                   #$DE#$CA + //FMULP ST(2) ST(0).  v_j^2 ch(r_j) in ST(0), a_j
                              //sh(r_j) in ST(1), ch(r_j) in ST(2)
		               #$DE#$C1 + //FADDP. v_j^2 ch(r_j) + a_j sh(r_j) in ST(0),
                              //cosh(r_j) in ST(1)
		               WriteDdP(w[j]);
                end;
          'J': begin
                 ddC:= ddC + #$D9#$EA#$DE#$C9 +  //FLDL2E ,FMULP. r_j *ln_2e in ST(0)
                   c_2ToX_S + // e^{r_j} in ST(0)
                   #$D8#$C8 + //FMUL ST(0) ST(0). e^{2r_j} in ST(0)
                   c_Load1 +  // 1 in ST(0), e^{2r_j} in ST(1)
                   #$D8#$E9 + //FSUBR ST(0) ST(1). e^{2r_j} - 1 in ST(0),
                               //e^{2r_j} in ST(1)
                   #$D9#$C9 + //FXCH. e^{2r_j} in ST(0), e^{2r_j} - 1 in ST(1)
                   c_Load1 +  //1 in ST(0), e^{2r_j} in ST(1), e^{2r_j} - 1 in ST(2)
                   #$DE#$C1 + //FADDP. e^{2r_j}+1 in ST(0), e^{2r_j} - 1 in ST(2)
                   #$DE#$F9 + // FDIVP.  tanh(r_j) in ST(0)
                   #$D9#$C0 + // FLD ST(0).  tanh(r_j) in ST(0) and in ST(1)
                   #$D8#$C8 + // FMUL ST(0) ST(0). tanh(r_j)^2 in ST(0),
                               //tanh(r_j) in ST(1)
                   c_Load1 +  // 1 in ST(0), tanh(r_j)^2 in ST(1), tanh(r_j)$in ST(2)
                   #$DE#$E1 + // FBUBRP ST(1) ST(0). 1-tanh(r_j)^2 in ST(0),
                               //tanh(r_j) in ST(1)
                   LoadDd(u[j])+ // v_j in ST(0), 1-th(r_j)^2 in ST(1),  th(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0). v_j in ST(0) and ST(1),  1-th(r_j)^2 in
                               //ST(2),  th(r_j) in ST(3)
                   #$D8#$C8 + //FMUL ST(0) ST(0).  v_j^2 in ST(0), v_j in ST(1),
                               //1-th(r_j)^2 in ST(2),  th(r_j) in ST(3)
                   #$D9#$C9 + // FXCH. v_j in ST(0), v_j^2 in ST(1), 1-th(r_j)^2
                               //in ST(2),  th(r_j) in ST(3)
                   #$D8#$CA + //FMUL ST(0) ST(2). v_j(1-th^2) enST(0), v_j^2 in
                               //ST(1), 1-th(r_j)^2 in ST(2),  th(r_j) in ST(3)
                   WriteDdP(u[j])+ // v_j^2 in ST(0), 1-th(r_j)^2 in ST(1),
                               //th(r_j) in ST(2)
                   #$D8#$C0 + //FADD ST(0) ST(0) 2u_j^2 in ST(0), 1-th(r_j)^2 in
                               //ST(1),  th(r_j) in ST(2)
                   #$D8#$CA + //FMUL ST(0) ST(2). 2u_j^2 th(r_j)  in ST(0),
                               //1-th(r_j)^2 in ST(1),  th(r_j) in ST(2)
                   #$D8#$C9 + //FMUL ST(0) ST(1). 2u_j^2 th(r_j)(1-th(r_j)^2 )
                               //in ST(0),   1-th(r_j)^2 in ST(1),  th(r_j) in ST(2)
                   LoadDd(w[j])+
                   #$DE#$CA + // FMULP ST(2) ST(0). 2u_j^2 th(r_j)(1-th(r_j)^2 )
                               //in ST(0),  a_j( 1-th(r_j)^2) in ST(1),  th(r_j) in ST(2)
                   #$DE#$E9 + //FSUBP.  -2u_j^2 th(r_j)(1-th(r_j)^2 ) + a_j(
                               //1-th(r_j)^2) in ST(0),  th(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'K': begin
                 ddC:= ddC + #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
 		               #$D9#$E1 + // FABS. |r_j| in ST(0), r_j in ST(1)
		               c_1_comp + // FABS. |r_j| in ST(0), r_j in ST(1)
                   #$77#$2A + // JBE. if |r_j| < 1, jumps 42 bytes
		               #$DD#$D8  + // FSTP ST(0). r_j in ST(0)
		               c_Load1 +  // 1 in ST(0), r_j in ST(1)
                   #$D9#$E0 + // FCHS. -1 in ST(0), r_j in ST(1)
		               c_LoadPi +  // pi in ST(0), -1 in ST(1), r_j in ST(2)
		               #$D9#$FD + // FSCALE pi/2 in ST(0), -1 in ST(1),  r_j in ST(2)
                   #$DD#$D9 + // FSTP ST(1). pi/2 in ST(0), r_j in ST(1)
		               #$D9#$C9 + //FXCH. r_j in ST(0), pi/2 in ST(1),
		               c_zero_comp +
		               #$DD#$D8 + // FSTP ST(0). pi/2 in ST(0)
                   #$72#$02 + // JB. if r_j > 0, jumps 2 bytes
		               #$D9#$E0 + // FCHS, -pi/2 in ST(0)
		               c_Load0 +  // 0 in ST(0),pi/2 o  -pi/2 in ST(0)
		               WriteDdP(u[j])+
		               c_Load0 +  // 0 in ST(0), pi/2 o  -pi/2 in ST(0)
                   WriteDdP(w[j])+
                   #$EB#$30 + //JMP. jumps 48  bytes
                   #$D8#$C8 + // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
                   c_load1 + // 1 in ST(0),   r_j^2 in ST(1), r_j in ST(2)
		               #$DE#$E1 + // FSUBRP ST(1) ST(0). 1-r_j^2 in ST(0), r_j in ST(1)
                   #$D9#$FA + // FSQRT. sqrt(1-r_j^2) in ST(0), r_j in ST(1)
                   LoadDd(u[j])+  // v_j in ST(0), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). v_j/sqrt(1-r_j^2) in ST(0),
                              //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D9#$C0 + // FLD ST(0). v_j/sqrt(1-r_j^2) in ST(0) and in
                              //ST(1), sqrt(1-r_j^2) in ST(2), r_j in ST(3)
		               WriteDdP(u[j])+
		               #$D8#$C8 + // FMUL  ST(0). v_j^2/(1-r_j^2) in ST(0),
                              //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$CA + // FMUL  ST(0) ST(2). v_j^2r_j/(1-r_j^2) in ST(0),
                              //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$F1 + // FDIV ST(0) ST(1). v_j^2r_j/(1-r_j^2) ^(3/2),
                              //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               LoadDd(w[j])+
		               #$D8#$F2 + // FDIV ST(0) ST(2). a_j/sqrt(1-r_j^2) in ST(0),
                              //v_j^2r_j/(1-r_j^2) ^(3/2) in ST(1),  sqrt(1-r_j^2) in ST(2), r_j in ST(3)
		               #$DE#$C1 + //FADDP. a_j/sqrt(1-r_j^2)  + v_j^2r_j/(1-r_j^2)) in ST(0),
                              //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               WriteDdP(w[j])+ //sqrt(1-r_j^2) in ST(0), r_j in ST(1)
		               #$D9#$F3 ; //FPATAN. arcsen(r_j) in ST(0)
               end;
          'L': begin
                 ddC:= ddC + #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
                   #$D9#$E1 + // FABS. |r_j| in ST(0), r_j in ST(1)
		               c_1_comp + // FABS. |r_j| in ST(0), r_j in ST(1)
		               #$77#$22 + // JBE. if |r_j| < 1, jumps 34 bytes
                   #$DD#$D8 	+  // FSTP ST(0).
                   c_zero_comp +
		               #$DD#$D8 	+  // FSTP ST(0).
		               #$72#$04 + // JB. if r_j > 0, jumps 4 bytes
		               c_LoadPi  +  // pi in ST(0)
		               #$EB#$02 + // JMP. jumps 2 bytes
                   c_Load0 +
		               c_Load0 +  // 0 in ST(0), pi  o  0 in ST(1)
		               WriteDdP(u[j])+
		               c_Load0 +  // 0 in ST(0), pi  o  0 in ST(1)
		               WriteDdP(w[j])+

                   #$EB#$36 + //JMP. jumps 54  bytes
		               #$D8#$C8+ // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
    		           c_load1 + // 1 in ST(0),   r_j^2 in ST(1), r_j in ST(2)
                   #$DE#$E1 + // FSUBRP ST(1) ST(0). 1-r_j^2 in ST(0), r_j in ST(1)
                   #$D9#$FA + // FSQRT. sqrt(1-r_j^2) in ST(0), r_j in ST(1)
                   LoadDd(u[j])+  // v_j in ST(0), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
 		               #$D8#$F1+ // FDIV ST(0) ST(1). v_j/sqrt(1-r_j^2) in ST(0),
                             //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D9#$C0+ // FLD ST(0). v_j/sqrt(1-r_j^2) in ST(0) and in
                             //ST(1), sqrt(1-r_j^2) in ST(2), r_j in ST(3)
		               #$D9#$E0 + // FCHS.
                   WriteDdP(u[j])+
		               #$D8#$C8+ // FMUL  ST(0). v_j^2/(1-r_j^2) in ST(0),
                              //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$CA+ // FMUL  ST(0) ST(2). v_j^2r_j/(1-r_j^2) in ST(0),
                             //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$F1+ // FDIV ST(0) ST(1). v_j^2r_j/(1-r_j^2) ^(3/2),
                             //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   LoadDd(w[j])+
                   #$D8#$F2+ // FDIV ST(0) ST(2). a_j/sqrt(1-r_j^2) in ST(0),
                             //v_j^2r_j/(1-r_j^2) ^(3/2) in ST(1),  sqrt(1-r_j^2) in ST(2), r_j in ST(3)
		               #$DE#$C1+ //FADDP. a_j/sqrt(1-r_j^2)  + v_j^2r_j/(1-r_j^2)^(3/2) in ST(0),
                             //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D9#$E0 + // FCHS.
                   WriteDdP(w[j])+ //sqrt(1-r_j^2) in ST(0), r_j in ST(1)
                   #$D9#$C9 + //FXCH. r_j in ST(0), sqrt(1-r_j^2) in ST(1)
                   #$D9#$F3 ; //FPATAN. arccos(r_j) in ST(0)
               end;
          'M': begin
                 ddC:= ddC + #$D9#$C0 + // FLD ST(0). r_j in ST(0) and in ST(1)
                   #$D8#$C8 +  // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
                   c_Load1 +  // 1 in ST(0), r_j^2 in ST(1), r_j in ST(2)
                   #$DE#$C1 + //FADDP. 1+r_j^2 in ST(0), r_j in ST(1)
		               LoadDd(u[j])+  // v_j in ST(0), 1+r_j^2 in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). v_j/(1+r_j^2) in ST(0),  1+r_j^2 in ST(1), r_j in ST(2)
                   LoadDd(w[j])+ // a_j in ST(0), v_j/(1+r_j^2) in ST(1),  1+r_j^2 in ST(2), r_j in ST(3)
                   #$DE#$F2 +  // FDIVRP. ST(2), ST(0). v_j/(1+r_j^2) in ST(0),  a_j/(1+r_j^2) in ST(1), r_j in ST(2)
                   #$D9#$C0 + // FLD ST(0). v_j/(1+r_j^2) in ST(0) and ST(1),  a_j/(1+r_j^2) in ST(2), r_j in ST(3)
                   WriteDdP(u[j])+ // v_j/(1+r_j^2) in  ST(1),  a_j/(1+r_j^2) in ST(2), r_j in ST(3)
                   #$D8#$C8 + // FMUL ST(0) ST(0). v_j^2/(1+r_j^2)^2 in  ST(0),  a_j/(1+r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$CA + // FMUL ST(0) ST(0). r_ju_j^2/(1+r_j^2)^2 in  ST(0),  a_j/(1+r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$C0 + // FADD ST(0) ST(0). 2r_ju_j^2/(1+r_j^2)^2 in  ST(0),  a_j/(1+r_j^2) in ST(1), r_j in ST(2)
		               #$DE#$E9 + // FSUBP ST(1) ST(0), a_j/(1+r_j^2) - 2r_ju_j^2/(1+r_j^2)^2 in  ST(0),   r_j in ST(1)
		               WriteDdP(w[j])+ // r_j in ST(0)
		               c_Load1 +  // 1 in ST(0), r_j in ST(1)
		               #$D9#$F3;  // atan(r_j) in ST(0)
               end;
          'N': begin
                 ddC:= ddC + c_zero_comp +
                   #$76#$1E +  // JBE. if r_j > =0, jumps 30 bytes
		               #$D9#$E0 + // FCHS -r_j in ST(0)
                   LoadDd(u[j])+ // v_j in ST(0), - r_j in ST(1)
		               #$D9#$E0  + //FCHS.  -v_j in ST(0), -r_j in ST(1)
                   WriteDdP(u[j])+
		               LoadDd(w[j])+ // a_j in ST(0), ,- r_j in ST(1)
                   #$D9#$E0 +  //FCHS.  -a_j in ST(0), -r_j in ST(1)
		               WriteDdP(w[j]);
               end;
          'O': begin
                 ddC:= ddC + c_mod_pi + c_zero_comp +
                   #$75#$1A +  //JNE, if r_j <> 0 jumps 26 bytes
                   #$DD#$D8 + // POP ST(0)
		                LoadDd(uSpecialValues[2]) +  //eMaxNum in ST(0)
                    c_Load0 +
                    WriteDdP(u[j])+
          	        c_Load0 +
		                WriteDdP(w[j])+
		                #$EB#$36+ //JMP. jumps 54 bytes

                    #$D9#$F2 + // FPTAN.  1 in ST(0), tan(r_j) in ST(1)
                    #$DC#$F1 +  //FDIV ST(1) ST(0). 1 in ST(0), cot(r_j) in ST(1)
                    #$D9#$C1 +  // FLD ST(1). cot(r_j) in ST(0), 1 in ST(1), cot(r_j) in ST(2)
                    #$D8#$C8 + //FMUL ST(0), ST(0).  cot(r_j)^2 in ST(0), 1 in ST(1), cot(r_j) in ST(2)
                    #$DE#$C1 + //FADDP.  1+cot(r_j)^2 in ST(0),  cot(r_j) in ST(1)
                    LoadDd(u[j])+ // v_j in ST(0), 1+cot(r_j)^2 in ST(1),  cot(r_j) in ST(2)
                    #$D9#$C0 + // FLD ST(0). v_j in ST(0) and ST(1),  1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                    #$D8#$C8 +  //FMUL ST(0) ST(0).  v_j^2 in ST(0), v_j in ST(1), 1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                    #$D9#$C9 + // FXCH. v_j in ST(0), v_j^2 in ST(1), 1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                    #$D8#$CA +  //FMUL ST(0) ST(2). v_j(1+cot^2) in ST(0), v_j^2 in ST(1), 1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
		                #$D9#$E0  + //FCHS. -v_j(1+cot^2) enST(0), v_j^2 in ST(1), 1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                    WriteDdP(u[j])+ // v_j^2 in ST(0), 1+cot(r_j)^2 in ST(1),  cot(r_j) in ST(2)
                    #$D8#$C0 +  //FADD ST(0) ST(0) 2u_j^2 in ST(0), 1+cot(r_j)^2 in ST(1),  cot(r_j) in ST(2)
		                #$D8#$CA +  //FMUL ST(0) ST(2). 2u_j^2 cot(r_j)  in ST(0),  1+cot(r_j)^2 in ST(1),  cot(r_j) in ST(2)
                    #$D8#$C9 +  //FMUL ST(0) ST(1). 2u_j^2 cot(r_j)(1+cot(r_j)^2) in ST(0), 1+cot(r_j)^2 in ST(1), cot(r_j) in ST(2)
		                LoadDd(w[j])+
                    #$DE#$CA + // FMULP ST(2) ST(0). 2u_j^2 cot(r_j)(1+cot(r_j)^2) in ST(0), a_j( 1+cot(r_j)^2) in ST(1),  cot(r_j) in ST(2)
                    #$DE#$E1 + //FSUBRP.  2u_j^2 cot(r_j)(1+cot(r_j)^2) - a_j( 1+cot(r_j)^2) in ST(0),  cot(r_j) in ST(1)
		                WriteDdP(w[j]);
                  end;
          'P' : begin
                  ddC:= ddC +
                    c_1_comp +
		                #$72#$16 + // JB. if r_j  > 1, jumps 22 bytes
                    #$DD#$D8 +  // FSTP ST(0).
                    c_Load0 +
		                c_Load0 + // 0 in ST(0) and in ST(1)
		                WriteDdP(u[j]) +
                    c_Load0 +  // 0 in ST(0) and in st(1)
                    WriteDdP(w[j])+
                    #$EB#$3C + //JMP. jumps 60  bytes
		                #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
                    #$D8#$C8+ // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
    		            c_load1 + // 1 in ST(0),   r_j^2 in ST(1), r_j in ST(2)
		                #$DE#$E9 + // FSUBP. r_j^2-1 in ST(0), r_j in ST(1)
		                #$D9#$FA + // FSQRT. sqrt(r_j^2-1) in ST(0), r_j in ST(1)
                    LoadDd(u[j])+  // v_j in ST(0), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
                    #$D8#$F1 + // FDIV ST(0) ST(1). v_j/sqrt(r_j^2-1) in ST(0), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
                    #$D9#$C0 + // FLD ST(0). v_j/sqrt(r_j^2-1) in ST(0) and in ST(1), sqrt(r_j^2-1) in ST(2), r_j in ST(3)
                    WriteDdP(u[j])+
                    #$D8#$C8 + // FMUL  ST(0). v_j^2/(r_j^2-1) in ST(0), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
                    #$D8#$CA + // FMUL  ST(0) ST(2). v_j^2r_j/(r_j^2-1) in ST(0), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
		                #$D8#$F1 + // FDIV ST(0) ST(1). v_j^2r_j/(r_j^2-1) ^(3/2), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
                    LoadDd(w[j])+
		                #$D8#$F2 + // FDIV ST(0) ST(2). a_j/sqrt(r_j^2-1) in ST(0), v_j^2r_j/(r_j^2-1) ^(3/2) in ST(1),  sqrt(r_j^2-1) in ST(2), r_j in ST(3)
                    #$DE#$E1 + //FSUBRP. a_j/sqrt(r_j^2-1)  - v_j^2r_j/(r_j^2-1) ^(3/2) in ST(0),  sqrt(r_j^2-1) in ST(1), r_j in ST(2)
		                WriteDdP(w[j])+ //sqrt(r_j^2-1) in ST(0), r_j in ST(1)
		                #$DE#$C1 + //FADDP. r_j+ sqrt(r_j^2-1) in ST(0)
		                c_ln;    //acosh(r_j) in ST(0)
                  end;
          'Q': begin
                 ddC:= ddC + #$D9#$C0 + // FLD ST(0). r_j in ST(0) and in ST(1)
		               #$D9#$E1 + // FABS. |r_j| in ST(0), r_j in ST(1)
		               c_1_comp + // FABS. |r_j| in ST(0), r_j in ST(1)
                   #$77#$2C + // JBE. if |r_j| < 1, jumps 44 bytes
                   #$DD#$D8 +  // FSTP ST(0). r_j in ST(0)
                   c_zero_comp +                       		             
		               #$77#$0A +  //JA, if  r_j < 0 jumps 10 bytes
                   #$DD#$D8 +  // FSTP ST(0).  
	                 LoadDd(uSpecialValues[2]) + //eMaxNum in ST(0)
 	 	               #$EB#$08 + //JMP. jumps 8 bytes
                   #$DD#$D8 +  // FSTP ST(0).
		               LoadDd(uSpecialValues[3]) + //-eMaxNum in ST(0)
		               c_Load0 +
		               WriteDdP(u[j]) +
		               c_Load0 +  
		               WriteDdP(w[j])+
  		             #$EB#$4E + //JMP. jumps 78  bytes
                   #$D8#$C8 +  // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
                   c_Load1 +  // 1 in ST(0), r_j^2 in ST(1), r_j in ST(2)
 		               #$DE#$E1 + //FSUBRP. 1-r_j^2 in ST(0), r_j in ST(1)
		               c_Load1 +  // 1 in ST(0), 1-r_j^2 in ST(1), r_j in ST(2)
		               #$DE#$F1 + //FDIVRP. 1/(1-r_j^2) in ST(0), r_j in ST(1)
                   LoadDd(u[j])+  // v_j in ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$C9 + // FMUL ST(0) ST(1). v_j/(1-r_j^2)  in ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D9#$C0 + // FLD ST(0). v_j/(1-r_j^2)  in ST(0) and ST(1), 1/(1-r_j^2) in ST(2), r_j in ST(3)
                   WriteDdP(u[j])+ // v_j/(1-r_j^2) in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$C8 + // FMUL ST(0) ST(0). v_j^2/(1-r_j^2)^2 in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$C0 + // FADD ST(0) ST(0). 2u_j^2/(1-r_j^2)^2 in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$CA + // FMUL ST(0) ST(2).  2r_ju_j^2/(1-r_j^2)^2 in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               LoadDd(w[j])+ // a_j in ST(0),  2r_ju_j^2/(1-r_j^2)^2 in  ST(1), 1/(1-r_j^2) in ST(2), r_j in ST(3)
		               #$D8#$CA + //FMUL ST(0) ST(2).  a_j/(1-r_j^2) in ST(0), 2r_ju_j^2/(1-r_j^2)^2 in  ST(1), 1/(1-r_j^2) in ST(2), r_j in ST(3)
		               #$DE#$C1 + //FADDP.  a_j/(1-r_j^2) +2r_ju_j^2/(1-r_j^2)^2 in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               WriteDdP(w[j])+ // 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D9#$C9 + // FXCH.r_j in ST(0), 1/(1-r_j^2) in ST(1).
		               c_Load1 +  // 1 in ST(0), r_j in ST(1), 1/(1-r_j^2) in ST(2).
		               #$DE#$C1 + //FADDP. 1 + r_j in ST(1), 1/(1-r_j^2) in ST(2).
		               #$D8#$C8 +  // FMUL  ST(0).(1+ r_j)^2 in ST(0), 1/(1-r_j^2) in ST(1).
                   #$DE#$C9 + //FMULP. .(1+ r_j)^2 /(1-r_j^2) in ST(0).
                   c_ln + 			//2atanh(r_j) in ST(0)
		               c_Load1 +
                   #$D9#$E0 + //FCHS. -1 in ST(0), 2atanh(r_j) in ST(1)
                   #$D9#$C9 + //FXCH. 2atanh(r_j) in ST(0), -1 in ST(1)
		               #$D9#$FD + // FSCALE atanh(r_j) in ST(0), -1 in ST(1)
                   #$DD#$D9 ; // FSTP ST(1). atanh(r_j) in ST(0)
               end;
          'R' : begin
                 ddC:= ddC +
		               #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
                   #$D8#$C8+ // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
                   c_load1 + // 1 in ST(0),   r_j^2 in ST(1), r_j in ST(2)
                   #$DE#$C1 + //FADDP. 1+r_j^2 in ST(0), r_j in ST(1)
                   #$D9#$FA + // FSQRT. sqrt(1+r_j^2 ) in ST(0), r_j in ST(1)
		               LoadDd(u[j])+  // v_j in ST(0), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). v_j/sqrt(1+r_j^2 ) in ST(0), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
		               #$D9#$C0 + // FLD ST(0). v_j/sqrt(1+r_j^2 ) in ST(0) and in ST(1), sqrt(1+r_j^2 ) in ST(2), r_j in ST(3)
                   WriteDdP(u[j])+
                   #$D8#$C8 + // FMUL  ST(0). v_j^2/(1+r_j^2 ) in ST(0), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
                   #$D8#$CA + // FMUL  ST(0) ST(2). v_j^2r_j/(1+r_j^2 ) in ST(0), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). v_j^2r_j/(1+r_j^2 ) ^(3/2), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
                   LoadDd(w[j])+
                   #$D8#$F2 + // FDIV ST(0) ST(2). a_j/sqrt(1+r_j^2 ) in ST(0), v_j^2r_j/(1+r_j^2 ) ^(3/2) in ST(1),  sqrt(1+r_j^2 ) in ST(2), r_j in ST(3)
                   #$DE#$E1 + //FSUBRP. a_j/sqrt(1+r_j^2 )  - v_j^2r_j/(1+r_j^2 ) ^(3/2) in ST(0),  sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
		               WriteDdP(w[j])+ //sqrt(1+r_j^2 ) in ST(0), r_j in ST(1)
                   #$DE#$C1 + //FADDP. r_j+ sqrt(1+r_j^2 ) in ST(0)
                   c_ln;    //asinh(r_j) in ST(0)
                 end;
          'S' : begin
                 j:= j+1;
                   ddC:= ddC+ #$50#$B8; {#$50 push eax; #$B8 mov eax imm32}
                   AddIntToCode(j, ddC);
                   ddC:= ddC+ #$89#$05     {MOV r/m32,r32, mov m32 eax}
                           + sn(@currentJ);
                    ddC:= ddC + #$58; {pop eax}

                    ddC := ddC + #$BA; {#$BA mov edx imm32}
                    i:= i+2;
                    ch := actorList[i];
                    AddIntToCode(Integer(ord(ch)), ddC);
                    ddC:= ddC+ #$50#$B8; {#$50 push eax; #$B8 mov eax imm32}
                    AddAddressToCode(@self, ddC);
                    inc(lstn);
                    if high(n) < lstn then
                      SetLength(n, lstn+2);
                    n[lstn]:= Length(ddC)+1;
                    ddC := ddC + #$E8 + kZeroPointerStr32 + #$58;
                              {#$E8 call} {#$58 pop eax}
                    emptyRegisters:= c_LastReg;
                  end;
              else
                ;
              end;
            end;
        end;
    end;

  ddC := ddC + #$C3;
  PutCodeInExeBlock(ddC, aaBlock);
  @AA:= @aaBlock[0];
  if ((lastIntegral1 >= 0) or (lastIntegral2 >= 0)) and (lstn >=0) then
    for i:= 0 to lstn do begin
      k:= Int64(@TEvaluator_32.EvaluateBSplinesFunctionFirstAndSecondDer)-(Integer(@AA) + n[i] + SizeOf(pointer) );
      c:= @k;
      for j := 0 to SizeOf(Integer) - 1 do
        aaBlock[n[i]+j] := Byte(c[j]);
    end;
  prot:= PAGE_EXECUTE_READWRITE;
  VirtualProtect(aaBlock, SizeOf(aaBlock), PAGE_EXECUTE_READ, @prot);
  ddC:= '';
  n:= nil;
end;



destructor TEvaluator_32.Destroy;
begin
  fC := '';
  dC:= '';
  ddC:= '';
  data := nil;
  intList1:= nil;
  intList2:= nil;
  theArg:= nil;
  minCorner:= nil;
  jump1:= nil;
  invJump1:= nil;
  jump2:= nil;
  invJump2:= nil;
  u:= nil;
  if ffBlock <> nil then begin
    VirtualFree(ffBlock, 0, MEM_RELEASE);
    ffBlock:= nil
  end;
  if vvBlock <> nil then begin
    VirtualFree(vvBlock, 0, MEM_RELEASE);
    vvBlock:= nil
  end;
  if aaBlock <> nil then begin
    VirtualFree(aaBlock, 0, MEM_RELEASE);
  aaBlock:= nil
  end;
  inherited Destroy;
end;


function TEvaluator_32.gradF(const x: Vector3): Vector3;
begin
  vv([x[1],x[2],x[3],1,0,0]);
  result[1]:= u[1];
  vv([x[1],x[2],x[3],0,1,0]);
  result[2]:= u[1];
  vv([x[1],x[2],x[3],0,0,1]);
  result[3]:= u[1];
end;

function TEvaluator_32.gradF(const x: Vector2): Vector2;
begin
  vv([x[1],x[2],1,0]);
  result[1]:= u[1];
  vv([x[1],x[2],0,1]);
  result[2]:= u[1];
end;

procedure InitializeCodeVars;
begin
  uSpecialValues[0] := halfPi;
  uSpecialValues[1] := twoPi;
  uSpecialValues[2] := eMaxNum;
  uSpecialValues[3] := -eMaxNum;
  uSpecialValues[4] := eMinNum;
  uSpecialValues[5] := -eMinNum;
  uSpecialValues[6] := eMaxExp;
  uSpecialValues[7] := 0.5;
  c_div_S := c_zero_comp + #$75#$0C#$DD#$D8 + LoadDd(uSpecialValues[2]) +
    #$DE#$C9#$EB#$02#$DE#$F9; {Safe division. 20 bytes. Requiere ST(c_LastReg) libre}
  c_ln_S := c_zero_comp + #$72#$0C#$DD#$D8 + LoadDd(uSpecialValues[6]) +
    #$D9#$E0#$EB#$0A + c_ln;
    {Safe ln(ST(0)). 26 bytes. Requires ST(c_LastReg) empty}
  c_2toX_S := #$D9#$C0#$D9#$E1 + LoadDd(uSpecialValues[6]) + #$DF#$F1 +
    #$DD#$D8#$72#$1A + c_2toX + #$EB#$16 + c_zero_comp + #$DD#$D8#$72#$08
    +
    LoadDd(uSpecialValues[4]) + #$EB#$06 + LoadDd(uSpecialValues[2]);
     {Safe 2^ST(0). 64 bytes. Requires ST(6) and ST(c_LastReg) empty.}
end;

PROCEDURE TEvaluator_32.LeaveRoomInFPUStack(neededPlaces : Integer; var cod: AnsiSTring);
BEGIN
  IF emptyRegisters >= neededPlaces THEN
    Exit;
  IF emptyRegisters = 0 THEN BEGIN
      cod := cod
       + #$D9#$F6 {fdecstp: ST(c_LastReg] becomes ST(0)}
       + #$83#$EC+ c_stackDisplacDoubleHStr
                                  {subtract c_stackDisplacDouble from ESP}
          +#$DD#$1C#$24;  {fstp [ESP]}
      inc(emptyRegisters);
  END;
  IF emptyRegisters < neededPlaces then
    REPEAT
      cod := cod + #$D9 + chr($C0 + c_LastReg - emptyRegisters)
          {push last occupied ST to ST(0)}
      + #$83#$EC+ c_stackDisplacDoubleHStr
                                  {subtract c_stackDisplacDouble from ESP}
      +#$DD#$1C#$24;  {fstp [ESP]}
      cod := cod + #$DD + chr($C0 + c_LastReg - emptyRegisters);
         {last occupied place contains the former value: declare it empty}
      inc(emptyRegisters);
    UNTIL emptyRegisters = neededPlaces;
END;


PROCEDURE TEvaluator_32.PopFromAppStack(var cod: AnsiSTring);
VAR
  m : byte;
BEGIN
  m := c_LastReg +1 - emptyRegisters;
  cod := cod + #$DD + chr($D8 + m) {copy ST(0) to ST(m), which
     is the first free register, and pop st(0)}

   + #$DD#$04#$24 {fld [ESP]}
   + #$83#$C4 + c_stackDisplacDoubleHStr {add c_stackDisplacDouble to ESP};
  {if using Extended type, use 'c_stackDisplacExtendedHStr' instead}

  cod := cod + #$D9 + chr($C8 + m); {exchange ST(m) with ST(0)}
  dec(emptyRegisters);
END;


procedure TEvaluator_32.PutCodeInExeBlock(const cod: AnsiString;
     var exeBlock: PByte);
var
  len: Integer;
  i: Integer;
  prot: Word;
begin
  len:= Length(cod);
  prot:= PAGE_EXECUTE_READWRITE;
  for i:= 0 to len-1 do
    exeBlock[i]:= Byte(cod[i+1]);
end;

function TEvaluator_32.SecondD(const XV: array of Double): SecondDeriv;
begin
  result.fa0:= aa(XV);
  result.dfa0:= u[1];
  result.ddfa0:= w[1];
end;

procedure TEvaluator_32.SetUWLength;
var
  i, j, maxJ : Integer;
  ch : AnsiChar;
  escape : Boolean;
begin
  i := 0;
  j:= 0;
  maxJ:=0;
  while i < lengthActorList do begin
    inc(i);
    ch := actorList[i];
    escape := (ch = kLastChar);
    if escape or (ch in ['p', '1', 'S']) then begin
      if escape then
        inc(i)
      else if ch= 'S' then
        i:= i+2;
      inc(j);
      if j > maxJ then
        maxJ:= j;
    end
    else if (ch in ['+', '-', '*', '/', '^']) then
      dec(j)
  end; {del while}
  SetLength(u, maxJ+1);
  SetLength(w, maxJ+1);
end;

{function TEvaluator_32.theF(const x: Vector3): Double;
begin
  result:= ff([x[1], x[2], x[3]]);
end; }

procedure TEvaluator_32.UpdateParameters(parameters:
          array of Double);
{Changes the value of the parameters stored in 'data', so that they
 will have the new values in the following evaluations}
  var
    h,i: Integer;
begin
  h:= high(parameters);
  if (h >-1) and (h < parNum) then for i:=0 to high(parameters) do
    data[i+1]:= parameters[i];
end;

procedure Finalize;
var
  i: Integer;
begin
  gTheF1.Free;
  gTheF2.Free;
  gTheF3.Free;
  gTheF4.Free;
  for i:= 0 to gNumOfCPU-1 do
   gTheF1Clones[i].Free;
  gTheF1Clones:= nil;
  for i:=0 to gLastImplPart do
    gImplParts[i].Free;
  gImplParts:= nil;

end;

procedure Initialize;
var
  i: Integer;
begin
  gnumOfCPU:= NumberOfProcessors;
  gTheF1:= TEvaluator_32.Create;
  gTheF2:= TEvaluator_32.Create;
  gTheF3:= TEvaluator_32.Create;
  gTheF4:= TEvaluator_32.Create;
  SetLength(gTheF1Clones, gNumOfCPU);
  for i:= 0 to gNumOfCPU-1 do
    gTheF1Clones[i]:= TEvaluator_32.Create;
  gLastImplPart:= -1;
end;

initialization
  Initialize;
  InitializeCodeVars;
finalization
  Finalize;

end.
