{-----------------------------------------------------------------------------
 Unit Name: FImplicitCurveLine
 Project Name : ImplicitCurve2D
 Author:    Angel Montesinos Amilibia
 November 2006 - February 2009
 Purpose: Computation of a curve given as F(x,y)=0
-----------------------------------------------------------------------------}
{
Copyright (C) 2006-2009 Angel Montesinos Amilibia

This unit is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This unit is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

unit FImplicitCurveLine;

interface
uses
  Graphics,
  VectorsAndUtil;
const
  cMaxPPLine = 61;
  cMaxNumFlags = 800;
  cTol = 1.0e-5;
  cMinErr = 1.0e-30;
  cMinDenom = cMinErr * cMinErr;

type
  InRange2DFunction = function(var depth: Double; p: Vector2): Boolean of
    object;
  StoreSegmentProcedure2D = procedure(prevX, currentX: Vector2; key:
    Integer) of object;

  TImplicitCurveLine = class(TObject)
    {The finality of this class is to draw a curve defined implicitly in a
    rectangle of the plane as the locus where a smooth given function 'F' of two
    real variables vanish.

    The goal is to get it very fast while giving, in most cases, a visually
    correct solution, at the pixel level. I don't have made a mathematical
    study of the algorithm but the results are very good in general.

    The algorithm works as follows. We divide the domain rectangle in a
    customizable number of small equal rectangles, 'sqr(ppLine)', say 900
    rectangles if 'ppLine= 30'. Then compute the value of 'F' at nine
    points in the interior of each rectangle and record the point 'p' where 'F'
    has smallest absolute value. From 'p' we apply the routine 'GetNearestZero'
    that seeks a nearby point 'q' where 'F' vanishes (that is, it has an absolute
    value very small). If it finds it, then taking 'q' as seed, we apply
    repeatedly the routine 'GetNextPoint' to obtain a curve arc (really a succession
    of nearby points) that satisfy 'abs(F) < verySmall'. At each successful
    call of 'GetNextPoint' we check, using the routine 'CheckCameBack', whether
    the curve has reached a point already encountered, and stop the process
    in that case.
    }
  private
    theF: R2ToR1Function;
    theFGrad: R2toR2Function;
    InRange: InRange2DFunction;
    Store: StoreSegmentProcedure2D;
    alreadyUsed: packed array[0..cMaxNumFlags, 0..cMaxNumFlags] of boolean;
    coordRatios: array[1..3] of Double;
    coordVert: array[1..2, 0..cMaxPPLine] of Double;
    minCorner, maxCorner, step2D, flagStep, halfEdges, center: Vector2;
    ppLine, flagsPLine: Integer;
    verySmallF, radius: Double;
    minErr: Double;
    function CheckCameBack(iIni, jIni, iFini, jFini: Integer): Boolean;
    procedure FollowLine(ini: vector2; u1, u2: Integer);
    function GetNearestZero(thePoint, dirSignaller: Vector2; natural:
      Boolean; var theZero: Vector2): Boolean;
    function GetNextPoint(pos: Vector2; var vel, nextPos: Vector2): Boolean;
    function GetVel(pos: Vector2; var vel: Vector2): Boolean;
    function GetZeroInSegment(starter, dir: Vector2; b, fa, fb: Double;
      var target: Vector2): Double;
    function GradientOfF(const pp: Vector2): Vector2;
    function InitialPointInRectangle(ci, cj: Integer;
      var startingPoint: Vector2): Boolean;
    procedure SetEnvironment(theF0: R2ToR1Function;
      theF0grad: R2toR2Function; InRange0: InRange2DFunction;
      Store0: StoreSegmentProcedure2D;
      minCorner0, maxCorner0: Vector2; ppLine0, flagsPLine0: Integer;
      radius0: double);
  public
    procedure ComputeImplicitCurve(theF0: R2ToR1Function; theF0grad:
      R2toR2Function; InRange0: InRange2DFunction;
      Store0: StoreSegmentProcedure2D;
      minCorner0, maxCorner0: Vector2; ppLine0, flagsPLine0: Integer;
      radius0: double);
  end;

var
  theImplicitCurveLine: TImplicitCurveLine;

implementation
uses
  Windows, Math;

const
  kIncr4Points= 0.0002;
  k2Incr4Points= 2*kIncr4Points;
  k12Incr4Points = 12*kIncr4Points;

{------------------------------------------------}
          {Routines private to this unit}
{------------------------------------------------}

procedure FreeImplicitCurveLine;
begin
  theImplicitCurveLine.Free;
  theImplicitCurveLine:= nil;
end;

procedure InitializeTImplicitCurveLine;
{Create an instance of TImplicitCurveLine. Set the
coordinate ratios, relative to the dimensions of each small
rectangle, of the nine points where we evaluate the function
in order to start looking for an initial point}
begin
  theImplicitCurveLine:= TImplicitCurveLine.Create;
  theImplicitCurveLine.coordRatios[1]:= 1 / 6;
  theImplicitCurveLine.coordRatios[2]:= 3 / 6;
  theImplicitCurveLine.coordRatios[3]:= 5 / 6;
end;

function myFloor(x: Double): Integer;
{The 'floor' function of the unit 'math' provided by Delphi
uses the current rounding. We need the rounding given by
'trunc'}
begin
  if x >= 0 then
    myFloor:= system.trunc(x)
  else
    myFloor:= system.trunc(x) - 1;
end;

{----------end of private routines----------------}
{-------------------------------------------------}

{------------------------------------------------}
         {Routines visible from outside}
{------------------------------------------------}

function TImplicitCurveLine.CheckCameBack(iIni, jIni, iFini, jFini: Integer):
  Boolean;
{We provide the domain rectangle  with a virtual set of "pxels" of a
 customizable density 'flagsPLine.' Suppose that we have found that
 a segment of endpoints 'x' and 'y' is an approsimate solution of our
 problem. Let (iIni,jIni) be the coordinates of 'x' in these "pixel"
 units, and (iFini,jFini) the same for 'y'. then 'CheckCameBack' gives
 True if the segment 'xy' passes by a point already marked as a solution.
 Also, it marks as solution the "pixels" that are nearby to the segment 'xy'}
var
  pitch: Double;
  height, k, l, increment: Integer;
begin
  result:= FALSE;
  alreadyUsed[iIni, jIni]:= TRUE; {Mark the initial point as used}
  if (iFini <> iIni) or (jFini <> jIni) then
    {the segment has more than a point} begin
    {Choose the coordinate direction, horizontal or vertical, on which the
    segment leans}
    if abs(iFini - iIni) >= abs(jFini - jIni) then begin
      if iFini > iIni then
        increment:= 1
      else
        increment:= -1;
      pitch:= (jFini - jIni) / (iFini - iIni); {tangent of the angle
      that the segment makes with the chosen coordinate direction}
      k:= 0;
      repeat
        k:= k + increment;
        l:= iIni + k;
        height:= round(jIni + k * pitch);
        if alreadyUsed[l, height] then
          result:= TRUE {The point was already marked}
        else
          alreadyUsed[l, height]:= TRUE; {Mark it otherwise}
      until (l = iFini);
    end else {the same in the other coordinate direction} begin
      if jFini > jIni then
        increment:= 1
      else
        increment:= -1;
      pitch:= (iFini - iIni) / (jFini - jIni);
      k:= 0;
      repeat
        k:= k + increment;
        l:= jIni + k;
        height:= round(iIni + k * pitch);
        if alreadyUsed[height, l] then
          result:= TRUE
        else
          alreadyUsed[height, l]:= TRUE;
      until (l = jFini);
    end
  end;
end;

procedure TImplicitCurveLine.ComputeImplicitCurve(theF0: R2ToR1Function;
  theF0grad: R2toR2Function;
  InRange0: InRange2DFunction;
  Store0: StoreSegmentProcedure2D; { verySmallF0: Double;}
  minCorner0, maxCorner0: Vector2; ppLine0, flagsPLine0: Integer;
  radius0: double);
{This is the meaning of the arguments. 'theF0' is the function that defines the
 curve. 'theF0grad' is the gradient of 'theF0' or NIL if it is not known.
 'InRange0' is a function that gives True when the point passed to it is in the
 domain or False otherwise; also, it gives the 'depth' of the point in the domain,
 so that a positive 'depth' means that the point is in and a negative 'depth',
 that it is out the domain. The absolute value of 'depth' increases as the
 point separates from the domain boundary. 'Store0' is a function that may
 store the current point of the curve, the previous one, and the key of the
 current point. The key is 1 if the point initiates a part of the curve, or
 zero otherwise. 'verySmallF0' is so that if the value 'fx' of 'theF0' at 'x' is
 such that 'abs(fx) < verySmallF0' then 'x' is taken as a point of the curve.
 'minCorner' and 'maxCorner' define the domain rectangle. The square of
 'ppLine0' is the number of rectangles on which the domain is divided on
 each coordinate direction for finding seed points. 'flagsPLine0' gives the
 per-dimension density of data saying that a point has been  marked as belonging
 to the curve. 'radius0' is the domain length that when projected on the
 screen gives a length less than a pixel.

 'minLength0' can be zero or a positive number; if it is zero,
 it is ignored; otherwise, the routine 'FollowLine' stores a new point when
 the separation from the last stored point is greater than 'minLength0.' This
 is useful when we need to spare memory or to draw the curve very fast,
 because in this manner we reduce greatly the number of points that make the
 curve to draw.}
var
  i, j, u1, u2: Integer;
  initialPointFound, inDomain: Boolean;
  pInitial: vector2;
begin
  {Store data provided by the caller}
  SetEnvironment(theF0, theF0grad, InRange0, Store0, {verySmallF0,} minCorner0,
    maxCorner0,
    ppLine0, flagsPLine0, radius0);

  {Visit all the rectangles on which we divide the domain for finding a
  seed point. if succesful, begin following the track}
  for i:= 0 to ppLine - 1 do begin
    for j:= 0 to ppLine - 1 do begin
      initialPointFound:= InitialPointInRectangle(i, j, pInitial);
      if initialPointFound then begin
        u1:= myFloor((pInitial[1] - minCorner[1]) / flagStep[1]);
        u2:= myFloor((pInitial[2] - minCorner[2]) / flagStep[2]);
        inDomain:= (u1 >= 0) and (u1 < flagsPLine) and (u2 >= 0)
          and (u2 < flagsPLine);
        if inDomain and (not alreadyUsed[u1, u2]) then
          {Follow the solution from 'pInitial' and store it}
          FollowLine(pInitial, u1, u2);
      end;
    end;
  end;
end;

procedure TImplicitCurveLine.FollowLine(ini: vector2; u1, u2: Integer);
{It is supposed that 'ini' is a TPoint in the subset where 'theF' vanishes.
 First, we apply 'GetFirstPair2D' to obtain two points in that subset that
 signal two marching directions. We apply a marching algorithm, whose essence
 is  'GetNextPoint2D', in both directions.}
const
  allowForRoundingErrors = 1.0E-18;
  maxInSameRect = 50;
var
  i, j, k, iCurrent, jCurrent, iFormer, jFormer: Integer;
  leader, prePos, curPos, vel: vector2;
  value, depth, initialDepth, formerDepth: Double;
  rectChange, keepGoing, goesOut: Boolean;
  repeated, sameRect: Integer;

begin {of  FollowLine2D}
  try
    if not GetVel(ini, leader) then begin
      Store(ini, ini, 1); {Save 'ini' as the initial point of a stream
      that consists of one point.}
      Exit;
    end;
    if Assigned(InRange) then
      InRange(initialDepth,ini)
    else
      initialDepth:= min(halfEdges[1] - abs(ini[1] - center[1]), halfEdges[2]
      - abs(ini[2] - center[2])); {"depth" of 'ini' in the domain rectangle}
    if initialDepth > 0 then {else, 'ini' would be out of domain}
      for i:= 1 to 2 do begin
        Store(ini, ini, 1); {Save 'ini' as the initial point of a stream of them.
        That is the meaning of the value 1 for the third argument}
        depth:= initialDepth;

        {we change the sense of the previous pass}
        leader[1]:= -leader[1];
        leader[2]:= -leader[2];

        prePos:= ini;
        vel:= leader;
        repeated:= 0; {Number of times on which a repeated point has been found}
        sameRect:= 0; {We don't permit to make more than 'maxInSameRect' steps
        in the same small rectangle}
        keepGoing:= GetNextPoint(prePos, vel, curPos);
        if not keepGoing then
          break;
        iCurrent:= u1;
        jCurrent:= u2;
        repeat
          goesOut:= FALSE;
          formerDepth:= depth;
          if Assigned(InRange) then
            InRange(depth, curPos)
          else
            depth:= min(halfEdges[1] - abs(curPos[1] - center[1]),
                    halfEdges[2] - abs(curPos[2] - center[2]));
          if depth < 0 then begin
            value:= (formerDepth - allowForRoundingErrors) / (formerDepth -
              depth);
            {Compute the approximate position where the curve exits the domain}
            k:= 0;
            while (depth < 0) and (k < 20) do begin
              inc(k);
              for j:= 1 to 2 do
                curPos[j]:= prePos[j] + (1-k/20)*value * (curPos[j] - prePos[j]);
              if Assigned(InRange) then
                InRange(depth,curPos)
              else
                depth:= min(halfEdges[1] - abs(curPos[1] - center[1]),
                    halfEdges[2] - abs(curPos[2] - center[2]));
            end;
            goesOut:= TRUE;
          end;
          if depth > 0 then begin
            iformer:= iCurrent;
            jFormer:= jCurrent;
            iCurrent:= myFloor((curPos[1] - minCorner[1]) / flagStep[1]);
            jCurrent:= myFloor((curPos[2] - minCorner[2]) / flagStep[2]);
            if not ((iCurrent < 0) or (jCurrent < 0) or (iCurrent > cMaxNumFlags) or
                (jCurrent > cMaxNumFlags)) then begin
              Store(prePos, curPos, 0);
              rectChange:= abs(iCurrent - iFormer) + abs(jCurrent - jFormer) > 1;
              if rectChange then begin
                if CheckCameBack(iFormer, jFormer, iCurrent, jCurrent) then
                  repeated:= repeated + 1 {One more repeated point has been found}
                else
                  repeated:= 0;
                keepGoing:= (repeated < 4) and (not goesOut);
                sameRect:= 0;
              end
              else begin
                inc(sameRect);
                keepGoing:= (sameRect < maxInSameRect);
              end;
              if not keepGoing then
                break;
              prePos:= curPos;
              keepGoing:= GetNextPoint(prePos, vel, curPos);
            end
          else
            goesOut:= True;
          end;
        until (not keepGoing) or (repeated > 3) or goesOut;
        alreadyUsed[u1, u2]:= FALSE;
      end;
    alreadyUsed[u1, u2]:= True;
  except
    alreadyUsed[u1, u2]:= True;
  end;
end;

function TImplicitCurveLine.GetZeroInSegment(starter, dir: Vector2;
  b, fa, fb: Double; var target: Vector2): Double;
{We have 0 < b, fa= theF(starter), fb= theF(starter + b*dir), and fa*fb < 0.
 Then, we get a zero of theF in a point 'starter + t*dir' with a < t < b.}
var
  f1, f2, t1, t2, t: Double;
  i, count: Integer;
begin
  t1:= 0;
  t2:= b;
  f1:= fa;
  f2:= fb;
  target:= starter;
  count:= 0;
  while (f1 * f2 < 0) and (count < 100) do begin
    inc(count);
    t:= (t1 + t2) / 2;
    for i:= 1 to 2 do
      target[i]:= starter[i] + t * dir[i];
    result:= theF(target);
    if (result = f1) or (result = f2) then begin
      {this means that we have reached the max resolution of evaluator}
      verySmallF:= 10 * abs(result);
      break;
    end;
    if abs(result) < verySmallF then
      break;
    if result * f1 > 0 then begin
      f1:= result;
      t1:= t;
    end else begin
      f2:= result;
      t2:= t;
    end;
  end;
end;

function TImplicitCurveLine.GetNearestZero(thePoint, dirSignaller:
  vector2; natural: Boolean; var theZero: vector2):
  Boolean;
{It is supposed that 'dirSignaller' was a solution and that 'thePoint'
 is a rough guess about the next point in the solution curve. then, if
 'natural = False', we seek for a solution along the line
    t  -->  thePoint + t * n
 where n is orthogonal to the vector joining 'dirSignaller' to 'thePoint'.
 if 'natural = True', we seek for a solution along that line, but now 'n'
 denotes the gradient of the function at 'thePoint' and 'signaller' is ignored.}

var
  i, count: Integer;
  moving, prevMoving, gradient, director, shift: vector2;
  f, prevF, t, step, previousStep, average: Double;
  badPoint: Boolean;
begin
  result:= FALSE;
  gradient:= GradientOfF(thePoint);
  if natural then
    director:= gradient
  else begin
    director[1]:= -dirSignaller[2] + thePoint[2];
    director[2]:= dirSignaller[1] - thePoint[1];
  end;
  moving:= thePoint;
  f:= theF(moving);
  average:= abs(f);
  count:= 1;
  if average < verySmallF then begin
    {'thePoint' is already the solution}
    result:= TRUE;
    theZero:= moving;
    exit;
  end
  else begin
    verySmallF:= Max(minErr, average * 0.1 / count); {a rough error control}
    verySmallF:= Min(verySmallF, 1.0e-10);
  end;
  step:= 0.3 * hypot(step2D[1], step2D[2]); {in the look for the
  solution, a step greater than 'step' will be rejected}

  repeat
    prevF:= f;
    prevMoving:= moving;
    previousStep:= Step;
    gradient:= GradientOfF(moving);
    if natural then
      director:= gradient;
    t:= gradient[1] * director[1] + gradient[2] * director[2];
    if abs(t) < cMinDenom then
      exit;
    {guess the following better point}
    t:= -f / t;
    if (count mod 5) = 0 then
      t:= 2*t;
    for i:= 1 to 2 do begin
      shift[i]:= t * director[i];
      moving[i]:= prevMoving[i] + shift[i];
    end;
    step:= hypot(shift[1], shift[2]);
    f:= theF(moving);
    average:= average + abs(f); {the sum of absolute values of the function
    encountered till the moment}

    badPoint:= ((count mod 5) <> 0) and (step > 0.9 * previousStep); {the sequence must converge}
    inc(count);
    if (abs(f) < verySmallF) then begin
      badPoint:= False;
      break;
    end;
    if (not badPoint) and (f * prevF < -cMinDenom) then begin
      f:= GetZeroInSegment(prevMoving, director, t, prevF, f, moving);
      badPoint:= (abs(f) > verySmallF);
      break;
    end;
  until (abs(f) < verySmallF) or badPoint;
  if not badPoint then begin
    result:= True;
    theZero:= moving;
    verySmallF:= Max(minErr, average * 0.01 / count); {a rough error control}
    verySmallF:= Min(verySmallF, 1.0e-10);
  end
  {else
    beep(100,100);}
end;

function TImplicitCurveLine.GetNextPoint(pos: Vector2; var vel, nextPos:
  Vector2): Boolean;
{'pos' is a solution and 'vel' an approximate tangent to the curve at 'pos'. In
 'nextPos' is returned another solution in the approximate direction and sense
 of 'vel' and in 'vel' the new value of the velocity at the new point}
var
  k, count: Integer;
  y, z: vector2;
  a, t: Double;
begin
  result:= GetVel(pos, y);
  if result then begin
    a:= y[1] * vel[1] + y[2] * vel[2];
    if a < 0 then begin {change sense of direction}
      y[1]:= -y[1];
      y[2]:= -y[2];
    end;
    count:= 0;
    t:= 5 / 4; {initialization: on the first pass of the following repeat,
    this value will be equal to one}
    repeat
      inc(count);
      t:= 0.8 * t; {reduce step}
      for k:= 1 to 2 do
        z[k]:= pos[k] + t * y[k];
      result:= GetNearestZero(z, pos, False, nextPos);
    until result or (count = 8);
    if result then
      vel:= y
  end;
end;

function TImplicitCurveLine.GetVel(pos: Vector2; var vel: Vector2): Boolean;
{The tangent to the curve 'F=0' is orthogonal to the gradient of 'F'. We compute
 a vector 'vel' orthogonal to gradient of F. The norm of 'vel' is made to
 depend on the length 'a' of the gradient of 'F' and on the parameter 'radius'.
 The dependence has been chosen in an empiric manner that
 has given us good results.}
var
  grad: Vector2;
  a: Double;
begin
  grad:= GradientOfF(pos);
  a:= hypot(grad[1], grad[2]);
  result:= a > 1.0e-24;
  if result then begin
    a:= radius * (6 + 5 * a) / (a * (2 + a)); {empirical formula}
    vel[1]:= -a * grad[2];
    vel[2]:= a * grad[1];
  end;
end;

function TImplicitCurveLine.GradientOfF(const pp: Vector2): Vector2;
  var
    i: Integer;
    y: Vector2;
    f1, f2, f3, f4: Extended;
begin
  if Assigned(theFGrad) then
    result:= theFGrad(pp)
  else for i:= 1 to 2 do begin
    y:= pp;
    y[i]:= pp[i] + kIncr4Points;
    f1:= theF(y);
    y[i]:= pp[i] + k2Incr4Points;
    f2:= theF(y);
    y[i]:= pp[i] - kIncr4Points;
    f3:= theF(y);
    y[i]:= pp[i] - k2Incr4Points;
    f4:= theF(y);
    result[i]:= (8*(f1- f3) + f4 - f2)/k12Incr4Points;
  end;
end;

function TImplicitCurveLine.InitialPointInRectangle(ci, cj: Integer;
  var startingPoint: vector2): Boolean;
{Find the best among nine points in the rectangle and from it, find
a nearby solution point}
var
  position: array[1..9] of vector2;
  hereStarts: vector2;
  i, j, k, theBest, count: Integer;
  f, badness, leastBadness, d: Double;
begin
  result:= FALSE;
  leastBadness:= 1.0E30;
  theBest:= 1;
  count:= 0;
  k:= 0;
  for i:= 1 to 3 do
    for j:= 1 to 3 do begin
      k:= k + 1;
      position[k, 1]:= coordVert[1, ci] + step2D[1] * coordRatios[i];
      position[k, 2]:= coordVert[2, cj] + step2D[2] * coordRatios[j];
      f:= theF(position[k]);
      badness:= abs(f);
      if (badness < leastBadness) then begin
        leastBadness:= badness;
        theBest:= k;
        inc(count);
      end;
    end; {of the j loop}
  if (count > 0) and InRange(d, position[thebest]) then begin
    hereStarts:= position[thebest];
    if getNearestZero(position[thebest], startingPoint, True, hereStarts) then
      begin
      result:= TRUE;
      startingPoint:= hereStarts;
    end;
  end;
end; {InitialPointInRectangle2D}

procedure TImplicitCurveLine.SetEnvironment(theF0: R2ToR1Function;
  theF0grad: R2toR2Function; InRange0: InRange2DFunction;
  Store0: StoreSegmentProcedure2D; {verySmallF0: Double;}
  minCorner0, maxCorner0: Vector2; ppLine0, flagsPLine0: Integer;
  radius0: double);
{Store data given by the caller, initialize things and compute
the average of the absolute value of the function. The
admissible error is made proportional to that average.}
const
  samplesPerLine = 10;
  maxSegmentPixels = 8; {Max number of pixels of each segment.          }
  littleReal = 1.0E-15;
var
  epsi, place: vector2;
  a, average, maxAbsValue: Double;
  i, j, totalCount: Integer;
begin
  theF:= theF0;
  theFgrad:= theF0grad;
  InRange:= InRange0;
  Store:= Store0;
  {verySmallF := verySmallF0;}
  minCorner:= minCorner0;
  maxCorner:= maxCorner0;
  ppLine:= min(cMaxPPLine, ppLine0);
  flagsPLine:= min(cMaxNumFlags, flagsPLine0);
  radius:= radius0;
  for i:= 1 to 2 do begin
    halfEdges[i]:= (maxCorner[i] - minCorner[i]) / 2;
    center[i]:= (maxCorner[i] + minCorner[i]) / 2;
    step2D[i]:= 2 * halfEdges[i] / ppLine;
    flagStep[i]:= 2 * halfEdges[i] / flagsPLine;
    coordVert[i, 0]:= minCorner[i];
    for j:= 1 to ppLine do
      coordVert[i, j]:= coordVert[i, j - 1] + step2D[i];
  end;

  {clear the used-points map}
  for i:= 0 to flagsPLine - 1 do
    for j:= 0 to flagsPLine - 1 do
      alreadyUsed[i, j]:= FALSE;

  {obtain an average of the absolute value of 'theF'}
  for i:= 1 to 2 do
    epsi[i]:= halfEdges[i] / samplesPerLine;
  average:= littleReal;
  maxAbsValue:= 0;
  totalCount:= sqr(samplesPerLine) + 1;
  for i:= 1 to samplesPerLine do
    for j:= 1 to samplesPerLine do begin
      place[1]:= minCorner[1] + (2 * i - 1) * epsi[1];
      place[2]:= minCorner[2] + (2 * j - 1) * epsi[2];
      try
        a:= theF(place);
        a:= abs(a);
        if a > maxAbsValue then
          maxAbsValue:= a;
        average:= average + a;
      except
        dec(totalCount);
      end;
    end;
  radius:= hypot(halfEdges[1],halfEdges[2])/flagsPLine;
  average:= (average - maxAbsValue) / totalCount;
  MinErr:= average * cMinErr;
  verySmallF:= Max(minErr, average * 1.0e-6);
  verySmallF:= Min(verySmallF, 1.0e-10);
end; {SetEnvironment}

initialization;
  InitializeTImplicitCurveLine;
finalization;
  FreeImplicitCurveLine;

end.

