
{-----------------------------------------------------------------------------
 Unit Name: FMapR3R6
 Project Name : ImmersionR3R6
 Author:    Angel Montesinos-Amilibia
 Date: 2009
 Purpose: Implements class 'TMapR3R6' that defines several properties of smooth
    maps from R^3 to R^6
-----------------------------------------------------------------------------}
{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

unit FMapR3R6;
interface
uses
  Windows, Graphics,
  OCFunctDDDoubleSafe, VectorsAndUtil, FDrawParamSurf;

const
  cSamples = 20;

type
  BackFunction = Function(const xDom: Vector3; var value: Double): Boolean of object;

  TMapR3R6 = class
    xx: Vector6; {the point of the surface where we are working}
    xp: array[1..3] of vector6; {first derivatives of parameterization}
    xs: array[1..3, 1..3] of Vector6; {second derivatives of parameterization}
    tt: array[1..3] of Vector6; {orthonormal basis of tangent space}
    ttc: array[1..3] of Vector3; {components of the 'tt[i]' in the coord basis 'xp'}
    nn: array[1..3] of Vector6; {orthonormal basis of the normal space}
    bb: matrix333; {Components in the normal orthonormal basis
                    of the vectors defining the second ff}
    bb1m2, bb12mTwo3: Vector3; {some useful linear combinations of the vectors 'bb'}
    meanCurv: Vector3;
    gg: matrix33; {metric of immersion}
    fm: matrix33;  {focal locus matrix}
    siSamp, coSamp, si2Samp, co2Samp: array[0..cSamples] of Double; {precomputed
        values of sines and cosines}
    meanCurvNorm, stp: Double;
    si, co, si2, co2: array of Double; {variable precomputed values of sines and cosines}
    function AsymptoticFunctionSphericCoord(const pp: Vector2): Double;
    function AsymptoticFunctionSphericCoordGrad(const pp: Vector2): Vector2;
    function AsymptoticFunctionStereogrCoord(const pp: Vector2): Double;
    function AsymptoticFunctionStereogrCoordGrad(const pp: Vector2): Vector2;
    function AsymptoticSecondDegreeInvar(const pp: Vector2): Double;
    function AsymptoticSecondDegreeInvarGrad(const pp: Vector2): Vector2;
    procedure ComputeCoeffsFFF;
    procedure ComputeCube;
    function ComputeOrthonormalTangentBasis: Boolean;
    procedure ComputeSecondFFGeneratingVectors(var ff: matrix333; var ff1m2, ff12mTwo3,mc:
                Vector3);
    function ComputeVeroneseData(xDom: Vector3; grid: Integer): Boolean;
    function ComputeVeroneseInvariants(xDom: Vector3; var SecFFRank, VeronRank: Integer;
                var Hnorm, MeanCurvEnerg: Double): Boolean;
    procedure ComputeVeroneseVertexes(var pd: PointDataGrid);
    procedure DefineOrthonormalNormalBasis;
    procedure FirstAndSecondVeroneseDer(const u,v: Double; var xu, xv,
                xuu, xuv, xvv: Vector3);
    procedure FirstFFVeronese(const pp: Vector2; var E, F, G: Double);
    procedure FirstVeroneseDer(const u,v: Double; var xu, xv: Vector3);
    function FocalLocusFunction(u: Vector3): Double;
    function FocalLocusFunctionGrad(u: Vector3): Vector3;
    function GaussCurvVeronese(const pp: Vector2): Double;
    procedure InitializeSiCo;
    procedure InitializeSiCoSamples;
    function NormalCompVeronese1(const pp: Vector2): Double;
    function NormalCompVeronese1Grad(const pp: Vector2): Vector2;
    function NormalCompVeronese2(const pp: Vector2): Double;
    function NormalCompVeronese2Grad(const pp: Vector2): Vector2;
    function NormalCompVeronese3(const pp: Vector2): Double;
    function NormalCompVeronese3Grad(const pp: Vector2): Vector2;
    procedure PosFirstDerivativesFR3R6(const p: vector3);
    procedure PosFirstAndSecondDerivativesFR3R6(const p: vector3);
    function SingularPointInSegment(var sp: Vector2; pre,cur: Vector2): Boolean;
    function TwistingMatrixRank(const pp: Vector2; tol: Double): Integer;
  public
    gr: Integer;
    min3DNorm, max3DNorm: Vector3;
    backF: BackFunction; {a generic function that may represent an invariant
       of the second fundamental form}
    function CheckRangeStereogrCoord(var depth: Double; p: Vector2): Boolean;
    constructor Create;
    destructor Destroy; override;
    function MeanCurvEnergy(const xDom: Vector3; var value: Double): Boolean;
    function NormOfH(const xDom: Vector3; var value: Double): Boolean;
    function SecondFF(const p: Vector2): vector3;
  end;

  {See the detailed explanations about the invariants of the second fundamental
   form, the Veronese of curvature and the Focal locus given in the memory
   "ComputationsOn3Subvarieties.pdf" in this same folder}

var
  MapR3R6: TMapR3R6;

implementation
uses
  Classes, SysUtils, Math, Forms,
  FImmersionR3R6;

const
  zero2: Vector2=(0,0);

function TMapR3R6.AsymptoticFunctionSphericCoord(const pp: Vector2): Double;
var
  X1, X2: Double;
  si1, si2, co1, co2: Extended;
  am: matrix33;
  a,b: Integer;
begin
  sincos(pp[1], si1, co1);
  sincos(pp[2], si2, co2);
  X1:= si1*co2;
  X2:= si1*si2;
  for a:= 1 to 3 do
    for b:= 1 to 3 do
      am[a,b]:= X1*bb[1,a,b] + X2*bb[2,a,b] + co1*bb[3,a,b];
  result:= Determinant3(am);
end;


function TMapR3R6.AsymptoticFunctionSphericCoordGrad(const pp: Vector2): Vector2;
var
  X1, X2, X3, Y1, Y2, Y3: Double;
  si1, co1, si2, co2: Extended;
  am, am1, am2, am3: matrix33;
  a,b: Integer;
begin
  sincos(pp[1], si1, co1);
  sincos(pp[2], si2, co2);
  X1:= si1*co2;
  X2:= si1*si2;
  X3:= co1;
  for a:= 1 to 3 do
     for b:= 1 to 3 do
        am[a,b]:= X1*bb[1,a,b] + X2*bb[2,a,b] + X3*bb[3,a,b];
  Y1:= co1*co2;
  Y2:= co1*si2;
  Y3:= -si1;
  am1:= am;
  am2:= am;
  am3:= am;
  for b:= 1 to 3 do begin
    am1[1,b]:= Y1*bb[1,1,b] + Y2*bb[2,1,b] + Y3*bb[3,1,b];
    am2[2,b]:= Y1*bb[1,2,b] + Y2*bb[2,2,b] + Y3*bb[3,2,b];
    am3[3,b]:= Y1*bb[1,3,b] + Y2*bb[2,3,b] + Y3*bb[3,3,b];
  end;
  result[1]:= Determinant3(am1) + Determinant3(am2)+ Determinant3(am3);

  Y1:= -co1*si2;
  Y2:= co1*co2;
  Y3:= 0;
  am1[1]:= am[1];
  am2[2]:= am[2];
  am3[3]:= am[3];
  for b:= 1 to 3 do begin
    am1[1,b]:= Y1*bb[1,1,b] + Y2*bb[2,1,b] + Y3*bb[3,1,b];
    am2[2,b]:= Y1*bb[1,2,b] + Y2*bb[2,2,b] + Y3*bb[3,2,b];
    am3[3,b]:= Y1*bb[1,3,b] + Y2*bb[2,3,b] + Y3*bb[3,3,b];
  end;
  result[2]:= Determinant3(am1) + Determinant3(am2)+ Determinant3(am3);
end;

function TMapR3R6.AsymptoticFunctionStereogrCoord(const pp: Vector2): Double;
var
  X1, X2, X3, denInv: Double;
  am: matrix33;
  a,b: Integer;
begin
  denInv:= 1/(1+sqr(pp[1])+sqr(pp[2]));
  X1:= 2*pp[1]*denInv;
  X2:= 2*pp[2]*denInv;
  X3:= (1-sqr(pp[1])-sqr(pp[2]))*denInv;
  for a:= 1 to 3 do
    for b:= 1 to 3 do
      am[a,b]:= X1*bb[1,a,b] + X2*bb[2,a,b] + X3*bb[3,a,b];
  result:= Determinant3(am);
end;


function TMapR3R6.AsymptoticFunctionStereogrCoordGrad(const pp: Vector2): Vector2;
var
  X1, X2, X3, Y1, Y2, Y3, denInv, denInv2: Double;
  am, am1, am2, am3: matrix33;
  a,b: Integer;
begin
  denInv:= 1/(1+sqr(pp[1])+sqr(pp[2]));
  denInv2:= sqr(denInv);
  X1:= 2*pp[1]*denInv;
  X2:= 2*pp[2]*denInv;
  X3:= (1-sqr(pp[1])-sqr(pp[2]))*denInv;
  for a:= 1 to 3 do
    for b:= 1 to 3 do
      am[a,b]:= X1*bb[1,a,b] + X2*bb[2,a,b] + X3*bb[3,a,b];
  Y1:= 2*(1- sqr(pp[1])+ sqr(pp[2]))*denInv2;
  Y2:= -4*pp[1]*pp[2]*denInv2;
  Y3:= -4*pp[1]*denInv2;
  am1:= am;
  am2:= am;
  am3:= am;
  for b:= 1 to 3 do begin
    am1[1,b]:= Y1*bb[1,1,b] + Y2*bb[2,1,b] + Y3*bb[3,1,b];
    am2[2,b]:= Y1*bb[1,2,b] + Y2*bb[2,2,b] + Y3*bb[3,2,b];
    am3[3,b]:= Y1*bb[1,3,b] + Y2*bb[2,3,b] + Y3*bb[3,3,b];
  end;
  result[1]:= Determinant3(am1) + Determinant3(am2)+ Determinant3(am3);

  Y2:= 2*(1- sqr(pp[2])+ sqr(pp[1]))*denInv2;
  Y1:= -4*pp[1]*pp[2]*denInv2;
  Y3:= -4*pp[2]*denInv2;
  am1[1]:= am[1];
  am2[2]:= am[2];
  am3[3]:= am[3];
  for b:= 1 to 3 do begin
    am1[1,b]:= Y1*bb[1,1,b] + Y2*bb[2,1,b] + Y3*bb[3,1,b];
    am2[2,b]:= Y1*bb[1,2,b] + Y2*bb[2,2,b] + Y3*bb[3,2,b];
    am3[3,b]:= Y1*bb[1,3,b] + Y2*bb[2,3,b] + Y3*bb[3,3,b];
  end;
  result[2]:= Determinant3(am1) + Determinant3(am2)+ Determinant3(am3);
end;

function TMapR3R6.AsymptoticSecondDegreeInvar(const pp: Vector2): Double;
var
  X1, X2, X3, denInv: Double;
  am: matrix33;
  a,b: Integer;
begin
  denInv:= 1/(1+sqr(pp[1])+sqr(pp[2]));
  X1:= 2*pp[1]*denInv;
  X2:= 2*pp[2]*denInv;
  X3:= (1-sqr(pp[1])-sqr(pp[2]))*denInv;
  for a:= 1 to 3 do
    for b:= 1 to 3 do
      am[a,b]:= X1*bb[1,a,b] + X2*bb[2,a,b] + X3*bb[3,a,b];
  result:= am[1,1]*am[2,2]- am[1,2]*am[2,1] + am[2,2]*am[3,3]- am[2,3]*am[3,2]+
           am[1,1]*am[3,3]- am[1,3]*am[3,1];
end;

function TMapR3R6.AsymptoticSecondDegreeInvarGrad(const pp: Vector2): Vector2;
var
  X1, X2, X3, X1p, X2p, X3p, denInv, u, v, u2, v2, u2pv2: Double;
  am, amp: matrix33;
  a,b: Integer;
begin
  u:= pp[1];
  v:= pp[2];
  u2:= sqr(u);
  v2:= sqr(v);
  u2pv2:= u2+v2;
  denInv:= 1/(1+u2pv2);
  X1:= 2*u*denInv;
  X2:= 2*v*denInv;
  X3:= (1-u2pv2)*denInv;
  for a:= 1 to 3 do
    for b:= 1 to 3 do
      am[a,b]:= X1*bb[1,a,b] + X2*bb[2,a,b] + X3*bb[3,a,b];

  denInv:= sqr(denInv);
  X1p:= 2*(1-u2+v2)*denInv;
  X2p:= -4*u*v*denInv;
  X3p:= -4*u*denInv;
  for a:= 1 to 3 do
    for b:= 1 to 3 do
      amp[a,b]:= X1p*bb[1,a,b] + X2p*bb[2,a,b] + X3p*bb[3,a,b];
  result[1]:=  amp[1,1]*am[2,2]+ am[1,1]*amp[2,2]- am[1,2]*amp[2,1]- amp[1,2]*am[2,1]
             + amp[2,2]*am[3,3]+ am[2,2]*amp[3,3]- am[2,3]*amp[3,2]- amp[2,3]*am[3,2]
             + amp[3,3]*am[1,1]+ am[3,3]*amp[1,1]- am[3,1]*amp[1,3]- amp[3,1]*am[1,3];

  X2p:= 2*(1+u2-v2)*denInv;
  X1p:= -4*u*v*denInv;
  X3p:= -4*v*denInv;
  for a:= 1 to 3 do
    for b:= 1 to 3 do
      amp[a,b]:= X1p*bb[1,a,b] + X2p*bb[2,a,b] + X3p*bb[3,a,b];
  result[2]:=  amp[1,1]*am[2,2]+ am[1,1]*amp[2,2]- am[1,2]*amp[2,1]- amp[1,2]*am[2,1]
             + amp[2,2]*am[3,3]+ am[2,2]*amp[3,3]- am[2,3]*amp[3,2]- amp[2,3]*am[3,2]
             + amp[3,3]*am[1,1]+ am[3,3]*amp[1,1]- am[3,1]*amp[1,3]- amp[3,1]*am[1,3];
end;

function TMapR3R6.CheckRangeStereogrCoord(var depth: Double; p: Vector2): Boolean;
begin
  depth:= 1.00001 - sqr(p[1]) - sqr(p[2]);
  result:= depth > 0;
end;

procedure TMapR3R6.ComputeCoeffsFFF;
{gives gg[1,1]=xu.xu , gg[1,2]= gg[2,1]=xu.xv, gg[2,2]= xv.xv, coefficients of
 the first fund. form}
var
  i, j: Integer;
begin
  for i:= 1 to 3 do
    for j:= i to 3 do begin
      gg[i, j]:= EuclInnerProd(xp[i], xp[j]);
      if j > i then
        gg[j, i]:= gg[i, j];
    end;
end;

procedure TMapR3R6.ComputeCube;
{Compute the parallelopiped of sides parallel to the axes than contains tightly
the Veronese. It is given in the affine coordinates centered at the mean curvature
vector and basis 'nn'}
var
  i, j, a: Integer;
  y: Vector3;
begin
  for i:= 1 to 3 do begin
    min3DNorm[i]:= 1.0e20;
    max3DNorm[i]:= -1.0e20;
  end;
  for i:= 0 to cSamples do
    for j:= 0 to cSamples do
      for a:= 1 to 3 do begin
        y[a]:=- bb12mTwo3[a]*(1/3+co2Samp[i]) +
                    sqr(siSamp[i])*( bb1m2[a]*co2Samp[j]+ bb[1,2,a]*si2Samp[j] )+
                    si2Samp[i]*(bb[1,3,a]*coSamp[j] + bb[2,3,a]*siSamp[j]);
        min3DNorm[a]:= min(min3DNorm[a], y[a]);
        max3DNorm[a]:= max(max3DNorm[a], y[a]);
      end;
end;

function TMapR3R6.ComputeOrthonormalTangentBasis: Boolean;
{We apply the Gram-Schmidth algorithm for obtaining an orthonormal basis 'tt' from
 the coordinate basis 'xp' of the tangent space at the point currently considered.
 Also, the components 'ttc' of the vectors in 'tt' in the 'xp' basis are calculated.}
var
  aux, c1, c2: Double;
  i, j: Integer;
begin
  aux:= EuclNorm(xp[1]);
  result:= aux > 1.0e-15;
  if result then begin
    aux:= 1 / aux;
    ttc[1, 1]:= aux;
    ttc[1, 2]:= 0;
    ttc[1, 3]:= 0;
    for i:= 1 to 6 do
      tt[1, i]:= xp[1, i] * aux;

    ttc[2, 2]:= gg[1, 1];
    ttc[2, 1]:= -gg[1, 2];
    ttc[2, 3]:= 0;
    for i:= 1 to 6 do
      tt[2, i]:= gg[1, 1] * xp[2, i] - gg[1, 2] * xp[1, i];
    aux:= EuclNorm(tt[2]);
    result:= aux > 1.0e-15;
    if result then begin
      aux:= 1 / aux;
      for i:= 1 to 6 do
        tt[2, i]:= tt[2, i] * aux;
      ttc[2, 1]:= ttc[2, 1] * aux;
      ttc[2, 2]:= ttc[2, 2] * aux;

      c1:= -EuclInnerProd(tt[1], xp[3]);
      c2:= -EuclInnerProd(tt[2], xp[3]);
      ttc[3, 1]:= c1 * ttc[1, 1] + c2 * ttc[2, 1];
      ttc[3, 2]:= c1 * ttc[1, 2] + c2 * ttc[2, 2];
      ttc[3, 3]:= 1;
      for i:= 1 to 6 do
        tt[3, i]:= xp[3, i] + c1 * tt[1, i] + c2 * tt[2, i];
      aux:= EuclNorm(tt[3]);
      result:= aux > 1.0e-15;
      if result then begin
        aux:= 1 / aux;
        for i:= 1 to 6 do
          tt[3, i]:= tt[3, i] * aux;
        for i:= 1 to 3 do
          ttc[3, i]:= ttc[3, i] * aux;
      end;
    end;
  end;
end;

procedure TMapR3R6.ComputeSecondFFGeneratingVectors(var ff: matrix333; var ff1m2, ff12mTwo3, mc:
      Vector3);
{The normal components 'ff' of the derivatives of the 'tt' with respect to the 'tt'
vector fields are calculated. They are stored as their components in the normal
basis 'nn'. Some useful linear combinations of the 'ff' vectors are stored in
'ff1m2, ff12mTwo3, mc' }
var
  dd: array[1..3, 1..3] of Vector3;
  i, j, a, b, k: Integer;
begin
  for i:= 1 to 3 do
    for j:= i to 3 do begin
      for k:= 1 to 3 do
        dd[i, j, k]:= EuclInnerProd(xs[i, j], nn[k]);
      if i < j then
        dd[j, i]:= dd[i, j];
    end;
  for i:= 1 to 3 do
    for j:= i to 3 do begin
      for k:= 1 to 3 do begin
        ff[i, j, k]:= 0;
        for a:= 1 to 3 do
          for b:= 1 to 3 do
            ff[i, j, k]:= ff[i, j, k] + ttc[i, a] * ttc[j, b] * dd[a, b, k];
      end;
      if i < j then
        ff[j, i]:= ff[i, j]; {the matrix of the 2nd fund. form is symmetric}
    end;
  for i:= 1 to 3 do begin
    ff12mTwo3[i]:= (ff[1,1,i]+ ff[2,2,i] - 2*ff[3,3,i])/4;
    ff1m2[i]:= (ff[1,1,i]- ff[2,2,i])/2;
  end;
  for i:= 1 to 3 do  {compute mean curvature vector}
    mc[i]:= (ff[1,1,i]+ ff[2,2,i]+ ff[3,3,i])/3;
end;

function TMapR3R6.ComputeVeroneseData(xDom: Vector3; grid: Integer): Boolean;
begin
  if gr <> grid then begin
    gr:= grid;
    InitializeSiCo;
  end;
  PosFirstAndSecondDerivativesFR3R6(xDom);
  ComputeCoeffsFFF;
  result:= ComputeOrthonormalTangentBasis;
  ImmersionR3R6.InfoBar.SimpleText:= '';
  if result then begin
    DefineOrthonormalNormalBasis;
    ComputeSecondFFGeneratingVectors(bb, bb1m2,bb12mTwo3,MeanCurv);
    ComputeCube;
  end
  else
    ImmersionR3R6.Wash;
end;

function TMapR3R6.ComputeVeroneseInvariants(xDom: Vector3; var SecFFRank, VeronRank: Integer;
       var Hnorm, MeanCurvEnerg: Double): Boolean;
{Compute several invariants of the Veronese}
const
  cTol= 1.0e-15;
var
  ff: matrix333;
  ff1m2, ff12mTwo3, mc: Vector3;
  i: Integer;
  a: Matrix0703;
begin
  PosFirstAndSecondDerivativesFR3R6(xDom);
  ComputeCoeffsFFF;
  result:= ComputeOrthonormalTangentBasis;
  if result then begin
    DefineOrthonormalNormalBasis;
    ComputeSecondFFGeneratingVectors(ff, ff1m2, ff12mTwo3, mc);
    HNorm:= EuclNorm(mc);
    MeanCurvEnerg:= (EuclNorm(ff[1,1]) + EuclNorm(ff[2,2]) + EuclNorm(ff[3,3])+
      2*(EuclNorm(ff[1,2]) + EuclNorm(ff[1,3]) + EuclNorm(ff[2,3])))/3;
    for i:= 1 to 3 do begin
      a[1,i]:= ff12mTwo3[i];
      a[2,i]:= ff1m2[i];
      a[3,i]:= ff[1,2,i];
      a[4,i]:= ff[1,3,i];
      a[5,i]:= ff[2,3,i];
    end;
    VeronRank:= MatrixRank(a, 5,3, cTol);
    if VeronRank < 3 then begin
      for i:= 1 to 3 do
        a[6,i]:= ff[3,3,i];
      SecFFRank:= MatrixRank(a, 6, 3, cTol);
    end
    else
      SecFFRank:= 3;
  end;
end;

procedure TMapR3R6.ComputeVeroneseVertexes(var pd: PointDataGrid);
{Computes the components of the vertexes in the affine orthonormal
 frame with origin in the mean curvature vector and basis nn}
var
  am: matrix33;
  sinPhi, cosPhi, sin2Phi, cos2Phi, sinTheta, cosTheta, sin2Theta, cos2Theta: Extended;
  i, j, a, b: Integer;
  etau, etav: Vector3;
begin
  for i:= 0 to gr do
    for j:= 0 to gr do begin
      for a:= 1 to 3 do begin
        {first, the position of the Veronese point}
        pd[i, j].pos[a]:= - bb12mTwo3[a]*(1/3+co2[i]) +
                    sqr(si[i])*( bb1m2[a]*co2[j]+ bb[1,2,a]*si2[j] )+
                    si2[i]*(bb[1,3,a]*co[j] + bb[2,3,a]*si[j]);

        {now the derivatives of the Veronese with respect to the two parameters
         that we use. See the notes "ComputationsOn3Subvarieties.pdf"}
        etau[a]:= co2[i]*(bb[1,3,a]*co[j] + bb[2,3,a]*si[j]) + 0.5*si2[i]*(2*bb12mTwo3[a]+
                  bb1m2[a]*co2[j]+ bb[1,2,a]*si2[j]);
        etav[a]:= si[i]*(bb[1,2,a]*co2[j] - bb1m2[a]*si2[j]) + co[i]*
                  (bb[2,3,a]*co[j] - bb[1,3,a]*si[j]);
      end;
      pd[i, j].nor:= CrossAndNormalize(etau, etav); {the unit normal to the Veronese
      surface. A failure is marked by returning a normal whose first component is 3}
      if pd[i,j].nor[1] > 2 then repeat  {if the parameterization is singular at
         the current point, obtain the normal at a non-sigular point very near it}
        sincos(i*stp + random*1.0e-3,sinTheta,cosTheta);
        sin2Theta:= 2*sinTheta*cosTheta;
        cos2Theta:= sqr(cosTheta)- sqr(sinTheta);
        sincos(j*stp + random*1.0e-3,sinPhi,cosPhi);
        sin2Phi:= 2*sinPhi*cosPhi;
        cos2Phi:= sqr(cosPhi)- sqr(sinPhi);
        for a:= 1 to 3 do begin
          etau[a]:= cos2Theta*(bb[1,3,a]*cosPhi + bb[2,3,a]*sinPhi) + 0.5*sin2Theta
              *(2*bb12mTwo3[a]+ bb1m2[a]*cos2Phi+ bb[1,2,a]*sin2Phi);
          etav[a]:= sinTheta*(bb[1,2,a]*cos2Phi - bb1m2[a]*sin2Phi) + cosTheta*
                  (bb[2,3,a]*cosPhi - bb[1,3,a]*sinPhi);
        end;
        pd[i, j].nor:= CrossAndNormalize(etau, etav);
      until pd[i,j].nor[1] < 2;
    end;
end;


constructor TMapR3R6.Create;
var
  i: Integer;
begin
  inherited;
  for i:= 1 to 6 do
    theFs[i]:= TOCFunctDDDoubleSafe.Create;
  gr:= ImmersionR3R6.grid;
  InitializeSiCoSamples;
  InitializeSiCo;
  backF:= NormOfH;
end;

procedure TMapR3R6.DefineOrthonormalNormalBasis;
{Based on the Gram-Schmidht algorithm applied to the canonical basis of r^6}
var
  aux, a, energ: Double;
  i, i1, i2, i3, j: Integer;
  en: array[1..6] of Vector6;
  v: Vector6;
  good: Boolean;
begin
  a:= 0;
  for i:= 1 to 6 do begin
    for j:= 1 to 6 do
      en[i, j]:= ord(j = i) - tt[1, i] * tt[1, j] - tt[2, i] * tt[2, j] - tt[3,
        i] * tt[3, j];
    energ:= EuclEnergy(en[i]);
    if a < energ then begin
      a:= energ;
      i1:= i;
    end;
  end;
  good:= (a > 1.0e-20);
  if good then begin
    a:= 1 / sqrt(a);
    for j:= 1 to 6 do
      nn[1, j]:= en[i1,j] * a; {we have already the first normal unit vector}

    a:= 0;
    for i:= 1 to 6 do begin
      aux:= EuclInnerProd(en[i],nn[1]);
      for j:= 1 to 6 do 
        en[i,j]:= en[i,j] - aux*nn[1,j];
      energ:= EuclEnergy(en[i]);
      if (a < energ) and (i <> i1) then begin
        a:= energ;
        i2:= i;
      end;
    end;
    good:= (a > 1.0e-20);
    if good then begin
      a:= 1 / sqrt(a);
      for j:= 1 to 6 do
        nn[2, j]:= en[i2,j] * a; {we have already the second normal unit vector}

      a:= 0;
      for i:= 1 to 6 do begin
        aux:= EuclInnerProd(en[i],nn[2]);
        for j:= 1 to 6 do
          en[i,j]:= en[i,j] - aux*nn[2,j];
        energ:= EuclEnergy(en[i]);
        if (a < energ) and (i<>i1) and (i<>i2) then begin
          a:= energ;
          i3:= i;
        end;
      end;
      good:= (a > 1.0e-20);
      if good then begin
        a:= 1 / sqrt(a);
        for j:= 1 to 6 do
          nn[3, j]:= en[i3,j] * a; {we have already the third normal unit vector}
      end;
    end;
  end;

end;

destructor TMapR3R6.Destroy;
var
  i: Integer;
begin
  for i:= 1 to 6 do
    FreeAndNil(theFs[i]);
  si:= nil;
  co:= nil;
  si2:= nil;
  co2:= nil;
  inherited;
end;

procedure TMapR3R6.FirstAndSecondVeroneseDer(const u,v: Double; var xu, xv,
  xuu, xuv, xvv: Vector3);
var
  u2, v2, u4, v4, uv , u2pv2: Double;
  xxp, xyp, xzp, yyp, yzp, zzp,
  denInv: Double;
  a,b: Integer;
begin
  u2:= sqr(u);
  v2:= sqr(v);
  u4:= sqr(u2);
  v4:= sqr(v2);
  uv:= u*v;
  u2pv2:= u2 + v2;
  denInv:= 1/(1+ u2pv2);
  denInv:= denInv*sqr(denInv);
  xxp:= 8*u*(1 - u2 + v2);
  yyp:= -16*u*v2;
  zzp:= 8*u*(u2pv2 - 1);
  xyp:= 4*v*(1 - 3*u2 + v2);
  xzp:= 2*(1 + u4 -6*u2- v4);
  yzp:= 4*uv*(u2pv2 - 3);
  for a:= 1 to 3 do
    xu[a]:= (xxp*bb[1,1,a] + 2*xyp*bb[1,2,a] + 2*xzp*bb[1,3,a] +
            yyp*bb[2,2,a] + 2*yzp*bb[2,3,a] + zzp*bb[3,3,a])*denInv;

  yyp:= 8*v*(1- v2 + u2);
  xxp:= -16*u2*v;
  zzp:= 8*v*(u2pv2 - 1);
  xyp:= 4*u*(1-3*v2+ u2);
  yzp:= 2*(1 + v4 - 6*v2- u4);
  xzp:= 4*uv*(u2pv2 -3);
  for a:= 1 to 3 do
    xv[a]:= (xxp*bb[1,1,a] + 2*xyp*bb[1,2,a] + 2*xzp*bb[1,3,a] +
            yyp*bb[2,2,a] + 2*yzp*bb[2,3,a] + zzp*bb[3,3,a])*denInv;

  denInv:= 1/(1+ u2pv2);
  denInv:= sqr(sqr(denInv));

  xxp:= 8*(3*u4 - 8*u2*(1+v2)+ sqr(1+v2));
  xyp:= 48*uv*(u2-v2-1);
  xzp:= -4*u*(9+ u4+ 6*v2-3*v4-2*u2*(7+v2));
  yyp:= -16*v2*(1-5*u2+v2);
  yzp:= 4*v*(v4 - 2*(1+u2)*v2-3*(1-6*u2+u4));
  zzp:= -8*(1+3*u4-v4+2*u2*(v2-4));
  for a:= 1 to 3 do
    xuu[a]:= (xxp*bb[1,1,a] + 2*xyp*bb[1,2,a] + 2*xzp*bb[1,3,a] +
            yyp*bb[2,2,a] + 2*yzp*bb[2,3,a] + zzp*bb[3,3,a])*denInv;

  yyp:= 8*(3*v4 - 8*v2*(1+u2)+ sqr(1+u2));
  xyp:= 48*uv*(v2-u2-1);
  yzp:= -4*v*(9+ v4+ 6*u2-3*u4-2*v2*(7+u2));
  xxp:= -16*u2*(1-5*v2+u2);
  xzp:= 4*u*(u4 - 2*(1+v2)*u2-3*(1-6*v2+v4));
  zzp:= -8*(1+3*v4-u4+2*v2*(u2-4));
  for a:= 1 to 3 do
    xvv[a]:= (xxp*bb[1,1,a] + 2*xyp*bb[1,2,a] + 2*xzp*bb[1,3,a] +
            yyp*bb[2,2,a] + 2*yzp*bb[2,3,a] + zzp*bb[3,3,a])*denInv;

  xxp:= 32*uv*(2*u2- v2- 1);
  xyp:= -4*(3*(u4+v4)-1+2*v2 + u2*(2-18*v2));
  xzp:= 4*v*(v4-2*(1+u2)*v2-3*(1-6*u2+u4));
  yyp:= -32*uv*(1+u2-2*v2);
  yzp:= 4*u*(u4 - 2*u2*(1+v2)- 3*(1-6*v2+v4));
  zzp:= -32*uv*(u2pv2-2);
  for a:= 1 to 3 do
    xuv[a]:= (xxp*bb[1,1,a] + 2*xyp*bb[1,2,a] + 2*xzp*bb[1,3,a] +
            yyp*bb[2,2,a] + 2*yzp*bb[2,3,a] + zzp*bb[3,3,a])*denInv;
end;

procedure TMapR3R6.FirstFFVeronese(const pp: Vector2; var E, F, G: Double);
var
  xu, xv: Vector3;
begin
  FirstVeroneseDer(pp[1], pp[2], xu, xv);
  E:= EuclInnerProd(xu,xu);
  F:= EuclInnerProd(xu,xv);
  G:= EuclInnerProd(xv,xv);
end;

procedure TMapR3R6.FirstVeroneseDer(const u,v: Double; var xu, xv: Vector3);
var
  u2, v2, u4, v4, uv , u2pv2: Double;
  xxp, xyp, xzp, yyp, yzp, zzp,
  denInv: Double;
  a,b: Integer;
begin
  u2:= sqr(u);
  v2:= sqr(v);
  u4:= sqr(u2);
  v4:= sqr(v2);
  uv:= u*v;
  u2pv2:= u2 + v2;
  denInv:= 1/(1+ u2pv2);
  denInv:= denInv*sqr(denInv);
  xxp:= 8*u*(1 - u2 + v2);
  yyp:= -16*u*v2;
  zzp:= 8*u*(u2pv2 - 1);
  xyp:= 4*v*(1 - 3*u2 + v2);
  xzp:= 2*(1 + u4 -6*u2- v4);
  yzp:= 4*uv*(u2pv2 - 3);
  for a:= 1 to 3 do
    xu[a]:= (xxp*bb[1,1,a] + 2*xyp*bb[1,2,a] + 2*xzp*bb[1,3,a] +
            yyp*bb[2,2,a] + 2*yzp*bb[2,3,a] + zzp*bb[3,3,a])*denInv;

  yyp:= 8*v*(1- v2 + u2);
  xxp:= -16*u2*v;
  zzp:= 8*v*(u2pv2 - 1);
  xyp:= 4*u*(1-3*v2+ u2);
  yzp:= 2*(1 + v4 - 6*v2- u4);
  xzp:= 4*uv*(u2pv2 -3);
  for a:= 1 to 3 do
    xv[a]:= (xxp*bb[1,1,a] + 2*xyp*bb[1,2,a] + 2*xzp*bb[1,3,a] +
            yyp*bb[2,2,a] + 2*yzp*bb[2,3,a] + zzp*bb[3,3,a])*denInv;
end;




function TMapR3R6.FocalLocusFunction(u: Vector3): Double;
{the function used for computing the focal locus. It is up to the sign,
the determinant of  u.alpha - g , where  'g' is the first fundamental
form and alpha, the second fundamental form}
var
  i,j: Integer;
begin
  for i:= 1 to 3 do
    for j:= i to 3 do begin
      fm[i,j]:= u[1]*bb[i,j,1]+ u[2]*bb[i,j,2]+u[3]*bb[i,j,3]- ord(i=j);
      if j>i then
        fm[j,i]:= fm[i,j];
    end;
  result:= Determinant3(fm);
end;

function TMapR3R6.FocalLocusFunctionGrad(u: Vector3): Vector3;
{the gradient of FocalLocusFunction. It is assumed that FocalLocusFunction has
 been already calculated, so that the matrix 'mf' has already been initialized
 with its value at 'u'}
var
  m: matrix33;
  v: Vector3;
  i,j, k: Integer;
begin
  m:= fm;
  for k:= 1 to 3 do
    for i:= 1 to 3 do begin
      result[k]:= 0;
      v:= fm[i];
      for j:= 1 to 3 do
        m[i,j]:= bb[i,j,k];
      result[k]:= result[k] +  Determinant3(m);
      m[i]:= v;
  end;
end;

function TMapR3R6.GaussCurvVeronese(const pp: Vector2): Double;
  var
    xu, xv, xuu, xuv, xvv, normal: Vector3;
    den, q, eII, fII, gII : Double;
    i: Integer;
begin
  FirstAndSecondVeroneseDer(pp[1], pp[2], xu, xv, xuu, xuv, xvv);
  normal := EuclCrossProd(xu, xv);
  den := EuclEnergy(normal);
  q:= sqrt(den);
  if den > 1.0e-15 then begin
    for i := 1 to 3 do
      normal[i] := normal[i] / q;
    eII := EuclInnerProd(normal, xuu);
    fII := EuclInnerProd(normal, xuv);
    gII := EuclInnerProd(normal, xvv);
    result:= (eII*gII-sqr(fII))/den;
  end
  else
    result:= kTheError;
end;




procedure TMapR3R6.InitializeSiCo;
{initialize values of sine and cosine of several angles that will be heavily used}
var
  sine, cosine: Extended;
  i: Integer;
begin
  stp:= pi / gr;
  SetLength(si, gr+1);
  SetLength(co, gr+1);
  SetLength(si2, gr+1);
  SetLength(co2, gr+1);
  for i:=0 to gr do begin
    sincos(i*stp,sine, cosine);
    si[i]:= sine;
    co[i]:= cosine;
    sincos(2*i*stp,sine, cosine);
    si2[i]:= sine;
    co2[i]:= cosine;
  end;
end;

procedure TMapR3R6.InitializeSiCoSamples;
{initialize values of sine and cosine of several angles that will be heavily used}
var
  st: Double;
  sine, cosine: Extended;
  i: Integer;
begin
  st:= pi / cSamples;
  for i:= 0 to cSamples do begin
    sincos(i * st, sine, cosine);
    siSamp[i]:= sine;
    coSamp[i]:= cosine;
    sincos(2*i * st, sine, cosine);
    si2Samp[i]:= sine;
    co2Samp[i]:= cosine;
  end;
end;

function TMapR3R6.MeanCurvEnergy(const xDom: Vector3; var value: Double): Boolean;
{if all goes well, returns in 'value' the mean curvature energy at 'xDom'}
const
  cTol= 1.0e-15;
var
  ff: matrix333;
  ff1m2, ff12mTwo3, mc: Vector3;
begin
  PosFirstAndSecondDerivativesFR3R6(xDom);
  ComputeCoeffsFFF;
  result:= ComputeOrthonormalTangentBasis;
  if result then begin
    DefineOrthonormalNormalBasis;
    ComputeSecondFFGeneratingVectors(ff, ff1m2, ff12mTwo3, mc);
    value:= (EuclNorm(ff[1,1]) + EuclNorm(ff[2,2]) + EuclNorm(ff[3,3])+
      2*(EuclNorm(ff[1,2]) + EuclNorm(ff[1,3]) + EuclNorm(ff[2,3])))/3;
  end;
end;

function TMapR3R6.NormalCompVeronese1(const pp: Vector2): Double;
var
  xu, xv: Vector3;
begin
  FirstVeroneseDer(pp[1], pp[2], xu, xv);
  Result:= xu[2]*xv[3] - xu[3]*xv[2];
end;

function TMapR3R6.NormalCompVeronese1Grad(const pp: Vector2): Vector2;
var
  xu, xv, xuu, xuv, xvv: Vector3;
begin
  FirstAndSecondVeroneseDer(pp[1], pp[2], xu, xv, xuu, xuv, xvv);
  result[1]:= xuu[2]*xv[3]+ xu[2]*xuv[3] - xuu[3]*xv[2]- xu[3]*xuv[2];
  result[2]:= xuv[2]*xv[3]+ xu[2]*xvv[3] - xuv[3]*xv[2]- xu[3]*xvv[2];
end;

function TMapR3R6.NormalCompVeronese2(const pp: Vector2): Double;
var
  xu, xv: Vector3;
begin
  FirstVeroneseDer(pp[1], pp[2], xu, xv);
  Result:= xu[3]*xv[1] - xu[1]*xv[3];
end;

function TMapR3R6.NormalCompVeronese2Grad(const pp: Vector2): Vector2;
var
  xu, xv, xuu, xuv, xvv: Vector3;
begin
  FirstAndSecondVeroneseDer(pp[1], pp[2], xu, xv, xuu, xuv, xvv);
  result[1]:= xuu[3]*xv[1]+ xu[3]*xuv[1] - xuu[1]*xv[3]- xu[1]*xuv[3];
  result[2]:= xuv[3]*xv[1]+ xu[3]*xvv[1] - xuv[1]*xv[3]- xu[1]*xvv[3];
end;

function TMapR3R6.NormalCompVeronese3(const pp: Vector2): Double;
var
  xu, xv: Vector3;
begin
  FirstVeroneseDer(pp[1], pp[2], xu, xv);
  Result:= xu[1]*xv[2] - xu[2]*xv[1];
end;

function TMapR3R6.NormalCompVeronese3Grad(const pp: Vector2): Vector2;
var
  xu, xv, xuu, xuv, xvv: Vector3;
begin
  FirstAndSecondVeroneseDer(pp[1], pp[2], xu, xv, xuu, xuv, xvv);
  result[1]:= xuu[1]*xv[2]+ xu[1]*xuv[2] - xuu[2]*xv[1]- xu[2]*xuv[1];
  result[2]:= xuv[1]*xv[2]+ xu[1]*xvv[2] - xuv[2]*xv[1]- xu[2]*xvv[1];
end;


function TMapR3R6.NormOfH(const xDom: Vector3; var value: Double): Boolean;
{if all goes well, returns in 'value' the norm of the mean curvature vector
 at 'xDom'}
const
  cTol= 1.0e-15;
var
  ff: matrix333;
  ff1m2, ff12mTwo3, mc: Vector3;
begin
  PosFirstAndSecondDerivativesFR3R6(xDom);
  ComputeCoeffsFFF;
  result:= ComputeOrthonormalTangentBasis;
  if result then begin
    DefineOrthonormalNormalBasis;
    ComputeSecondFFGeneratingVectors(ff, ff1m2, ff12mTwo3, mc);
    value:= EuclNorm(mc);
  end;
end;

procedure TMapR3R6.PosFirstAndSecondDerivativesFR3R6(const p: vector3);
{wrap for the function value and that of its first and second derivatives}
var
  s12, s13, s23: Double;
  fd: SecondDeriv;
  i: Integer;
begin
  for i:= 1 to 6 do begin
    fd:= theFs[i].SecondD([p[1], p[2], p[3], 1, 0, 0, 0, 0, 0]);
    xx[i]:= fd.fa0;
    xp[1, i]:= fd.dfa0;
    xs[1, 1, i]:= fd.ddfa0;
    fd:= theFs[i].SecondD([p[1], p[2], p[3], 0, 1, 0, 0, 0, 0]);
    xp[2, i]:= fd.dfa0;
    xs[2, 2, i]:= fd.ddfa0;
    fd:= theFs[i].SecondD([p[1], p[2], p[3], 0, 0, 1, 0, 0, 0]);
    xp[3, i]:= fd.dfa0;
    xs[3, 3, i]:= fd.ddfa0;
    fd:= theFs[i].SecondD([p[1], p[2], p[3], 1, 1, 0, 0, 0, 0]);
    s12:= fd.ddfa0;
    fd:= theFs[i].SecondD([p[1], p[2], p[3], 1, 0, 1, 0, 0, 0]);
    s13:= fd.ddfa0;
    fd:= theFs[i].SecondD([p[1], p[2], p[3], 0, 1, 1, 0, 0, 0]);
    s23:= fd.ddfa0;
    xs[1, 2, i]:= 0.5 * (s12 - xs[1, 1, i] - xs[2, 2, i]);
    xs[2, 1, i]:= xs[1, 2, i];
    xs[1, 3, i]:= 0.5 * (s13 - xs[1, 1, i] - xs[3, 3, i]);
    xs[3, 1, i]:= xs[1, 3, i];
    xs[2, 3, i]:= 0.5 * (s23 - xs[2, 2, i] - xs[3, 3, i]);
    xs[3, 2, i]:= xs[2, 3, i];
  end;
end;

procedure TMapR3R6.PosFirstDerivativesFR3R6(const p: vector3);
{wrap for the function value and that of its first derivatives}
var
  fd: FirstDeriv;
  i: Integer;
begin
  for i:= 1 to 6 do begin
    fd:= theFs[i].FirstD([p[1], p[2], p[3], 1, 0, 0]);
    xx[i]:= fd.fa0;
    xp[1, i]:= fd.dfa0;
    fd:= theFs[i].FirstD([p[1], p[2], p[3], 0, 1, 0]);
    xp[2, i]:= fd.dfa0;
    fd:= theFs[i].FirstD([p[1], p[2], p[3], 0, 0, 1]);
    xp[3, i]:= fd.dfa0;
  end;
end;

function TMapR3R6.SingularPointInSegment(var
           sp: Vector2; pre, cur: Vector2): Boolean;
var
  p2: Double;
  par1, par2: Vector2;
  xu, xv, a1, a2, a: Vector3;
  i: Integer;
begin
  result:= False;
  FirstVeroneseDer(pre[1], pre[2], xu, xv);
  a1:= EuclCrossProd(xu, xv);
  FirstVeroneseDer(cur[1], cur[2], xu, xv);
  a2:= EuclCrossProd(xu, xv);
  result:= (a1[2]*a2[2] <= 0);
  if result then begin
    par1:= pre;
    par2:= cur;
    while abs(a1[2]) > 1.0e-10 do begin
      for i:= 1 to 2 do
        sp[i]:= (par1[i] + par2[i])/2;
      FirstVeroneseDer(sp[1], sp[2], xu, xv);
      a:= EuclCrossProd(xu, xv);
      if a[2]*a1[2] >=0 then begin
        par1:= sp;
        a1:= a;
      end
      else begin
        par2:= sp;
        a2:= a
      end;
    end;
     result:= abs(a1[3]) < 1.0e-3;
  end;
end;

function TMapR3R6.SecondFF(const p: Vector2): vector3;
var
  si1, co1, si2, co2: Extended;
  co12, co22, si12, si22: Double;
  a: Integer;
begin
  sincos(p[1], si1, co1);
  si12:= 2*si1*co1;
  co12:= sqr(co1) - sqr(si1);

  sincos(p[2], si2, co2);
  si22:= 2*si2*co2;
  co22:= sqr(co2) - sqr(si2);
  for a:= 1 to 3 do
    result[a]:= - bb12mTwo3[a]*(1/3+co12) +
                  sqr(si1)*( bb1m2[a]*co22+ bb[1,2,a]*si22)+
                  si12*(bb[1,3,a]*co2 + bb[2,3,a]*si2);
end;

function TMapR3R6.TwistingMatrixRank(const pp: Vector2; tol: Double): Integer;
var
  X1, X2, X3, denInv, tr, aux: Double;
  m, crosses: matrix33;
  m73: matrix0703;
  a,b: Integer;
begin
  denInv:= 1/(1+sqr(pp[1])+sqr(pp[2]));
  X1:= 2*pp[1]*denInv;
  X2:= 2*pp[2]*denInv;
  X3:= (1-sqr(pp[1])-sqr(pp[2]))*denInv;
  for a:= 1 to 3 do
    for b:= 1 to 3 do
      m[a,b]:= X1*bb[1,a,b] + X2*bb[2,a,b] + X3*bb[3,a,b];
  result:= 3;
  aux:= Determinant3(m);
  if abs(aux) < tol then begin
    result:= 2;
    for a:= 1 to 3 do
      crosses[a]:= EuclCrossProd(m[a], m[a mod 3+1]);
    aux:= 0;
    for a:= 1 to 3 do
      for b:= 1 to 3 do
        aux:= aux + abs(crosses[a,b]);
    if aux < tol then begin
      result:= 1;
      aux:= 0;
      for a:= 1 to 3 do
        for b:= 1 to 3 do
        aux:= aux + abs(m[a,b]);
      if aux < tol then
        result:= 0;
    end
  end
end;

end.

