
{-----------------------------------------------------------------------------
 Unit Name: FRenderImplicitSurfaceWithThreads
 Project Name : ImmersionR3R6
 Author:    Angel Montesinos-Amilibia
 Date: 2009
 February, 2005 - January 2009
 Purpose: Implements class TRenderImplicitSurfaceWithThreads, for rendering
    a surface given as the locus of zeros of a function in R^3. It launches
    so many threads as CPU are in the running machine in order to accelerate
    the rendering.
-----------------------------------------------------------------------------}
{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

{See documentation in the unit of same name in the project "Superficies"}

unit FRenderImplicitSurfaceWithThreads;

interface
uses
  Classes, Windows, Graphics, Messages,
  FGammaPicker, VectorsAndUtil, ImplicitRenderThread, FDrawParamSurf;

type

TRenderImplicitSurfaceWithThreads = class(TObject)
  private
    min3D, max3D, edges3D, halfEdges, center3D: Vector3;
    c0Abs, vpCenterDist: Double;
    cube: array of Segment3;
    procedure ComputeCubeKey(from: Vector3; var ck: CubeKey);
    function DefinedUnitVector(place: TPoint; var theVector: Vector3): Boolean;
    procedure DilateCube(sc: Double);
    procedure FindRay(var ray: Vector3; x,y: Double);
    procedure InitializeZBuffer;
    function ProjectToPaneRealAndInv(x: Vector3): Vector3; overload;
    procedure SetFocus;
    procedure SetProjParameters;
    

  public
    rc: TRect; {the effective rectangle where the surface is drawn. Note
                that we create the initial bitmap as to accomodate the
                maximum rectangle allowed}
    iPts: array[0..1] of InterpPoints;
    vpCubeKey: CubeKey;
    span3D: Double;
    cubePt: Vector2Int;
    stop, busy: Boolean;
    drawn: Boolean;
    viewPoint, focus: Vector3;
    c0, c1, c2, c3, cx, cy: Double;
    wH, wV, wN: Vector3;
    numOfCPU, grain: Integer;
    function ClipSegmentToRect3D(r: TRect; var x, y: Vector3; yDirector: Boolean): Boolean;
    function ComputeAndDraw(hor, vert: Integer; ComputeVPFoc: Boolean): AnsiString;
    constructor Create(hor, vert: Integer);
    destructor Destroy; override;
    procedure DrawSegment(can: TCanvas; vec1, vec2: Vector3; maxPart: Integer;  baseColor: TColor);
    function DrawSurface(grain: Integer): AnsiString;
    procedure InitializeCubeViewpointAndFocus;
    function MoveSurface(Shift: TShiftState; prevPt, curPt: TPoint): AnsiString;
    procedure ScaleCoefficients(sc: Double);
  end;

var
  rist: TRenderImplicitSurfaceWithThreads;



implementation
  uses
    Math, SysUtils, mmSystem, Forms,
    Stutz_Get_CPU_data, FImmersionR3R6, FMapR3R6;
  const
    cTolZBuffer = 0.995;
    cLinesXY = 1;
    cLinesYZ = 2;
    cLinesZX = 4;
    cBullet = chr(149);

    cIncr = 0.0002;
    cTwoIncr = 2*cIncr;
    cTwelveIncr = 12* cIncr;
    cInvMaxByte2= 1/(255*255);

function SqrEuclDistance(x, y : Vector3): Double;
begin
  result:= sqr(x[1]-y[1]) + sqr(x[2]-y[2]) + sqr(x[3]-y[3]);
end;

function TRenderImplicitSurfaceWithThreads.ClipSegmentToRect3D(r: TRect; var x, y: Vector3; yDirector: Boolean): Boolean;
{If yDirector = True, clips the line  t -> x + t*y inside 'r' by giving on
exit on 'x' and 'y' the points where that line cuts the boundary of 'r', on the
understanding that one of them may be at the infinity, what we will substitute
by a very far point in the line.
If director = False, clips inside 'r' the segment from 'x' to 'y', giving on
exit, on 'x' and 'y', the first and last points of that segment that lie in 'r'.
Result is true if the clip is nonempty.
See the paper entitled "ClippingASegmentOrLine.pdf" that accompanies the source
of this program.}
const
  cPInf= 1.0e6;   {This is plus infinity for our practical purposes}
  cMInf= -1.0e6;   {And this is minus infinity}
var
  a,b, pa, pb: Vector3;
  minT, maxT, t, t0, t1, k0, k1, k0n, k1n, aux: Double;
  i, sgn: Integer;

  procedure ReduceInterval; {A frequently used task}
  begin
    if t0 > t1 then begin
      t:= t0;
      t0:= t1;
      t1:= t;
    end;
    minT:= max(minT, t0);
    maxT:= min(maxT, t1);
    result:= ( maxT >= minT );
  end;

begin
  result:= False;
  a:= x;
  if yDirector then
    for i:= 1 to 3 do
      b[i]:= x[i] + y[i]
  else
    b:= y;
  pa:= ProjectToPaneRealAndInv(a);
  pb:= ProjectToPaneRealAndInv(b);

  {These are the false infinite ends that we use as initializations when
   'director' is true}
  if yDirector then begin
    minT:= cMInf;
    maxT:= cPInf;
  end
  else begin  {the initial interval is the initial limit}
    minT:= 0;
    maxT:= 1;
  end;

  {We first study the special cases}
  if abs(pa[3]-pb[3]) < 1.0e-12 then begin
    if abs(pa[1]-pb[1]) > 1.0e-12 then begin
      t0:= (pa[1]-r.left)/(pa[1]-pb[1]);
      t1:= (pa[1]-r.right+1)/(pa[1]-pb[1]);
      ReduceInterval;
      if  result and (abs(pa[2]-pb[2]) > 1.0e-12) then begin
        t0:= (pa[2]-r.top)/(pa[2]-pb[2]);
        t1:= (pa[2]-r.bottom-1)/(pa[2]-pb[2]);
        ReduceInterval;
      end;
    end
  end
  {Now the usual cases. First for the horizontal coordinate}
  else begin
    sgn:= 1- 2* ord( pa[3]*pb[3]/(pa[3]-pb[3]) < 0);
    aux:= pa[1]-pb[1];
    if (abs(aux) < 1.0e-10) then begin
      {the line has constant horizontal projection. Hence}
      if (pa[1] < r.left) or (pa[1] > r.Right-1) then
        Exit;
    end
    else begin
      k0n:= r.Left*(pa[3]-pb[3])+ pa[1]*pb[3]-pb[1]*pa[3];
      k0:= k0n/aux;
      k1n:=(r.Right-1)*(pa[3]-pb[3])+ pa[1]*pb[3]-pb[1]*pa[3];
      k1:= k1n/aux;
      if (k0 <= 0) and (k1 <= 0) then begin
        result:= False;
        Exit;
      end;
      if (k0 > 0) and (k1 > 0) then begin
        t0:= (pa[1]-r.left)*pb[3]/k0n;
        t1:= (pa[1]-r.right+1)*pb[3]/k1n;
        ReduceInterval;
      end
      else if (k0 > 0) and (k1 < 0) then begin
        aux:= (pa[1]-r.left)*pb[3]/k0n;
        if sgn = 1 then begin
          t0:= aux;
          t1:= cPInf;
        end
        else begin
          t0:= cMInf;
          t1:= aux;
        end;
        ReduceInterval;
      end
      else if (k0 < 0) and (k1 > 0) then begin
        aux:= (pa[1]-r.right+1)*pb[3]/k1n;
        if sgn = 1 then begin
          t0:= aux;
          t1:= cPInf;
        end
        else begin
          t0:= cMInf;
          t1:= aux;
        end;
        ReduceInterval;
      end;

      {And now for the vertical coordinate}

      if result then begin
        aux:= pa[2]-pb[2];
        if (abs(aux) < 1.0e-10) then begin
          {the line has constant vertical projection. Hence}
          if (pa[2] < r.top) or (pa[2] > r.Bottom-1) then
            Exit;
        end
        else if result then begin
          k0n:= r.top*(pa[3]-pb[3])+ pa[2]*pb[3]-pb[2]*pa[3];
          k0:= k0n/aux;
          k1n:=(r.bottom-1)*(pa[3]-pb[3])+ pa[2]*pb[3]-pb[2]*pa[3];
          k1:= k1n/aux;
          if (k0 <= 0) and (k1 <= 0) then begin
            result:= False;
            Exit;
          end;
          if (k0 > 0) and (k1 > 0) then begin
            t0:= (pa[2]-r.top)*pb[3]/k0n;
            t1:= (pa[2]-r.bottom+1)*pb[3]/k1n;
            ReduceInterval;
          end
          else if (k0 > 0) and (k1 < 0) then begin
            aux:= (pa[2]-r.top)*pb[3]/k0n;
            if sgn = 1 then begin
              t0:= aux;
              t1:= cPInf;
            end
            else begin
              t0:= cMInf;
              t1:= aux;
            end;
            ReduceInterval;
          end
          else if (k0 < 0) and (k1 > 0) then begin
            aux:= (pa[2]-r.bottom+1)*pb[3]/k1n;
            if sgn = 1 then begin
              t0:= aux;
              t1:= cPInf;
            end
            else begin
              t0:= cMInf;
              t1:= aux;
            end;
            ReduceInterval;
          end;
        end;
      end;
    end;
  end;
  if result then for i:= 1 to 3 do begin
    x[i]:= (1-minT)*a[i]+ minT*b[i];
    y[i]:= (1-maxT)*a[i]+ maxT*b[i];
  end;
end;


function TRenderImplicitSurfaceWithThreads.ComputeAndDraw(hor, vert: Integer;
          ComputeVPFoc: Boolean): AnsiString;
  var
    ms: AnsiString;
begin
  hor:= max(hor,vert);
  result:= '';
  busy:= True;
  rc:= Rect(0,0,hor, vert);
  drawn:= False;
  SetBufferSizes(hor, False);
  if bmp.width <> hor then begin
    bmp.Width:= hor;
    bmp.Height:= hor;
  end;
  try
    if ComputeVPFoc then begin
      SetProjParameters;
      SetFocus;
    end;

    ms:= DrawSurface(1); {draw surface with 'grain = 1', that is as
                                   fine grained as possible}
    if ms<>'' then
      raise Exception.Create(ms);
  except on E: Exception do
    result:= E.message;
  end;
end;

procedure TRenderImplicitSurfaceWithThreads.ComputeCubeKey(from: Vector3; var ck: CubeKey);
{the cubekey 'ck' is used to facilitate the computation of the segment on which
 a line cuts the cube on which the surface must be computed}
var
  i: Integer;
begin
  for i := 1 to 3 do begin
    ck[i] := 2;
    if from[i] > max3D[i] then
      ck[i] := 1;
    if from[i] < min3D[i] then
      ck[i] := 0;
  end;
end;


constructor TRenderImplicitSurfaceWithThreads.Create(hor, vert: Integer);
var
  i: Integer;
begin
  inherited Create;
  if high(zBuffer) < hor - 1 then begin
    SetLength(zBuffer,hor);
    SetLength(map,hor);
    for i:= 0 to hor-1 do begin
      SetLength(zBuffer[i],vert);
      SetLength(map[i],vert);
    end;
  end;
  drawn:= False;
  numOfCPU:= NumberOfProcessors;
  grain:=1;
  InitializeCubeViewpointAndFocus;
end;

function TRenderImplicitSurfaceWithThreads.DefinedUnitVector(place: TPoint; var theVector: Vector3): Boolean;
{We use an inverse stereographic projection, with respect to the point
 'r*wN' to the sphere of center the origin and radius 'r'.}
const
  cReductor = 0.3; {The lesser, the more sensitive will be the movement of
                    the surface}
var
  b, a, k, bMa: Vector3;
  t,r: double;
  i: Integer;
begin
  r:= cReductor*span3D/2;
  FindRay(k,place.X, place.y); {vector whose direction is determined by the mouse
                       position 'place'. }
  t:= (vpCenterDist - r)/EuclInnerProd(k, wN); {So, the line 't*k'
               cuts the plane orthogonal to 'wN' passing by
               the origin when 't' has this value}
  {Now, we calculate the inverse stereographic projection of 't*k'
   to the sphere}
  for i:= 1 to 3 do begin
    b[i]:= t*k[i];
    a[i]:= (vpCenterDist + r)*wN[i];
    bMa[i]:= b[i]-a[i];
  end;
  t:= -2*r*EuclInnerProd(wN,bMa)/EuclInnerProd(bMa, bMa);
  for i:= 1 to 3 do
    theVector[i]:= -vpCenterDist*wN[i]+(1-t)*a[i] + t*b[i]; {This is the inverse
    stereographic projection relative to the sphere center. We now normalize
    'theVector'}
  r:= EuclNorm(theVector);
  result:= r > 1.0e-12;
  if result then for i:= 1 to 3 do
    theVector[i]:= theVector[i]/r;
end;

destructor TRenderImplicitSurfaceWithThreads.Destroy;
begin
  Cube:= nil;
  inherited;
end;

procedure TRenderImplicitSurfaceWithThreads.DilateCube(sc: Double);
{as answer to a request from the caller. We dilate the cube and also
 put the viewpoint farther so that the cube is seen as before}
var
  i: Integer;
begin
  for i:= 1 to 3 do begin
    halfEdges[i]:= sc*halfEdges[i];
    edges3D[i]:= 2*halfEdges[i];
    max3D[i]:= center3D[i]+ halfEdges[i];
    min3D[i]:= center3D[i]- halfEdges[i];
    viewpoint[i]:= center3D[i] + sc*(viewPoint[i]- center3D[i]);
  end;
  vpCenterDist:= sc*vpCenterDist;
  span3D:= sc*span3D;
end;

procedure TRenderImplicitSurfaceWithThreads.DrawSegment(can: TCanvas; vec1, vec2: Vector3;
     maxPart: Integer;  baseColor: TColor);
{draw a segment using the zbuffer to check its visibility}
const
  cTolZBuffer = 0.995;
var
  i, num: Integer;
  d, ddif, b,span, lambda: Double;
  pixR1, pixR2, pix: Vector3;
  pt1, pt2: Tpoint;
  vis1, vis2: boolean;
  x1, x2: Vector3;
begin
  x1:= vec1;
  x2:= vec2;
  if not ClipSegmentToRect3D(rc, x1, x2,False) then
    Exit;
  pixR1 := ProjectToPaneRealAndInv(x1);
  pixR2 := ProjectToPaneRealAndInv(x2);
  span:= hypot(pixR1[1]-pixR2[1], pixR1[2]-pixR2[2]);
  num:= 1 + trunc(span/max(2,maxPart));
  ddif:= pixR2[3]-pixR1[3];
  pixR2[1]:= pixR2[1]- pixR1[1];
  pixR2[2]:= pixR2[2]- pixR1[2];

  pt2.x:= round(pixR1[1]);
  pt2.y:= round(pixR1[2]);
  b:= zBuffer[pt2.x,pt2.y]+(pixR1[1]-pt2.x)*(zBuffer[pt2.x+1,pt2.y]-
          zBuffer[pt2.x,pt2.y])+ (pixR1[2]-pt2.y)*(zBuffer[pt2.x,pt2.y+1]-
          zBuffer[pt2.x,pt2.y]);
  vis2 := (pixR1[3] > cTolZBuffer*b);
  for i:= 1 to num do begin
    pt1:= pt2;
    vis1:= vis2;
    lambda:= i/num;
    d:= pixR1[3] + lambda*ddif;
    pix[1]:= pixR1[1]+ lambda*pixR2[1];
    pix[2]:= pixR1[2]+ lambda*pixR2[2];
    pt2.x:= round(pix[1]);
    pt2.y:= round(pix[2]);
    b:= zBuffer[pt2.x,pt2.y]+(pix[1]-pt2.x)*(zBuffer[pt2.x+1,pt2.y]-
          zBuffer[pt2.x,pt2.y])+ (pix[2]-pt2.y)*(zBuffer[pt2.x,pt2.y+1]-
          zBuffer[pt2.x,pt2.y]);
    vis2 := (d > cTolZBuffer*b);
    if vis1 and vis2 then begin
      can.moveto(pt1.x,pt1.y);
      can.lineto(pt2.x,pt2.y);
    end;
  end;
end;


function TRenderImplicitSurfaceWithThreads.DrawSurface(grain: Integer): AnsiString;
{Prepare the launching of the threads that will do the effective drawing and
launch them}
  var
    i,j,hMin, hMax: Integer;
    c: TColor;
begin
  result:= '';
  if ImmersionR3R6.thrNumber = 0 then begin
    try
      stop:= False;
      ComputeCubeKey(viewpoint,vpCubeKey);
      InitializeZBuffer;
      c:= gammas[0,0];
      bmp.Canvas.Brush.Color:= (c and $00FF0000) shr 16 + (c and $0000FF00)+ (c and $000000FF) shl 16;
      bmp.canvas.FillRect(rc);
      for i:= 0 to rc.Right-1 do
        for j:= 0 to rc.Bottom-1 do
          map[i,j]:=0;
      drawn:= True;
      hMin:= 0;
      hMax:= rc.bottom div numOfCPU;
      SetLength(thrs,numOfCPU);
      thrs[0]:= TImplicitRenderThread.CreateIRT(True, MapR3R6.bb, min3D, max3D, hMin, hMax-1,grain);
      if numOfCPU > 1 then
        for i:= 1 to numOfCPU-1 do begin
          hMin:= hMax;
          hMax:= rc.bottom*(i+1) div numOfCPU;
          thrs[i]:= TImplicitRenderThread.CreateIRT(True, MapR3R6.bb, min3D, max3D, hMin, hMax-1,grain);
        end;
      try
        for i:= 0 to numOfCPU-1 do begin
          thrs[i].FreeOnTerminate:= True;
          thrs[i].Priority:= tpHigher;
          thrs[i].OnTerminate:= ImmersionR3R6.HandleThreadDone;
        end;
        busy:= True;
        for i:= 0 to numOfCPU-1 do begin
          inc(ImmersionR3R6.thrNumber);
          thrs[i].Resume;
        end;
      except
        for i:= 0 to numOfCPU-1 do
          thrs[i].Free;
        ImmersionR3R6.thrNumber:=0;
        busy:= False;
      end;
    except on E: Exception do
      result:= E.message;
    end;
  end
  else begin
    for i:= 0 to numOfCPU-1 do
      if Assigned(thrs[i]) then begin
        thrs[i].Suspend;
        thrs[i].Free;
      end;
    busy:= False;
  end;
end;

procedure TRenderImplicitSurfaceWithThreads.FindRay(var ray: Vector3; x,y: Double);
{Returns in 'ray' a vector such that the entire line
     t -> viewpoint + t*ray
 projects on the point (x,y) of the bitmap}
var
  k: Integer;
  ah, av: Double;
begin
  ah := (x - cx)/c0;
  av := (y - cy)/c0;
  for k := 1 to 3 do
    ray[k] := ah * wH[k] + av * wV[k] + wN[k];
end;


procedure TRenderImplicitSurfaceWithThreads.InitializeZBuffer;
  var
    i, j: Integer;
begin
  for i:= 0 to rc.right-1 do
    for j:= 0 to rc.bottom-1 do
      zBuffer[i,j]:=0;
end;


function TRenderImplicitSurfaceWithThreads.MoveSurface(Shift: TShiftState; prevPt, curPt: TPoint): AnsiString;
{move the surface according with the state 'Shift' of the mouse while it was
 dragged from 'prevPt' to 'curPt'}
var
  e1, e2, b: Vector3;
  nrm, twoVe2, e2b, vb, scale: Double;
  i, mode: Integer;
  good: Boolean;
begin
  result := '';
  if busy or (not drawn) then
    Exit;
  busy:= True;
  good := DefinedUnitVector(prevPt, e1);

  if ssCtrl in Shift then
    mode := 1
  else if ssShift in Shift then
    mode := 2
  else
    mode := 0;
  case mode of
    1: begin {translate surface}
        cx := cx - prevPt.x + curPt.x;
        cy := cy - prevPt.y + curPt.y;
      end;
    2: begin  {dilate or shrink then cube so that we will see a greater or
               lesser part ot the focal locus}
        scale := (1 + (prevPt.y - curPt.y) / 500);
        if ((c0abs > 1) or (scale > 1)) and ((c0abs < 10000) or (scale <
          1)) then
            DilateCube(scale);
      end;
  else begin   {rotate surface}
      good := good and DefinedUnitVector(curPt, e2);
      if good then begin
        for i := 1 to 3 do
          b[i] := e1[i] + e2[i];
        nrm := EuclNorm(b);
        for i := 1 to 3 do
          b[i] := b[i] / nrm;
        e2b := EuclInnerProd(e2, b);
        twoVe2 := 2 * EuclInnerProd(e2, wH);
        vb := EuclInnerProd(b, wH);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          wH[i] := vb * b[i] - twoVe2 * e2[i] + wH[i];
        twoVe2 := 2 * EuclInnerProd(e2, wV);
        vb := EuclInnerProd(b, wV);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          wV[i] := vb * b[i] - twoVe2 * e2[i] + wV[i];
        twoVe2 := 2 * EuclInnerProd(e2, wN);
        vb := EuclInnerProd(b, wN);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          wN[i] := vb * b[i] - twoVe2 * e2[i] + wN[i];
        c3 := EuclInnerProd(center3D, wN) - vpCenterDist;
        c1 := EuclInnerProd(center3D, wH);
        c2 := EuclInnerProd(center3D, wV);
        for i := 1 to 3 do
          viewPoint[i] := c3 * wN[i] + c1 * wH[i] + c2 * wV[i];
        vpCenterDist:= EuclDistance(viewPoint,center3D);
        SetFocus;
      end;
    end;
  end;
  try
    ComputeCubeKey(viewpoint,vpCubeKey);
    InitializeZBuffer;
    DrawSurface(grain); {draw the locus using the 'grain' resolution estimated
        along the previous computation in order to change speed for resolution}
    drawn:= True;
  except on E: Exception do
    result:= E.message;
  end;
end;

function TRenderImplicitSurfaceWithThreads.ProjectToPaneRealAndInv(x: Vector3): Vector3;

begin
  result[3] := 1/(EuclInnerProd(wN,x) - c3);
  result[1] := (EuclInnerProd(wH,x) - c1) * c0 *result[3] + cx;
  result[2] := (EuclInnerProd(wV,x) - c2) * c0 *result[3] + cy;
end;

procedure TRenderImplicitSurfaceWithThreads.InitializeCubeViewpointAndFocus;
var
  i: Integer;
begin
  for i:= 1 to 3 do begin
    min3D[i]:= -5;
    max3D[i]:= 5;
    edges3D[i]:= 10;
    halfEdges[i]:= 5;
    center3D[i]:= 0;
  end;
  span3D:=EuclNorm(edges3D);
  vpCenterDist:= 15;

  wN[1]:= -1;
  wN[2]:= 0;
  wN[3]:= 0;
  wH[1]:= 0;
  wH[2]:= 1;
  wH[3]:= 0;
  wV[1]:= 0;
  wV[2]:= 0;
  wV[3]:= 1;
  
  viewPoint[1]:= 15;
  viewPoint[2]:= 0;
  viewPoint[3]:= 0;

  SetFocus;
  SetProjParameters;
end;

procedure TRenderImplicitSurfaceWithThreads.ScaleCoefficients(sc: Double);
begin
  c0:= c0*sc;
  cx:= cx*sc;
  cy:= cy*sc;
end;

procedure TRenderImplicitSurfaceWithThreads.SetFocus;
{define the position of the focus that illuminates the surface}
var
  r: Double;
  i: Integer;
begin
  r := 0.3 * vpCenterDist;
  for i := 1 to 3 do
    focus[i] := viewpoint[i] - r * (wH[i] + wV[i]);
end;


procedure TRenderImplicitSurfaceWithThreads.SetProjParameters;
{Set the coefficients that define the projection of R^3 to the bitmap}
const
  farthness: Double = 3.5;
var
  i, j, k: Integer;
  maximum, minimum: Vector2Single;
  pos: Vector3;
  
  procedure Sieve(u: Vector3);
  var
    z: Vector3;
    ii: Integer;
  begin
    z := ProjectToPaneRealAndInv(u);
    for ii := 1 to 2 do begin
      if z[ii] > maximum[ii] then
        maximum[ii] := z[ii];
      if z[ii] < minimum[ii] then
        minimum[ii] := z[ii]
    end;
  end; {Sieve}
begin
  c0 := 1;
  c1 := EuclInnerProd(viewPoint, wH);
  c2 := EuclInnerProd(viewPoint, wV);
  c3 := EuclInnerProd(viewPoint, wN);
  cx := rc.left + rc.right div 2;
  cy := rc.top + rc.bottom div 2;
  maximum[1] := -maxInt;
  maximum[2] := -maxInt;
  minimum[1] := maxInt;
  minimum[2] := maxInt;
  for i := 0 to 1 do
    for j := 0 to 1 do
      for k := 0 to 1 do begin
        pos[1] := min3D[1] + i * edges3D[1];
        pos[2] := min3D[2] + j * edges3D[2];
        pos[3] := min3D[3] + k * edges3D[3];
        Sieve(pos);
      end;
  c0 := 0.95 * min(rc.right/ (maximum[1] - minimum[1]), rc.bottom/
    (maximum[2] - minimum[2]));
  c0abs := abs(c0);
end;


procedure Initialize;
begin
  rist:= TRenderImplicitSurfaceWithThreads.Create(kWidthMainBox,kWidthMainBox);
end;

procedure Finalize;
begin
  FreeAndNil(rist);
end;

initialization
   Initialize;
finalization
   Finalize;

end.
