

{-----------------------------------------------------------------------------
 Unit Name: FunctionParser_32
 Author:    Angel Montesinos Amilibia
 October 11, 2011 -
 Purpose: Class TFunctionParser_32. Parses a text for a "Realfunction" type.
    Encodes as a string an rpn notation of the function. That rpn string may be
    passed to object code (for x68-64) by means of the accompanying class
    TEvaluator_32. Three object codes may so be obtained: for evaluationg the
    function, its first and its second derivatives. This unit has an evaluator
    for constant expressions, that is without independent variables, for instance
    like  'cos(3 pi/5)/ sqrt(22)'.

 History: I am indebted to Dmitry Kaschenko and Vladimir Safin for the general
    idea of translating an evaluation tree to direct X86 opcode that I learned
    by studying their component "FormulaCompiler".

 For source code you may email me to:
     montesin@uv.es
-----------------------------------------------------------------------------}
{

Unit Name: FunctionParser_32
parser  for real functions of several variables of Double type.

Copyright (C) 2006-2011 Angel Montesinos Amilibia

This unit is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This unit is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can obtain more details by emailing me to:
     montesin@uv.es

Or writing me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andres Estelles, 1
     46100 Burjasot (Valencia)
     Spain
}


unit FunctionParser_32;

interface

uses
  ResStrings, Globals;

const
  cN1 = 512; {The number of steps that make the grid for the
    interpolation of a function given by an integral depending on one
    variable}
  cN2 = 128; {The number of steps that make the grid for the
    interpolation of a function given by an integral depending on two
    variables}
  cQ = 15; {The greater this integer, the better will be the
    approximation of integrals by B-Splines, but with more CPU time}

  cNodes1 = cN1 + 2*cQ + 4;
  cNodes2 = cN2 + 2*cQ + 4;

  {Functions defined by integrals are allowed. Such a function may
  involve at least one and at most two independent variables
  chosen between the allowed variables "allowedvars" of the main formula.
  These variables may appear in the integrand or in the extremes of
  the integral. Then "var1" is the position in "allowedvars" (counting
  from zero) of the first variable, and "var2" the position of the second,
  if there is one (if not, var2 = -1.)

  An integral in the variables, say 'x' and 'y' must be written in the follo-
  wing form

      int( f(x,y,r) ; m(x,y) ; M(x,y) ),

  where 'f(x,y,r)' is a valid formula for a function in the variables 'x','y',
  'r'. This formula stands for the integrand, and 'r' is always the variable of
  integration;

  where 'm(x,y)' and  'M(x,y)' are valid formulas for two functions in the
  variables 'x', 'y'. They represents the inferior and superior limits of the
  integration, respectively.

  Warning: It is not alllowed to write integrals inside integrals, that is
  in the formulas for the integrand of the integration limits. Nor are allowed
  parameters inside integrals.}

type
  ArrayD = array of Double;
  ArrayarrayD = array of arrayD;
  RealFunction = function (const X: array of Double): Double;
  ArrayNodes1 = array [0..cNodes1] of Double;
  ArrayNodes2 = array [0..cNodes2] of Double;
  ArrayArrayNodes2 = array[0..cNodes2, 0..cNodes2] of Double;
  PArrayNodes1 = ^arrayNodes1;
  PArrayArrayNodes2 = ^arrayarrayNodes2;

  IntegralData1 = record
    var1: Integer;
    data: array[0..2] of ArrayNodes1;
  end;

  IntegralData2= record
    var1, var2: Integer; {'var1' and 'var2' code the independent variables from
                  'allowedVars' on which the integrand or its upper and lower
                  limits may depend.}
    data: array[0..2,0..2] of arrayarrayNodes2; {A grid of data that is computed
           when parsing the function. These data serve for interpolating by
           B-splines the value of the integral as function of the vars
           whose position in 'allowedVars' is 'var1' and 'var2'}
  end;
  IntegralDataList1= array of IntegralData1;
       {For all integrals with one variable that may appear in the formula}
  IntegralDataList2= array of IntegralData2;
       {For all integrals with two variable that may appear in the formula}

  TFunctionParser_32= class(TObject)
  private
    allowedVars, allowedPars, allVars: AnsiString; {variable names that may appear in 'rawText'.
                 must be a subset of [p,q,r,s,t,u,v,w,x,y,z]. However, if there
                 are integrals, for instance   int( f(x,y,r) ; m(x,y) ; M(x,y) ),
                 the 'r' parameter in f(x,y,r) shall be interpreted not as the
                 independent variable 'r' that may appear in other parts of
                 the formula but as the integration variable. That is
                 int( f(x,y,r) ; m(x,y) ; M(x,y) is a function of (x,y) given
                 as the integral of      f(x,y,r) dr       between r= m(x,y)
                 and r= M(x,y) }
    rawText: AnsiString; {the text of the function}
    formula, textReduced: AnsiString;
    fCode: AnsiString; {object code of a function without integrals nor parameters}
    dataLast: byte;
    emptyRegisters: Byte; {at each moment in evaluation of the parsed formula}
    VBlock: PByte;
    procedure AddAddressToCode(p: PByte; var cod: AnsiString);
    procedure AddIntToCode(n: Integer; var cod: AnsiString);
    procedure AdjustIntegralDataForBSplines(prov1: PArrayNodes1; prov2: PArrayArrayNodes2;
                dim: Integer);
    function BuildIntegralDataList(tripleArgument: AnsiString):Integer;
    procedure CheckFunction;
    procedure CodifyFunction;
    function ComputeExponent(var n: Integer; ini, fini: Integer): Boolean;
    procedure ComputeIntegralData(integrand, inferior,superior:
        TFunctionParser_32);
    function ConvertIntegral(var hotSpot: Integer): AnsiChar;
    procedure ConvertPowers;
    procedure GetRoomAndCoeffsFirstDeriv;
    procedure GetRoomAndCoeffsSecondDeriv;
    procedure InitializeCodeVars;
    procedure LeaveRoomInFPUStack(neededPlaces: Integer;
               var cod: AnsiSTring);
    function PieceBeginning(const last: Integer; soughtChars: AnsiString):
        Integer;
    procedure PopFromAppStack(var cod: AnsiSTring);
    procedure PutCodeInExeBlock(const cod: AnsiString; var exeBlock: PByte);
    procedure ReduceSymbols(integralsAllowed: Boolean);
    procedure UnifyChars;
  public
    data: array of double; {stores the constants that appear in the formula}
    minCorner, maxCorner, jump1, jump2, invJump1, invJump2: array of double;
    intList1: IntegralDataList1;
    intList2: IntegralDataList2;
    lastIntegral1, lastIntegral2, maxJ: Integer;
    argNum, {length(allowedVars)}
      parNum : Integer; {Length(allowedParams)}
    actorList: AnsiString; {the RPN (reverse Polish notation) string that represents
                            the function}
    lengthActorList: Integer;
    waitingInApplStack: Integer; {Number of doubles that have been pushed to
        the application stack from the FPU registers to leave room for doing
        operations.}

    usedVarsAndParams: AnsiString; {really appearing in 'rawText'}
    V: RealFunction;
    constructor Create;
    destructor Destroy; override;
    procedure ParseNewText(t, vars, pars: AnsiString;
        integralsAllowed: Boolean; minC, maxC: array of double);
    {'t' is the text of the function (capitalization is unimportant); 'vars' is
     a string with some of the characters in 'PQRSTUVWXYZ'. ' pars '
     is a string with some of the characters in 'PQRSTUVWXYZ' that do not
     appear in 'vars'. The characters in 'vars' will be treated as independent
     variables, that is, if vars= 'xyw', the function will be called as
     f(3, 7.21, 2.4) and this will mean that the value of the variable x is 3,
     that of y is 7.21 and that of w is 2.4. However, the function f may have
     a text as  t = 'cos(p x) + q y^2 w'. Then 'pars' would be 'pq'. The value
     of the parameters p and q will be passed to the function using the
     procedure UpdateParameters.}
  end;

  ImplicitFunctionList= array[0..kMaxImplParts] of TFunctionParser_32;

procedure ReadConstantField(theStr: AnsiString; var val: double);

  var
    theParser: TFunctionParser_32;


implementation
uses
  SysUtils, Math, Forms, Windows,
  Utilities;

const
  {Don't change these constants! They are needed to convert a grid of
   points to another grid that, being used for evaluation of B-splines,
   makes the initial grid the interpolation points, that is the curve
   or surface defined by the B-splines passes by the initial points
   with great approximation that may be increased as one likes.
   For a detailed expalanation of the use of B-splines, read the
   attached paper "AproxFuncBSplinesEnglish.pdf"}
  cA= -0.0755861459167170;
  cB= 2.28831641118603;
  cU= -0.01372542929733973;
  cV= -0.36134122590022017;



  pi: double = 3.1415926535897932384;
  halfPi: double = 1.5707963267948966192313;
  twoPi: double = 6.283185307179586476925;
  recipTwoPi: double = 0.1591549430918953357689;
  recipPi: double = 0.3183098861837906715378;
  kMaxNumOfNames = 44; {The number of strings of length less}
                       {than 8 that the parser recognizes as}
                       {function names. For instance, asin  }
                       {and arcsin are both recognized.     }
  upCaseLetters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  variables = 'PQRSTUVWXYZ';
  digits = '0123456789';

  eMaxNum = 1.0E24;
  eMinNum = 1.0E-24;
  eMaxExp = 75;


  {char constants}
  kReturn = char(13);
  kLineFeed = char(10);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';
  kLastChar = #$FF;

  kEscapePointerStr32 = #$00#$00#$00#$00;
  kEscapePointerStr64 = #$00#$00#$00#$00#$00#$00#$00#$00;

  {Code constants. Complete information about the instruction sets x86
   x87, etc. may be downloaded from the sites of Intel and AMD. We use
   the instruction names that appear there and the format of hexadecimal
   characters. A byte is shown here as part of a string, with a sintaxis
   as #$D7, where #$ tells that the two symbols that follow make a character,
   writen as a byte or two hexadecimal digits}

  c_stackDisplacDouble = 8;
  c_stackDisplacDoubleHStr= #$08; {hexadecimal string of c_stackDisplacDouble}
  c_stackDisplacExtended = 12;
  c_stackDisplacExtendedHStr= #$0C; {hexadecimal string of c_stackDisplacExtended}
  c_LastReg = 7; {last register in the FPU}

  c_fstp0ESPd = #$DD#$9C#$24#$00#$00#$00#$00; //fstp st(0) to [ESP] as Double
  c_fldESPd = #$DD#$84#$24#$00#$00#$00#$00;  //fld from [ESP] as a Double.

  c_LoadPi = #$D9#$EB;
  c_Load1 = #$D9#$E8;
  c_Load0 = #$D9#$EE;
  c_zero_comp = #$D9#$EE#$DF#$F1;
    {4 bytes. ST(cLastReg) must be empty. Compares 0 with x=ST(0). Then, JA jumps
     if 0>x; JNA, if 0 <= x; JB,  if < x; JNB, if 0 >= x; JNE, if 0<>x}
  c_1_comp = #$D9#$E8#$DF#$F1;
    {4 bytes.  ST(cLastReg) must be empty. Compares 1 with x=ST(0). Then, JA jumps
     if 1>x; JNA, if 1 <= x; JB, if 1 < x; JNB, if 1 >= x; JNE if 1<>x}
  c_mod_pi = #$D9#$EB#$D9#$C9#$D9#$F8#$DD#$D9;
    {8 bytes. ST(cLastReg) must be empty. Leaves in ST(0) the value of the
     initial value of ST(0) mod pi}
  c_mod_2pi = #$D9#$EB#$D8#$C0#$D9#$C9#$D9#$F8#$DD#$D9;
    {10 bytes. ST(cLastReg) must be empty}
  c_ln = #$D9#$E8#$D9#$C9#$D9#$F1#$D9#$ED#$DE#$C9;
    {10 bytes. Requires ST(0) > 0 and ST(cLastReg) empty}
  c_2toX = #$D9#$C0#$D9#$FC#$D9#$C9#$D8#$E1#$D9#$F0#$D9#$E8#$DE#$C1#$D9#$C9#$D9
    +
    #$E8#$D9#$FD#$DD#$D9#$DE#$C9;
    {24 bytes. Requires two last registers empty}
  c_div2 = c_Load1 + #$D9#$E0#$D9#$C9#$D9#$FD#$DD#$D9;


  {Warnings and messages}

  eBadNumber = 'Hay un nmero mal escrito.';
               {A number is miswritten.}
  eBadCharacters = 'Ha utilizado un carcter no permitido.';
               {You have used an illegal character.}
  eBadFunctionOrVariables = 'Hay una funcin o variable mal escrita.';
               {You mispelled a function or variable}
  eBadParentheses = 'Compruebe, por favor, los parntesis.';
               {Check the parentheses, please.}
  eTooMuchConstants = 'No puedo manejar tantas constantes.';
               {I cannot digest that much constants.}
  eSyntaxError = 'Hay un defecto de sintaxis en la frmula.';
               {There is a sintax fault in the formula.}
  eErrorInParsing = 'Lo siento: no entiendo esa frmula.';
               {Analysis of function failed}
  eExpressionTooLarge = 'Expresin demasiado larga.';
               {The expression is too long}
  eNothingWritten = 'Quizs haya olvidado escribir algo.';
               {There was nothing to work on!}
  eBadVar = 'Las variables independientes han de ser: ';
               {The allowed variables in this formula are:}
  eBadVarSet = 'Las variables permitidas han de estar entre ''p'' y ''z''';
               {The allowed variables must be in the range 'p' to 'z'}
  eBadParamSet = 'Un parmetro no puede ser tambin una variable';
               {No parameter may be named like any variable}
  eNoIntegralsInsideIntegrals = 'No se permiten integrales dentro de integrales';
               {Integrals inside integrals are not allowed}
  eBadSplineArgument = 'Mal argumento de B-spline';
               {Bad B-spline argument}
  eBadIntegralSyntax= 'Sintaxis errnea de una integral';
               {Bad integral syntax}
  eAtMosttwoVarsInInt = 'Las integrales pueden depender a lo sumo de dos variables';
               {Integrals may depend on two variables at most}
  eNaughtIntegral = 'Encontrada una integral intil:  no depende de ninguna variable';
                {A naught integral found: it doesn''t depend on any variables}
type

  pieceClass = (openParent, closeParent, operationSymb,
    negation, constantSymb, varSymb, functionSymb, badThing);

  WrittenFunction = record {The name for a function that may be recognized by
                           the parser. For instance 'cos' or '*'}
    theText: AnsiString;
    equivChar: AnsiChar; {When parsing, that function will be changed to a single
                      character. For instance, 'cos' becomes 'A'}
  end;



var
  uKnownFunc: array[1..kMaxNumOfNames] of WrittenFunction; {Functions that
                      will be recognized by the parser}
  uAttribution: array['#'..#$9F] of pieceClass; {A look-up table of the type of
             character for characters from '#' (num 35) to #$9F (num 159) that
             perhaps may appear sometime in a formula}
  uSpecialValues : array[0..7] of Double; {array of special values that are
                 occasionally injected in the opcode string that makes the
                 function. For example, uSpecialValues[1]= 2*pi}

  qq : array[-cQ..cQ] of Double; {used in the adjusting of the control points
                                  for B-splines}

  c_div_S, c_ln_S, c_2toX_S : AnsiString; {Code vars. Initialized at the
                                 beginning of the application}
  escapePointerStr: AnsiString;

{--------------------------------------------------------------}
          {Auxiliar routines private to this unit}
{--------------------------------------------------------------}



function ExpressionString(theText: AnsiString): AnsiString;
  forward;

function AtomString(theText: AnsiString): AnsiString;
{parses a simple factor in an expression, verifying that it is not
 a variable or parameter. Converts it to RPN}
var
  base: AnsiString;
  len: Integer;
  ch: AnsiChar;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else begin
    ch := theText[1];
    if ch = '(' then begin
      {parse the text between parentheses}
      base := Copy(theText, 2, len - 2);
      result := ExpressionString(base);
    end
    else begin
      {It is expected that 'ch' is not a variable name nor a parameter name.
       We assert this just in case}
      Assert(not (pos(ch, 'qrstuvwxyz') <> 0)); {the 'p' character is admitted
       because it would denote here the constant 'pi'}

      case ch of
        '1', 'p':
          Result := ch;
        kLastChar:
          Result := theText;
        'A'..'Z': begin
            {'theText' is a function. Hence, after 'ch', there must be a
            parenthesis and the last char in 'theText' must be a closing
            parenthesis}
            base := Copy(theText, 3, len - 3); {what is between the parentheses}

            if ch='S' then
              result:= ch + base {integrals have a special syntax here. The in
                 tegral has been converted to the data for its evaluation, and
                 what follows the 'S' here is 'kLastChar' followed by a number 'n'
                 that says that this is integral number 'n'.}
            else
              result := ExpressionString(base) + ch; {invert order for RPN notation}
          end;
      else
        ;
      end;
    end;
  end;
end;


function ConvertPowerSymbol(x: AnsiString; n: Integer): AnsiString;
{for optimization, an expression as 'x^n', where 'x' is any expression
 and 'n' an integer, is converted to the best composition of products
 and squares. Por instance  x^6= D(D(x)*x), where 'D' is the code for the
 square function}
var
  z: AnsiString;

  function ConvertPowerSymbolProv(xx: AnsiString; nn: Integer): AnsiString;
  var
    zz: AnsiString;
    m, np: Integer;
  begin
    zz := xx;
    np := abs(nn);
    m := np div 2;
    if np = 0 then
      Result := '1'
    else if nn = 1 then
      Result := xx
    else if m > 0 then begin
      result := ConvertPowerSymbolProv(zz, m);
      result := 'D(' + result + ')';
      if odd(np) then
        result := result + '*' + zz;
    end
    else
      Result := zz;
  end;

begin
  z := ConvertPowerSymbolProv(x, n);
  Result := '(' + z + ')'; {in order to prevent interferences, we enclose the
              result between  parentheses}
  if n < 0 then  {invert the result}
    Result := '(' + '1/' + Result + ')';
end;

procedure finalizeParser;
begin
  theParser.Free;
end;

function IdentifyFunction(theFunctionText: AnsiString): AnsiChar;
{If 'theFunctionText' is among the known functions, IdentifyFunction }
{returns the capital letter that corresponds to it. else, it    }
{returns '?'.                                                        }
var
  i: Integer;
  foundFunction: Boolean;
begin
  i := 0;
  repeat
    i := i + 1;
    foundFunction := (theFunctionText = uKnownFunc[i].theText);
  until (i = kMaxNumOfNames) or foundFunction;
  if foundFunction then
    Result := uKnownFunc[i].equivChar
  else
    Result := '?';
end;

procedure InitializeParser;
var
  character: char;
  u, v: Double;
  i: Integer;
begin
  case  SizeOf(Pointer) of
    4:
      escapePointerStr:= kEscapePointerStr32;
    8:
      escapePointerStr := kEscapePointerStr64;
  else
    Application.Terminate;
  end;
  uKnownFunc[1].theText := 'COS';
  uKnownFunc[1].equivChar := 'A';

  uKnownFunc[2].theText := 'SIN';
  uKnownFunc[2].equivChar := 'B';

  uKnownFunc[3].theText := 'SEN';
  uKnownFunc[3].equivChar := 'B';

  uKnownFunc[4].theText := 'EXP';
  uKnownFunc[4].equivChar := 'C';

  uKnownFunc[5].theText := 'SQR';
  uKnownFunc[5].equivChar := 'D';

  uKnownFunc[6].theText := 'SQRT';
  uKnownFunc[6].equivChar := 'E';

  uKnownFunc[7].theText := 'LN';
  uKnownFunc[7].equivChar := 'F';

  uKnownFunc[8].theText := 'LOG';
  uKnownFunc[8].equivChar := 'F';

  uKnownFunc[9].theText := 'TG';
  uKnownFunc[9].equivChar := 'G';

  uKnownFunc[10].theText := 'TAN';
  uKnownFunc[10].equivChar := 'G';

  uKnownFunc[11].theText := 'SH';
  uKnownFunc[11].equivChar := 'H';

  uKnownFunc[12].theText := 'SINH';
  uKnownFunc[12].equivChar := 'H';

  uKnownFunc[13].theText := 'CH';
  uKnownFunc[13].equivChar := 'I';

  uKnownFunc[14].theText := 'COSH';
  uKnownFunc[14].equivChar := 'I';

  uKnownFunc[15].theText := 'TH';
  uKnownFunc[15].equivChar := 'J';

  uKnownFunc[16].theText := 'TANH';
  uKnownFunc[16].equivChar := 'J';

  uKnownFunc[17].theText := 'ARCSEN';
  uKnownFunc[17].equivChar := 'K';

  uKnownFunc[18].theText := 'ASEN';
  uKnownFunc[18].equivChar := 'K';

  uKnownFunc[19].theText := 'ARCSIN';
  uKnownFunc[19].equivChar := 'K';

  uKnownFunc[20].theText := 'ASIN';
  uKnownFunc[20].equivChar := 'K';

  uKnownFunc[21].theText := 'ARCCOS';
  uKnownFunc[21].equivChar := 'L';

  uKnownFunc[22].theText := 'ACOS';
  uKnownFunc[22].equivChar := 'L';

  uKnownFunc[23].theText := 'ARCOS';
  uKnownFunc[23].equivChar := 'L';

  uKnownFunc[24].theText := 'ARCTG';
  uKnownFunc[24].equivChar := 'M';

  uKnownFunc[25].theText := 'ARCTAN';
  uKnownFunc[25].equivChar := 'M';

  uKnownFunc[26].theText := 'ATAN';
  uKnownFunc[26].equivChar := 'M';

  uKnownFunc[27].theText := 'ABS';
  uKnownFunc[27].equivChar := 'N';

  uKnownFunc[28].theText := 'COTG';
  uKnownFunc[28].equivChar := 'O';

  uKnownFunc[29].theText := 'COTAN';
  uKnownFunc[29].equivChar := 'O';

  uKnownFunc[30].theText := 'COT';
  uKnownFunc[30].equivChar := 'O';

  uKnownFunc[31].theText := 'ARGCH';
  uKnownFunc[31].equivChar := 'P';

  uKnownFunc[32].theText := 'ACOSH';
  uKnownFunc[32].equivChar := 'P';

  uKnownFunc[33].theText := 'ARCCOSH';
  uKnownFunc[33].equivChar := 'P';

  uKnownFunc[34].theText := 'ARCTANH';
  uKnownFunc[34].equivChar := 'Q';

  uKnownFunc[35].theText := 'ARGTANH';
  uKnownFunc[35].equivChar := 'Q';

  uKnownFunc[36].theText := 'ARGTH';
  uKnownFunc[36].equivChar := 'Q';

  uKnownFunc[37].theText := 'ATANH';
  uKnownFunc[37].equivChar := 'Q';

  uKnownFunc[38].theText := 'ARCSINH';
  uKnownFunc[38].equivChar := 'R';

  uKnownFunc[39].theText := 'ASINH';
  uKnownFunc[39].equivChar := 'R';

  uKnownFunc[40].theText := 'ARGSH';
  uKnownFunc[40].equivChar := 'R';

  uKnownFunc[41].theText := 'ARGSINH';
  uKnownFunc[41].equivChar := 'R';

  uKnownFunc[42].theText := 'ARGCOSH';
  uKnownFunc[42].equivChar := 'P';

  uKnownFunc[43].theText := 'INT';
  uKnownFunc[43].equivChar := 'S';

  uKnownFunc[44].theText := 'INTEGRAL';
  uKnownFunc[44].equivChar := 'S';

  for character := '#' to 'z' do
    case character of
      '(':
        uAttribution[character] := openParent;
      '#':
        uAttribution[character] := negation;
      ')':
        uAttribution[character] := closeParent;
      '+', '-', '*', '/', '^',';':
        uAttribution[character] := operationSymb;
      '1', 'p':
        uAttribution[character] := constantSymb;
      'q'..'z':
        uAttribution[character] := varSymb;
      'A'..'Z':
        uAttribution[character] := functionSymb;
    else
      uAttribution[character] := badThing;
    end; {of case}

  uSpecialValues[0] := Pi/2;
  uSpecialValues[1] := 2*Pi;
  uSpecialValues[2] := eMaxNum;
  uSpecialValues[3] := -eMaxNum;
  uSpecialValues[4] := eMinNum;
  uSpecialValues[5] := -eMinNum;
  uSpecialValues[6] := eMaxExp;
  uSpecialValues[7] := 1/2;

  u:= cA;
  v:= cB;
  qq[0]:= u + v;
  for i:= 1 to cQ do begin {initialize the coefficients that will serve to
      convert a grid of interpolating points to a grid of control points
      for a B-spline}
    u:= u*cU;
    v:= v*cV;
    qq[i]:= u+v;
    qq[-i]:= qq[i];
  end;
  theParser:= TFunctionParser_32.Create;
end;

function FactorizeText(aF: TFunctionParser_32; targetString: AnsiString; {}
  integralsAllowed: Boolean; var mayEndInFunction: Boolean): AnsiString;
{A block of consecutive alphabetic characters has been found and it is followed
 by a parenthesis if and only if 'mayEndInFunction = True'. Then, decompose it
 as a product of variables (and a final function if it is followed by a
 parenthesis. For instance, if  'vars' were  'ux' and the block were 'XCOS'
 followed by a parenthesis, then the result will be 'X*COS'}
var
  numberOfChars, hotSpot, count, position: Integer;
  currentChar: AnsiChar;
  presumedText: AnsiString;
  resultingString: AnsiString;
begin
  result := '';
  resultingString := '';
  numberOfChars := Length(targetString);
  hotSpot := numberOfChars;
  if mayEndInFunction and (numberOfChars > 1) then begin
    {we try to recognize a standard function first in the whole string,
     then in it minus the first char, then it minus the two first
     chars, etc.}
    count := numberOfChars;
    hotSpot := 1;
    repeat
      presumedText := Copy(targetString, hotSpot, count);
      currentChar := IdentifyFunction(presumedText);
      hotSpot := hotSpot + 1;
      count := count - 1;
    until (currentChar <> '?') or (count = 0);
    if (currentChar <> '?') then begin
      hotSpot := hotSpot - 2;
      if (currentChar= 'S') and not integralsAllowed then
      {user has submitted a text with the subexpression 'INT(' that
      can only be interpreted as meaning the beginning of
      the expression of an integral. If integrals are not allowed,
      then interrupt the parsing by raising an exception}
        raise Exception.Create(eNoIntegralsInsideIntegrals);

      mayEndInFunction := TRUE; {this is redundant}
      if (hotSpot > 0) then
        {hotSpot = 0 would mean that the function was recognized in
         the first pass and the routine would end here. Thus, here we
         are in a situation on which the initial part of 'targetString'
         must be the product of variables with perhaps the pair 'pi'}
        resultingString := Concat('*', currentChar)
           {and leave the parsing of the first part of 'targetString'
            to the code following the line below that begins by
                 'if hotSpot > 0'    }
      else begin
        resultingString := currentChar;
        result := resultingString;
        Exit;
      end;
    end
    else begin {No function was recognnized. Hence, it cannot
     end in a function, and we are yet to factorize all the string}
      hotSpot := numberOfChars;
      mayEndInFunction := FALSE;
    end;
  end {of IF mayEndInFunction ...}
  else
    mayEndInFunction := FALSE;
  if hotSpot > 0 then begin {we will go reading characters backwards,
     in order to discover an 'I' that cannot be a variable nor a
     parameter and in consequence should be preceded by a 'p'. Here
     we take the first step}
    currentChar := targetString[hotSpot];
    position := pos(currentChar, aF.allVars);
    if (position <> 0) then begin
      {we encode the variable by its position in the list of 'allVars'
      preceded by kLastchar}

      {We do not add a '*' after the position of the char because that
       times symbol had been added already}
      resultingString := Concat(kLastChar, chr(position), resultingString);
      if pos(currentChar, aF.usedVarsAndParams)=0 then
        {update the collection of variables and parameters actually used
         in the formula}
        aF.usedVarsAndParams:= aF.usedVarsAndParams + currentChar;
    end
    else begin {check if maybe it is a 'pi'}
      if (currentChar = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
        = 'P') then begin
        hotSpot := hotSpot - 1;
        resultingString := 'p' + resultingString;
        {Thus, the appearance of a 'p' will mean from now on that a
         factor equal to pi is present}
      end
      else {fatal error}
        raise Exception.Create(eBadFunctionOrVariables);
    end;
    hotSpot := hotSpot - 1;
  end;
  while hotSpot > 0 do begin  {We take the remaining steps}
    currentChar := targetString[hotSpot];
    position := pos(currentChar, aF.allVars);
    if (position <> 0) then begin
      resultingString := Concat(kLastChar, chr(position), '*',
        resultingString);
      if pos(currentChar, aF.usedVarsAndParams)=0 then
        aF.usedVarsAndParams:= aF.usedVarsAndParams + currentChar;
    end
    else begin {check if maybe it is a 'pi'}
      if (currentChar = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
        = 'P') then begin
        hotSpot := hotSpot - 1;
        resultingString := 'p*' + resultingString;
      end
      else {fatal error}
        raise Exception.Create(eBadFunctionOrVariables);
    end;
    hotSpot := hotSpot - 1;
  end;
  result := resultingString;
end;

procedure FillMargins(prov1: PArrayNodes1; prov2: PArrayArrayNodes2; dim: Integer);
var
  i, j: Integer;
begin
  if dim = 1 then begin
    for i:= 1 to cQ+2 do
      prov1^[cQ+2-i]:= 2*prov1^[cQ+2] - prov1^[cQ+2+i];
    for i:= 1 to cQ+3 do
      prov1^[cN1+cQ+1+i]:= 2*prov1[cN1+cQ+1] - prov1[cN1+cQ+1-i];
  end
  else begin
    for j:= cQ+2 to cN2+cQ+1 do begin
      for i:= 1 to cQ+2 do
        prov2^[j,cQ+2-i]:= 2*prov2^[j,cQ+2] - prov2^[j,cQ+2+i];
      for i:= 1 to cQ+3 do
        prov2^[j, cN2+cQ+1+i]:= 2*prov2^[j,cN2+cQ+1] - prov2^[j, cN2+cQ+1-i];
    end;
    for j:= 0 to cN2+2*cQ+4 do begin
      for i:= 1 to cQ+2 do
        prov2^[cQ+2-i,j]:= 2*prov2^[cQ+2,j] - prov2^[cQ+2+i,j];
      for i:= 1 to cQ+3 do
        prov2^[cN2+cQ+1+i,j]:= 2*prov2^[cN2+cQ+1,j] - prov2^[cN2+cQ+1-i,j];
    end;
  end;
end;

function GetEffectiveVars(allowedVars, list1, list2, list3: AnsiString;
               var var1, var2: Integer): integer;
{Used only for the functions that define an integral after they have been
 converted to a codified RPN notation}
  var
    charSet: Set of AnsiChar;
    i, ll: Integer;
    c: AnsiChar;
begin
  charSet:= [];
  if length(list1) > 0 then for i:= 1 to length(list1) do begin
    c:= list1[i];
    if c<>'R' then {letter 'R' denotes the integration variable and cannot be
                    considered a variable of the integrated function}
      charSet:= charSet + [c];
  end;
  if length(list2) >0 then for i:= 1 to length(list2) do
    charSet:= charSet + [ list2[i] ];
  if length(list3) >0 then for i:= 1 to length(list3) do
    charSet:= charSet + [ list3[i] ];
  {Now in 'charSet' there is the set of all characters appearing in the three
   functions that make the integral other than 'R'}
  ll:= Length(allowedVars);
  result:=0;
  var1:= -1;
  var2:= -1;
  if ll > 0 then for i:= 1 to ll do
    if allowedVars[i] in charSet then begin
      inc(result);
      if result=1 then var1:= i-1  {take note of the position of the first var
           in the list of allowed variables}
      else if result = 2 then var2:= i-1; {id for the second var}
      {if result = 2 we do not stop the loop, because the caller may be
       interested in the number 'result' of vars that effectively appear in
       the integral.}
    end;
end;

function inNumber(targetStr: AnsiString;
  var startChar: AnsiChar;
  var startPos: Integer): double;
{The parser has found a character that suggests the beginning of
 a number. This function parses the characters beginning in position 'startPos';
 if it encounters a number followed by other and separated of it by a blank
 space, then multiplies both. Returns the product of all those consecutive
 numbers, returns in 'startPos' the last position where the last character
 of the number was found, and in 'startChar' that last character.}
var
  valid, alreadyBlank, alreadyDot, foundOther: boolean;
  charPos, stringLength: Integer;
  theChar: AnsiChar;
  numStr: AnsiString;
  theNumber: double;
begin
  stringLength := length(targetStr);
  alreadyDot := false;
  valid := true;
  alreadyBlank := false;
  foundOther := False;
  numStr := startChar;
  charPos := startPos;
  theChar := startChar;
  theNumber := 1.0; {used as seed for multiply to it the numbers found}
  while (charPos < stringLength) and valid and (not foundOther) do begin
    inc(charPos);
    theChar := targetStr[charPos];
    case theChar of
      '0', '1', '2', '3', '4', '5', '6', '7', '8', '9':
        if not alreadyBlank then
          numStr := numStr + theChar {add 'theChar' to the number string}
        else begin {a valid number ended and another has begun}
          valid := true;
          theNumber := theNumber * StrToFloat(numStr); {multiply the number that
             ended by the saved numbers obtained in this way}
          numStr := theChar; {We begin again parsing a number}
          alreadyDot := false;
          alreadyBlank := false;
        end;
      '.':
        if alreadyDot or alreadyBlank then
        {in the case 'alreadyDot= True' the number would have two dots. In
         the case of having found previously a blank a new number would
         begin with a dot, but this is not permitted. Only formats as
         7.34, 234, or 0.52  are permitted}
          valid := False
        else begin
          alreadyDot := True; {the number must be followed by decimal chars}
          numStr := numStr + theChar; {add the dot to the number string}
        end;
      ' ':  {I think now that this case is a relic from previous versions,
             because 'TOCFunctDDDoubleSafeIntegrals.UnifyChars' will have
             eaten all blanks}
        if numStr[length(numStr)] = '.' then {the number would have a blank
             after a dot and this is not permitted}
          valid := False
        else
          alreadyBlank := true; {a blank has been encountered. Perhaps there is
          another number after this. So, keep going}
    else begin
        {the number has ended without flaws}
        foundOther := true;
        charPos := charPos - 1; {return to the char that
                                 ended the number string}
        theChar := targetStr[charPos];
      end;
    end;
  end;
  if valid then begin
    Result := theNumber * StrToFloat(numStr); {multiply the number by other
    previously found and return this value}
    startPos := charPos;
    startChar := theChar;
  end
  else
    raise Exception.Create(eBadNumber);
end;

function IntegrateRationalExtr(anF: RealFunction; var args: array of Double;
   r: Integer; tMin, tMax: Double): Double;
{Implementation of a method for numerical integration described in the
 book "Introduction to Numerical Analysis", by J. Stoer and R. Bulirsch,
 Springer, 1991}
const
  cMaxM = 24;
  cNum = 10;
  cTol = 1.0e-8;
var
  i,j,n,m: Integer;
  t0, h, a,b, err: Double;
  finished: Boolean;
  S: array[0..1] of Double;
  den: array[0..cMaxM] of Integer;
  T: array[0..cMaxM,-1..cMaxM] of double;
  jump: Double;
begin
  jump:= (tMax-tMin)/cNum;
  result:= 0;
  for i:= 0 to cMaxM do
    T[i,-1]:= 0;
  for n:= 1 to cNum do begin
    t0:= tMin + (n-1)*jump;
    args[r]:= t0;
    S[0]:= anF(args);
    args[r]:= t0+ jump;
    S[0]:= (S[0] + anF(args))/2;
    S[1]:= S[0];
    den[0]:= 1;
    T[0,0]:= jump*S[0];
    den[1]:= 2;
    h:= jump/den[1];
    args[r]:= t0 + h;
    S[1]:= S[1] + anF(args);
    T[1,0]:= h*S[1];
    den[2]:= 3;
    h:= jump/den[2];
    args[r]:= t0 + h;
    S[0]:= S[0] + anF(args);
    args[r]:= t0 + 2*h;
    S[0]:= S[0] + anF(args);
    T[2,0]:= h*S[0];
    a:= T[1,0]-T[0,-1];
    b:= sqr(den[1]/den[0])*(T[0,0]-T[0,-1])-a;
    if b<>0 then
      T[1,1]:= T[1,0] + a*(T[1,0]-T[0,0])/b
    else
      T[1,1]:= T[1,0];

    a:= T[2,0]-T[1,-1];
    b:= sqr(den[2]/den[1])*(T[1,0]-T[1,-1])-a;
    if (b<>0) then
      T[2,1]:= T[2,0] + a*(T[2,0]-T[1,0])/b
    else
      T[2,1]:= T[2,0];

    a:= T[2,1]-T[1,0];
    b:= sqr(den[2]/den[0])*(T[1,1]-T[1,0])-a;
    if b<>0 then
      T[2,2]:= T[2,1] + a*(T[2,1]-T[1,1])/b
    else
      T[2,2]:= T[2,1];
    err:= abs(T[2,2]*cTol);
    m:= 2;
    finished:= False;
    while not finished do begin
      inc(m);
      den[m]:= 2*den[m-2];
      h:= jump/den[m];
      j:= m mod 2;
      for i:= 0 to (den[m] div 2 -1) do begin
        args[r]:= t0 + + (2*i+1)*h;
        S[j]:= S[j] + anF(args);
      end;
      T[m,0]:= h*S[j];
      for i:= 1 to m-1 do begin
        a:= T[m,i-1]-T[m-1,i-2];
        b:= sqr(den[m]/den[m-i])*(T[m-1,i-1]-T[m-1,i-2])-a;
        if b<>0 then begin
          T[m,i]:= T[m,i-1]+ a*(T[m,i-1]-T[m-1,i-1])/b;
          finished:= (m>6) and (abs(T[m,i]-T[m-1,i]) < err);
          if finished then
            result:= result+ T[m,i];
        end
        else if a=0 then begin
          finished:= True;
          result:= result+ T[m,i-1];
        end
        else
          T[m,i]:= T[m,i-1];
        if finished then
          break;
      end;
      if not finished then begin
        a:= T[m,m-1]-T[m-1,m-2];
        b:= sqr(den[m]/den[0])*(T[m-1,m-1]-T[m-1,m-2])-a;
        if b<>0 then begin
          T[m,m]:= T[m,m-1]+ a*(T[m,m-1]-T[m-1,m-1])/b;
          finished:= (m >= cMaxM);
          if finished then
            result:= result+ T[m,m];
        end
        else begin
          finished:= True;
          result:= result+ T[m,m-1];
        end;
      end;
    end;
  end;
end;


function PieceEnd(const first: Integer; const theText, soughtChars:
  AnsiString; var symbol: AnsiChar): Integer;
{This function acts upon 'theText'. It is assumed that 'theText[first]'
 is not in the set [+,-,*,/,^], but possibly is '#'. Then, it scans forward
 'theText' until in reaches one of the symbols in 'soughtChars' while being at
 zero parentheses depth. Result is length[theText] if not found; if it
 was found, result is the position of the char previous to that symbol.
 In the first case, it returns symbol= '?', otherwise  symbol is the
 char previous}
var
  i, len, depth: Integer;
  ch: AnsiChar;
  found: Boolean;
begin
  symbol := '?';
  len := Length(theText);
  i := first - 1;
  depth := 0;
  found := False;
  while (i < len) and not found do begin
    inc(i);
    ch := theText[i];
    if (ch = '(') then
      inc(depth)
    else if (ch = ')') then
      dec(depth);
    if ((pos(ch, soughtChars) <> 0) and (depth = 0)) or (depth < 0) then
      {if depth < 0, then a parenthesis has closed and this is equivalent
       to the end of any piece}
      found := true;
  end;
  if found then begin
    result := i - 1;
    symbol := ch;
  end
  else
    result := len;
end;

function FactorString(theText: AnsiString): AnsiString;
{Parses a factor. A factor may be decomposed further if it is a power,
 that is it has the form (x+y)^(2 z). Then the program must parse
 (x+y), then  parse (2 z) and then apply the power function}
var
  neg: Boolean;
  redStr, anAtom: AnsiString;
  len, first, last: Integer;
  symbol, prevSymbol: AnsiChar;
begin
  if (theText = '') then
    Result := ''
  else begin
    neg := (theText[1] = '#');
    if neg then begin
      len := Length(theText) - 1;
      redStr := Copy(theText, 2, len);
    end
    else begin
      len := Length(theText);
      redStr := theText;
    end;
    if len = 0 then
      Result := ''
    else begin
      last := PieceEnd(1, redStr, '^', symbol);
      if last = len then {there is not a caret in 'theText'. Hence, it is
                          an atom}
        Result := AtomString(redStr)
      else begin {there is a caret}
        anAtom := Copy(redStr, 1, last);
        prevSymbol := symbol;
        result := AtomString(anAtom);
        repeat {perhaps there are more carets, for instance in (x+y)^3^z}
          first := last + 2;
          last := PieceEnd(first, redStr, '^', symbol);
          anAtom := Copy(redStr, first, last - first + 1);
          if (prevSymbol = '^') then
            Result := AtomString(anAtom) + Result + 'F*C'
               {'F*C', in reverse Polish notation must be understood as
                 apply Logarithm, multiply and then apply Exp. That is
                 we are applying the known formula x^z = exp(z ln(x))}
          else
            Result := AtomString(anAtom) + Result + prevSymbol;
          prevSymbol := symbol;
        until last = len;
      end;
      if neg then Result := Result + '#';
    end;
  end;
end;

function TermString(theText: AnsiString): AnsiString;
{Parses a term. A factor may be decomposed if it is a product
 or a quotient that is it has the form (x+y)*(2-z). Then the program
 must parse (x+y), then  parse (2-z) and then apply the times function}
var
  aFactor: AnsiString;
  len, first, last: Integer;
  symbol, prevSymbol: AnsiChar;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else begin
    last := PieceEnd(1, theText, '*/', symbol);
    if last = len then
      Result := FactorString(theText)
    else begin
      aFactor := Copy(theText, 1, last);
      prevSymbol := symbol;
      result := FactorString(aFactor);
      repeat {perhaps there are more factors, for instance in (x+y)zw}
        first := last + 2;
        last := PieceEnd(first, theText, '*/', symbol);
        aFactor := Copy(theText, first, last - first + 1);
        Result := Result + FactorString(aFactor) + prevSymbol;
        prevSymbol := symbol;
      until last = len;
    end;
  end;
end; {of TermString}

function ExpressionString(theText: AnsiString): AnsiString;
{Parses an expression. An expression may be decomposed if it is a summ
 or a subtraction that is it has the form (x+y) + (2-z). Then the program
 must parse (x+y), then  parse (2-z) and then apply the plus function}
var
  aTerm: AnsiString;
  len, first, last: Integer;
  symbol, prevSymbol: AnsiChar;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else begin
    last := PieceEnd(1, theText, '+-', symbol);
    if last = len then
      Result := TermString(theText)
    else begin
      aTerm := Copy(theText, 1, last);
      prevSymbol := symbol;
      result := TermString(aTerm);
      repeat
        first := last + 2;
        last := PieceEnd(first, theText, '+-', symbol);
        aTerm := Copy(theText, first, last - first + 1);
        Result := Result + TermString(aTerm) + prevSymbol;
        prevSymbol := symbol;
      until last = len;
    end;
  end;
end;



function LoadXd(i: Integer): AnsiString;
{Reads from the function stack the independent variable 'i' of Double
 type and stores it in ST(0). That is if the function is f(x,y,z) and
 i=2, then at the call to f(x,y,z), the system passes to f  a stack
 with the values of the parameters x,y,z. Then, LoadXd(2) reads from
 the stack the value of y and stores it in st(0)}
begin
  if i = 1 then
    result:= #$DD#$00
  else if i > 1 then
    result:= #$DD#$40 + chr(SizeOf(Double) * (i-1));
end;

function sn(n: Pointer): AnsiString;
{Converts a pointer to a string in the format that the CPU needs to
 read correctly a memory address}
var
  c: PByte;
  i: Integer;
begin
  c:= @n;
  Result := '';
  for i := 0 to SizeOf(Pointer) - 1 do
    Result := Result + AnsiChar(c[i]);
end;

function LoadDd(const a): AnsiString;
{Reads a Double from the address of the var 'a' and stores it in ST(0)}
begin
  result:= #$DD#$05 + sn(@a);
end;

function WriteDd(const a): AnsiString;
{Writes ST(0), as an Double, into the memory location @a, but doesn't pop}
begin
  result:= #$DD#$15 + sn(@a);
end;

function WriteDdP(const a): AnsiString;
{Writes ST(0), as an Double, into the memory location @a, and pops ST(0)}
begin
  result:= #$DD#$1D + sn(@a);
end;

{-------------------------------------------------------------------}
{Procedures and functions of the class TFunctionParser_32}
{-------------------------------------------------------------------}


procedure TFunctionParser_32.AddAddressToCode(p: PByte; var
   cod: AnsiString);
{Adds to 'cod' the address written in 'p', in the format that is
required by the CPU}
var
  s: AnsiString;
  i: Integer;
begin
  s := '';
  for i := 0 to SizeOf(Pointer) - 1 do
    s := s + AnsiChar(p[i]);
  cod := cod + s;
end;

procedure TFunctionParser_32.AddIntToCode(n: Integer; var
   cod: AnsiString);
{Adds to 'cod' the code of the integer 'n', in the format that is
required by the CPU}
var
  c: PByte;
  s: AnsiString;
  i: Integer;
begin
  c := @n;
  s := '';
  for i := 0 to SizeOf(Integer) - 1 do
    s := s + AnsiChar(c[i]);
  cod := cod + s;
end;

procedure TFunctionParser_32.AdjustIntegralDataForBSplines(prov1: PArrayNodes1;
        prov2: PArrayArrayNodes2; dim: Integer);
{From data stored in 'provData' that consists of values at equal intervals of the
 function that is to be interpolated, the procdedure builds the control values
'intData' that will be directly used in the evaluation by de Boor algorithm.
 See the attached paper "AproxFuncBSplinesEnglish.pdf"}
  var
    i, j, k: Integer;
    dd: PArrayArrayNodes2;
begin
  if dim=1 then
    for i:= cQ-2 to cN1+cQ+2 do begin
      intList1[lastIntegral1].Data[0,i]:= 0;
      for k:=-cQ to cQ do
        intList1[lastIntegral1].Data[0,i]:= intList1[lastIntegral1].Data[0,i]+qq[k]*prov1^[i+k+2];
    end
  else begin
    New(dd);
    for j:= 0 to cN2 + 2*cQ+ 4 do
      for i:= cQ-2 to cN2 + cQ+ 2 do begin
        dd^[i,j]:= 0;
        for k:= -cQ to cQ do
          dd^[i,j]:= dd^[i,j]+ qq[k]*prov2^[i+k+2,j];
      end;
    for i:= cQ-2 to cN2+ cQ+2 do
      for j:= cQ-2 to cN2+cQ+2 do begin
        intList2[lastIntegral2].Data[0,0,i,j] := 0;
        for k:= -cQ to cQ do
          intList2[lastIntegral2].Data[0,0,i,j]:=
                   intList2[lastIntegral2].Data[0,0,i,j]+ qq[k]*dd^[i,j+k+2];
      end;
    Dispose(dd);
  end;
end;


function TFunctionParser_32.BuildIntegralDataList(tripleArgument:
     AnsiString):Integer;
{An integral is build by three functions. The integrand and the limits. Here,
 we separate the strings containing those functions and parse them but now
 specifying that no integrals are allowed. Then compute the data for the
 corresponding B-splines. Result is   lastIntegral1 + (lastIntegral2 +1) shl 4.
 This means among other things that de maximum number of integrals of each type
 in a single function is 16.}
  var
    len, i, i0: Integer;
    c: AnsiChar;
    funcText: AnsiString;
    integrand, inferior, superior: TFunctionParser_32;
    mc: array of double;
begin
  result:= -1;
  len:= Length(tripleArgument);
  i:=0;
  mc:= nil;
  inferior := TFunctionParser_32.create;
  superior := TFunctionParser_32.create;
  integrand := TFunctionParser_32.create;
  try
    repeat  {Find the semicolon}
      inc(i);
      c:= tripleArgument[i];
    until (c=';') or (i = len);
    if c=';' then begin
      funcText:= Copy(tripleArgument,1,i-1);
      {Parse integrand}
      integrand.ParseNewText(funcText, allowedVars+'r','', False, mc, mc);
    end;
    if (i > len-3) then
      raise exception.Create(eBadIntegralSyntax);
    i0:= i+1;
    repeat  {Find the semicolon}
      inc(i);
      c:= tripleArgument[i];
    until (c=';') or (i = len);
    if c=';' then begin
      funcText:= Copy(tripleArgument,i0,i-i0);
      {Parse inferior limit}
      inferior.ParseNewText(funcText, allowedVars, '', False, mc, mc);
    end;
    if (i > len-1) then
      raise exception.Create(eBadIntegralSyntax);
    i0:= i+1;
    funcText:= Copy(tripleArgument,i0,len-i0+1);
    {Parse superior limit}
    superior.ParseNewText(funcText, allowedVars,'', False, mc, mc);

    ComputeIntegralData(integrand, inferior,superior);

    {Once the integral data have been computed, these functions are
     useless}
    finally
      integrand.Free;
      inferior.Free;
      superior.Free;
  end;
  result:= lastIntegral1+1 + (lastIntegral2+1) shl 4;
end;


procedure TFunctionParser_32.CheckFunction;
{Checks the sintaxis of the function, by comparing a char with the
 previous one}
var
  lastRead, reducedLength: integer;
  species, speciesPrev: pieceClass;
  charac: AnsiChar;
  stopHere: boolean;

  function hasError: boolean;
  var
    foundError, usual: boolean;
  begin
    foundError := false;
    if species = badThing then
      foundError := true
    else begin
      usual := (species = operationSymb) or (species = closeParent);
      case speciesPrev of
        openParent:
          foundError := usual;
        closeParent, constantSymb, varSymb:
          foundError := not usual;
        negation, operationSymb:
          foundError := usual or (species = negation);
        functionSymb:
          foundError := (species <> openParent);
      end; {of case}
    end; {of else}
    result := foundError;
  end; {of hasError}

begin {CheckFunction}
  lastRead := 0;
  reducedLength := Length(textReduced);
  speciesPrev := openParent;
  repeat
    lastRead := lastRead + 1;
    charac := textReduced[lastRead];
    if charac = kLastChar then begin
      inc(lastRead);
      species := constantSymb;
    end
    else
      species := uAttribution[charac];
    stopHere := hasError;
    speciesPrev := species;
  until (lastRead = reducedLength) or stopHere;
  if (not stopHere) then begin
    species := closeParent;
    stopHere := hasError;
  end;
  if stopHere then
    raise Exception.Create(eSyntaxError);
end; {CheckFunction}

procedure TFunctionParser_32.CodifyFunction;
{See documentation at BuildFirstDerivative}
var
  i, j: Integer;
  k: Int64;
  prot: Word;
  ch: AnsiChar;
  c: PAnsiChar;
  escape: Boolean;
  n: array of Integer;
  lstn: Integer;
begin
  j:= $1000;
  if VBlock = nil then begin
    prot:= PAGE_EXECUTE_READWRITE;
    VBlock:= VirtualAlloc(nil, j, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(VBlock, j, PAGE_EXECUTE_READWRITE, @prot);
  end;
  lstn:= -1;
  waitingInApplStack:= 0;
  fCode := '';
  emptyRegisters := c_LastReg + 1;
  i := 0;
  while i < lengthActorList do begin
    inc(i);
      ch := actorList[i];
      escape := (ch = kLastChar);
      if escape or (ch = 'p') or (ch = '1') then
        begin
          if escape then inc(i);
          ch := actorList[i];
          LeaveRoomInFPUStack(1, fCode);
          if escape then
            begin
              if (ord(ch) > argNum) then
                fCode := fCode + LoadDd(data[ord(ch) - argNum])
              else if (ord(ch)<= argNum) then
                fCode := fCode + LoadXd(ord(ch));
            end
          else if (ch = 'p') then
            fCode := fCode + c_LoadPi
          else
            fCode := fCode + c_Load1;
          dec(emptyRegisters);
        end
      else
        begin
          if (ch in ['+', '-', '*', '/', '^']) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack(fCode);
              case ch of
                '+' :
                  fCode := fCode + #$DE#$C1;
                '-' :
                  fCode := fCode + #$DE#$E9;
                '*' :
                  fCode := fCode + #$DE#$C9;
                '/' :
                  begin
                    LeaveRoomInFPUStack(1, fCode);
                    fCode := fCode + c_div_S;
                  end;
                '^' :
                  begin
                    LeaveRoomInFPUStack(1, fCode);
                    fCode := fCode + c_zero_comp +
                      #$72#$08#$DD#$D8#$DD#$D8#$D9#$EE#$EB#$42#$D9#$F1 +
                      c_2toX_S;
                  end;
              end;
              inc(emptyRegisters);
            end
      else {opCode of a function}  begin
        if (ch in ['A', 'B', 'C', 'E', 'F', 'G', 'M']) then
          LeaveRoomInFPUStack(1, fCode)
        else if (ch in ['H', 'I', 'J', 'K', 'L', 'O', 'P', 'Q', 'R']) then
          LeaveRoomInFPUStack(2, fCode)
        else if (ch ='S') then begin
          LeaveRoomInFPUStack(c_LastReg+1, fCode);
        end;
        case ch of
           '#' :
                  fCode := fCode + #$D9#$E0;
                'A' :
                  fCode := fCode + c_mod_2pi + #$D9#$FF;
                'B' :
                  fCode := fCode + c_mod_2pi + #$D9#$FE;
                'C' :
                  fCode := fCode + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S;
                'D' :
                  fCode := fCode + #$D8#$C8;
                'E' :
                  fCode := fCode + c_zero_comp +
                    #$72#$06 +
                    #$DD#$D8#$D9#$EE#$EB#$02#$D9#$FA;
                'F' :
                  fCode := fCode + c_ln_S;
                'G' :
                  fCode := fCode + c_mod_pi + #$D9#$F2#$DD#$D8;

                'H' :
                  fCode := fCode + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D9#$C0#$D9#$E8#$DE +
                    #$F1#$DE#$E9+ c_div2;
                'I' :
                  fCode := fCode + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D9#$C0#$D9#$E8#$DE +
                    #$F1#$DE#$C1+ c_div2;
                'J' :
                  fCode := fCode + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D8#$C8#$D9#$E8#$D8#$E9#$D9#$C9 +
                    #$D9#$E8#$DE#$C1#$DE#$F9;
                'K' :
                  fCode := fCode + #$D9#$C0#$D9#$E1 +
                    c_1_comp +
                    #$DD#$D8#$72#$0E#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1#$D9#$FA#$D9#$F3 +
                    #$EB#$10 + c_zero_comp + #$DD#$D8 +
                    LoadDd(uSpecialValues[0]) +
                    #$72#$02#$D9#$E0;
                'L' :
                  fCode := fCode + #$D9#$C0#$D9#$E1 +
                    c_1_comp +
                    #$DD#$D8#$72#$10#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1#$D9#$FA#$D9#$C9 +
                    #$D9#$F3#$EB#$0E + c_zero_comp +
                    #$DD#$D8#$72#$04#$D9#$EB#$EB#$02 +
                    #$D9#$EE;
                'M' : fCode := fCode + #$D9#$E8#$D9#$F3;
                'N' : fCode := fCode + #$D9#$E1;
                'O' :
                  fCode := fCode + c_mod_pi +
                    #$D9#$FB#$D9#$C9 + c_div_S;
                'P' :
                  fCode := fCode + c_1_comp +
                    #$77#$18#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E9#$D9#$FA#$DE#$C1 +
                    c_ln + #$EB#$04#$DD#$D8#$D9#$EE;
                'Q' :
                  fCode := fCode +
                    #$D9#$C0#$D9#$E1 + c_1_comp + #$DD#$D8#$76#$24
                    +
                    #$D9#$C0#$D9#$E8#$DE#$C1#$D9#$C9#$D9#$E8#$DE#$E1#$DE#$F9 +
                    c_ln + c_div2 + #$EB#$16
                    + c_zero_comp +
                    #$DD#$D8#$77#$08 +LoadDd(uSpecialValues[2]) +
                    #$EB#$06 + LoadDd(uSpecialValues[3]);
                'R' :
                  fCode := fCode + c_zero_comp +
                    #$D9#$E1#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$C1
                    +
                    #$D9#$FA#$DE#$C1 + c_ln +
                      #$76#$02#$D9#$E0;
        else
          ;
        end;
      end;
    end;
  end;

  fCode := fCode + #$C3;
  PutCodeInExeBlock(fCode, VBlock);
  @V := @VBlock[0];
  fCode:= '';
  n:= nil;
end;

function TFunctionParser_32.ComputeExponent(var n: Integer; ini, fini:
  Integer): Boolean;
  {Gives true if the exponent is an integer and then the exponent value
   is assigned to 'n'}
var
  c1, c2: AnsiChar;
  negate: boolean;
  firstExp, lastExp, shift: Integer;
  exponent: double;
begin
  firstExp := ini;
  lastExp := fini;
  c1 := textReduced[firstExp];
  c2 := textReduced[lastExp];
  if (c1 = '(') then begin
    result := (c2 = ')');
    if result then begin
      inc(firstExp);
      dec(lastExp);
      c1 := textReduced[firstExp];
      c2 := textReduced[lastExp];
    end;
  end;
  negate := (c1 = '#');
  if negate then begin
    inc(firstExp);
    c1 := textReduced[firstExp];
  end;

  if (lastExp = firstExp) and (c1 = '1') then begin
    n := 1;
    Result := True;
  end
  else if ((lastExp - firstExp) = 1) and (c1 = kLastChar) and (ord(c2) >
    argNum+parNum)
    then begin {the exponent is a constant or parameter}
    shift := ord(c2) - argNum;
    exponent := data[shift];
    n := round(exponent);
    Result := (n = exponent);
  end
  else
    result := False;
  if result and negate then n := -n;
end;


procedure TFunctionParser_32.ComputeIntegralData(
             integrand, inferior,superior: TFunctionParser_32);
{Computes the integral values at a grid of points. That grid with the
 computed values will be then the interpolating points for a curve or
 surface of B-splines. Approximate evaluation of the integral will be
 done afterwards by using the de Boor algorithm. See the attached paper
 "AproxFuncBSplinesEnglish.pdf"}
var
  provData1: PArrayNodes1;
  provData2: PArrayArrayNodes2;
  i, j, k, var1, var2: Integer;
  x0, infE, supE: Double;
  argum: array of double;
begin
  {Check the number of effective independent variables that appear in the
  functions defining the integral. That number must be 1 or 2}
  i:= GetEffectiveVars(integrand.allowedVars, integrand.usedVarsAndParams, inferior.usedVarsAndParams, superior.usedVarsAndParams,
    var1, var2);
  if i>2 then
    raise Exception.Create(eAtMostTwoVarsInInt);
  if i=0 then
    raise Exception.Create(eNaughtIntegral); {the integral is constant. The user
                                              should compute it by other means}

  if var2 < 0 then begin
    inc(lastIntegral1);
    if High(intList1) < lastIntegral1 then
      SetLength(intList1, lastIntegral1+1);
    IntList1[lastIntegral1].var1:= var1;
  end
  else begin
    inc(lastIntegral2);
    if High(intList2) < lastIntegral2 then
      SetLength(intList2, lastIntegral2+1);
    IntList2[lastIntegral2].var1:= var1;
    IntList2[lastIntegral2].var2:= var2;
  end;

  SetLength(argum, argNum+1);
  for i:= 0 to argNum do
    argum[i]:=0;

  if var2 = -1 then begin {there is only and independent variable. The domain of
                           the function defined by the integral is a segment}
    New(provData1);

    {Calculate values of the integral at cN1+1 evenly spaced points, such that
     the distance between minCorner[var1] and the first point is half the
      separation between the points of the grid, and the same at the end}
     argum[var1]:= minCorner[var1]-jump1[var1]/2;
    for i:= cQ+2 to cN1+cQ+1 do begin
      argum[var1]:= argum[var1] + jump1[var1];
      infE:= inferior.V(argum);
      supE:= superior.V(argum);
      provData1^[i]:=IntegrateRationalExtr(integrand.V, argum, argNum, infE, supE);
    end;
    FillMargins(provData1, nil, 1);
    AdjustIntegralDataForBSplines(provData1, nil, 1);
  end
  else begin
    New(provData2);
    x0:= minCorner[var1]-jump2[var1]/2;
    argum[var2]:= minCorner[var2]-jump2[var2]/2;
    for j:= cQ+2 to cN2+cQ+1 do begin
      argum[var2]:= argum[var2] + jump2[var2];
      argum[var1]:= x0 ;
      for i:= cQ+2 to cN2+cQ+1 do begin
        argum[var1]:= argum[var1] + jump2[var1];
        infE:= inferior.V(argum);
        supE:= superior.V(argum);
        provData2^[i,j]:=IntegrateRationalExtr(integrand.V, argum, argNum, infE, supE);
      end;
    end;
    FillMargins(nil, provData2, 2);
    AdjustIntegralDataForBSplines(nil, provData2, 2);
  end;
  {Prepare the B-splines control points that make the calculted
   values interpolation values with good approximation}


  argum:= nil;
  if var2 < 0 then
    dispose(provData1)
  else
    dispose(provData2);
end;

function TFunctionParser_32.ConvertIntegral(var hotSpot: Integer): AnsiChar;
{Extract the string containing the definition of the integral and convert the
 integral to a grid of control points of a B-Spline, ready to be evaluated by the
 de Boor algorithm. The result is the place of the integral in the list of
 integrals that appear in the function}
var
  tripleArgument: AnsiString;
  reducedLength, n, depth: Integer;
  c: AnsiChar;
begin
  result:= chr(0);
  reducedLength := Length(textReduced);
  if reducedLength < 8 then
    Exit;
  depth:= 1;
  tripleArgument:= '';
  repeat
    inc(hotSpot);
    c:= textReduced[hotSpot];
    if c='(' then Inc(depth);
    if c= ')' then Dec(depth);
    if depth > 0 then
      tripleArgument := tripleArgument+c;
  until (depth = 0) or (hotSpot=reducedLength);
  if (depth =0) and (tripleArgument <> '') then begin
    {the triple argumen string has been extracted. Build the B-splines}
    n:= BuildIntegralDataList(tripleArgument);
    result:= AnsiChar(n);
  end;
end;

procedure TFunctionParser_32.ConvertPowers;
{Scan the 'textReduced' for the appearance of carets '^', and send
 the correponding exponents to convert them to optimal composition of
 squares and multiplies if they are integer}
var
  newString: AnsiString;
  reducedLength, firstBase, lastBase, beginsExp, endsExp, i, n: Integer;
  c: AnsiChar;
  good: Boolean;
begin
  reducedLength := Length(textReduced);
  if reducedLength < 3 then
    Exit;
  i := 1;
  repeat  {scan  for '^'}
    inc(i);
    c := textReduced[i];
    if (c = '^') then begin
      beginsExp := i + 1;

      {Find the end of the exponent, 'endsExp'}
      endsExp := PieceEnd(beginsExp, textReduced, '+-*/^', c);

      {Check if exponent is an integer}
      good := computeExponent(n, beginsExp, endsExp);

      if good then begin  {Exponent is an integer}
        lastBase := i - 1;
        firstBase := PieceBeginning(lastBase, '+-*/^#');
        newString := Copy(textReduced, firstBase, lastBase - firstBase + 1);

        {Convert the power expression to an optimal composition of squares and
         multiplies}
        newString := ConvertPowerSymbol(newString, n);
        Insert(newString, textReduced, endsExp + 1);
        Delete(textReduced, firstBase, endsExp - firstBase + 1);
        reducedLength := Length(textReduced);
        i := firstBase + Length(newString) - 1;
      end
    end;
  until (i > reducedLength - 1);
end;

constructor TFunctionParser_32.Create;
begin
  inherited Create;
  rawText := '';
  fCode := '';
  InitializeCodeVars;
end;

destructor TFunctionParser_32.Destroy;
begin
  rawText := '';
  fCode := '';
  data := nil;
  minCorner:= nil;
  maxCorner:= nil;
  jump1:= nil;
  invJump1:= nil;
  jump2:= nil;
  invJump2:= nil;
  intList1:= nil;
  intList2:= nil;
  if VBlock <> nil then begin
    VirtualFree(VBlock, 0, MEM_RELEASE);
    VBlock:= nil;
  end;
  inherited Destroy;
end;

procedure TFunctionParser_32.GetRoomAndCoeffsFirstDeriv;
{Computes and stores the interpolating points needed for calculating
 the first derivatives of the integrals involved in the function}
  var
    i, j, a, b, v1, v2: Integer;
begin
  if lastIntegral1 >= 0 then
    for i:= 0 to lastIntegral1 do begin
      v1:= IntList1[i].var1;
      for a:= cQ-1 to cNodes1 do
        IntList1[i].data[1,a]:=
             (IntList1[i].data[0,a]-IntList1[i].data[0,a-1])*invJump1[v1];
    end;
  if lastIntegral2 >= 0 then
    for i := 0 to lastIntegral2 do begin
      v1:= IntList2[i].var1;
      v2:= IntList2[i].var2;
      for a:= cQ-1 to cNodes2 do
        for b:= cQ-2 to cNodes2 do
          IntList2[i].data[1,0,a,b]:=
             (IntList2[i].data[0,0,a,b]-IntList2[i].data[0,0,a-1,b])*invJump2[v1];
      for a:= cQ-2 to cNodes2 do
        for b:= cQ-1 to cNodes2 do
          IntList2[i].data[0,1,a,b]:=
             (IntList2[i].data[0,0,a,b]-IntList2[i].data[0,0,a,b-1])*invJump2[v2];
    end;
end;

procedure TFunctionParser_32.GetRoomAndCoeffsSecondDeriv;
{Reserves the memory necessary for storing the interpolating points needed for
 calculating the second derivatives of the integrals involved in the function}
  var
    i,j,a,b,len, v1, v2: Integer;
begin
  If lastIntegral1 >= 0 then
    for i:= 0 to lastIntegral1 do begin
      v1:= IntList1[i].var1;
      for a:= cQ to cNodes1 do
        IntList1[i].data[2,a]:=
             (IntList1[i].data[1,a]-IntList1[i].data[1,a-1])*invJump1[v1];
    end;
  if lastIntegral2 >= 0 then
    for i:= 0 to lastIntegral2 do begin
      v1:= IntList2[i].var1;
      v2:= IntList2[i].var2;
      for a:= cQ to cNodes2 do
        for b:= cQ-2 to cNodes2 do
          IntList2[i].data[2,0,a,b]:=
              (IntList2[i].data[1,0,a,b]-IntList2[i].data[1,0,a-1,b])*invJump2[v1];
      for a:= cQ-2 to cNodes2 do
        for b:= cQ to cNodes2 do
          IntList2[i].data[0,2,a,b]:=
              (IntList2[i].data[0,1,a,b]-IntList2[i].data[0,1,a,b-1])*invJump2[v2];
      for a:= cQ-1 to cNodes2 do
        for b:= cQ-1 to cNodes2 do
          IntList2[i].data[1,1,a,b]:=
              (IntList2[i].data[1,0,a,b]-IntList2[i].data[1,0,a,b-1])*invJump2[v2];
    end;
end;

procedure TFunctionParser_32.InitializeCodeVars;
begin
  c_div_S := c_zero_comp + #$75#$0C#$DD#$D8 + LoadDd(uSpecialValues[2]) +
    #$DE#$C9#$EB#$02#$DE#$F9;
  {Safe division. 20 bytes. Requires ST(c_LastReg) clear.
      By "safe" I mean that if the denominator is zero, the result is an enormous
      number, so that perhaps a 'division by zero' is avoided}

  c_ln_S := c_zero_comp + #$72#$0C#$DD#$D8 + LoadDd(uSpecialValues[6]) +
    #$D9#$E0#$EB#$0A + c_ln;
  {Safe ln(ST(0)). 26 bytes. Requires ST(c_LastReg) clear. By safe I mean that
     if argument is less or equal to zero, the result is a negative number
     of the order of the maximum exponent A that makes 1.0eA a real of single type}

  c_2toX_S := #$D9#$C0#$D9#$E1 + LoadDd(uSpecialValues[6]) + #$DF#$F1 +
    #$DD#$D8#$72#$1A + c_2toX + #$EB#$16 + c_zero_comp + #$DD#$D8#$72#$08
    +
    LoadDd(uSpecialValues[4]) + #$EB#$06 + LoadDd(uSpecialValues[2]);
  {Safe 2^ST(0). 64 bytes. Requires two last FPU registers empty. By safe,
      I mean that when the exponent is greater than some quantity, the result
      is an enormous number or an enormously little number}
end;

procedure TFunctionParser_32.ParseNewText(t, vars, pars: AnsiString;
    integralsAllowed: Boolean; minC, maxC: array of double);
{Parses and builds the function given in the string 't'. Each
 character in the string 'vars' identifies one of the possible arguments
 on which the function may depend. Each character in the string 'pars'
 identifies one of the possible parameters that may appear in the function. If
 'integralsAllowed' is True, 't' may contain integrals and then the arguments
 'minC' and 'maxC' describe the domain where the variables can move. If
 'integralsAllowed' is False, then 'minC' and 'maxC' are useless, and the
 appearance of an integral in 't' will raise an exception.

 Warning: don't repeat chars in 'vars' or 'pars'.}
var
  i,len: Integer;
  ch: AnsiChar;
begin
  rawText := t;
  formula := t;
  allowedVars:= '';
  lastIntegral1:= -1;
  lastIntegral2:= -1;
  if vars = '' then begin
    {if user forgot initialize 'vars' then all chars between p and z
     will be admitted as variables}
    allowedVars := variables;
    argNum := Length(variables);
  end
  else begin
    allowedVars := vars;
    StrUpper(PAnsiChar(allowedVars));
    argNum := Length(allowedVars);

    {Check that the variable names are among those allowed}
    for i := 1 to argNum do begin
      ch := allowedVars[i];
      if pos(ch, variables)=0 then
        raise Exception.Create(eBadVarSet);
    end;
  end;
  allowedPars := pars;

  {Check that the parameter names are among those allowed and
   that none of them has been included also as a variable}
  if pars <> '' then begin
    StrUpper(PAnsiChar(allowedPars));
    parNum := Length(allowedPars);
    for i := 1 to parNum do begin
      ch := allowedPars[i];
      if (pos(ch,allowedVars)<>0) or (pos(ch, variables) = 0) then
        raise Exception.Create(eBadParamSet);
    end;
  end;
  allVars:= allowedVars + allowedPars;

  {Get room for the maximum value numbers: for position, velocity
  and acceleration}
  usedVarsAndParams:= '';
  data := nil;

  {Get room for parameters and for seven constants more. If the formula
   needs more, it will be obtained latter}
  SetLength(data, parNum + 8);
  {data[0] is not used}
  data[0] := 0;

  if parNum > 0 then for i:= 1 to parNum do
    data[i]:=0; {initialize the values of parameters as zero. Any
    other update is to be done via a call to 'UpdateParameters'}

  dataLast := parNum; {last place in 'data' that has been occupied}

  if integralsAllowed then begin  {get room for several variables}
    len:= Length(minC);
    SetLength(minCorner,len);
    SetLength(maxCorner,len);
    SetLength(jump1,len);
    SetLength(invJump1,len);
    SetLength(jump2,len);
    SetLength(invJump2,len);
    if len >0 then
      for i:= 0 to len-1 do begin
        minCorner[i]:= minC[i];
        maxCorner[i]:= maxC[i];
        jump1[i]:= (maxCorner[i]-minCorner[i])/cN1;
        invJump1[i]:= 1/jump1[i];
        jump2[i]:= (maxCorner[i]-minCorner[i])/cN2;
        invJump2[i]:= 1/jump2[i];
      end;
  end;

  UnifyChars;
  ReduceSymbols(integralsAllowed);
  formula:= '';
  CheckFunction;
  ConvertPowers;
  actorList := expressionString(textReduced); {this makes the RPN expression}
  textReduced:= '';
  lengthActorList := Length(actorList);
  CodifyFunction; {this builds the object code to be excuted by the function
                   'V' that evaluates the parsed function}
  if (lastIntegral1 >= 0) or (lastIntegral2 >= 0) then begin
    GetRoomAndCoeffsFirstDeriv;
    GetRoomAndCoeffsSecondDeriv;
  end;
end;

function TFunctionParser_32.PieceBeginning(const last: Integer; soughtChars: AnsiString):
  Integer;
{This function acts upon 'textReduced'. It is assumed that 'textReduced[last]'
is not in the set [+,-,*,/,^]. Then, it scans backwards
'textReduced' until in reaches one of the symbols in 'endChars' while being at
<=zero parentheses depth. Result is 1 (if not found) or the index following that symbol.}
var
  i, depth: Integer;
  ch: AnsiChar;
  found: Boolean;
begin
  i := last + 1;
  depth := 0;
  found := False;
  while (i > 1) and not found do begin
    dec(i);
    ch := textReduced[i];
    if (ch = ')') then
      inc(depth)
    else if (ch = '(') then
      dec(depth);
    if ((pos(ch, soughtChars) <> 0) and (depth = 0)) or (depth < 0) then
      found := true;
  end;
  if found then
    result := i + 1
  else
    result := 1;
end;

PROCEDURE TFunctionParser_32.LeaveRoomInFPUStack(neededPlaces : Integer; var cod: AnsiSTring);
BEGIN
  IF emptyRegisters >= neededPlaces THEN
    Exit;
  IF emptyRegisters = 0 THEN BEGIN
      cod := cod
       + #$D9#$F6 {fdecstp: ST(c_LastReg] becomes ST(0)}
       + #$83#$EC+ c_stackDisplacDoubleHStr
                                  {subtract c_stackDisplacDouble from ESP}
          +#$DD#$1C#$24;  {fstp [ESP]}
      inc(emptyRegisters);
  END;
  IF emptyRegisters < neededPlaces then
    REPEAT
      cod := cod + #$D9 + chr($C0 + c_LastReg - emptyRegisters)
          {push last occupied ST to ST(0)}
      + #$83#$EC+ c_stackDisplacDoubleHStr
                                  {subtract c_stackDisplacDouble from ESP}
      +#$DD#$1C#$24;  {fstp [ESP]}
      cod := cod + #$DD + chr($C0 + c_LastReg - emptyRegisters);
         {last occupied place contains the former value: declare it empty}
      inc(emptyRegisters);
    UNTIL emptyRegisters = neededPlaces;
END;


PROCEDURE TFunctionParser_32.PopFromAppStack(var cod: AnsiSTring);
VAR
  m : byte;
BEGIN
  m := c_LastReg +1 - emptyRegisters;
  cod := cod + #$DD + chr($D8 + m) {copy ST(0) to ST(m), which
     is the first free register, and pop st(0)}

   + #$DD#$04#$24 {fld [ESP]}
   + #$83#$C4 + c_stackDisplacDoubleHStr {add c_stackDisplacDouble to ESP};
  {if using Extended type, use 'c_stackDisplacExtendedHStr' instead}

  cod := cod + #$D9 + chr($C8 + m); {exchange ST(m) with ST(0)}
  dec(emptyRegisters);
END;

procedure TFunctionParser_32.PutCodeInExeBlock(const cod: AnsiString;
     var exeBlock: PByte);
var
  len: Integer;
  i: Integer;
  prot: Word;
begin
  len:= Length(cod);
  prot:= PAGE_EXECUTE_READWRITE;
  for i:= 0 to len-1 do
    exeBlock[i]:= Byte(cod[i+1]);
  VirtualProtect(exeBlock, len + SizeOf(exeBlock), PAGE_EXECUTE_READ, @prot);
end;

procedure TFunctionParser_32.ReduceSymbols(integralsAllowed: Boolean);
{Substitutes the name of each function by a single character; any number or
parameter by an "escape character", 'kLastChar', followed by a byte in certain range, that
signals the place where the actual number is stored in the array 'data'; an
independent variable by 'kLastChar', followed by another byte in other range.
Also, each integral is processed to build an array of data that
will define the B-splines that will be used in evaluating it, and it is
substituted by 'S('+kLastChar+ n+')', where 'n' means the byte that signals
the place of that integral in the list of integrals that appear in the function}
var
  numberOfChars, hotSpot, reducedLength: Integer;
  currentChar: AnsiChar;
  letterStr: AnsiString;
  convertedStr, resultingString: AnsiString;
  numberBegins, inLetter, mayHaveFunction, foundEqualNumber: Boolean;
  realValue: double;
  aByte: Byte;
begin {ReduceSymbols}
  resultingString := '';
  reducedLength := Length(textReduced);
  numberOfChars := reducedLength;
  hotSpot := 0;
  repeat
    hotSpot := hotSpot + 1;
    currentChar := textReduced[hotSpot];
    inLetter := (pos(currentChar, upCaseLetters) <> 0);
    numberBegins := (pos(currentChar, digits) <> 0) or (currentChar = '.');
    if (not inLetter) and (not numberBegins) then {it is a symbol as '+' or ')'}
      ResultingString := Concat(resultingString, currentChar);
    if numberBegins then begin
      realValue := 0;
      if currentChar = '.' then {a number cannot begin with a dot}
        currentChar:= '?'
      else
        realValue := inNumber(textReduced, currentChar, hotSpot);
      if (currentChar <> '?') then begin {a number with value 'realValue'
                                          appears in the formula}
        if realValue = 1 then
          resultingString := resultingString + '1' {the 1 is treated differently}
        else begin {check if the same number had appeared already in the formula}
          foundEqualNumber := False;
          if dataLast > parNum then
            for aByte := parNum+1 to dataLast do
              if data[aByte] = realValue then begin
                foundEqualNumber := True;
                break;
              end;
          if foundEqualNumber then {mark the position that it had previously
                                   iin 'data'}
            resultingString := Concat(resultingString, kLastChar, chr(aByte
              +argNum))
          else begin
            inc(dataLast);
            if (dataLast + argNum = $FF - 1) then
              raise Exception.Create(eTooMuchConstants);
            if Length(data) <= dataLast then {make room}
              SetLength(data, dataLast + 10);
            data[dataLast] := realValue;
            resultingString := Concat(resultingString, kLastChar,
              chr(dataLast + argNum))
          end;
        end
      end
      else
        raise Exception.Create(eBadNumber);
    end; {of numberBegins}
    if inLetter then begin
      letterStr := currentChar;
      while (hotSpot < numberOfChars) and inLetter do begin
        hotSpot := hotSpot + 1;
        currentChar := textReduced[hotSpot];
        inLetter := (pos(currentChar, upCaseLetters) <> 0);
        if inLetter then
          letterStr := Concat(letterStr, currentChar);
      end;
      mayHaveFunction := (currentChar = '('); {mayHaveFunction = True says that
          after the letters comes an opening parenthesis, that is these letters
          may be the identifier of a known function}

      if not inLetter then  {'currentChar' has to be processed afterwards}
        hotSpot := hotSpot - 1;

      {Check if the final part of letterStr is a known function, substitute it
       by its corresponding letter, and factorize the initial part}
      convertedStr := FactorizeText(self,letterStr, integralsAllowed, mayHaveFunction);
      resultingString := Concat(resultingString, convertedStr);

      if (currentChar = '(') and (not mayHaveFunction) then {the parenthesis
         initiates a factor that multiplies the last factor}
        ResultingString := Concat(resultingString, '*');

      if (convertedStr='S') and (currentChar = '(') and mayHaveFunction then begin
        inc(hotSpot);
        currentChar:= ConvertIntegral(hotSpot);  {prepare B-splines data}
        resultingString:= Concat(resultingString,'('+kLastChar+currentChar+')');
      end;
    end; {of inLetter}
  until (hotSpot >= numberOfChars);
  textReduced := resultingString;
end; {ReduceSymbols}

procedure TFunctionParser_32.UnifyChars;
{This is a tidy-up policy. All alphabetic chars are capitalized, blanks that
 separate symbols are suppressed except when they are followed by an alphabetic
 or a numeric character, and in that cases it is changed to '*'. Brackets or
 braces are converted to parentheses. The consistency of parentheses is
 partially checked}
var
  place, depth, anInteger, theLength: Integer;
  currentChar, precChar, lastAdded, toAdd: AnsiChar;
begin {UnifyChars}
  theLength := Length(formula);
  formula := UpperCase(formula); {All alphabetic chars are capitalized}
  textReduced := '';
  if theLength = 0 then
    raise Exception.Create(eNothingWritten);

  {initializations}
  lastAdded := '('; {this initialization does not produce any interference}
  currentChar := '('; {for consistency with the 'lastAdded' value}
  toAdd := ' '; {the formula could begin with any number of blanks}
  place := 0;
  depth := 0;
  repeat
    precChar := currentChar;
    place := place + 1;
    currentChar := formula[place];
    case currentChar of
      'A'..'Z': begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) or
            ((pos(lastAdded, upCaseLetters) <> 0) and (precChar = ' ')) then
            textReduced := Concat(textReduced, '*'); {only a product can be
             accepted as separator from the previous part of the text. Add
             '*'}
          toAdd := CurrentChar; {save this char to add it after the end of "case"}
        end;
      '0'..'9': begin
          if (lastAdded = ')') or (pos(lastAdded, upCaseLetters) <> 0) or
            ((pos(lastAdded, digits) <> 0) and (precChar = ' ')) then
            textReduced := Concat(textReduced, '*'); {similar to the
            previous case}
          toAdd := CurrentChar;
        end;
      '(', '[', '{': begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) then
            textReduced := Concat(textReduced, '*');
          toAdd := '(';
          depth := depth + 1; {we will check parenthesis consistence in the
          sense that a character cannot be at a negative depth}
        end;
      '*':
        if (lastAdded = '*') then begin
          {perhaps the user want to build a power expression. So, delete
           the previous '*' from the reduced text and add a caret}
          anInteger := Length(textReduced);
          delete(textReduced, anInteger, 1);
          textReduced := concat(textReduced, '^');
          lastAdded := '^';
          toAdd := ' '; {in fact we will add nothing}
        end
        else
          toAdd := '*';
      ')', ']', '}': begin
          toAdd := ')';
          depth := depth - 1;
        end;
      ',', '.', '''': {if user writes  7,56, or 7.56 or 7'56 it will be
                       converted to 7.56. Thus, we are not enforcing totally
                       the law that the decimal separator must be the dot.}
        toAdd := '.';
      '-':
        if (lastAdded = '(') then
          toAdd := '#' {we distinguish the operator "change sign" that we
                        denote by '#', from the subtract operator. The
                        first one demands one argument; the other, demands two.}
        else
          toAdd := '-';
      '/', '\':
        toAdd := '/';  {Thus,  'cos(x)\sin(x)' will be converted to 'COS(X)/SIN(X)'}
      ' ', '^',';':
        toAdd := currentChar; {the semicolon ';' is used in expressions for
                               integrals}
      '+':
        if (lastAdded = '(') then
          toAdd := ' '
        else
          toAdd := '+';
        {Note that a string as '+cos(x)- sin(y)' should be converted to
         'COS(X)-SIN(Y)'. This is the reason for having initialized 'lastAdded'
         as '('   }

      kReturn, kTab, kLineFeed: {Eat all these characters}
        toAdd := ' ';
    else
      raise Exception.Create(eBadCharacters);
    end; {of CASE}

    if toAdd <> ' ' then begin
      textReduced := concat(textReduced, toAdd);
      lastAdded := toAdd;
      {therefore, when 'toAdd' was a blank, nothing is added}
    end;
  until (place = theLength) or (depth < 0);
  if depth <> 0 then
    raise Exception.Create(eBadParentheses);
end;




{----------------------------------------------------------------------}
     {Routines visible from other units}
{----------------------------------------------------------------------}

procedure ReadConstantField(theStr: AnsiString; var val: double);
{An expression may not depend on any variables or parameters.
 Then, its vallue is constant and this procedure parses the
 expression, evaluates it and returns its value}
var
  forConstFunct: TFunctionParser_32;
  mc: array of double;
begin
  val := 0;
  mc:= nil;
  forConstFunct:= TFunctionParser_32.create;
  try
    forConstFunct.ParseNewText(theStr, '', '', False, mc, mc);
    val := forConstFunct.V(mc);
  finally
    forConstFunct.Free;
  end;
end;




initialization
  InitializeParser;
finalization
  FinalizeParser;
end.

