{-----------------------------------------------------------------------------
 Unit Name: ImplicitRenderThread
 Project Name : ImmersionR3R6
 Author:    Angel Montesinos-Amilibia
 Date: 2009
 February, 2005 - January 2009
 Purpose: Implements the class TImplicitRenderThread that draws an implicit
          surface on a contiguous set of rows of a bitmap.
-----------------------------------------------------------------------------}
{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

{See the notes "ComputationsOn3Subvarieties" about the computations of the

 Focal locus}

unit ImplicitRenderThread;

interface

uses
  Classes, Graphics,
  FGammaPicker, VectorsAndUtil, FDrawParamSurf, OCFunctDDDoubleSafe;


type
  CubeKey = array[1..3] of 0..2;
  ReflectedLightFunction = function(x,y: Double; var light: word; var inv: Single): Boolean of object;
  dr = procedure of object;
      
  TImplicitRenderThread = class(TThread)
  private
    minCorner, maxCorner, edges3D: Vector3;
    span3D, aa0: Double;
    rowWidthMinus1, firstRow, lastRow, grain: Integer;
    drawSurf: dr;
    bb: matrix333;
    nn, nnA: matrix33; {the matrix of the focal locus and its adjoint at 'viewpoint'}
    procedure DrawCoarseSurfaceGrainSmoothed;
    procedure DrawSurface;
    procedure FindRay(var ray: Vector3; x,y: Double);
    function FocalLocusSurfNormal(u: Vector3): Vector3;
    function GetNearestPoint(dir: Vector3; a,b: Double; var x, nor: Vector3): Boolean;
    function PickedPt(x,y: Double; var pt, nor: Vector3): Boolean;

    function ReflectedLightNoShadows(x,y: Double; var light: word; var inv: Single): Boolean;
    procedure SegmentInCube(ori, dir: Vector3; keyOri: CubeKey; var a, b: Double);

  protected
    procedure Execute; override;
  public
    vPCubeKey: CubeKey;
    c0, c3, cx, cy: Double;
    focus, vwPt, wH, wV, wN: Vector3;
    frl: ReflectedLightFunction;
    constructor CreateIRT(CreateSuspended: Boolean; bb0: matrix333;
         minCorner0, maxCorner0: Vector3; firstRow0, lastRow0, grain0: Integer);
  end;

implementation


uses
  Math, SysUtils, DateUtils,
  FImmersionR3R6, FRenderImplicitSurfaceWithThreads;


const
  cTolZBuffer = 0.995;

function EuclInnerProd(x,y: Vector3): Double;
begin
  result:= x[1]*y[1] + x[2]*y[2] + x[3]*y[3];
end;

function EuclNorm(x: Vector3): Double;
begin
  result:= sqrt(sqr(x[1]) + sqr(x[2]) + sqr(x[3]));
end;

function TImplicitRenderThread.FocalLocusSurfNormal(u: Vector3): Vector3;
{The gradient of FocalLocusFunction. It is assumed that FocalLocusFunction has
been already calculated, so that the matrix 'mf' has already its value at 'u'}
var
  a: Double;
  m, mA: matrix33;
  i, j: Integer;
begin
  for i:= 1 to 3 do
    for j:= i to 3 do begin
      m[i,j]:= bb[i,j,1]*u[1]+ bb[i,j,2]*u[2]+bb[i,j,3]*u[3]- ord(i=j);
      if j>i then
        m[j,i]:= m[i,j];
    end;
  mA[1,1]:= m[2,2]*m[3,3] - sqr(m[2,3]);
  mA[1,2]:= m[1,3]*m[2,3] - m[1,2]*m[3,3];
  mA[1,3]:= m[1,2]*m[2,3] - m[2,2]*m[1,3];
  mA[2,2]:= m[1,1]*m[3,3] - sqr(m[1,3]);
  mA[2,3]:= m[1,2]*m[1,3] - m[1,1]*m[2,3];
  mA[3,3]:= m[1,1]*m[2,2] - sqr(m[1,2]);
  mA[2,1]:= mA[1,2];
  mA[3,1]:= mA[1,3];
  mA[3,2]:= mA[2,3];

  for i:= 1 to 3 do
    result[i]:= bb[1,1,i]*mA[1,1] + bb[2,1,i]*mA[2,1] + bb[3,1,i]*mA[3,1]+
                bb[1,2,i]*mA[1,2] + bb[2,2,i]*mA[2,2] + bb[3,2,i]*mA[3,2]+
                bb[1,3,i]*mA[1,3] + bb[2,3,i]*mA[2,3] + bb[3,3,i]*mA[3,3];
  a:= EuclNorm(result);
  if a > 1.0e-30 then
    for i:= 1 to 3 do
      result[i]:= result[i]/a
  else
    result[1]:= 3;
end;

constructor TImplicitRenderThread.CreateIRT(CreateSuspended: Boolean;
         bb0: matrix333;
         minCorner0, maxCorner0: Vector3; firstRow0, lastRow0, grain0: Integer);
var
  i, j: Integer;
begin
  inherited Create(CreateSuspended);

  {initialize everything}
  minCorner:= minCorner0;
  maxCorner:= maxCorner0;
  for i:= 1 to 3 do
    edges3D[i]:= maxCorner[i] - minCorner[i];
  firstRow:= firstRow0;
  lastRow:= lastRow0;
  grain:= grain0;
  bb:= bb0;
  drawSurf:= DrawSurface;
  cx:= rist.cx;
  cy:= rist.cy;
  c0:= rist.c0;
  c3:= rist.c3;

  rowWidthMinus1:= bmp.width - 1;
  wH:= rist.wH;
  wV:= rist.wV;
  wN:= rist.wN;
  focus:= rist.focus;
  vwPt:= rist.viewPoint;
  vPCubeKey:= rist.vpCubeKey;
  span3D:= rist.span3D;

  {Value of the Focal locus function and its adjoint matrix at the viewpoint}
  for i:= 1 to 3 do
    for j:= i to 3 do begin
      nn[i,j]:= bb[i,j,1]*vwPt[1]+ bb[i,j,2]*vwPt[2]+bb[i,j,3]*vwPt[3]- ord(i=j);
      if j>i then
        nn[j,i]:= nn[i,j];
    end;
  nnA[1,1]:= nn[2,2]*nn[3,3] - sqr(nn[2,3]);
  nnA[1,2]:= nn[1,3]*nn[2,3] - nn[1,2]*nn[3,3];
  nnA[1,3]:= nn[1,2]*nn[2,3] - nn[2,2]*nn[1,3];
  nnA[2,2]:= nn[1,1]*nn[3,3] - sqr(nn[1,3]);
  nnA[2,3]:= nn[1,2]*nn[1,3] - nn[1,1]*nn[2,3];
  nnA[3,3]:= nn[1,1]*nn[2,2] - sqr(nn[1,2]);
  nnA[2,1]:= nnA[1,2];
  nnA[3,1]:= nnA[1,3];
  nnA[3,2]:= nnA[2,3];

  {The degree zero coefficient in the polynomial aa0 + aa1*t+ aa2*t^2+ aa3*t^3
   that must vanish along a ray passing by the viewpoint}
  aa0:= nn[1,1]*nnA[1,1] + nn[1,2]*nnA[1,2]+ nn[1,3]*nnA[1,3];
end;


procedure TImplicitRenderThread.DrawCoarseSurfaceGrainSmoothed;
{This routine may be usen only if 'grain >= 2' and
 'rowWidthMinus1 < bm.width' and 'lastRow < bm.height'. For
 instance, it is useable for moving implicit surfaces.}
var
    i,j,k,m,n,anver, present, past: Integer;
    sh, fu, fuu, fv, fvv, zu, zv, zuu, zvv, u, v, z0, ind, zInd: Double;
    line: array of pTColorArray;
    cl: TColor;
    bad: Boolean;
    index: Integer;
    by: Word;
begin
  SetLength(line,grain);
  sh:= (grain-1)/2;

  {We store in the array 'map' the values of illumination in a grid of points
   regularly separated by 'grain' pixels. That is, in 'Map[i0,j0]' we store the
   illumination key of point [grain*i0 + sh, grain+j0 + sh], where 'sh = grain/2'.
   Thus we have the illumination at the centers of the squares of that grid. As
   for the zBuffer values, we store at 'zBuffer[grain*i0, grain+j0]' the zBuffer
   of [grain*i0 + sh, grain+j0 + sh]. This is because we have now no size
   restriction for the array zBuffer.}
  j:=firstRow;
  repeat
    i:= 0;
    repeat
      frl(i+sh,j+sh,by, zBuffer[i,j]);
      map[i,j]:= by;
      i:= i+grain;
    until i > rowWidthMinus1;
    j:= j+grain;
  until j > lastRow;

  {First, we draw the two upper 'grain x grain' corners without approximations}
  j:=firstRow;
  for k:= 0  to grain-1 do
    line[k]:= bmp.scanline[j+k];
  i:= 0;
  for m:= 0 to grain-1 do
    for k:= 0 to grain-1 do begin
    {Calculate the map value and zBuffer of each point in the rectangle}
      frl(i+m,j+k,by,zBuffer[i+m,j+k]);

      {Apply the color that corresponds}
      line[k]^[i+m]:= gammas[ (by and $8000) shr 15 , by and $1FF];

      {Store value to 'map'. The zBuffer value has been stored already}
      map[i+m,j+k]:= by;
    end;
  m:= rowWidthMinus1 - (rowWidthMinus1 mod grain);
  for i:= m to rowWidthMinus1 do
    for k:= 0 to grain-1 do begin
      frl(i,j+k,by,zBuffer[i,j+k]);
      line[k]^[i]:= gammas[ (by and $8000) shr 15 , by and $1FF];
      map[i,j+k]:= by;
    end;
  {And the remaining of the first 'grain' rows}

  i:= grain;
  repeat
    by:= map[i,j];
    z0:= zBuffer[i,j];

    {a square is 'bad' if the left upper corner is background and some of
    the others are not}
    if (by = 0) then
      bad:= (map[i-grain,j]<>0) or(map[i+grain,j]<>0) or
          (map[i,j+grain]<>0)
    else
      {or if the left upper corner is not background and some of
       the other are, or differ from it in more than 20}
      bad:= (map[i-grain,j]<>0) or (map[i+grain,j]=0) or
          (map[i,j+grain]=0) or
          (abs(by-map[i-grain,j])> 20)or(abs(by-map[i+grain,j])> 20) or
          (abs(by-map[i,j+grain])> 20);
    if bad then for m:= 0 to grain-1 do
      {Draw the square without approximations}
      for k:= 0 to grain-1 do begin
        frl(i+m,j+k,by,zBuffer[i+m,j+k]);
        line[k]^[i+m]:= gammas[ (by and $8000) shr 15 , by and $1FF];
        map[i+m,j+k]:= by;
      end
    else begin
      {the square is not 'bad'. Apply to all of its points the same values,
       that are already stored}
      cl:= gammas[ (by and $8000) shr 15 , by and $1FF];
      for m:= 0 to grain-1 do
        for k:= 0 to grain-1 do begin
          line[k]^[i+m]:= cl;
          zBuffer[i+m,j+k]:= z0;
          map[i+m,j+k]:= by;
        end;
    end;
    i:= i+grain;
  until i > rowWidthMinus1-grain;

{Now the lower corners. See the upper corners}
  n:= (lastRow-firstRow) mod grain;
  for k:= 0  to n do
    line[k]:= bmp.scanline[lastRow-n+k];
  i:= 0;
  j:= lastRow-n;
  for m:= 0 to grain-1 do
    for k:= 0 to n do begin
      frl(i+m,j+k,by,zBuffer[i+m,j+k]);
      line[k]^[i+m]:= gammas[ (by and $8000) shr 15 , by and $1FF];
      map[i+m,j+k]:= by;
    end;
  m:= rowWidthMinus1 - (rowWidthMinus1 mod grain);
  for i:= m to rowWidthMinus1 do
    for k:= 0 to n do begin
      frl(i,j+k,by,zBuffer[i,j+k]);
      line[k]^[i]:= gammas[ (by and $8000) shr 15 , by and $1FF];
      map[i,j+k]:= by;
    end;

 {And the remining of the last rows}

  i:= grain;
  repeat
    by:= map[i,j];
    z0:= zBuffer[i,j];
    if (by = 0) then
      bad:= (map[i-grain,j]<>0) or(map[i+grain,j]<>0) or
          (map[i,j-grain]<>0)
    else
      bad:= (map[i-grain,j]<>0) or (map[i+grain,j]=0) or
          (map[i,j-grain]=0) or
          (abs(by-map[i-grain,j])> 20)or(abs(by-map[i+grain,j])> 20) or
          (abs(by-map[i,j-grain])> 20);
    if bad then for m:= 0 to grain-1 do
      for k:= 0 to n do begin
        frl(i+m,j+k,by,zBuffer[i+m,j+k]);
        line[k]^[i+m]:= gammas[ (by and $8000) shr 15 , by and $1FF];
        map[i+m,j+k]:= by;
      end
    else begin
      cl:= gammas[ (by and $8000) shr 15 , by and $1FF];
      for m:= 0 to grain-1 do
        for k:= 0 to n do begin
          line[k]^[i+m]:= cl;
          zBuffer[i+m,j+k]:= z0;
          map[i+m,j+k]:= by;
        end;
    end;
    i:= i+grain;
  until i > rowWidthMinus1-grain;

  {Now the remaining of the first and last columns. The method is
   similar to that of the upper row}
  j:= firstRow + grain;
  repeat
    i:=0;
    by:= map[i,j];
    z0:= zBuffer[i,j];
    for k:= 0  to grain-1 do
      line[k]:= bmp.scanline[j+k];
    if (by = 0) then
      bad:= (map[i,j+ grain]<>0) or(map[i+grain,j]<>0) or
          (map[i,j-grain]<>0)
    else
      bad:= (map[i+grain,j]=0) or (map[i,j+grain]=0) or
          (map[i,j-grain]=0) or
          (abs(by-map[i+grain,j])> 20)or(abs(by-map[i,j+grain])> 20) or
          (abs(by-map[i,j-grain])> 20);
    if bad then for m:= 0 to grain-1 do
      for k:= 0 to grain-1 do begin
        frl(i+m,j+k,by,zBuffer[i+m,j+k]);
        line[k]^[i+m]:= gammas[ (by and $8000) shr 15 , by and $1FF];
        map[i+m,j+k]:= by;
      end
    else begin
      cl:= gammas[ (by and $8000) shr 15 , by and $1FF];
      for m:= 0 to grain-1 do
        for k:= 0 to grain-1 do begin
          line[k]^[i+m]:= cl;
          zBuffer[i+m,j+k]:= z0;
          map[i+m,j+k]:= by;
        end;
    end;
    i:= rowWidthMinus1 - (rowWidthMinus1 mod grain);
    by:= map[i,j];
    z0:= zBuffer[i,j];
    if (by = 0) then
      bad:= (map[i,j+ grain]<>0) or(map[i-grain,j]<>0) or
          (map[i,j-grain]<>0)
    else
      bad:= (map[i-grain,j]=0) or (map[i,j+grain]=0) or
          (map[i,j-grain]=0) or
          (abs(by-map[i-grain,j])> 20)or(abs(by-map[i,j+grain])> 20) or
          (abs(by-map[i,j-grain])> 20);
    if bad then for m:= i to rowWidthMinus1 do
      for k:= 0 to grain-1 do begin
        frl(m,j+k,by,zBuffer[m,j+k]);
        line[k]^[m]:= gammas[ (by and $8000) shr 15 , by and $1FF];
        map[m,j+k]:= by;
      end
    else begin
      cl:= gammas[ (by and $8000) shr 15 , by and $1FF];
      for m:= i to rowWidthMinus1 do
        for k:= 0 to grain-1 do begin
          line[k]^[m]:= cl;
          zBuffer[m,j+k]:= z0;
          map[m,j+k]:= by;
        end;
    end;
    j:= j + grain
  until j > lastRow-grain;

  {Finally, the central region}
  j:= firstRow + grain;
  repeat
    i:= grain;
    for k:= 0 to grain-1 do
      line[k]:= bmp.scanline[j+k];
    repeat
      by:= map[i,j];
      z0:= zBuffer[i,j];
      if (by = 0) then
        bad:= (map[i+grain,j]<>0) or
              (map[i,j-grain]<>0) or (map[i,j+grain]<>0) or (map[i-grain,j]<>0)
      else
        bad:= (map[i+grain,j]=0) or
              (map[i,j-grain]=0) or (map[i,j+grain]=0) or (map[i-grain,j]=0) or
              (abs(by-map[i+grain,j])> 20) or
              (abs(by-map[i,j-grain])> 20) or (abs(by-map[i,j+grain])> 20) or
              (abs(by-map[i-grain,j])> 20);
      if bad then for m:= 0 to grain-1 do
        for k:= 0 to grain-1 do begin
          frl(i+m,j+k,by, zBuffer[i+m,j+k]);
          line[k]^[i+m]:= gammas[ (by and $8000) shr 15 , by and $1FF];
          map[i+m,j+k]:= by;
        end
      else begin
        {make an interpolation using the surrounding values. First we
         compute the approximate derivatives}
        fu:= 0.5*(map[i+grain,j]-map[i-grain,j]);
        fv:= 0.5*(map[i,j+grain]-map[i,j-grain]);
        fuu:= 0.5*(map[i+grain,j]+map[i-grain,j])- by;
        fvv:= 0.5*(map[i,j+grain]+map[i,j-grain])-by;
        zu:= 0.5*(zBuffer[i+grain,j]-zBuffer[i-grain,j]);
        zv:= 0.5*(zBuffer[i,j+grain]-zBuffer[i,j-grain]);
        zuu:= 0.5*(zBuffer[i+grain,j]+zBuffer[i-grain,j])- z0;
        zvv:= 0.5*(zBuffer[i,j+grain]+zBuffer[i,j-grain])-z0;

        {and now the interpolation}
        for m:= 0 to grain-1 do begin
          u:= (m-sh)/grain;
          ind:= by+u*(fu + u* fuu);
          zInd:= z0+u*(zu + u* zuu);
          for k:= 0 to grain-1 do begin
            v:= (k-sh)/grain;
            index:= round(ind+ v*(fv + v* fvv));
            zBuffer[i+m,j+k]:= zInd+ v*(zv + v* zvv);
            cl:= gammas[ (index and $8000) shr 15 , index and $1FF];
            line[k]^[i+m]:= cl;
            map[i+m,j+k]:= by;
          end;
        end;
      end;
      i:= i + grain;
    until i +grain > rowWidthMinus1;
    j:= j + grain
  until j + grain > lastRow;
  line:= nil;
end;

procedure TImplicitRenderThread.DrawSurface;
{Draw the surface with maximum resolution, that is computing each pixel
 separately. Save 'zBuffer' and 'map' values}
var
  j,k: Integer;
  line: pTColorArray;
  index: Word;
begin
  for k := firstRow to lastRow do begin
    if Terminated then
      break;
    line := bmp.ScanLine[k];
    for j := 0 to rowWidthMinus1 do begin
      if frl(j,k,index, zBuffer[j,k]) then begin
        line^[j] := gammas[ (index and $8000) shr 15 , index and $1FF];
        map[j, k] := index;
      end
      else begin
        line^[j]:= gammas[0,0];
        map[j, k] := 0;
      end;
    end;
  end;
end;

procedure TImplicitRenderThread.Execute;
var
  first, last: TDateTime;
begin
  first:= Time;
  if (grain<=0) or Terminated then
    Exit;
  frl:= ReflectedLightNoShadows;
  if grain = 1 then
    DrawSurf
  else
    DrawCoarseSurfaceGrainSmoothed;
  last:= Time;

  {adjust the 'grain' to the time the computation takes}
  rist.grain:=  1 + round(MilliSecondsBetween(last, first)/50);
end;

procedure TImplicitRenderThread.FindRay(var ray: Vector3; x,y: Double);
{Returns in 'ray' a vector such that the entire line
     t -> viewpoint + t*ray
 projects on the point (x,y) of the bitmap}
var
  k: Integer;
  ah, av: Double;
begin
  ah := (x - cx)/c0;
  av := (y - cy)/c0;
  for k := 1 to 3 do
    ray[k] := ah * wH[k] + av * wV[k] + wN[k];
end;

function TImplicitRenderThread.GetNearestPoint(dir: Vector3; a,b: Double; var x, nor: Vector3): Boolean;
{We consider the line  't  -->  viewPoint + t*dir' and find (if there is one) the least
value 't0' in [a,b] such that 'x := viewPoint + t0*dir'  belongs to the surface.
In 'nor' is returned the unit normal to the surface at 'x'. Iff al goes well,
result is True}

var
  a1, a2, a3, t: Double;
  i,j: Integer;
  pp, pA: matrix33;
begin
  for i:= 1 to 3 do
    for j:= i to 3 do begin
      pp[i,j]:= bb[i,j,1]*dir[1]+ bb[i,j,2]*dir[2]+bb[i,j,3]*dir[3];
      if j>i then
        pp[j,i]:= pp[i,j];
    end;
  pA[1,1]:= pp[2,2]*pp[3,3] - sqr(pp[2,3]);
  pA[1,2]:= pp[1,3]*pp[2,3] - pp[1,2]*pp[3,3];
  pA[1,3]:= pp[1,2]*pp[2,3] - pp[2,2]*pp[1,3];
  pA[2,2]:= pp[1,1]*pp[3,3] - sqr(pp[1,3]);
  pA[2,3]:= pp[1,2]*pp[1,3] - pp[1,1]*pp[2,3];
  pA[3,3]:= pp[1,1]*pp[2,2] - sqr(pp[1,2]);
  pA[2,1]:= pA[1,2];
  pA[3,1]:= pA[1,3];
  pA[3,2]:= pA[2,3];
  a1:= pp[1,1]*nnA[1,1] + pp[2,1]*nnA[2,1] + pp[3,1]*nnA[3,1] +
        pp[1,2]*nnA[1,2] + pp[2,2]*nnA[2,2] + pp[3,2]*nnA[3,2] +
        pp[1,3]*nnA[1,3] + pp[2,3]*nnA[2,3] + pp[3,3]*nnA[3,3] ;
  a2:= nn[1,3]*pA[1,3] + nn[2,3]*pA[2,3] + nn[3,3]*pA[3,3] +
        nn[1,2]*pA[1,2] + nn[2,2]*pA[2,2] + nn[3,2]*pA[3,2] +
        nn[1,1]*pA[1,1] + nn[2,1]*pA[2,1] + nn[3,1]*pA[3,1];
  a3:=  pp[1,1]*pA[1,1] + pp[2,1]*pA[2,1] + pp[3,1]*pA[3,1];
  result:= CubicPolynomialRoot(aa0,a1,a2,a3, a, b, t);
  if result then begin
    for i:= 1 to 3 do
      x[i]:= vwPt[i] + t*dir[i];
    nor:= FocalLocusSurfNormal(x);
    result:= (nor[1] < 2);
  end
end;

function TImplicitRenderThread.PickedPt(x,y: Double; var pt, nor: Vector3): Boolean;
{Given a point of coordinates (x,y) in the bitmap rectangle, find the visible
 point of the surface that projects onto (x,y), if there is one}
var
  a, b: Double;
  ray: Vector3;
begin
  FindRay(ray, x,y);
  SegmentInCube(vwPt, ray, VpCubeKey, a, b);
  result:= (a < b) and GetNearestPoint(ray, a, b, pt, nor);
end;

function TImplicitRenderThread.ReflectedLightNoShadows(x,y: Double;
    var light: word; var inv: Single): Boolean;
const
  dullness: Double = 0.3;
  OneMinusDullness: Double = 1 - 0.3;
  FocusIntens: Double = 0.60;
  OneMinusFocusIntens: Double = 0.4;
var
  lambda, sigma, omega, compF, compAmb, f2x, m2x: Double;
  normal, thePos, toVp, toFocus: Vector3;
  i: Integer;
begin
  inv:= 0;
  light:=0;
  result:= PickedPt(x,y, thePos, normal);
  if result then begin
    for i := 1 to 3 do begin
      toVp[i] := vwPt[i] - thePos[i];
      toFocus[i] := focus[i] - thePos[i];
    end;
    inv:= 1/(EuclInnerProd(wN,thePos)-c3);
    m2x := EuclNorm(toVp);
    sigma := EuclInnerProd(toVp, Normal) / m2x;
    {the sign of sigma says whether we are looking to one side or the other}

    {Compute ambient component of illumination}
    compAmb := abs(sigma);

    f2x := EuclNorm(toFocus);
    lambda := EuclInnerProd(toFocus, normal) / f2x;
    omega := 2 * lambda * sigma; {the sign of 'omega' says that we are looking
                          to the illuminated (by the focus) side of the surface}
    if (omega > 0) then begin
        {We are looking to the side illuminated by the focus}
      lambda := abs(lambda);
      omega := (1 + omega - EuclInnerProd(toFocus, toVp) / (m2x * f2x)) /2;

        {Compute focus component of illumination}
      compF :=  lambda *(OneMinusDullness* sqr(omega) + dullness * (1 +
          abs(sigma)) / 2);
    end
    else
      compF := 0;
    omega := Min(1,FocusIntens * compF + OneMinusFocusIntens * compAmb);

      {add wheighed components and encode the reverse character in the most
       significant bit}
      light:= (ord(sigma<0) shl 15) + 1 + round((cGammaN-2)*omega);
  end;
end;

procedure TImplicitRenderThread.SegmentInCube(ori, dir: Vector3; keyOri: CubeKey; var a, b:
  Double);
{Gives the values 'a' <= 'b' of 't' such that the point  ori + t*dir  is on the
 boundary of the cube. If the result satisfies, 'b < a' then that line does not
 cut the cube}
const
  veryBig: Double = 1.0E+30;
  yetBigger: Double = 2.0E+30;
var
  ta, tb: Double;
  i: Integer;
begin
  a := -veryBig;
  b := veryBig;
  tb := yetBigger;
  for i := 1 to 3 do
    if keyOri[i] = 2 then begin
      if dir[i] > 0 then
        tb := (maxCorner[i] - ori[i]) / dir[i]
      else if dir[i] < 0 then
        tb := (minCorner[i] - ori[i]) / dir[i];
      if tb < b then
        b := tb;
    end
    else begin
      ta := (minCorner[i] + keyOri[i] * edges3D[i] - ori[i]) /
        dir[i];
      tb := (minCorner[i] + (1 - keyOri[i]) * edges3D[i] - ori[i]) /
        dir[i];
      if ta > a then
        a := ta;
      if tb < b then
        b := tb;
    end;
end;

end.
 
