
{-----------------------------------------------------------------------------
 Unit Name: OCFunctDDDoubleSafe
 Author:    Angel Montesinos Amilibia
 November 21, 2006 - October 21, 2008
 Purpose: Class TOpCodeFunction. Compiles, in direct X86 opcodes for evaluation,
    a function given as a string. It is at least as fast as the same string
    compiled by Delphi at compile time.

 History: I am indebted to Dmitry Kaschenko and Vladimir Safin for the general
    idea of translating an evaluation tree to direct X86 opcode that I learned
    by studying their component "FormulaCompiler".

 For source code you may email me to:
     montesin@uv.es
-----------------------------------------------------------------------------}


{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos-Amilibia


This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

unit OCFunctDDDoubleSafe;

{See documentation of the unit 'OCFunctDDDoubleSafeIntegrals.pas' in the
 program Superficies. That unit is equal to this but for the addition
 of integrals}

interface

uses
  VectorsAndUtil;
                 
type
  {firstDeriv = record
    fa0: Double;
    dfa0: Double;
  end;

  secondDeriv = record  //used for the following: given a function in R^n, a
    fa0: Double;      //curve a:R --> R^n, f0 will be the value of
    dfa0 : Double;    //f(a(t)) at t=0, dfa0 will be f(a(t))'(0) (that is the
    ddfa0 : Double;   //value at t=0 of the derivative of f(a(t)); and ddfa0
  end;                  //will be f(a(t))''(0).}

  TOCFunctDDDoubleSafe = class(TObject)
  PRIVATE
    rawText : AnsiString; {the text of the function}
    data : array of Double; {stores the value of the
      parameters and constants that appear in the formula}
    u: array of Double; {Stores velocities along evaluation of FoA_prime, FoA_second}
    w: array of Double; {Stores accelerations along evaluaation of FoA_second}
    functionCode, dC, ddc  : AnsiString; {the function and Directional derivative opCodes}
    argNum : Integer; {Length[allowedVars]}
    parNum : Integer; {Length[allowedParams]}
    actorList : AnsiString; {the RPN string that represents the function}
    lengthActorList : Integer;
    procedure AddIntToCode(n : Integer; var cod: AnsiSTring);
    procedure LeaveRoomInFPUStack(neededPlaces : Integer; var cod: AnsiSTring);
    procedure SetUWLength;
    function LoadVd(i: Integer): AnsiString;
    function LoadAd(i: Integer): AnsiString;
    procedure PopFromAppStack(var cod: AnsiSTring);
    procedure PutCodeInExeBlock(cod: AnsiString; var exeBlock: PChar);
    procedure CodifyFunction;
  PUBLIC
    usedVarsAndParams: AnsiString;
    V : function(const X : array of Double) : Double;
    VBlock: PChar;
    FoA_prime: function (const XV: array of Double): Double;
      {'XV' must be an array of Double whose lenth is two times 'argNum'.
      Gives as result the value of the function at the first 'argNum' arguments.
      Leaves at 'u[1]' the directional derivative of the function with respect
      to the last 'argNum' entries of 'XV'.}
    DBlock: PChar;
    FoA_second: function (const XV: array of Double): Double;
      {'XV' must be an array of Double whose lenth is two times 'argNum'.
      Gives as result the value of the function at the first 'argNum' arguments.
      Leaves at 'u[1]' the directional derivative of the function with respect
      to the last 'argNum' entries of 'XV'.}
    DDBlock: PChar;
    function FirstD(const XV: array of Double): FirstDeriv;
      {'XV' must be an array of Double whose length is two times 'argNum'.
      Let a(t) be a curve such that a(0) equals the point whose coordinates
      are the first 'argNum' arguments, a'(0) the vector ... the second group
      of 'argNum' arguments. Gives as result the value of the function at the first 'argNum' arguments,
      f(a(0)), (f(a(t))'(0)}
    function SecondD(const XV: array of Double): SecondDeriv;
      {'XV' must be an array of Double whose length is three times 'argNum'.
      Let a(t) be a curve such that a(0) equals the point whose coordinates
      are the first 'argNum' arguments, a'(0) the vector ... the second group
      of 'argNum' arguments, a''(0) the vector... the third group. Gives as result the value of the function at the first 'argNum' arguments,
      f(a(0)), (f(a(t))'(0), (f(a(t))''(0)}
    constructor Create;
    destructor Destroy; OVERRIDE;
    function ParseNewText(t, vars, pars : AnsiString) : AnsiString;
    {'vars' and 'pars' are two disjoint subsets of the letters 'p'..'z'
      and are the names of variables and parameters to be used in the
      formula}
    procedure BuildFirstDerivative;
    procedure BuildSecondDerivative;
    procedure UpdateParameters(parameters: array of Double);
  end;

  procedure ReadConstantField(theStr: AnsiString; var val: Double);

  
  var
    theFs: array[1..6] of TOCFunctDDDoubleSafe;

implementation
uses
  SysUtils, Math, Windows;

const
  kMaxNumOfNames = 42; {The number of strings of length less}
                             {than 8 that the parser recognizes as}
                             {function names. For instance, asin  }
                             {and arcsin are both recognized.     }
  upCaseLetters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  variables = 'PQRSTUVWXYZ';
  digits = '0123456789';

  eThereIsNoError = 0;
  eBadNumber = 1; {A number is miswritten.}
  eBadCharacters = 2; {You have used an illegal character.}
  eBadFunctionOrVariables = 3; {You mispelled a function or variable}
  eBadParentheses = 4; {Check the parentheses, please.}
  eTooMuchConstants = 5; {I cannot digest that much constants.}
  eSintaxError = 6; {There is a sintax fault in the formula.}
  eErrorInParsing = 7; {I did fail in analysing the function.}
  eExpressionTooLarge = 8; {The expression is too long}
  eNothingWritten = 9; {There was nothing to work on!}
  eBadVar = 10; {The allowed variables in this formula are:}
  eBadVarSet = 11; {The allowed variables must be in the range 'p' to 'z'}
  eBadParamSet = 12; {No parameter may be named as any variable}
  eMaxNum = 1.0E24;
  eMinNum = 1.0E-24;
  eMaxExp = 75;


  {char constants}
  kReturn = char(13);
  kLineFeed = char(10);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';
  kLastChar = #$FF;
  kLastMinusOne= #$FE;

  {Code constants}

  c_stackDisplacDouble = 8;
  c_stackDisplacDoubleHStr= #$08; {hexadecimal string of c_stackDisplacDouble}
  c_stackDisplacExtended = 12;
  c_stackDisplacExtendedHStr= #$0C; {hexadecimal string of c_stackDisplacExtended}
  c_LastReg = 7; {last register in the FPU}

  c_fstp0ESPd = #$DD#$9C#$24#$00#$00#$00#$00; //fstp st(0) to [ESP] as Double
  c_fldESPd = #$DD#$84#$24#$00#$00#$00#$00;  //fld from [ESP] as a Double.

  c_LoadPi = #$D9#$EB;
  c_Load1 = #$D9#$E8;
  c_Load0 = #$D9#$EE;


  c_zero_comp = #$D9#$EE#$DF#$F1;
    {4 bytes. ST(cLastReg) must be empty. Compares 0 with x=ST(0). Then, JA jumps
     if 0>x; JNA jumps if 0 <= x; JB, if 0 < x; JNB, if 0 >= x; JNE if 0<>x}
  c_1_comp = #$D9#$E8#$DF#$F1;
    {4 bytes.  ST(cLastReg) must be empty. Compares 1 with x=ST(0). Then, JA jumps
     if 1>x; JNA jumps if 1 <= x; JB, if 1 < x; JNB, if 1 >= x; JNE if 1<>x}
  c_mod_pi = #$D9#$EB#$D9#$C9#$D9#$F8#$DD#$D9;
    {8 bytes. ST(cLastReg) must be empty}
  c_mod_2pi = #$D9#$EB#$D8#$C0#$D9#$C9#$D9#$F8#$DD#$D9;
    {10 bytes. ST(cLastReg) must be empty}
 // c_push_eax = #$50;
 // c_pop_eax = #$58;
  c_ln = #$D9#$E8#$D9#$C9#$D9#$F1#$D9#$ED#$DE#$C9;
    {10 bytes. Requires ST(0) > 0 and ST(cLastReg) empty}
  c_2toX = #$D9#$C0#$D9#$FC#$D9#$C9#$D8#$E1#$D9#$F0#$D9#$E8#$DE#$C1#$D9#$C9#$D9
    +
    #$E8#$D9#$FD#$DD#$D9#$DE#$C9;
    {24 bytes. Requires ST6) and ST(cLastReg) empty}
  c_div2 = c_Load1 + #$D9#$E0#$D9#$C9#$D9#$FD#$DD#$D9; 
      {10 bytes. Requires ST(cLastReg) empty}

type

  pieceClass = (openParent, closeParent, operationSymb, {}
    negation, constantSymb, varSymb, functionSymb, badThing);

  WrittenFunction = record
    theText : AnsiString;
    equivChar : char;
  end;



var
  uKnownFunc : array[1..kMaxNumOfNames] of WrittenFunction;
  uFormula : AnsiString;
  uAttribution : array['#'..#$9F] of pieceClass;
  uReduced : AnsiString;
  uSpecialValues : array[0..7] of Double;
  dataLast : byte;
  uStdMessages : array[1..12] of AnsiString;
  allowedVars, allowedPars, allVars: AnsiString; {variable names that may appear in 'rawText'.
                 must be a subset of [p,q,r,s,t,u,v,w,x,y,z]}
  emptyRegisters: Byte;
  c_div_S, c_ln_S, c_2toX_S : AnsiString; {code vars}

constructor TOCFunctDDDoubleSafe.Create;
begin
  inherited Create;
  rawText := '';
  functionCode := '';
  data := nil;
end;

destructor TOCFunctDDDoubleSafe.Destroy;
begin
  if self <> nil then
    begin
      rawText := '';
      functionCode := '';
      u:= nil;
      w:= nil;
      data := nil;
      if VBlock <> nil then begin
        VirtualFree(VBlock, 0, MEM_RELEASE);
        VBlock:= nil
      end;
      if DBlock <> nil then begin
        VirtualFree(DBlock, 0, MEM_RELEASE);
        DBlock:= nil
      end;
      if DDBlock <> nil then begin
        VirtualFree(DDBlock, 0, MEM_RELEASE);
        DDBlock:= nil
      end;
    end;
  inherited Destroy;
end;

function inNumber(targetStr : AnsiString;
  var startChar : char;
  var startPos : Integer) : Double;
var
  valid, alreadyBlank, alreadyDot, foundOther : boolean;
  charPos, stringLength : Integer;
  theChar, antiguoSeparadorDecimal : char;
  numStr : AnsiString;
  theNumber : Double;
begin
  antiguoSeparadorDecimal := DecimalSeparator;
  DecimalSeparator := '.';
  stringLength := length(targetStr);
  alreadyDot := false;
  valid := true;
  alreadyBlank := false;
  foundOther := False;
  numStr := Concat('', startChar);
  charPos := startPos;
  theChar := startChar;
  theNumber := 1.0;
  while (charPos < stringLength) and valid and (not foundOther) do
    begin
      inc(charPos);
      theChar := targetStr[charPos];
      case theChar of
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' :
          if not alreadyBlank then
            numStr := numStr + theChar
          else
            begin
              valid := true;
              theNumber := theNumber * StrToFloat(numStr);
              numStr := theChar;
              alreadyDot := false;
              alreadyBlank := false;
            end;
        '.' :
          if alreadyDot or alreadyBlank then
            valid := False
          else
            begin
              alreadyDot := True;
              numStr := numStr + theChar;
            end;
        ' ' :
          if numStr[length(numStr)] = '.' then
            valid := False
          else
            alreadyBlank := true;
      else
        begin
          foundOther := true;
          charPos := charPos - 1;
          theChar := targetStr[charPos];
        end;
      end;
    end;
  if valid then
    begin
      inNumber := theNumber * StrToFloat(numStr);
      startPos := charPos;
      startChar := theChar;
    end
  else
    startChar := '?';
  DecimalSeparator := AntiguoSeparadorDecimal;
end;

function UnifyCharsODEinUV(var isFirstOrder: Boolean): Integer;
var
  place, depth, anInteger, theLength: Integer;
  currentChar, precChar, lastAdded, toAdd: Char;
begin {UnifyCharsSpecial}
  isFirstOrder:= True;
  result:= eThereIsNoError;
  theLength := Length(uFormula);
  uFormula := UpperCase(uFormula);
  uReduced := '';
  if theLength = 0 then begin
    Result := eNothingWritten;
    Exit;
  end;
  lastAdded := '(';
  currentChar := '(';
  toAdd := ' ';
  place := 0;
  depth := 0;
  repeat
    precChar := currentChar;
    place := place + 1;
    currentChar := uFormula[place];
    case currentChar of
      'A'..'Z': begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) or
            ((pos(lastAdded, upCaseLetters) <> 0) and (precChar = ' ')) then
            uReduced := Concat(uReduced, '*');
          toAdd := CurrentChar;
        end;
      '0'..'9': begin
          if (lastAdded = ')') or (pos(lastAdded, upCaseLetters) <> 0) or
            ((pos(lastAdded, digits) <> 0) and (precChar = ' ')) then
            uReduced := Concat(uReduced, '*');
          toAdd := CurrentChar;
        end;
      '(', '[', '{': begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) then
            uReduced := Concat(uReduced, '*');
          toAdd := '(';
          depth := depth + 1;
        end;
      '*':
        if (lastAdded = '*') then begin
          anInteger := Length(uReduced);
          delete(uReduced, anInteger, 1);
          uReduced := concat(uReduced, '^');
          lastAdded := '^';
          toAdd := ' ';
        end
        else
          toAdd := '*';
      ')', ']', '}': begin
          toAdd := ')';
          depth := depth - 1;
        end;
      ',', '.':
        toAdd := '.';
      '''': begin
          if lastAdded = 'U' then begin
            anInteger := Length(uReduced);
            delete(uReduced, anInteger, 1);
            if (place < theLength) and (uFormula[place+1]= '''') then begin
              inc(place);
              isFirstOrder:= False;
              uReduced := concat(uReduced, '(Y)');
              lastAdded:= ')';
              toAdd:= ' ';
            end
            else begin
              uReduced := concat(uReduced, '(W)');
              lastAdded:= ')';
              toAdd:= ' ';
            end;
          end
          else if lastAdded = 'V' then begin
            anInteger := Length(uReduced);
            delete(uReduced, anInteger, 1);
            if (place < theLength) and (uFormula[place+1]= '''') then begin
              inc(place);
              isFirstOrder:= False;
              uReduced := concat(uReduced, '(Z)');
              lastAdded:= ')';
              toAdd:= ' ';
            end
            else begin
              uReduced := concat(uReduced, '(X)');
              lastAdded:= ')';
              toAdd:= ' ';
            end
          end
          else
            toAdd:= '.';
        end;
      '-':
        if (lastAdded = '(') then
          toAdd := '#'
        else
          toAdd := '-';
      '/', '\':
        toAdd := '/';
      ' ', '^',';':
        toAdd := currentChar;
      '+':
        if (lastAdded = '(') then
          toAdd := ' '
        else
          toAdd := '+';
      kReturn, kTab, kLineFeed:
        toAdd := ' ';
      else begin
        result := eBadCharacters;
        Exit;
      end;
    end; {of CASE}

    if toAdd <> ' ' then begin
      uReduced := concat(uReduced, toAdd);
      lastAdded := toAdd;
    end;
  until (place = theLength) or (depth < 0);
  if depth <> 0 then
    result := eBadParentheses;
  if (place < theLength) then
    result:= eExpressionTooLarge;
end;

function UnifyChars : Integer;
var
  place, depth, anInteger, theLength : Integer;
  currentChar, precChar, lastAdded, toAdd : Char;
begin {UnifyChars}
  UnifyChars := eThereIsNoError;
  theLength := Length(uFormula);
  uFormula := UpperCase(uFormula);
  uReduced := '';
  if theLength = 0 then
    begin
      UnifyChars := eNothingWritten;
      Exit;
    end;
  lastAdded := '(';
  currentChar := '(';
  toAdd := ' ';
  place := 0;
  depth := 0;
  repeat
    precChar := currentChar;
    place := place + 1;
    currentChar := uFormula[place];
    case currentChar of
      'A'..'Z' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) or
            ((pos(lastAdded, upCaseLetters) <> 0) and (precChar = ' ')) then
            uReduced := Concat(uReduced, '*');
          toAdd := CurrentChar;
        end;
      '0'..'9' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, upCaseLetters) <> 0) or
            ((pos(lastAdded, digits) <> 0) and (precChar = ' ')) then
            uReduced := Concat(uReduced, '*');
          toAdd := CurrentChar;
        end;
      '(', '[', '{' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) then
            uReduced := Concat(uReduced, '*');
          toAdd := '(';
          depth := depth + 1;
        end;
      '*' :
        if (lastAdded = '*') then
          begin
            anInteger := Length(uReduced);
            delete(uReduced, anInteger, 1);
            uReduced := concat(uReduced, '^');
            lastAdded := '^';
            toAdd := ' ';
          end
        else
          toAdd := '*';
      ')', ']', '}' :
        begin
          toAdd := ')';
          depth := depth - 1;
        end;
      ',', '.', '''' :
        toAdd := '.';
      '-' :
        if (lastAdded = '(') then
          toAdd := '#'
        else
          toAdd := '-';
      '/', '\' :
        toAdd := '/';
      ' ', '^' :
        toAdd := currentChar;
      '+' :
        if (lastAdded = '(') then
          toAdd := ' '
        else
          toAdd := '+';
      kReturn, kTab, kLineFeed :
        toAdd := ' ';
    else
      begin
        UnifyChars := eBadCharacters;
        Exit;
      end;
    end; {of CASE}

    if toAdd <> ' ' then
      begin
        uReduced := concat(uReduced, toAdd);
        lastAdded := toAdd;
      end;
  until (place = theLength) or (depth < 0);
  if depth <> 0 then
    UnifyChars := eBadParentheses
  else if (place < theLength) then
    UnifyChars := eExpressionTooLarge;
end; {of UnifyChars}

function IdentifyFunction(theFunctionText : AnsiString) : char;
{If 'theFunctionText' is among the known functions, IdentifyFunction }
{returns the capital letter that corresponds to it. else, it    }
{returns '?'.                                                        }
var
  i : Integer;
  foundFunction : Boolean;
begin
  i := 0;
  repeat
    i := i + 1;
    foundFunction := (theFunctionText = uKnownFunc[i].theText);
  until (i = kMaxNumOfNames) or foundFunction;
  if foundFunction then
    IdentifyFunction := uKnownFunc[i].equivChar
  else
    IdentifyFunction := '?';
end;

function FactorizeText(aF:TOCFunctDDDoubleSafe; targetString : AnsiString; {}
  var mayEndInFunction : Boolean; isODE: Boolean) : AnsiString;
var
  numberOfChars, hotSpot, count, position: Integer;
  currentChar : char;
  presumedText : AnsiString;
  resultingString : AnsiString;
begin {FactorizeText}
  FactorizeText := '';
  resultingString := '';
  numberOfChars := Length(targetString);
  hotSpot := numberOfChars;
  if mayEndInFunction and (numberOfChars > 1) then
    begin
      count := numberOfChars;
      hotSpot := 1;
      repeat
        presumedText := Copy(targetString, hotSpot, count);
        currentChar := IdentifyFunction(presumedText);
        hotSpot := hotSpot + 1;
        count := count - 1;
      until (currentChar <> '?') or (count = 0);
      if (currentChar <> '?') then
        begin
          hotSpot := hotSpot - 2;
          mayEndInFunction := TRUE;
          if (hotSpot > 0) then
            resultingString := Concat('*', currentChar)
          else
            begin
              resultingString := currentChar;
              FactorizeText := resultingString;
              Exit;
            end;
        end
      else
        begin
          hotSpot := numberOfChars;
          mayEndInFunction := FALSE;
        end;
    end {of IF mayEndInFunction ...}
  else
    mayEndInFunction := FALSE;
  if hotSpot > 0 then
    begin
      currentChar := targetString[hotSpot];
      position := pos(currentChar, allVars);
      if (position <> 0) then begin
        resultingString := Concat(kLastChar, chr(position), resultingString);
        if pos(currentChar, aF.usedVarsAndParams)=0 then
          aF.usedVarsAndParams:= aF.usedVarsAndParams + currentChar;
      end
      else  begin {check if maybe it is a 'pi'}
          if (currentChar = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
            = 'P') then
            begin
              hotSpot := hotSpot - 1;
              resultingString := 'p' + resultingString;
            end
          else {fatal error}
            Exit;
        end;
      hotSpot := hotSpot - 1;
    end;
  while hotSpot > 0 do
    begin
      currentChar := targetString[hotSpot];
      position := pos(currentChar, allVars);
      if (position <> 0) then begin
        resultingString := Concat(kLastChar, chr(position), '*',
                 resultingString);
        if pos(currentChar, aF.usedVarsAndParams)=0 then
          aF.usedVarsAndParams:= aF.usedVarsAndParams + currentChar;
      end
      else
        begin {check if maybe it is a 'pi'}
          if (currentChar = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
            = 'P') then
            begin
              hotSpot := hotSpot - 1;
              resultingString := 'p*' + resultingString;
            end
          else {fatal error}
            Exit;
        end;
      hotSpot := hotSpot - 1;
    end; {of WHILE...}
  FactorizeText := resultingString;
end; {FactorizeText}


function ReduceSymbols(aF : TOCFunctDDDoubleSafe; isOde: Boolean) : Integer;
var
  numberOfChars, hotSpot, reducedLength : Integer;
  currentChar : char;
  letterStr : AnsiString;
  convertedStr, resultingString : AnsiString;
  numberBegins, inLetter, mayHaveFunction, foundEqualNumber : Boolean;
  realValue : Double;
  aByte : Byte;
begin {ReduceSymbols}
  reduceSymbols := eThereIsNoError;
  resultingString := '';
  reducedLength := Length(uReduced);
  numberOfChars := reducedLength;
  hotSpot := 0;
  repeat
    hotSpot := hotSpot + 1;
    currentChar := uReduced[hotSpot];
    inLetter := (pos(currentChar, upCaseLetters) <> 0);
    numberBegins := (pos(currentChar, digits) <> 0) or (currentChar = '.');
    if (not inLetter) and (not numberBegins) then
      ResultingString := Concat(resultingString, currentChar);
    if numberBegins then
      begin
        realValue := 0;
        if currentChar = '.' then
          currentChar := '?'
        else
          realValue := inNumber(uReduced, currentChar, hotSpot);
        if (currentChar <> '?') and (currentChar <> '.') then
          begin
            if realValue = 1 then
              resultingString := resultingString + '1'
            else
              begin
                foundEqualNumber := False;
                if dataLast > aF.parNum then
                  for aByte := aF.parNum+1 to dataLast do
                    if aF.data[aByte] = realValue then
                      begin
                        foundEqualNumber := True;
                        break;
                      end;
                if foundEqualNumber then
                  resultingString := Concat(resultingString, kLastChar, chr(aByte
                    +aF.argNum))
                else
                  begin
                    inc(dataLast);
                    if (dataLast + aF.argNum = $FF - 1) then
                      begin
                        ReduceSymbols := eTooMuchConstants;
                        Exit;
                      end;
                    if Length(aF.data) <= dataLast then
                      SetLength(aF.data, dataLast + 10);
                    aF.data[dataLast] := realValue;
                    resultingString := Concat(resultingString, kLastChar,
                      chr(dataLast + aF.argNum))
                  end;
              end
          end
        else
          begin
            ReduceSymbols := eBadNumber;
            Exit;
          end;
      end; {of numberBegins}
    if inLetter then
      begin
        letterStr := currentChar;
        while (hotSpot < numberOfChars) and inLetter do
          begin
            hotSpot := hotSpot + 1;
            currentChar := uReduced[hotSpot];
            inLetter := (pos(currentChar, upCaseLetters) <> 0);
            if inLetter then
              letterStr := Concat(letterStr, currentChar);
          end;
        mayHaveFunction := (currentChar = '(');
        if not inLetter then
          hotSpot := hotSpot - 1;
        convertedStr := FactorizeText(aF, letterStr, mayHaveFunction,isODE);
        if convertedStr = '' then
          begin
            ReduceSymbols := eBadFunctionOrVariables;
            Exit;
          end;
        resultingString := Concat(resultingString, convertedStr);
        if (currentChar = '(') and (not mayHaveFunction) then
          ResultingString := Concat(resultingString, '*');
      end; {of inLetter}

  until (hotSpot >= numberOfChars);
  uReduced := resultingString;
end; {ReduceSymbols}

function CheckFunction : Integer;
{Checks the syntax of the function}
var
  lastRead, reducedLength : integer;
  species, speciesAnt : pieceClass;
  charac : char;
  stopHere : boolean;

  function hasError : boolean;
  var
    foundError, usual : boolean;
  begin
    foundError := false;
    if species = badThing then
      foundError := true
    else
      begin
        usual := (species = operationSymb) or (species = closeParent);
        case speciesAnt of
          openParent :
            foundError := usual;
          closeParent, constantSymb, varSymb :
            foundError := not usual;
          negation, operationSymb :
            foundError := usual or (species = negation);
          functionSymb :
            foundError := (species <> openParent);
        end; {of case}
      end; {of else}
    hasError := foundError;
  end; {of hasError}

begin {CheckFunction}
  CheckFunction := eThereIsNoError;
  lastRead := 0;
  reducedLength := Length(uReduced);
  speciesAnt := openParent;
  repeat
    lastRead := lastRead + 1;
    charac := uReduced[lastRead];
    if charac = kLastChar then
      begin
        inc(lastRead);
        species := constantSymb;
      end
    else
      species := uAttribution[charac];
    stopHere := hasError;
    speciesAnt := species;
  until (lastRead = reducedLength) or stopHere;
  if (not stopHere) then
    begin
      species := closeParent;
      stopHere := hasError;
    end;
  if stopHere then
    CheckFunction := eSintaxError;
end; {CheckFunction}

function ConvertPowerSymbol(x : AnsiString; n : Integer) : AnsiString;
var
  z : AnsiString;

  function ConvertPowerSymbolProv(xx : AnsiString; nn : Integer) : AnsiString;
  var
    zz : AnsiString;
    m, np : Integer;
  begin
    zz := xx;
    np := abs(nn);
    m := np div 2;
    if np = 0 then
      Result := '1'
    else if nn = 1 then
      Result := xx
    else if m > 0 then
      begin
        result := ConvertPowerSymbolProv(zz, m);
        result := 'D(' + result + ')';
        if odd(np) then
          result := result + '*' + zz;
      end
    else
      Result := zz;
  end;

begin
  z := ConvertPowerSymbolProv(x, n);
  Result := '(' + z + ')';
  if n < 0 then
    Result := '(' + '1/' + Result + ')';
end;

function PieceEnd(const first : Integer; const theText, soughtChars :
  AnsiString; var symbol : char) : Integer;
{This function acts upon 'theText'. It is assumed that 'theText[first]'
is not in the set [+,-,*,/,^] (but possibly is '#'). Then, it scans forward
'theText' until in reaches one of the symbols in 'endChars' while being at
>=zero parentheses depth. Result is length[theText] if not found or the index
previous to that symbol.}
var
  i, len, depth : Integer;
  ch : char;
  found : Boolean;
begin
  symbol := '?';
  len := Length(theText);
  i := first - 1;
  depth := 0;
  found := False;
  while (i < len) and not found do
    begin
      inc(i);
      ch := theText[i];
      if (ch = '(') then
        inc(depth)
      else if (ch = ')') then
        dec(depth);
      if ((pos(ch, soughtChars) <> 0) and (depth = 0)) or (depth < 0) then
        found := true;
    end;
  if found then
    begin
      PieceEnd := i - 1;
      symbol := ch;
    end
  else
    PieceEnd := len;
end;

function PieceBeginning(const last : Integer; soughtChars : AnsiString) :
  Integer;
{This function acts upon 'uReduced'. It is assumed that 'uReduced[last]'
is not in the set [+,-,*,/,^]. Then, it scans backwards
'uReduced' until in reaches one of the symbols in 'endChars' while being at
<=zero parentheses depth. Result is 1 (if not found) or the index following that symbol.}
var
  i, depth : Integer;
  ch : char;
  found : Boolean;
begin
  i := last + 1;
  depth := 0;
  found := False;
  while (i > 1) and not found do
    begin
      dec(i);
      ch := uReduced[i];
      if (ch = ')') then
        inc(depth)
      else if (ch = '(') then
        dec(depth);
      if ((pos(ch, soughtChars) <> 0) and (depth = 0)) or (depth < 0) then
        found := true;
    end;
  if found then
    PieceBeginning := i + 1
  else
    PieceBeginning := 1;
end;

function computeExponent(var n : Integer; aF : TOCFunctDDDoubleSafe; ini, fini
  :
  Integer) : Boolean;
var
  c1, c2 : char;
  negate : boolean;
  firstExp, lastExp, shift : Integer;
  exponent : Double;
begin
  firstExp := ini;
  lastExp := fini;
  c1 := uReduced[firstExp];
  c2 := uReduced[lastExp];
  if (c1 = '(') then
    begin
      result := (c2 = ')');
      if result then
        begin
          inc(firstExp);
          dec(lastExp);
          c1 := uReduced[firstExp];
          c2 := uReduced[lastExp];
        end;
    end
  else
    result := (c1 = '1') or (c1 = kLastChar);
  negate := (c1 = '#');
  if negate then
    begin
      inc(firstExp);
      c1 := uReduced[firstExp];
    end;

  if (lastExp = firstExp) and (c1 = '1') then
    begin
      n := 1;
      Result := True;
    end
  else if ((lastExp - firstExp) = 1) and (c1 = kLastChar) and (ord(c2) > aF.argNum+aF.parNum)
    then
    begin
      shift := ord(c2) - aF.argNum;
      exponent := aF.data[shift];
      n := round(exponent);
      Result := (n = exponent);
    end
  else
    result := False;
  if result and negate then n := -n;
end;


procedure convertPowers(aF : TOCFunctDDDoubleSafe);
var
  newString : AnsiString;
  reducedLength, firstBase, lastBase, firstExp, lastExp, i, n: Integer;
  c : char;
  good : Boolean;
begin
  reducedLength := Length(uReduced);
  if reducedLength < 3 then
    Exit;
  i := 1;
  repeat
    inc(i);
    c := uReduced[i];
    if (c = '^') then
      begin
        firstExp := i + 1;
        lastExp := PieceEnd(firstExp, uReduced, '+-*/^', c);
        good := computeExponent(n, af, firstExp, lastExp);
        if good then
          begin
            lastBase := i - 1;
            firstBase := PieceBeginning(lastBase, '+-*/^#');
            newString := Copy(uReduced, firstBase, lastBase - firstBase + 1);
            newString := ConvertPowerSymbol(newString, n);
            Insert(newString, uReduced, lastExp + 1);
            Delete(uReduced, firstBase, lastExp - firstBase + 1);
            reducedLength := Length(uReduced);
            i := firstBase + Length(newString) - 1;
          end
      end;
  until (i > reducedLength - 1);
end;



function ExpressionString(theText : AnsiString) : AnsiString;
  FORWARD;

function AtomString(theText : AnsiString) : AnsiString;
var
  base : AnsiString;
  len : Integer;
  ch : char;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else
    begin
      ch := theText[1];
      if ch = '(' then
        begin
          base := Copy(theText, 2, len - 2);
          result := ExpressionString(base);
        end
      else
        begin
          Assert(not (pos(ch, 'qrstuvwxyz') <> 0));
          case ch of
            '1', 'p' :
              Result := ch;
            kLastChar :
              Result := theText;
            'A'..'Z' :
              begin
                base := Copy(theText, 3, len - 3);
                result := ExpressionString(base) + ch;
              end;
          else
            ;
          end;
        end;
    end;
end; {AtomString}

function FactorString(theText : AnsiString) : AnsiString;
var
  neg : Boolean;
  redStr, anAtom : AnsiString;
  len, first, last : Integer;
  symbol, prevSymbol : char;
begin
  if (theText = '') then
    Result := ''
  else
    begin
      neg := (theText[1] = '#');
      if neg then
        begin
          len := Length(theText) - 1;
          redStr := Copy(theText, 2, len);
        end
      else
        begin
          len := Length(theText);
          redStr := theText;
        end;
      if len = 0 then
        Result := ''
      else
        begin
          last := PieceEnd(1, redStr, '^', symbol);
          if last = len then
            Result := AtomString(redStr)
          else
            begin
              anAtom := Copy(redStr, 1, last);
              prevSymbol := symbol;
              result := AtomString(anAtom);
              repeat
                first := last + 2;
                last := PieceEnd(first, redStr, '^', symbol);
                anAtom := Copy(redStr, first, last - first + 1);
                if (prevSymbol = '^') then
                  Result := AtomString(anAtom) + Result + 'F*C'
                else
                  Result := AtomString(anAtom) + Result + prevSymbol;
                prevSymbol := symbol;
              until last = len;
            end;
          if neg then Result := Result + '#';
        end;
    end;
end; {of FactorString}

function TermString(theText : AnsiString) : AnsiString;
var
  aFactor : AnsiString;
  len, first, last : Integer;
  symbol, prevSymbol : char;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else
    begin
      last := PieceEnd(1, theText, '*/', symbol);
      if last = len then
        Result := FactorString(theText)
      else
        begin
          aFactor := Copy(theText, 1, last);
          prevSymbol := symbol;
          result := FactorString(aFactor);
          repeat
            first := last + 2;
            last := PieceEnd(first, theText, '*/', symbol);
            aFactor := Copy(theText, first, last - first + 1);
            Result := Result + FactorString(aFactor) + prevSymbol;
            prevSymbol := symbol;
          until last = len;
        end;
    end;
end; {of TermString}


function ExpressionString(theText : AnsiString) : AnsiString;
var
  aTerm : AnsiString;
  len, first, last : Integer;
  symbol, prevSymbol : char;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else
    begin
      last := PieceEnd(1, theText, '+-', symbol);
      if last = len then
        Result := TermString(theText)
      else
        begin
          aTerm := Copy(theText, 1, last);
          prevSymbol := symbol;
          result := TermString(aTerm);
          repeat
            first := last + 2;
            last := PieceEnd(first, theText, '+-', symbol);
            aTerm := Copy(theText, first, last - first + 1);
            Result := Result + TermString(aTerm) + prevSymbol;
            prevSymbol := symbol;
          until last = len;
        end;
    end;
end; {of ExpressionString}




function TOCFunctDDDoubleSafe.ParseNewText(t, vars, pars : AnsiString) :
  AnsiString;
var
  theErrorNumber, i : Integer;
  ch : char;
begin
  Result := '';
  rawText := t;
  uFormula := t;
  if vars = '' then
    begin
      allowedVars := variables;
      argNum := Length(variables);
    end
  else
    begin
      allowedVars := vars;
      StrUpper(PChar(allowedVars));
      argNum := Length(vars);
      for i := 1 to argNum do
        begin
          ch := allowedVars[i];
          if not (pos(ch, variables) <> 0) then
            begin
              Result := uStdMessages[eBadVarSet];
              Exit;
            end;
        end;
    end;
  allowedPars := pars;
  if pars <> '' then begin
    StrUpper(PChar(allowedPars));
    parNum := Length(allowedPars);
    for i := 1 to parNum do begin
      ch := allowedPars[i];
      if (pos(ch,allowedVars)<>0) or (pos(ch, variables) = 0) then begin
        Result := uStdMessages[eBadParamSet];
        Exit;
      end;
    end;
  end;
  allVars:= allowedVars + allowedPars;
  data := nil;
  SetLength(data, 8+ parNum);
  {data[0] is not used}
  data[0] := 0;

  if parNum >0 then for i:= 1 to parNum do
    data[i]:=0;
  dataLast := parNum;
  theErrorNumber := UnifyChars;
  if theErrorNumber = eThereIsNoError then
    begin
      theErrorNumber := ReduceSymbols(self, False);
      if theErrorNumber = eThereIsNoError then
        begin
          theErrorNumber := CheckFunction;
          if theErrorNumber = eThereIsNoError then
            begin
              ConvertPowers(self);
              actorList := expressionString(uReduced);
              lengthActorList := Length(actorList);
              CodifyFunction;
            end;
        end;
    end;
  if theErrorNumber <> eThereIsNoError then
    Result := uStdMessages[theErrorNumber];
end;



procedure TOCFunctDDDoubleSafe.AddIntToCode(n : Integer; var cod: AnsiSTring);
var
  c : PChar;
  s : AnsiString;
  i : Integer;
begin
  c := @n;
  s := '';
  for i := 0 to SizeOf(Integer) - 1 do
    s := s + c[i];
  cod := cod + s;
end;

function sn(n : Pointer) : AnsiString;
var
  c : PChar;
  pn, i : Integer;
begin
  pn := Integer(n);
  c := @pn;
  Result := '';
  for i := 0 to SizeOf(Integer) - 1 do
    Result := Result + c[i];
end;

function snd(p: Pointer; displ: Integer): AnsiSTring;
var
  c : PChar;
  pn, i : Integer;
begin
  pn := Integer(p)+ displ;
  c := @pn;
  Result := '';
  for i := 0 to SizeOf(Integer) - 1 do
    Result := Result + c[i];
end;


function LoadXd(i: Integer): AnsiString;
{Reads from the function stack the independent variable 'i' of Double
type and stores it in ST(0)}
begin
  if i = 1 then
    result:= #$DD#$00
  else if i > 1 then
    result:= #$DD#$40 + chr(SizeOf(Double) * (i-1));
end;

function TOCFunctDDDoubleSafe.LoadVd(i: Integer): AnsiString;
{Reads from the function stack the velocity 'i' of Double
type and stores it in ST(0)}
begin
  result:= #$DD#$40 + chr(SizeOf(Double )* (argNum + i-1));
end;

function TOCFunctDDDoubleSafe.LoadAd(i: Integer): AnsiString;
{Reads from the function stack the velocity 'i' of Double
type and stores it in ST(0)}
begin
  result:= #$DD#$40 + chr(SizeOf(Double )* (2*argNum + i-1));
end;

function LoadDd(const a): AnsiString;
{Reads from the address of the var 'a' as a Double and stores it in ST(0)}
begin
  result:= #$DD#$05 + sn(@a);
end;

function WriteDdP(const a): AnsiString;
{Writes ST(0), as an Double, into the memory location @a, and pops ST(0)}
begin
  result:= #$DD#$1D + sn(@a);
end;

function WriteDd(const a): AnsiString;
{Writes ST(0), as an Double, into the memory location @a, but not pops}
begin
  result:= #$DD#$15 + sn(@a);
end;

PROCEDURE TOCFunctDDDoubleSafe.LeaveRoomInFPUStack(neededPlaces : Integer; var cod: AnsiSTring);
BEGIN
  IF emptyRegisters >= neededPlaces THEN
    Exit;
  IF emptyRegisters = 0 THEN BEGIN
      cod := cod
       + #$D9#$F6 {fdecstp: ST(cLastReg] becomes ST(0)}
       + #$83#$EC+ c_stackDisplacDoubleHStr
                                  {subtract c_stackDisplacDouble from ESP}
          +#$DD#$1C#$24;  {fstp [ESP]}
      inc(emptyRegisters);
  END;
  IF emptyRegisters < neededPlaces then
    REPEAT
      cod := cod + #$D9 + chr($C0 + c_LastReg - emptyRegisters)
          {push last occupied ST to ST(0)}
      + #$83#$EC+ c_stackDisplacDoubleHStr
                                  {subtract c_stackDisplacDouble from ESP}
      +#$DD#$1C#$24;  {fstp [ESP]}
      cod := cod + #$DD + chr($C0 + c_LastReg - emptyRegisters);
         {last occupied place contains the former value: declare it empty}
      inc(emptyRegisters);
    UNTIL emptyRegisters = neededPlaces;
END;


PROCEDURE TOCFunctDDDoubleSafe.PopFromAppStack;
VAR
  m : byte;
BEGIN
  m := c_LastReg +1 - emptyRegisters;
  cod := cod + #$DD + chr($D8 + m) {copy ST(0) to ST(m), which
     is the first free register, and pop st(0)}

   + #$DD#$04#$24 {fld [ESP]}
   + #$83#$C4 + c_stackDisplacDoubleHStr {add c_stackDisplacDouble to ESP};
  {if using Extended type, use 'c_stackDisplacExtendedHStr' instead}

  cod := cod + #$D9 + chr($C8 + m); {exchange ST(m) with ST(0)}
  dec(emptyRegisters);
END;

procedure TOCFunctDDDoubleSafe.PutCodeInExeBlock(cod: AnsiString;
     var exeBlock: PChar);
var
  len: Integer;
  i: Integer;
  ba: ByteArray;
begin
  len:= Length(cod);
  SetLength(ba, len+2);
  for i:= 0 to len-1 do
    ba[i]:= ord(cod[i+1]);

  exeBlock:= VirtualAlloc(nil, len + SizeOf(exeBlock), MEM_COMMIT, PAGE_EXECUTE_READWRITE);
  for i:= 0 to len-1 do
    exeBlock[i]:= chr(ba[i]);
  ba:= nil;
end;

procedure TOCFunctDDDoubleSafe.CodifyFunction;
var
  i, n : Integer;
  ch : char;
  escape : Boolean;
begin
  functionCode := '';
  emptyRegisters := c_LastReg + 1;
  i := 0;
  SetUWLength;
  while i < lengthActorList do
    begin
      inc(i);
      ch := actorList[i];
      escape := (ch = kLastChar);
      if escape or (ch = 'p') or (ch = '1') then
        begin
          if escape then inc(i);
          ch := actorList[i];
          LeaveRoomInFPUStack(1, functionCode);
          if escape then
            begin
              if (ord(ch) > argNum) then
                functionCode := functionCode + LoadDd(data[ord(ch) - argNum])
              else if (ord(ch)<= argNum) then
                functionCode := functionCode + LoadXd(ord(ch));
            end
          else if (ch = 'p') then
            functionCode := functionCode + c_LoadPi
          else
            functionCode := functionCode + c_Load1;
          dec(emptyRegisters);
        end
      else
        begin
          if (ch in ['+', '-', '*', '/', '^']) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack(functionCode);
              case ch of
                '+' :
                  functionCode := functionCode + #$DE#$C1;
                '-' :
                  functionCode := functionCode + #$DE#$E9;
                '*' :
                  functionCode := functionCode + #$DE#$C9;
                '/' :
                  begin
                    LeaveRoomInFPUStack(1, functionCode);
                    functionCode := functionCode + c_div_S;
                  end;
                '^' :
                  begin
                    LeaveRoomInFPUStack(1, functionCode);
                    functionCode := functionCode + c_zero_comp +
                      #$72#$08#$DD#$D8#$DD#$D8#$D9#$EE#$EB#$42#$D9#$F1 +
                      c_2toX_S;
                  end;
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in ['A', 'B', 'C', 'E', 'F', 'G', 'M']) then
                LeaveRoomInFPUStack(1, functionCode)
              else if (ch in ['H', 'I', 'J', 'K', 'L', 'O', 'P', 'Q', 'R']) then
                LeaveRoomInFPUStack(2, functionCode);
              case ch of
                '#' :
                  functionCode := functionCode + #$D9#$E0;
                'A' :
                  functionCode := functionCode + c_mod_2pi + #$D9#$FF;
                'B' :
                  functionCode := functionCode + c_mod_2pi + #$D9#$FE;
                'C' :
                  functionCode := functionCode + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S;
                'D' :
                  functionCode := functionCode + #$D8#$C8;
                'E' :
                  functionCode := functionCode + c_zero_comp +
                    #$72#$06 +
                    #$DD#$D8#$D9#$EE#$EB#$02#$D9#$FA;
                'F' :
                  functionCode := functionCode + c_ln_S;
                'G' :
                  functionCode := functionCode + c_mod_pi + #$D9#$F2#$DD#$D8;

                'H' :
                  functionCode := functionCode + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D9#$C0#$D9#$E8#$DE +
                    #$F1#$DE#$E9+ c_div2;
                'I' :
                  functionCode := functionCode + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D9#$C0#$D9#$E8#$DE +
                    #$F1#$DE#$C1+ c_div2;
                'J' :
                  functionCode := functionCode + #$D9#$EA#$DE#$C9 +
                    c_2ToX_S + #$D8#$C8#$D9#$E8#$D8#$E9#$D9#$C9 +
                    #$D9#$E8#$DE#$C1#$DE#$F9;
                'K' :
                  functionCode := functionCode + #$D9#$C0#$D9#$E1 +
                    c_1_comp +
                    #$DD#$D8#$72#$0E#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1#$D9#$FA#$D9#$F3 +
                    #$EB#$10 + c_zero_comp + #$DD#$D8 +
                    LoadDd(uSpecialValues[0]) +
                    #$72#$02#$D9#$E0;
                'L' :
                  functionCode := functionCode + #$D9#$C0#$D9#$E1 +
                    c_1_comp +
                    #$DD#$D8#$72#$10#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1#$D9#$FA#$D9#$C9 +
                    #$D9#$F3#$EB#$0E + c_zero_comp +
                    #$DD#$D8#$72#$04#$D9#$EB#$EB#$02 +
                    #$D9#$EE;
                'M' : functionCode := functionCode + #$D9#$E8#$D9#$F3;
                'N' : functionCode := functionCode + #$D9#$E1;
                'O' :
                  functionCode := functionCode + c_mod_pi +
                    #$D9#$FB#$D9#$C9 + c_div_S;
                'P' :
                  functionCode := functionCode + c_1_comp +
                    #$77#$18#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E9#$D9#$FA#$DE#$C1 +
                    c_ln + #$EB#$04#$DD#$D8#$D9#$EE;
                'Q' :
                  functionCode := functionCode +
                    #$D9#$C0#$D9#$E1 + c_1_comp + #$DD#$D8#$76#$24
                    +
                    #$D9#$C0#$D9#$E8#$DE#$C1#$D9#$C9#$D9#$E8#$DE#$E1#$DE#$F9 +
                    c_ln + c_div2 + #$EB#$16
                    + c_zero_comp +
                    #$DD#$D8#$77#$08 +LoadDd(uSpecialValues[2]) +
                    #$EB#$06 + LoadDd(uSpecialValues[3]);
                'R' :
                  functionCode := functionCode + c_zero_comp +
                    #$D9#$E1#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$C1
                    +
                    #$D9#$FA#$DE#$C1 + c_ln +
                      #$76#$02#$D9#$E0;
              else
                ;
              end;
            end;
        end;
    end; {del while}
  functionCode := functionCode + #$C3;
  PutCodeInExeBlock(functionCode, VBlock);
  @V := @VBlock[0];
end;

procedure TOCFunctDDDoubleSafe.SetUWLength;
var
  i, j, n, maxJ : Integer;
  ch : char;
  escape : Boolean;
begin
  i := 0;
  j:= 0;
  maxJ:=0;
  while i < lengthActorList do begin
    inc(i);
    ch := actorList[i];
    escape := (ch = kLastChar);
    if escape or (ch = 'p') or (ch = '1') then begin
      if escape then
        inc(i);
      ch := actorList[i];
      inc(j);
      if j > maxJ then
        maxJ:= j;
    end
    else if (ch in ['+', '-', '*', '/', '^']) then
      dec(j)
  end; {del while}
  SetLength(u, maxJ+1);
  SetLength(w, maxJ+1);
end;

procedure TOCFunctDDDoubleSafe.BuildFirstDerivative;
var
  i, j, n : Integer;
  ch : char;
  escape : Boolean;
begin
  dC := '';
  emptyRegisters := c_LastReg + 1;
  i := 0;
  j:= 0;
  while i < lengthActorList do
    begin
      inc(i);
      ch := actorList[i];
      escape := (ch = kLastChar);
      if escape or (ch = 'p') or (ch = '1') then
        begin
          if escape then inc(i);
          ch := actorList[i];
          LeaveRoomInFPUStack(1, dC);
          j:= j+1;
          if escape then
            begin
              if (ord(ch) > argNum) then
                begin                  
                  dC:= dC + c_Load0;
                  dC:= dC + WriteDdP(u[j]);
                  dC := dC + LoadDd(data[ord(ch) - argNum]);
                end
              else if (ord(ch)<= argNum) then
                begin
                  dC:= dC + LoadVd(ord(ch));
                  dC:= dC + WriteDdP(u[j]);
                  dC := dC + LoadXd(ord(ch));
                end
            end
          else if (ch = 'p') then begin
            dC:= dC + c_Load0;
            dC:= dC + WriteDdP(u[j]);
            dC := dC + c_LoadPi;
          end
          else begin;
            dC:= dC + c_Load0;
            dC:= dC + WriteDdP(u[j]);
            dC := dC + c_Load1
          end;
          dec(emptyRegisters);
        end
      else
        begin
          if (ch in ['+', '-', '*', '/', '^']) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack(dC);
              if ch = '^' then
                LeaveRoomInFPUStack(3, dC)
              else if ch in ['*','/'] then
                LeaveRoomInFPUStack(2, dC)
              else
                LeaveRoomInFPUStack(1, dC);
              case ch of
                '+' : begin
                    dC:= dC + #$DE#$C1;
                    dC:= dC + LoadDd(u[j-1]);
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C1;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
                '-' : begin
                    dC:= dC + #$DE#$E9;
                    dC:= dC + LoadDd(u[j-1]);
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$E9;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
                '*' : begin
                    dC:= dC + #$D9#$C0#$D8#$CA#$D9#$CA;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + LoadDd(u[j-1]);
                    dC:= dC + #$DE#$CA#$DE#$C1;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
                '/' :
                  begin
                    dC:= dC + c_zero_comp;
                    dC:= dC + #$75#$14#$DD#$D8;
                    dC:= dC + LoadDd(uSpecialValues[2]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + c_Load0;
                    dC:= dC + WriteDdP(u[j-1]);
                    dC:= dC + #$EB#$28;
                    dC:= dC + c_Load1;
                    dC:= dC + #$DE#$F1#$D9#$C0#$D8#$CA#$D9#$CA;
                    dC:= dC + LoadDd(u[j-1]);
	                  dC:= dC + #$D8#$CA#$D9#$CA#$D8#$C8#$DE#$C9;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9;
	                  dC:= dC + #$DE#$E9;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
                '^' :
                  begin
                    dC:= dC + c_zero_comp;
                    dC:= dC + #$72#$0A#$DD#$D8#$DD#$D8;
                    dC:= dC + c_Load0 + c_Load0;
                    dC:= dC + #$EB#$70#$D9#$C0#$D8#$FA;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9#$D9#$CA#$D9#$C9;
                    dC:= dC + c_Load1;
	                  dC:= dC + #$D9#$C9#$D9#$F1#$D9#$C0#$D9#$CA#$DE#$C9;
                    dC:= dC + c_2toX_S + #$D9#$C9;
                    dC:= dC + LoadDd(u[j-1]);
	                  dC:= dC + #$DE#$C9#$D9#$ED#$DE#$C9#$DE#$C2#$DC#$C9#$D9#$C9;
                    dC:= dC + WriteDdP(u[j-1]);
                  end;
              end;
              inc(emptyRegisters);
              j:= j-1;
            end
          else {opCode of a function}
            begin
              if ch='#' then
                LeaveRoomInFPUStack(1, dC)
              else
                LeaveRoomInFPUStack(2, dC);
              case ch of
                '#' : begin
                    dC:= dC + #$D9#$E0;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$D9#$E0;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'A' : begin
                    dC:= dC + c_mod_2pi;
                    dC:= dC + #$D9#$FB#$D9#$C9#$D9#$E0;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'B' : begin
                    dC:= dC + c_mod_2pi; 
	                  dC:= dC + #$D9#$FB;
	                  dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'C' : begin
                    dC:= dC + #$D9#$EA#$DE#$C9;
                    dC:= dC + c_2ToX_S;
	                  dC:= dC + #$D9#$C0;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'D' : begin
                    dC:= dC + #$D9#$C0#$D8#$C8#$D9#$C9#$D8#$C0;
                    dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$C9;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'E' : begin
                    dC:= dC +c_zero_comp + #$72#$08;
	                  dC:= dC + #$DD#$D8 + c_Load0;
	                  dC:= dC + c_Load0;
	                  dC:= dC + #$EB#$0E#$D9#$FA#$D9#$C0#$D8#$C0;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$F1;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'F' : begin
                    dC:= dC +c_zero_comp + #$72#$0E;
	                  dC:= dC + #$DD#$D8;
	                  dC:= dC + LoadDd(uSpecialValues[6]);
                    dC:= dC + #$D9#$E0;
	                  dC:= dC + c_Load0;
	                  dC:= dC + #$EB#$16#$D9#$C0;
                    dC:= dC + c_ln;
                    dC:= dC + #$D9#$C9;
	                  dC:= dC + LoadDd(u[j]);
	                  dC:= dC + #$DE#$F1;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'G' : begin
                    dC:= dC + c_mod_pi + #$D9#$F2#$D9#$C1#$D8#$C8#$DE#$C1;
                    dC:= dC + LoadDd(u[j]) + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'H' : begin
                    dC:= dC + #$D9#$EA#$DE#$C9;
	                  dC:= dC + c_2ToX_S;
	                  dC:= dC + c_Load1;
	                  dC:= dC + #$D9#$E0#$D9#$C1#$D8#$F9#$D9#$FD#$D9#$CA#$D9#$FD;
                    dC:= dC + #$DD#$D9#$D9#$C0#$D8#$C2#$D9#$CA#$DE#$E9;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
	                  dC:= dC + WriteDdP(u[j]);
                  end;
                'I' : begin
                    dC:= dC + #$D9#$EA#$DE#$C9;
                    dC:= dC + c_2ToX_S;
	                  dC:= dC + c_Load1;
                    dC:= dC + #$D9#$E0#$D9#$C1#$D8#$F9#$D9#$FD#$D9#$CA#$D9#$FD;
                    dC:= dC + #$DD#$D9#$D9#$C0#$D8#$C2#$D9#$C9#$DE#$E2;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'J' : begin
                    dC:= dC + #$D9#$EA#$DE#$C9;
                    dC:= dC + c_2ToX_S;
                    dC:= dC + #$D8#$C8;
                    dC:= dC + c_Load1;
                    dC:= dC + #$D8#$E9#$D9#$C9;
                    dC:= dC + c_Load1;
                    dC:= dC + #$DE#$C1#$DE#$F9#$D9#$C0#$D8#$C8;
                    dC:= dC + c_Load1;
                    dC:= dC + #$DE#$E1;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'K' : begin
                    dC:= dC + #$D9#$C0#$D9#$E1 + c_1_comp;
                    dC:= dC + #$DD#$D8#$76#$18#$D9#$C0#$D8#$C8;
                    dC:= dC + c_Load1 + #$DE#$E1#$D9#$FA;
                    dC:= dC + c_Load1 + #$D8#$F1#$D9#$CA#$D9#$C9#$D9#$F3#$D9#$C9;
                    dC:= dC + #$EB#$16;
	                  dC:= dC + c_zero_comp;
                    dC:= dC + #$DD#$D8 + c_Load1 + #$D9#$E0 + c_LoadPi + #$D9#$FD;
                    dC:= dC + #$DD#$D9#$72#$02#$D9#$E0 + c_Load0;
                    dC:= dC + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9;
                    dC:= dC + WriteDdP(u[j]);
                  end;
                'L' : begin
                    dC:= dC + #$D9#$C0#$D9#$E1;
                    dC:= dC + c_1_comp;
                    dC:= dC + #$DD#$D8#$76#$18#$D9#$C0#$D8#$C8;
                    dC:= dC + c_Load1 + #$DE#$E1#$D9#$FA;
                    dC:= dC + c_Load1 +                                                             #$D8#$F1#$D9#$E0#$D9#$CA#$D9#$F3#$D9#$C9#$EB#$10;
                    dC:= dC + c_zero_comp + #$DD#$D8#$72#$04 + c_LoadPi + #$EB#$02;
                    dC:= dC + c_Load0 + c_Load0 + LoadDd(u[j]) + #$DE#$C9 +                               WriteDdP(u[j]);
                  end;
                'M' : begin
                    dC:= dC + #$D9#$C0 + c_Load1 + #$D9#$F3#$D9#$C9#$D8#$C8;
                    dC:= dC + c_Load1 + #$DC#$C1#$DE#$F1 + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9 + WriteDdP(u[j]);
                  end;
                'N' : begin
                    dC:= dC + c_zero_comp + #$76#$10#$D9#$E0 + LoadDd(u[j]);
                    dC:= dC + #$D9#$E0 + WriteDdP(u[j]) ;
                  end;
                'O' : begin
                    dC:= dC + c_mod_pi + c_zero_comp + #$74#$12#$D9#$F2#$DC#$F1;
                    dC:= dC + #$D9#$C9#$D9#$C0#$D8#$C8#$DE#$C2#$D9#$C9#$D9#$E0;
                    dC:= dC + #$EB#$06#$DD#$D8 + c_Load0 + c_Load0;
                    dC:= dC + LoadDd(u[j]) + #$DE#$C9 + WriteDdP(u[j]);
                  end;
                'P' : begin
                    dC:= dC + c_1_comp + #$73#$22#$D9#$C0#$D8#$C8 + c_Load1 +                                      #$DE#$E9;
                    dC:= dC + #$D9#$FA#$D9#$C0#$DE#$C2#$D9#$C9 + c_ln;
                    dC:= dC + #$D9#$C9 + c_Load1 + #$DE#$F1#$EB#$06#$DD#$D8;
                    dC:= dC + c_Load0 + c_Load0 + LoadDd(u[j]) + #$DE#$C9 +                               WriteDdP(u[j]);
                  end;
                'Q' : begin
                    dC:= dC + #$D9#$C0#$D9#$E1 + c_1_comp + #$DD#$D8#$76#$2E;
                    dC:= dC + #$D9#$C0 + c_Load1 + #$DE#$C1#$D9#$C9 + c_Load1;
                    dC:= dC + #$DE#$E1#$D9#$C0#$D8#$CA#$D9#$CA#$DE#$F1;
                    dC:= dC + c_ln + c_Load1 + #$D9#$E0#$D9#$C9#$D9#$FD#$D9#$CA;
                    dC:= dC + #$DE#$F9#$D9#$E0#$EB#$18 + c_zero_comp;
                    dC:= dC + #$DD#$D8#$77#$08 + LoadDd(uSpecialValues[2]);
                    dC:= dC + #$EB#$06 + LoadDd(uSpecialValues[3]) + c_Load0;
                    dC:= dC + LoadDd(u[j]) + #$DE#$C9 + WriteDdP(u[j]);
                  end;
                'R' : begin
                    dC:= dC + #$D9#$C0#$D8#$C8 + c_Load1 + #$DE#$C1#$D9#$FA;
                    dC:= dC + #$D9#$C0#$D9#$CA#$DE#$C1 + c_ln + #$D9#$C9;
                    dC:= dC + c_Load1 + #$DE#$F1 + LoadDd(u[j]);
                    dC:= dC + #$DE#$C9 + WriteDdP(u[j]);
                  end;
              else
                 ;
              end;
            end;
        end;
    end;
  dC:= dC + #$C3;       //return
  PutCodeInExeBlock(dC, DBlock);
  @FoA_Prime := @DBlock[0];
end;

procedure TOCFunctDDDoubleSafe.BuildSecondDerivative;
var
  i, j, n : Integer;
  ch : char;
  escape : Boolean;
begin
  ddC := '';
  emptyRegisters := c_LastReg + 1;
  i := 0;
  j:= 0;
  while i < lengthActorList do begin
    inc(i);
    ch := actorList[i];
    escape := (ch = kLastChar);
    if escape or (ch = 'p') or (ch = '1') then begin
      if escape then inc(i);
      ch := actorList[i];
      LeaveRoomInFPUStack(1, ddC);
      j:= j+1;
      if escape then begin
        if (ord(ch) > argNum) then
          ddC := ddC + 
              c_Load0 +  //PUSH  0 in ST(0)
              WriteDdP(u[j]) + // store ST(0) in u[j], and pop
              c_Load0 +  //PUSH  0 in ST(0)
              WriteDdP(w[j]) +  // store ST(0) in u[j], and pop
              LoadDd(data[ord(ch) - argNum])
        else if (ord(ch)<= argNum) then
          ddC := ddC  +
              LoadVd(ord(ch))+
              WriteDdP(u[j]) +
              LoadAd(ord(ch)) +
              WriteDdP(w[j]) +
              LoadXd(ord(ch))
      end
      else if (ch = 'p') then
        ddC := ddC + 
            c_Load0 +  //PUSH  0 in ST(0)
            WriteDdP(u[j]) + // store ST(0) in u[j], and pop
            c_Load0 +  //PUSH  0 in ST(0)
            WriteDdP(w[j])+  // store ST(0) in u[j], and pop
            c_LoadPi 
      else
        ddC := ddC + 
            c_Load0 +  //PUSH  0 in ST(0)
            WriteDdP(u[j]) + // store ST(0) in u[j], and pop
            c_Load0 +  //PUSH  0 in ST(0)
            WriteDdP(w[j]) + // store ST(0) in u[j], and pop
            c_Load1;
      dec(emptyRegisters);
    end
    else begin {opcode of a binary operator}
      if (ch in ['+', '-', '*', '/', '^']) then begin
        if (emptyRegisters = c_LastReg) then
          PopFromAppStack(ddC);
        if ch in ['+', '-']  then
          LeaveRoomInFPUStack(1, ddC)
        else
          LeaveRoomInFPUStack(3, ddC);
        case ch of
          '+': begin
             ddC:= ddC + #$DE#$C1 + // FADDP r_j + r_{j-1} in ST(0)
                LoadDd(u[j]) +
                LoadDd(u[j-1]) +
                #$DE#$C1 + // FADDP. u_j + u_{j-1} in ST(0), r_j + r_{j-1} in ST(1)
                WriteDdP(u[j-1]) +
                LoadDd(w[j]) +
                LoadDd(w[j-1]) +
                #$DE#$C1 + // FADDP. w_j + w_{j-1} in ST(0), r_j + r_{j-1} in ST(1)
                WriteDdP(w[j-1]);
               end;
          '-': begin
             ddC:= ddC + #$DE#$E9 + // FSUBP r_{j-1}-r_j in ST(0)
                LoadDd(u[j-1]) +
                LoadDd(u[j]) + // u_{j-1} in ST(0), u_j in ST(1), r_{j-1}-r_j in ST(2)
                #$DE#$E9 + // FSUBP. u_{j-1} - u_j in ST(0), r_{j-1}-r_j in ST(1)
                WriteDdP(u[j-1]) +
                LoadDd(w[j-1]) +
                LoadDd(w[j]) + // w_{j-1} in ST(0), w_j in ST(1), r_{j-1}-r_j in ST(2)
                #$DE#$E9 + // FSUBP. w_{j-1} - w_j in ST(0), r_{j-1}-r_j in ST(1)
                WriteDdP(w[j-1]);
               end;
          '*': begin
             ddC:= ddC + LoadDd(u[j])+ //u_j in ST(0),  r_j  in ST(1), r_{j-1} in ST(2)
                LoadDd(u[j-1]) + //u_{j-1}in ST(0), u_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                #$D9#$C0 + // FLD ST(0). u_{j-1}in ST(0) and ST(1), u_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D8#$CA + // FMUL ST(0) ST(2). u_{j_1}u_j in ST(0), u_{j-1} in ST(1), u_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D8#$C0 + // FADD ST(0) ST(0). 2u_{j_1}u_j in ST(0), u_{j-1} in ST(1), u_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D9#$CA + // FXCH ST(2).u_j in ST(0), u_{j-1} in ST(1), 2u_{j_1}u_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D8#$CC + // FMUL ST(0) ST(4). r_{j-1}u_j in ST(0), u_{j-1} in ST(1), 2u_{j_1}u_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D9#$C9 + // FXCH. u_{j-1} in ST(1), r_{j-1}u_j in ST(1),  2u_{j_1}u_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$D8#$CB + // FMUL ST(0) ST(3). r_ju_{j-1} in ST(0), r_{j-1}u_j in ST(1),  2u_{j_1}u_j in ST(2),  r_j  in ST(3), r_{j-1} in ST(4)
                #$DE#$C1 + // FADDP. r_ju_{j-1} +r_{j-1}u_j in ST(0), 2u_{j_1}u_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                WriteDdP(u[j-1])+ // 2u_{j_1}u_j in ST(0),  r_j  in ST(1), r_{j-1} in ST(2)
                LoadDd(w[j]) + // w_j in ST(0), 2u_{j_1}u_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                #$D8#$CB + // FMUL ST(0) ST(3).  w_j r_{j-1} in ST(0), 2u_{j_1}u_j in ST(1), r_j  in ST(2), r_{j-1} in ST(3)
                #$DE#$C1 + // FADDP. w_j r_{j-1} + 2u_{j_1}u_j in ST(0),  r_j  in ST(1), r_{j-1} in ST(2)
                LoadDd(w[j-1]) + // w_{j-1} in ST(0), w_j r_{j-1} + 2u_{j_1}u_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                #$D8#$CA + // FMUL ST(0) ST(2).  w_{j-1} r_j in ST(0), w_j r_{j-1} + 2u_{j_1}u_j in ST(1),  r_j  in ST(2), r_{j-1} in ST(3)
                #$DE#$C1 + // FADDP. w_j r_{j-1} + 2u_{j_1}u_j +w_{j-1} r_j in ST(0),  r_j  in ST(1), r_{j-1} in ST(2)
                WriteDdP(w[j-1])+ // FADDP. r_j in ST(0), r_{j-1} in ST(1)
                #$DE#$C9; // FMULP. r_j  r_{j-1} in ST(0)
               end;
          '/': begin
             ddC:= ddC + c_zero_comp +
                #$75#$1C +  //JNE, if r_j <> 0 jumps 28 bytes
                #$DD#$D8 + // POP ST(0). r_{j-1} (numerator) in ST(0)
                LoadDd(uSpecialValues[2]) +  //eMaxNum in ST(0), r_{j-1} in ST(1)
                #$DE#$C9 + //FMULP. r_{j-1*eMaxNum  in ST(0)
                c_Load0 +
                WriteDdP(u[j-1])+
                c_Load0 +
                WriteDdP(w[j-1])+
                #$EB#$48+ //JMP. jumps 72 bytes
                c_Load1 + //1 in ST(0), r_j in ST(1), r_{j-1} in ST(2)
                #$DE#$F1 + //FDIVRP.1/r_j in ST(0), r_{j-1} in ST(1)
                #$DC#$C9 + //FMUL ST(1) ST(0). 1/ r_j in ST(0), r_{j-1}/r_j in ST(1)
                LoadDd(u[j-1])+
                #$D8#$C9 + //FMUL ST(0) ST(1) u_{j-1}/r_j in ST(0), 1/r_j in ST(1), r_{j-1}/r_j in ST(2)
                LoadDd(u[j])+
                #$D8#$CA + //FMUL ST(0) ST(2) u_j/r_j in ST(0), u_{j-1}/r_j in ST(1), 1/r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D9#$C0 + //FLD ST(0). u_j/r_j in ST(0) and ST(1), u_{j-1}/r_j in ST(2), 1/ r_j in ST(3), r_{j-1}/r_j in ST(4)
                #$D8#$CC + //FMUL ST(0) ST(4). u_jr_{j-1}/r_j^2 in ST(0), u_j/r_j  ST(1), u_{j-1}/r_j in ST(2), 1/ r_j in ST(3), r_{j-1}/r_j in ST(4)
                #$D8#$EA + //FSUBR ST(0) ST(2). u_{j-1}/r_j - u_jr_{j-1}/r_j^2 in ST(0), u_j/r_j  ST(1), u_{j-1}/r_j in ST(2), 1/ r_j in ST(3), r_{j-1}/r_j in ST(4)
                WriteDdP(u[j-1])+ //u_j/r_j  ST(0), u_{j-1}/r_j in ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$DC#$C9 + //FMUL ST(1) ST(0). u_j/r_j  in ST(0), u_{j-1}u_j/r_j ^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D8#$C8 + //FMUL ST(0) ST(0). u_j^2/r_j^2  in ST(0), u_{j-1}u_j/r_j ^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D8#$CB + //FMUL ST(0) ST(3). u_j^2r_{j-1}/r_j^3 in ST(0), u_{j-1}u_j/r_j ^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$DE#$E1 + //FSUBRP. u_j^2r_{j-1}/r_j^3 -u_{j-1}u_j/r_j ^2 in  ST(0), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D8#$C0 + //FADD ST(0) ST(0). 2u_j^2r_{j-1}/r_j^3 -2u_{j-1}u_j/r_j ^2 in  ST(0), 1/ r_j in ST(1), r_{j-1}/r_j in ST(2)
                LoadDd(w[j-1])+
                #$D8#$CA + //FMUL ST(0) ST(2). w_{j-1}/r_j in ST(0), 2u_j^2r_{j-1}/r_j^3 -2u_{j-1}u_j/r_j ^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$DE#$C1 + //FADDP. w_{j-1}/r_j +2u_j^2r_{j-1}/r_j^3 -2u_{j-1}u_j/r_j ^2 in  ST(0), 1/ r_j in ST(1), r_{j-1}/r_j in ST(2)
                LoadDd(w[j])+ //w_j in ST(0), w_{j-1}/r_j +2u_j^2r_{j-1}/r_j^3 -2u_{j-1}u_j/r_j ^2 in  ST(1), 1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$D8#$CB + //FMUL ST(0) ST(3). w_j r_{j-1}/r_j in ST(0), w_{j-1}/r_j +2u_j^2r_{j-1}/r_j^3 -2u_{j-1}u_j/r_j ^2 in ST(1),   1/ r_j in ST(2), r_{j-1}/r_j in ST(3)
                #$DE#$CA + //FMULP ST(2) ST(0). w_{j-1}/r_j +2u_j^2r_{j-1}/r_j^3 -2u_{j-1}u_j/r_j ^2 in  ST(0), w_j r_{j-1}/r_j^2 in ST(1), r_{j-1}/r_j in ST(2)
                #$DE#$E1 + //FSUBRP. w_{j-1}/r_j-w_j r_{j-1}/r_j^2 +2u_j^2r_{j-1}/r_j^3 -2u_{j-1}u_j/r_j ^2 in ST(0), r_{j-1}/r_j in ST(1)
                WriteDdP(w[j-1]);
               end;
          else
            ;
        end;
        inc(emptyRegisters);
        j:= j-1;
      end
      else begin {opCode of a function}
        if ch in ['#','N'] then
          LeaveRoomInFPUStack(1, ddC)
        else if ch= 'C' then
          LeaveRoomInFPUStack(2, ddC)
        else
          LeaveRoomInFPUStack(3, ddC);
        case ch of
          '#': begin
                 ddC:= ddC + #$D9#$E0 + //FCHS, - r_j in  ST(0)
                   LoadDd(u[j]) + // u[j] in ST(0), -r_j in ST(1)
                   #$D9#$E0 + //FCHS,  -u[j] in ST(0), -r_j in ST(1)
                   WriteDdP(u[j]) +
                   LoadDd(w[j]) + // w[j] in ST(0), -r_j in ST(1)
                   #$D9#$E0 + //FCHS,  -w[j] in ST(0), -r_j in ST(1)
                   WriteDdP(w[j]);
               end;
          'A': begin
                 ddC:= ddC + c_mod_2pi +
                   #$D9#$FB + //FSINCOS. cos(r_j) in ST(0), sin(r_j) in ST(1)
                   #$D9#$C9 + //FXCH. interchange ST(0) and ST(1).
                   #$D9#$E0 + //FCHS. -sin(r_j) in ST(0), cos(r_j) in ST(1)
                   LoadDd(u[j]) + // u_j in ST(0), -sin(r_j) in ST(1), cos(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0).u_j in ST(0) and ST(1), -sin(r_j) in ST(2), cos(r_j) in ST(3)
                   #$D8#$C8 + // FMUL ST(0) ST(0).  u_j^2 in ST(0), u_j in ST(1), -sin(r_j) in ST(2), cos(r_j) in ST(3)
                   #$D8#$CB + // FMUL ST(0) ST(3).  u_j^2 cos(r_j) in ST(0), u_j in ST(1), -sin(r_j) in ST(2), cos(r_j) in ST(3)
                   #$D9#$C9 + //FXCH. u_j in ST(0), u_j^2 cos(r_j) in ST(1), -sin(r_j) in ST(2), cos(r_j) in ST(3)
                   #$D8#$CA + // FMUL ST(0) ST(2). -u_jsen(r_j) in ST(0), u_j^2 cos(r_j) in ST(1), -sin(r_j) in ST(2), cos(r_j) in ST(3)
                   WriteDdP(u[j])+ //u_j^2 cos(r_j) in ST(0), -sin(r_j) in ST(1), cos(r_j) in ST(2)
                   LoadDd(w[j]) + //w_j in ST(0), u_j^2 cos(r_j) in ST(1), -sin(r_j) in ST(2), cos(r_j) in ST(3)
                   #$DE#$CA + //FMULP ST(2) ST(0). u_j^2 cos(r_j) in ST(0), -w_j sin(r_j) in ST(1), cos(r_j) in ST(2)
                   #$DE#$E9 + //FSUBP ST(1) ST(0).  -u_j^2 cos(r_j) -w_j sin(r_j) in ST(0), cos(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'B': begin
                 ddC:= ddC + c_mod_2pi +
                   #$D9#$FB + //FSINCOS. cos(r_j) in ST(0), sin(r_j) in ST(1)
                   LoadDd(u[j]) + // u_j in ST(0), cos(r_j) in ST(1), sin(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0).u_j in ST(0) and ST(1), cos(r_j) in ST(2), sin(r_j) in ST(3)
                   #$D8#$C8 + // FMUL ST(0) ST(0).  u_j^2 in ST(0), u_j in ST(1), cos(r_j) in ST(2), sin(r_j) in ST(3)
                   #$D8#$CB + // FMUL ST(0) ST(3).  u_j^2 sin(r_j) in ST(0), u_j in ST(1), cos(r_j) in ST(2), sin(r_j) in ST(3)
                   #$D9#$C9 + //FXCH. u_j in ST(0), u_j^2 sin(r_j) in ST(1), cos(r_j) in ST(2), sin(r_j) in ST(3)
                   #$D8#$CA + // FMUL ST(0) ST(2). u_jcos(r_j) in ST(0), u_j^2 sin(r_j) in ST(1), cos(r_j) in ST(2), sin(r_j) in ST(3)
                   WriteDdP(u[j])+ //u_j^2 sin(r_j) in ST(0), cos(r_j) in ST(1), sin(r_j) in ST(2)
                   LoadDd(w[j]) + //w_j in ST(0), u_j^2 sin(r_j) in ST(1), cos(r_j) in ST(2), sin(r_j) in ST(3)
                   #$DE#$CA + //FMULP ST(2) ST(0). u_j^2 sin(r_j) in ST(0), w_j cos(r_j) in ST(1), sin(r_j) in ST(2)
                   #$DE#$E9 + //FSUBP ST(1) ST(0).  w_j sin(r_j) -u_j^2 cos(r_j) in ST(0), senj) in ST(1)
                   WriteDdP(w[j]);
               end;
          'C': begin
                 ddC:= ddC + #$D9#$EA + //FLDL2E, Load log in base 2 of e
                   #$DE#$C9 + //FMULP. r_j*log_2e  in ST(0)
                   c_2ToX_S + //exp(r_j) in ST(0)
                   LoadDd(u[j])+ //u_j in ST(0), exp(r_j) in ST(1)
                   #$D9#$C0 + // FLD ST(0). u_j in ST(0) and ST(1), exp(r_j) in ST(2)
                   #$D8#$CA + // FMUL ST(0) ST(2).  u_j exp(r_j) in ST(0), u_j in ST(1), exp(r_j) in ST(2)
                   WriteDdP(u[j]) + //  u_j in ST(0), exp(r_j) in ST(1)
                   #$D8#$C8 + // FMUL ST(0) ST(0). u_j^2 in ST(0), exp(r_j) in ST(1)
                   LoadDd(w[j]) + //w_j in ST(0), u_j^2 in ST(1), exp(r_j) in ST(2)
                   #$DE#$C1 + // ADDP. w_j +u_j^2 in ST(0), exp(r_j) in ST(1)
                   #$D8#$C9 + //FMULP. ( w_j +u_j^2 )exp(r_j) in ST(0), exp(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'D': begin
                 ddC:= ddC + #$D9#$C0 +  //FLD ST(0). r_j in ST(0)  and  in  ST(1)
                   #$D8#$C8 + //FMUL ST(0), ST(0). r_j^2 in ST(0), r_j in ST(1)
                   #$D9#$C9 + //FXCH. r_j in ST(0), r_j ^2en ST(1)
                   #$D8#$C0 + //ADD ST(0) ST(0). 2r_j in ST(0), r_j^2 in ST(1)
                   LoadDd(w[j])+ //w_j in ST(0), 2r_j in ST(1), r_j ^2 in ST(2)
                   #$D8#$C9 + // FMUL ST(0) ST(1). 2w_jr_j in ST(0), 2r_j in ST(1), r_j^2 in ST(2)
                   LoadDd(u[j])+ //u_j in ST(0), 2w_jr_j in ST(1), 2r_j in ST(2), r_j^2 in ST(3)
                   #$DC#$CA + // FMUL ST(2) ST(0) u_j in ST(0), 2w_jr_j in ST(1), 2r_ju_j in ST(2), r_j^2 in ST(3)
                   #$D9#$CA + // FXCH ST(2). 2r_ju_j in ST(0), 2w_jr_j in ST(1), u_j in ST(2), r_j^2 in ST(3)
                   WriteDdP(u[j]) + //2w_jr_j in ST(0), u_j in ST(1), r_j^2 in ST(2)
                   #$D9#$C9 + //FXCH. u_j in ST(0), 2w_jr_j in ST(1),  r_j^2 in ST(2)
                   #$D8#$C8 + // FMUL ST(0) ST(0). u_j^2 in ST(0), 2w_jr_j in ST(1),  r_j^2 in ST(2)
                   #$D8#$C0 + //ADD ST(0) ST(0). 2u_j^2 in ST(0), 2w_jr_j in ST(1),  r_j^2 in ST(2)
                   #$DE#$C1 + //ADDP. 2u_j^2 +2w_jr_j in ST(0),  r_j^2 in ST(1)
                   WriteDdP(w[j]);
               end;
          'E': begin
                 ddC:= ddC +c_zero_comp + #$72#$16 + //if ST(0)>0, jumps 22 bytes)
		               #$DD#$D8 + c_Load0 + //FSTP ST(0) and load 0. Substitute ST(0) value by 0
		               c_Load0 +  //0 in ST(0) and in ST(1)
                   WriteDdP(u[j]) +
		               c_Load0 +  //0 in ST(0) and in ST(1)
                   WriteDdP(w[j]) +
                   #$EB#$32 +  //jumps the following 50 bytes
                   #$D9#$FA + // FSQRT. sqrt{r_j} in ST(0)
		               #$D9#$C0 + // FLD ST(0).  sqrt{r_j} in ST(0) and  ST(1)
                   #$D8#$C0 + // FADD ST(0) ST(0). 2sqrt{r_j} in ST(0) , sqrt(r_j) in ST(1)
		               c_Load1 + // 1 in ST(0), 2sqrt{r_j} in ST(1) , sqrt(r_j) in ST(2)
                   #$DE#$F1 + // FDIVRP. 1/(2sqrt{r_j} ) in ST(0), sqrt(r_j) in ST(1)
                   #$D9#$C0 + // FLD ST(0). 1/(2sqrt{r_j} ) in ST(0) and ST(1), sqrt(r_j) in ST(2)
		               LoadDd(u[j])+ //u_j in ST(0), 1/(2sqrt{r_j} ) in ST(1) and ST(2), sqrt(r_j) in ST(3)
                   #$DE#$C9 + //FMULP. u_j/(2sqrt{r_j} ) in ST(0) , 1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   #$D9#$C0 +  //FLD ST(0). u_j/(2sqrt{r_j} ) in ST(0) and ST(1),  1/(2sqrt{r_j} ) in ST(2) , sqrt(r_j) in ST(3)
                   WriteDdP(u[j]) + // u_j/(2sqrt{r_j} ) in ST(0) ,  1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   #$D8#$C8 + // FMUL ST(0) ST(0). u_j^2/(4 r_j) in ST(0),  1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   #$D8#$C9 + // FMUL ST(0) ST(2). u_j^2/(8 r_jsqrt(r_j)) in ST(0),  1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   #$D8#$C0 + // FADD ST(0) ST(0).  u_j^2/(4 r_jsqrt(r_j)) in ST(0),  1/(2sqrt{r_j} ) in ST(1) , sqrt(r_j) in ST(2)
                   LoadDd(w[j])+ //w_j in ST(0), u_j^2/(4 r_jsqrt(r_j) ) in ST(1), 1/(2sqrt{r_j}) in ST(2) , sqrt(r_j) in ST(3)
  		             #$DE#$CA + //FMULP ST(2) ST(0). u_j^2/(4 r_jsqrt(r_j) ) in ST(0), w_j/(2sqrt{r_j}) in ST(1), sqrt(r_j) in ST(2)
		               #$DE#$E9 + //FSUBP.  w_j/(2sqrt{r_j})  - u_j^2/(4r_jsqrt(r_j)) in ST(0), sqrt(r_j) in ST(1)
		               WriteDdP(w[j]) ;
               end;
          'F': begin
                 ddC:= ddC +c_zero_comp + #$72#$1C + //if ST(0)>0, jumps 28 bytes
                   #$DD#$D8 + // FSTP ST(0) :pop  r_j
                   LoadDd(uSpecialValues[6])+ // load eMaxExp in ST(0)
                   #$D9#$E0 + // FCHS, -eMaxExp in ST(0)
                   c_Load0 + // 0 in ST(0), -eMaxExp in ST(1)
                   WriteDdP(u[j])+
                   c_Load0 +  //0 in ST(0) and in ST(1)
                   WriteDdP(w[j]) +
                   #$EB#$34 +  //jumps the following 52 bytes
                   #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
		               c_ln + // ln(r_j) in ST(0), r_j in ST(1)
                   #$D9#$C9 + // FXCH. r_j in ST(0), ln(r_j) in ST(1)
                   c_Load1 +  // 1 in ST(0), r_j in ST(1), ln(r_j) in ST(2)
                   #$DE#$F1 +  //FDIVRP 1/r_j in ST(0), ln(r_j) in ST(1)
	  	             LoadDd(u[j]) + //u_j in ST(0), 1/r_j in ST(1), ln(r_j) in ST(2)
                   #$D8#$C9 +  //FMUL ST(0) ST(1). u_j/r_j in ST(0), 1/r_j in ST(1), ln(r_j) in ST(2)
		               #$D9#$C0 + // FLD ST(0). u_j/r_j in ST(0) and in ST(1), 1/r_j in ST(2), ln(r_j) in ST(3)
                   WriteDdP(u[j]) + // u_j/r_j in ST(0), 1/r_j in ST(1), ln(r_j) in ST(2)
                   #$D8#$C8 + // FMUL ST(0) ST(0). u_j^2/r_j^2 in ST(0), 1/r_j in ST(1), ln(r_j) in ST(2)
		               LoadDd(w[j])+ //w_j in ST(0), u_j^2/r_j^2 in ST(1), 1/r_j in ST(2), ln(r_j) in ST(3)
		               #$DE#$CA + // FMULP ST(2) ST(0) . u_j^2/r_j^2 in ST(0), w_j/r_j in ST(1), ln(r_j) in ST(2)
		               #$DE#$E9 + //FSUBP. w_j/r_j -u_j^2/r_j^2  in ST(0), ln(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'G': begin
                 ddC:= ddC + c_mod_pi + #$D9#$F2 + // FPTAN.  1 in ST(0), tan(r_j) in ST(1)
		               #$D9#$C1 +  // FLD ST(1). tan(r_j) in ST(0), 1 in ST(1), tan(r_j) in ST(2)
                   #$D8#$C8 + //FMUL ST(0), ST(0).  tan(r_j)^2 in ST(0), 1 in ST(1), tan(r_j) in ST(2)
                   #$DE#$C1 + //FADDP.  1+tan(r_j)^2 in ST(0),  tan(r_j) in ST(1)
                   LoadDd(u[j])+ // u_j in ST(0), 1+tan(r_j)^2 in ST(1),  tan(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0). u_j in ST(0) and ST(1),  1+tan(r_j)^2 in ST(2),  tan(r_j) in ST(3)
 		               #$D8#$C8 +  //FMUL ST(0) ST(0).  u_j^2 in ST(0), u_j in ST(1), 1+tan(r_j)^2 in ST(2),  tan(r_j) in ST(3)
                   #$D9#$C9 + // FXCH. u_j in ST(0), u_j^2 in ST(1), 1+tan(r_j)^2 in ST(2),  tan(r_j) in ST(3)
                   #$D8#$CA +  //FMUL ST(0) ST(2). u_j(1+tan^2) enST(0), u_j^2 in ST(1), 1+tan(r_j)^2 in ST(2),  tan(r_j) in ST(3)
                   WriteDdP(u[j])+ // u_j^2 in ST(0), 1+tan(r_j)^2 in ST(1),  tan(r_j) in ST(2)
                   #$D8#$C0 +  //FADD ST(0) ST(0) 2u_j^2 in ST(0), 1+tan(r_j)^2 in ST(1),  tan(r_j) in ST(2)
                   #$D8#$CA +  //FMUL ST(0) ST(2). 2u_j^2 tan(r_j)  in ST(0),  1+tan(r_j)^2 in ST(1),  tan(r_j) in ST(2)
		               #$D8#$C9 +  //FMUL ST(0) ST(1). 2u_j^2 tan(r_j)(1+tan(r_j)^2 )  in ST(0),   1+tan(r_j)^2 in ST(1),  tan(r_j) in ST(2)
		               LoadDd(w[j])+
		               #$DE#$CA + // FMULP ST(2) ST(0). 2u_j^2 tan(r_j)(1+tan(r_j)^2 )  in ST(0),  w_j( 1+tan(r_j)^2) in ST(1),  tan(r_j) in ST(2)
 		               #$DE#$C1 + //FADDP.  2u_j^2 tan(r_j)(1+tan(r_j)^2 ) + w_j( 1+tan(r_j)^2) in ST(0),  tan(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'H': begin
                 ddC:= ddC + #$D9#$EA#$DE#$C9 + //FLDL2E ,FMULP. r_j *ln_2e in ST(0)
		               c_2ToX_S + // e^{r_j} in ST(0)
		               c_Load1 +  // 1 in ST(0),  exp (r_j) in ST(1)
		               #$D9#$E0 + // FCHS. -1 in ST(0), exp (r_j) in ST(1)
                   #$D9#$C1 + // FLF ST(1). exp(r_j) in ST(0), -1 in ST(1), exp (r_j) in ST(2)
                   #$D8#$F9 + // FDIVR ST(0) ST(1). -exp (-r_j) in ST(0), -1 in ST(1), exp (r_j) in ST(2)
                   #$D9#$FD + // FSCALE. -exp (-r_j)/2 in ST(0), -1 in ST(1), exp (r_j) in ST(2)
                   #$D9#$CA + // FXCH ST(2). exp (r_j) in ST(0), -1 in ST(1) , -exp (-r_j)/2 in ST(2)
                   #$D9#$FD + // FSCALE. exp (r_j)/2 in ST(0), -1 in ST(1), -exp (-r_j)/2 in ST(2)
		               #$DD#$D9 + // FSTP  ST(1). exp (r_j)/2 in ST(0) , -exp (-r_j)/2 in ST(1)
		               #$D9#$C0 + // FLD ST(0).  exp (r_j)/2 in ST(0) and in ST(1), -exp (-r_j)/2 in ST(2)
                   #$D8#$C2 + // FADD ST(0) ST(2). sinh(r_j) in ST(0), exp (r_j)/2 in ST(1), -exp (-r_j)/2 in ST(2)
                   #$D9#$CA + //FXCH ST(2). -exp (-r_j)/2 in ST(0), exp (r_j)/2 in ST(1),  sinh(r_j) in ST(2)
                   #$DE#$E9 + // FSUBP ST(1) ST(0). cosh(r_j) in ST(0), sinh(r_j) in ST(1)
		               LoadDd(u[j])+ // u_j in ST(0), cosh(r_j) in ST(1), sinh(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0). u_j in ST(0) and ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
		               #$D8#$C8 +  //FMUL ST(0) ST(0).  u_j^2 in ST(0), u_j in ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
  		             #$D8#$CB +  //FMUL ST(0) ST(3). u_j^2 sinh(r_j) in ST(0), u_j in ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
                   #$D9#$C9 +  // FXCH. u_j in ST(0), u_j^2 sinh(r_j) in ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
		               #$D8#$CA + //FMUL ST(0) ST(2). u_j ch(r_j) in ST(0), u_j^2 sh(r_j) in ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
                   WriteDdP(u[j])+ //u_j^2 sh(r_j) in ST(0), cosh(r_j) in ST(1), sinh(r_j) in ST(2)
                   LoadDd(w[j])+ //w_j in ST(0), u_j^2 sh(r_j) in ST(1), cosh(r_j) in ST(2), sinh(r_j) in ST(3)
                   #$DE#$CA + //FMULP ST(2) ST(0).  u_j^2 sh(r_j) in ST(0), w_j cosh(r_j) in ST(1), sinh(r_j) in ST(2)
                   #$DE#$C1 + //FADDP. u_j^2 sh(r_j) + w_j cosh(r_j) in ST(0), sinh(r_j) in ST(1)
                   WriteDdP(w[j]);
               end;
          'I': begin
                 ddC:= ddC + #$D9#$EA#$DE#$C9 + //FLDL2E ,FMULP. r_j *ln_2e in ST(0)
		               c_2ToX_S + // e^{r_j} in ST(0)
                   c_Load1 +  // 1 in ST(0),  exp (r_j) in ST(1)
		               #$D9#$E0 + // FCHS. -1 in ST(0), exp (r_j) in ST(1)
		               #$D9#$C1 + // FLF ST(1). exp(r_j) in ST(0), -1 in ST(1), exp (r_j) in ST(2)
                   #$D8#$F9 + // FDIVR ST(0) ST(1). -exp (-r_j) in ST(0), -1 in ST(1), exp (r_j) in ST(2)
                   #$D9#$FD + // FSCALE. -exp (-r_j)/2 in ST(0), -1 in ST(1), exp (r_j) in ST(2)
		               #$D9#$CA + // FXCH ST(2). exp (r_j) in ST(0), -1 in ST(1) , -exp (-r_j)/2 in ST(2)
                   #$D9#$FD + // FSCALE. exp (r_j)/2 in ST(0), -1 in ST(1), -exp (-r_j)/2 in ST(2)
                   #$DD#$D9 + // FSTP  ST(1). exp (r_j)/2 in ST(0) , -exp (-r_j)/2 in ST(1)
                   #$D9#$C0 + // FLD ST(0).  exp (r_j)/2 in ST(0) and in ST(1), -exp (-r_j)/2 in ST(2)
                   #$D8#$C2 + // FADD ST(0) ST(2). sinh(r_j) in ST(0), exp (r_j)/2 in ST(1), -exp (-r_j)/2 in ST(2)
		               #$D9#$CA + //FXCH ST(2). -exp (-r_j)/2 in ST(0), exp (r_j)/2 in ST(1),  sinh(r_j) in ST(2)
                   #$DE#$E9 + // FSUBP ST(1) ST(0). cosh(r_j) in ST(0), sinh(r_j) in ST(1)
		               #$D9#$C9 +  // FXCH. sinh(r_j) in ST(0), cosh(r_j) in ST(1)
                   LoadDd(u[j])+ // u_j in ST(0), sinh(r_j) in ST(1), cosh(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0). u_j in ST(0) and ST(1), sinh(r_j) in ST(2), cosh(r_j) in ST(3)
		               #$D8#$C8 +  //FMUL ST(0) ST(0).  u_j^2 in ST(0), u_j in ST(1), sinh(r_j) in ST(2), cosh(r_j) in ST(3)
                   #$D8#$CB +  //FMUL ST(0) ST(3). u_j^2 cosh(r_j) in ST(0), u_j in ST(1), sinh(r_j) in ST(2), cosh(r_j) in ST(3)
                   #$D9#$C9 +  // FXCH. u_j in ST(0), u_j^2 cosh(r_j) in ST(1), sinh(r_j) in ST(2), cosh(r_j) in ST(3)
		               #$D8#$CA + //FMUL ST(0) ST(2). u_j sh(r_j) in ST(0), u_j^2 ch(r_j) in ST(1), sinh(r_j) in ST(2), cosh(r_j) in ST(3)
		               WriteDdP(u[j])+ //u_j^2 ch(r_j) in ST(0), csinh(r_j) in ST(1), cosh(r_j) in ST(2)
                   LoadDd(w[j])+ //w_j in ST(0), u_j^2 ch(r_j) in ST(1), sh(r_j) in ST(2), ch(r_j) in ST(3)
                   #$DE#$CA + //FMULP ST(2) ST(0).  u_j^2 ch(r_j) in ST(0), w_j sh(r_j) in ST(1), ch(r_j) in ST(2)
		               #$DE#$C1 + //FADDP. u_j^2 ch(r_j) + w_j sh(r_j) in ST(0), cosh(r_j) in ST(1)
		               WriteDdP(w[j]);
               end;
          'J': begin
                 ddC:= ddC + #$D9#$EA#$DE#$C9 +  //FLDL2E ,FMULP. r_j *ln_2e in ST(0)
		               c_2ToX_S + // e^{r_j} in ST(0)
                   #$D8#$C8 + //FMUL ST(0) ST(0). e^{2r_j} in ST(0)
                   c_Load1 + // 1 in ST(0), e^{2r_j} in ST(1)
		               #$D8#$E9 + //FSUBR ST(0) ST(1). e^{2r_j} - 1 in ST(0), e^{2r_j} in ST(1)
                   #$D9#$C9  + //FXCH. e^{2r_j} in ST(0), e^{2r_j} - 1 in ST(1)
                   c_Load1 + //1 in ST(0), e^{2r_j} in ST(1), e^{2r_j} - 1 in ST(2)
                   #$DE#$C1 + //FADDP. e^{2r_j}+1 in ST(0), e^{2r_j} - 1 in ST(2)
                   #$DE#$F9 + // FDIVP.  tanh(r_j) in ST(0)
                   #$D9#$C0 + // FLD ST(0).  tanh(r_j) in ST(0) and in ST(1)
		               #$D8#$C8 +  // FMUL ST(0) ST(0). tanh(r_j)^2 in ST(0), tanh(r_j) in ST(1)
                   c_Load1 + //  1 in ST(0), tanh(r_j)^2 in ST(1), tanh(r_j)$in ST(2)
                   #$DE#$E1 + // FBUBRP ST(1) ST(0). 1-tanh(r_j)^2 in ST(0), tanh(r_j) in ST(1)
                   LoadDd(u[j])+ // u_j in ST(0), 1-th(r_j)^2 in ST(1),  th(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0). u_j in ST(0) and ST(1),  1-th(r_j)^2 in ST(2),  th(r_j) in ST(3)
 		               #$D8#$C8 +  //FMUL ST(0) ST(0).  u_j^2 in ST(0), u_j in ST(1), 1-th(r_j)^2 in ST(2),  th(r_j) in ST(3)
		               #$D9#$C9 + // FXCH. u_j in ST(0), u_j^2 in ST(1), 1-th(r_j)^2 in ST(2),  th(r_j) in ST(3)
                   #$D8#$CA +  //FMUL ST(0) ST(2). u_j(1-th^2) enST(0), u_j^2 in ST(1), 1-th(r_j)^2 in ST(2),  th(r_j) in ST(3)
                   WriteDdP(u[j])+ // u_j^2 in ST(0), 1-th(r_j)^2 in ST(1),  th(r_j) in ST(2)
                   #$D8#$C0 +  //FADD ST(0) ST(0) 2u_j^2 in ST(0), 1-th(r_j)^2 in ST(1),  th(r_j) in ST(2)
                   #$D8#$CA +  //FMUL ST(0) ST(2). 2u_j^2 th(r_j)  in ST(0),  1-th(r_j)^2 in ST(1),  th(r_j) in ST(2)
                   #$D8#$C9 +  //FMUL ST(0) ST(1). 2u_j^2 th(r_j)(1-th(r_j)^2 )  in ST(0),   1-th(r_j)^2 in ST(1),  th(r_j) in ST(2)
                   LoadDd(w[j])+
		               #$DE#$CA + // FMULP ST(2) ST(0). 2u_j^2 th(r_j)(1-th(r_j)^2 )  in ST(0),  w_j( 1-th(r_j)^2) in ST(1),  th(r_j) in ST(2)
                   #$DE#$E9 + //FSUBP.  -2u_j^2 th(r_j)(1-th(r_j)^2 ) + w_j( 1-th(r_j)^2) in ST(0),  th(r_j) in ST(1)
		               WriteDdP(w[j]);
               end;
          'K': begin
                 ddC:= ddC + #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
 		               #$D9#$E1 + // FABS. |r_j| in ST(0), r_j in ST(1)
		               c_1_comp + // FABS. |r_j| in ST(0), r_j in ST(1)
                   #$77#$2A + // JBE. if |r_j| < 1, jumps 42 bytes
		               #$DD#$D8  + // FSTP ST(0). r_j in ST(0)
		               c_Load1 +  // 1 in ST(0), r_j in ST(1)
                   #$D9#$E0 + // FCHS. -1 in ST(0), r_j in ST(1)
		               c_LoadPi +  // pi in ST(0), -1 in ST(1), r_j in ST(2)
                   #$D9#$FD + // FSCALE pi/2 in ST(0), -1 in ST(1),  r_j in ST(2)
                   #$DD#$D9 + // FSTP ST(1). pi/2 in ST(0), r_j in ST(1)
		               #$D9#$C9 + //FXCH. r_j in ST(0), pi/2 in ST(1),
		               c_zero_comp +
		               #$DD#$D8 + // FSTP ST(0). pi/2 in ST(0)
                   #$72#$02 + // JB. if r_j > 0, jumps 2 bytes
		               #$D9#$E0 + // FCHS, -pi/2 in ST(0)
		               c_Load0 +  // 0 in ST(0),pi/2  or  -pi/2 in ST(0)
		               WriteDdP(u[j])+
                   c_Load0 +  // 0 in ST(0), pi/2 o  -pi/2 in ST(0)
                   WriteDdP(w[j])+
                   #$EB#$30 + //JMP. jumps 48  bytes
                   #$D8#$C8 + // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
                   c_load1 + // 1 in ST(0),   r_j^2 in ST(1), r_j in ST(2)
                   #$DE#$E1 + // FSUBRP ST(1) ST(0). 1-r_j^2 in ST(0), r_j in ST(1)
                   #$D9#$FA + // FSQRT. sqrt(1-r_j^2) in ST(0), r_j in ST(1)
                   LoadDd(u[j])+  // u_j in ST(0), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). u_j/sqrt(1-r_j^2) in ST(0), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D9#$C0 + // FLD ST(0). u_j/sqrt(1-r_j^2) in ST(0) and in ST(1), sqrt(1-r_j^2) in ST(2), r_j in ST(3)
		               WriteDdP(u[j])+
		               #$D8#$C8 + // FMUL  ST(0). u_j^2/(1-r_j^2) in ST(0), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$CA + // FMUL  ST(0) ST(2). u_j^2r_j/(1-r_j^2) in ST(0), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$F1 + // FDIV ST(0) ST(1). u_j^2r_j/(1-r_j^2) ^(3/2), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               LoadDd(w[j])+
                   #$D8#$F2 + //FDIV ST(0) ST(2). w/sqrt(1-r_j^2) in ST(0), u_j^2r_j/(1-r_j^2)^(3/2) in ST(1), sqrt(1-r_j^2) in ST(2), r_j in ST(3)
		               #$DE#$C1 + //FADDP. w/sqrt(1-r_j^2) + u_j^2r_j/(1-r_j^2) ^(3/2) in ST(0), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               WriteDdP(w[j])+ //sqrt(1-r_j^2) in ST(0), r_j in ST(1)
                   #$D9#$F3 ; //FPATAN. arcsin(r_j) in ST(0)
               end;
          'L': begin
                 ddC:= ddC + #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
                   #$D9#$E1 + // FABS. |r_j| in ST(0), r_j in ST(1)
		               c_1_comp + // FABS. |r_j| in ST(0), r_j in ST(1)
		               #$77#$22 + // JBE. if |r_j| < 1, jumps 34 bytes
                   #$DD#$D8 	+  // FSTP ST(0).
                   c_zero_comp +
		               #$DD#$D8 	+  // FSTP ST(0).
		               #$72#$04 + // JB. if r_j > 0, jumps 4 bytes
		               c_LoadPi  +  // pi in ST(0)
		               #$EB#$02 + // JMP. jumps 2 bytes
                   c_Load0 +
		               c_Load0 +  // 0 in ST(0), pi  o  0 in ST(1)
		               WriteDdP(u[j])+
		               c_Load0 +  // 0 in ST(0), pi  o  0 in ST(1)
		               WriteDdP(w[j])+

                   #$EB#$36 + //JMP. jumps 54  bytes
		               #$D8#$C8+ // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
    		           c_load1 + // 1 in ST(0),   r_j^2 in ST(1), r_j in ST(2)
                   #$DE#$E1 + // FSUBRP ST(1) ST(0). 1-r_j^2 in ST(0), r_j in ST(1)
                   #$D9#$FA + // FSQRT. sqrt(1-r_j^2) in ST(0), r_j in ST(1)
                   LoadDd(u[j])+  // u_j in ST(0), sqrt(1-r_j^2) in ST(1), r_j in ST(2)
 		               #$D8#$F1+ // FDIV ST(0) ST(1). u_j/sqrt(1-r_j^2) in ST(0),
                             //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D9#$C0+ // FLD ST(0). u_j/sqrt(1-r_j^2) in ST(0) and in
                             //ST(1), sqrt(1-r_j^2) in ST(2), r_j in ST(3)
		               #$D9#$E0 + // FCHS.
                   WriteDdP(u[j])+
		               #$D8#$C8+ // FMUL  ST(0). u_j^2/(1-r_j^2) in ST(0),
                              //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$CA+ // FMUL  ST(0) ST(2). u_j^2r_j/(1-r_j^2) in ST(0),
                             //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$F1+ // FDIV ST(0) ST(1). u_j^2r_j/(1-r_j^2) ^(3/2),
                             //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   LoadDd(w[j])+
                   #$D8#$F2+ // FDIV ST(0) ST(2). w/sqrt(1-r_j^2) in ST(0),
                             //u_j^2r_j/(1-r_j^2) ^(3/2) in ST(1),  sqrt(1-r_j^2) in ST(2), r_j in ST(3)
		               #$DE#$C1+ //FADDP. w/sqrt(1-r_j^2)  + u_j^2r_j/(1-r_j^2)^(3/2) in ST(0),
                             //sqrt(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D9#$E0 + // FCHS.
                   WriteDdP(w[j])+ //sqrt(1-r_j^2) in ST(0), r_j in ST(1)
                   #$D9#$C9 + //FXCH. r_j in ST(0), sqrt(1-r_j^2) in ST(1)
                   #$D9#$F3 ; //FPATAN. arccos(r_j) in ST(0)
               end;
          'M': begin
                 ddC:= ddC + #$D9#$C0 + // FLD ST(0). r_j in ST(0) and in ST(1)
                   #$D8#$C8 +  // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
                   c_Load1 +  // 1 in ST(0), r_j^2 in ST(1), r_j in ST(2)
                   #$DE#$C1 + //FADDP. 1+r_j^2 in ST(0), r_j in ST(1)
                   LoadDd(u[j])+  // u_j in ST(0), 1+r_j^2 in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). u_j/(1+r_j^2) in ST(0),  1+r_j^2 in ST(1), r_j in ST(2)
                   LoadDd(w[j])+ // w_j in ST(0), u_j/(1+r_j^2) in ST(1),  1+r_j^2 in ST(2), r_j in ST(3)
                   #$DE#$F2 +  // FDIVRP. ST(2), ST(0). u_j/(1+r_j^2) in ST(0),  w/(1+r_j^2) in ST(1), r_j in ST(2)
                   #$D9#$C0 + // FLD ST(0). u_j/(1+r_j^2) in ST(0) and ST(1),  w/(1+r_j^2) in ST(2), r_j in ST(3)
                   WriteDdP(u[j])+ // u_j/(1+r_j^2) in  ST(1),  w/(1+r_j^2) in ST(2), r_j in ST(3)
                   #$D8#$C8 + // FMUL ST(0) ST(0). u_j^2/(1+r_j^2)^2 in  ST(0),  w/(1+r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$CA + // FMUL ST(0) ST(0). r_ju_j^2/(1+r_j^2)^2 in  ST(0),  w/(1+r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$C0 + // FADD ST(0) ST(0). 2r_ju_j^2/(1+r_j^2)^2 in  ST(0),  w/(1+r_j^2) in ST(1), r_j in ST(2)
                   #$DE#$E9 + // FSUBP ST(1) ST(0), w/(1+r_j^2) - 2r_ju_j^2/(1+r_j^2)^2 in  ST(0),   r_j in ST(1)
                   WriteDdP(w[j])+ // r_j in ST(0)
                   c_Load1 +  // 1 in ST(0), r_j in ST(1)
                   #$D9#$F3;  // atan(r_j) in ST(0)
               end;
          'N': begin
                 ddC:= ddC + c_zero_comp +
		               #$76#$1E +  // JBE. if r_j > =0, jumps 30 bytes
                   #$D9#$E0 + // FCHS -r_j in ST(0)
                   LoadDd(u[j])+ // u_j in ST(0), - r_j in ST(1)
                   #$D9#$E0  + //FCHS.  -u_j in ST(0), -r_j in ST(1)
                   WriteDdP(u[j])+
                   LoadDd(w[j])+ // w_j in ST(0), ,- r_j in ST(1)
                   #$D9#$E0 +  //FCHS.  -w_j in ST(0), -r_j in ST(1)
                   WriteDdP(w[j]);
                end;
          'O': begin
                 ddC:= ddC + c_mod_pi + c_zero_comp +
                   #$75#$1A +  //JNE, if r_j <> 0 jumps 26 bytes
                   #$DD#$D8 + // POP ST(0)
                   LoadDd(uSpecialValues[2]) +  //eMaxNum in ST(0)
                   c_Load0 +
                   WriteDdP(u[j])+
                   c_Load0 +
                   WriteDdP(w[j])+
                   #$EB#$36+ //JMP. jumps 54 bytes
                   #$D9#$F2 + // FPTAN.  1 in ST(0), tan(r_j) in ST(1)
                   #$DC#$F1 +  //FDIV ST(1) ST(0). 1 in ST(0), cot(r_j) in ST(1)
                   #$D9#$C1 +  // FLD ST(1). cot(r_j) in ST(0), 1 in ST(1), cot(r_j) in ST(2)
                   #$D8#$C8 + //FMUL ST(0), ST(0).  cot(r_j)^2 in ST(0), 1 in ST(1), cot(r_j) in ST(2)
                   #$DE#$C1 + //FADDP.  1+cot(r_j)^2 in ST(0),  cot(r_j) in ST(1)
                   LoadDd(u[j])+ // u_j in ST(0), 1+cot(r_j)^2 in ST(1),  cot(r_j) in ST(2)
                   #$D9#$C0 + // FLD ST(0). u_j in ST(0) and ST(1),  1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                   #$D8#$C8 +  //FMUL ST(0) ST(0).  u_j^2 in ST(0), u_j in ST(1), 1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                   #$D9#$C9 + // FXCH. u_j in ST(0), u_j^2 in ST(1), 1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                   #$D8#$CA +  //FMUL ST(0) ST(2). u_j(1+cot^2) in ST(0), u_j^2 in ST(1), 1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                   #$D9#$E0  + //FCHS. -u_j(1+cot^2) enST(0), u_j^2 in ST(1), 1+cot(r_j)^2 in ST(2),  cot(r_j) in ST(3)
                   WriteDdP(u[j])+ // u_j^2 in ST(0), 1+cot(r_j)^2 in ST(1),  cot(r_j) in ST(2)
                   #$D8#$C0 +  //FADD ST(0) ST(0) 2u_j^2 in ST(0), 1+cot(r_j)^2 in ST(1),  cot(r_j) in ST(2)
                   #$D8#$CA +  //FMUL ST(0) ST(2). 2u_j^2 cot(r_j)  in ST(0),  1+cot(r_j)^2 in ST(1),  cot(r_j) in ST(2)
                   #$D8#$C9 +  //FMUL ST(0) ST(1). 2u_j^2 cot(r_j)(1+cot(r_j)^2) in ST(0), 1+cot(r_j)^2 in ST(1), cot(r_j) in ST(2)
                   LoadDd(w[j])+
                   #$DE#$CA + // FMULP ST(2) ST(0). 2u_j^2 cot(r_j)(1+cot(r_j)^2) in ST(0), w_j( 1+cot(r_j)^2) in ST(1),  cot(r_j) in ST(2)
                   #$DE#$E1 + //FSUBRP.  2u_j^2 cot(r_j)(1+cot(r_j)^2) - w_j( 1+cot(r_j)^2) in ST(0),  cot(r_j) in ST(1)
                   WriteDdP(w[j]);
                end;
          'P': begin
                 ddC:= ddC +
                   c_1_comp +
		               #$72#$16 + // JB. if r_j  > 1, jumps 22 bytes
                   #$DD#$D8 +  // FSTP ST(0).
                   c_Load0 +
		               c_Load0 + // 0 in ST(0) and in ST(1)
		               WriteDdP(u[j]) +
                   c_Load0 +  // 0 in ST(0) and in st(1)
                   WriteDdP(w[j])+
                   #$EB#$3C + //JMP. jumps 60  bytes
		               #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
                   #$D8#$C8+ // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
    		           c_load1 + // 1 in ST(0),   r_j^2 in ST(1), r_j in ST(2)
		               #$DE#$E9 + // FSUBP. r_j^2-1 in ST(0), r_j in ST(1)
		               #$D9#$FA + // FSQRT. sqrt(r_j^2-1) in ST(0), r_j in ST(1)
                   LoadDd(u[j])+  // u_j in ST(0), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). u_j/sqrt(r_j^2-1) in ST(0), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
                   #$D9#$C0 + // FLD ST(0). u_j/sqrt(r_j^2-1) in ST(0) and in ST(1), sqrt(r_j^2-1) in ST(2), r_j in ST(3)
                   WriteDdP(u[j])+
                   #$D8#$C8 + // FMUL  ST(0). u_j^2/(r_j^2-1) in ST(0), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
                   #$D8#$CA + // FMUL  ST(0) ST(2). u_j^2r_j/(r_j^2-1) in ST(0), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
		               #$D8#$F1 + // FDIV ST(0) ST(1). u_j^2r_j/(r_j^2-1) ^(3/2), sqrt(r_j^2-1) in ST(1), r_j in ST(2)
                   LoadDd(w[j])+
		               #$D8#$F2 + // FDIV ST(0) ST(2). w/sqrt(r_j^2-1) in ST(0), u_j^2r_j/(r_j^2-1) ^(3/2) in ST(1),  sqrt(r_j^2-1) in ST(2), r_j in ST(3)
                   #$DE#$E1 + //FSUBRP. w/sqrt(r_j^2-1)  - u_j^2r_j/(r_j^2-1) ^(3/2) in ST(0),  sqrt(r_j^2-1) in ST(1), r_j in ST(2)
		               WriteDdP(w[j])+ //sqrt(r_j^2-1) in ST(0), r_j in ST(1)
		               #$DE#$C1 + //FADDP. r_j+ sqrt(r_j^2-1) in ST(0)
		               c_ln;    //acosh(r_j) in ST(0)
               end;
          'Q': begin
                 ddC:= ddC + #$D9#$C0 + // FLD ST(0). r_j in ST(0) and in ST(1)
		               #$D9#$E1 + // FABS. |r_j| in ST(0), r_j in ST(1)
		               c_1_comp + // FABS. |r_j| in ST(0), r_j in ST(1)
                   #$77#$2C + // JBE. if |r_j| < 1, jumps 44 bytes
                   #$DD#$D8 +  // FSTP ST(0). r_j in ST(0)
                   c_zero_comp +                       		             
		               #$77#$0A +  //JA, if  r_j < 0 jumps 10 bytes
                   #$DD#$D8 +  // FSTP ST(0).  
	                 LoadDd(uSpecialValues[2]) + //eMaxNum in ST(0)
 	 	               #$EB#$08 + //JMP. jumps 8 bytes
                   #$DD#$D8 +  // FSTP ST(0).
		               LoadDd(uSpecialValues[3]) + //-eMaxNum in ST(0)
		               c_Load0 +
		               WriteDdP(u[j]) +
		               c_Load0 +  
		               WriteDdP(w[j])+
  		             #$EB#$4E + //JMP. jumps 78  bytes
                   #$D8#$C8 +  // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
                   c_Load1 +  // 1 in ST(0), r_j^2 in ST(1), r_j in ST(2)
 		               #$DE#$E1 + //FSUBRP. 1-r_j^2 in ST(0), r_j in ST(1)
		               c_Load1 +  // 1 in ST(0), 1-r_j^2 in ST(1), r_j in ST(2)
		               #$DE#$F1 + //FDIVRP. 1/(1-r_j^2) in ST(0), r_j in ST(1)
                   LoadDd(u[j])+  // u_j in ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$C9 + // FMUL ST(0) ST(1). u_j/(1-r_j^2)  in ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D9#$C0 + // FLD ST(0). u_j/(1-r_j^2)  in ST(0) and ST(1), 1/(1-r_j^2) in ST(2), r_j in ST(3)
                   WriteDdP(u[j])+ // u_j/(1-r_j^2) in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D8#$C8 + // FMUL ST(0) ST(0). u_j^2/(1-r_j^2)^2 in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$C0 + // FADD ST(0) ST(0). 2u_j^2/(1-r_j^2)^2 in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
                   #$D8#$CA + // FMUL ST(0) ST(2).  2r_ju_j^2/(1-r_j^2)^2 in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               LoadDd(w[j])+ // w_j in ST(0),  2r_ju_j^2/(1-r_j^2)^2 in  ST(1), 1/(1-r_j^2) in ST(2), r_j in ST(3)
		               #$D8#$CA + //FMUL ST(0) ST(2).  w_j/(1-r_j^2) in ST(0), 2r_ju_j^2/(1-r_j^2)^2 in  ST(1), 1/(1-r_j^2) in ST(2), r_j in ST(3)
		               #$DE#$C1 + //FADDP.  w_j/(1-r_j^2) +2r_ju_j^2/(1-r_j^2)^2 in  ST(0), 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               WriteDdP(w[j])+ // 1/(1-r_j^2) in ST(1), r_j in ST(2)
		               #$D9#$C9 + // FXCH.r_j in ST(0), 1/(1-r_j^2) in ST(1).
		               c_Load1 +  // 1 in ST(0), r_j in ST(1), 1/(1-r_j^2) in ST(2).
		               #$DE#$C1 + //FADDP. 1 + r_j in ST(1), 1/(1-r_j^2) in ST(2).
		               #$D8#$C8 +  // FMUL  ST(0).(1+ r_j)^2 in ST(0), 1/(1-r_j^2) in ST(1).
                   #$DE#$C9 + //FMULP. .(1+ r_j)^2 /(1-r_j^2) in ST(0).
                   c_ln + 			//2atanh(r_j) in ST(0)
		               c_Load1 +
                   #$D9#$E0 + //FCHS. -1 in ST(0), 2atanh(r_j) in ST(1)
                   #$D9#$C9 + //FXCH. 2atanh(r_j) in ST(0), -1 in ST(1)
		               #$D9#$FD + // FSCALE atanh(r_j) in ST(0), -1 in ST(1)
                   #$DD#$D9 ; // FSTP ST(1). atanh(r_j) in ST(0)
               end;
          'R': begin
                 ddC:= ddC +
		               #$D9#$C0 + // FLD ST(0).  r_j in ST(0) and in ST(1)
                   #$D8#$C8+ // FMUL  ST(0). r_j^2 in ST(0), r_j in ST(1)
    		           c_load1 + // 1 in ST(0),   r_j^2 in ST(1), r_j in ST(2)
                   #$DE#$C1 + //FADDP. 1+r_j^2 in ST(0), r_j in ST(1)
                   #$D9#$FA + // FSQRT. sqrt(1+r_j^2 ) in ST(0), r_j in ST(1)
		               LoadDd(u[j])+  // u_j in ST(0), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). u_j/sqrt(1+r_j^2 ) in ST(0), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
		               #$D9#$C0 + // FLD ST(0). u_j/sqrt(1+r_j^2 ) in ST(0) and in ST(1), sqrt(1+r_j^2 ) in ST(2), r_j in ST(3)
                   WriteDdP(u[j])+
		               #$D8#$C8 + // FMUL  ST(0). u_j^2/(1+r_j^2 ) in ST(0), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
                   #$D8#$CA + // FMUL  ST(0) ST(2). u_j^2r_j/(1+r_j^2 ) in ST(0), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
                   #$D8#$F1 + // FDIV ST(0) ST(1). u_j^2r_j/(1+r_j^2 ) ^(3/2), sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
                   LoadDd(w[j])+
                   #$D8#$F2 + // FDIV ST(0) ST(2). w/sqrt(1+r_j^2 ) in ST(0), u_j^2r_j/(1+r_j^2 ) ^(3/2) in ST(1),  sqrt(1+r_j^2 ) in ST(2), r_j in ST(3)
                   #$DE#$E1 + //FSUBRP. w/sqrt(1+r_j^2 )  - u_j^2r_j/(1+r_j^2 ) ^(3/2) in ST(0),  sqrt(1+r_j^2 ) in ST(1), r_j in ST(2)
		               WriteDdP(w[j])+ //sqrt(1+r_j^2 ) in ST(0), r_j in ST(1)
                   #$DE#$C1 + //FADDP. r_j+ sqrt(1+r_j^2 ) in ST(0)
                   c_ln;    //asinh(r_j) in ST(0)
                end;
          else
            ;
          end;
        end;
      end;
    end;
  ddC:= ddC + #$C3;       //return
  PutCodeInExeBlock(ddC, DDBlock);
  @FoA_Second := @DDBlock[0];
end;


procedure InitializeCodeVars;
begin
  c_div_S := c_zero_comp + #$75#$0C#$DD#$D8 + LoadDd(uSpecialValues[2]) +
    #$DE#$C9#$EB#$02#$DE#$F9; {Safe division. 20 bytes. Requires ST(c_LastReg) empty}
  c_ln_S := c_zero_comp + #$72#$0C#$DD#$D8 + LoadDd(uSpecialValues[6]) +
    #$D9#$E0#$EB#$0A + c_ln;
    {Safe ln(ST(0)). 26 bytes. Requires ST(c_LastReg) empty}
  c_2toX_S := #$D9#$C0#$D9#$E1 + LoadDd(uSpecialValues[6]) + #$DF#$F1 +
    #$DD#$D8#$72#$1A + c_2toX + #$EB#$16 + c_zero_comp + #$DD#$D8#$72#$08
    +
    LoadDd(uSpecialValues[4]) + #$EB#$06 + LoadDd(uSpecialValues[2]);
     {Safe 2^ST(0). 64 bytes. Requires ST(6) and ST(c_LastReg) empty.}
end;

procedure InitializeParser;
var
  character : char;
begin
  uKnownFunc[1].theText := 'COS';
  uKnownFunc[1].equivChar := 'A';

  uKnownFunc[2].theText := 'SIN';
  uKnownFunc[2].equivChar := 'B';

  uKnownFunc[3].theText := 'SEN';
  uKnownFunc[3].equivChar := 'B';

  uKnownFunc[4].theText := 'EXP';
  uKnownFunc[4].equivChar := 'C';

  uKnownFunc[5].theText := 'SQR';
  uKnownFunc[5].equivChar := 'D';

  uKnownFunc[6].theText := 'SQRT';
  uKnownFunc[6].equivChar := 'E';

  uKnownFunc[7].theText := 'LN';
  uKnownFunc[7].equivChar := 'F';

  uKnownFunc[8].theText := 'LOG';
  uKnownFunc[8].equivChar := 'F';

  uKnownFunc[9].theText := 'TG';
  uKnownFunc[9].equivChar := 'G';

  uKnownFunc[10].theText := 'TAN';
  uKnownFunc[10].equivChar := 'G';

  uKnownFunc[11].theText := 'SH';
  uKnownFunc[11].equivChar := 'H';

  uKnownFunc[12].theText := 'SINH';
  uKnownFunc[12].equivChar := 'H';

  uKnownFunc[13].theText := 'CH';
  uKnownFunc[13].equivChar := 'I';

  uKnownFunc[14].theText := 'COSH';
  uKnownFunc[14].equivChar := 'I';

  uKnownFunc[15].theText := 'TH';
  uKnownFunc[15].equivChar := 'J';

  uKnownFunc[16].theText := 'TANH';
  uKnownFunc[16].equivChar := 'J';

  uKnownFunc[17].theText := 'ARCSEN';
  uKnownFunc[17].equivChar := 'K';

  uKnownFunc[18].theText := 'ASEN';
  uKnownFunc[18].equivChar := 'K';

  uKnownFunc[19].theText := 'ARCSIN';
  uKnownFunc[19].equivChar := 'K';

  uKnownFunc[20].theText := 'ASIN';
  uKnownFunc[20].equivChar := 'K';

  uKnownFunc[21].theText := 'ARCCOS';
  uKnownFunc[21].equivChar := 'L';

  uKnownFunc[22].theText := 'ACOS';
  uKnownFunc[22].equivChar := 'L';

  uKnownFunc[23].theText := 'ARCOS';
  uKnownFunc[23].equivChar := 'L';

  uKnownFunc[24].theText := 'ARCTG';
  uKnownFunc[24].equivChar := 'M';

  uKnownFunc[25].theText := 'ARCTAN';
  uKnownFunc[25].equivChar := 'M';

  uKnownFunc[26].theText := 'ATAN';
  uKnownFunc[26].equivChar := 'M';

  uKnownFunc[27].theText := 'ABS';
  uKnownFunc[27].equivChar := 'N';

  uKnownFunc[28].theText := 'COTG';
  uKnownFunc[28].equivChar := 'O';

  uKnownFunc[29].theText := 'COTAN';
  uKnownFunc[29].equivChar := 'O';

  uKnownFunc[30].theText := 'COT';
  uKnownFunc[30].equivChar := 'O';

  uKnownFunc[31].theText := 'ARGCH';
  uKnownFunc[31].equivChar := 'P';

  uKnownFunc[32].theText := 'ACOSH';
  uKnownFunc[32].equivChar := 'P';

  uKnownFunc[33].theText := 'ARCCOSH';
  uKnownFunc[33].equivChar := 'P';

  uKnownFunc[34].theText := 'ARCTANH';
  uKnownFunc[34].equivChar := 'Q';

  uKnownFunc[35].theText := 'ARGTANH';
  uKnownFunc[35].equivChar := 'Q';

  uKnownFunc[36].theText := 'ARGTH';
  uKnownFunc[36].equivChar := 'Q';

  uKnownFunc[37].theText := 'ATANH';
  uKnownFunc[37].equivChar := 'Q';

  uKnownFunc[38].theText := 'ARCSINH';
  uKnownFunc[38].equivChar := 'R';

  uKnownFunc[39].theText := 'ASINH';
  uKnownFunc[39].equivChar := 'R';

  uKnownFunc[40].theText := 'ARGSH';
  uKnownFunc[40].equivChar := 'R';

  uKnownFunc[41].theText := 'ARGSINH';
  uKnownFunc[41].equivChar := 'R';

  uKnownFunc[42].theText := 'ARGCOSH';
  uKnownFunc[42].equivChar := 'P';

  for character := '#' to 'z' do
    case character of
      '(' :
        uAttribution[character] := openParent;
      '#' :
        uAttribution[character] := negation;
      ')' :
        uAttribution[character] := closeParent;
      '+', '-', '*', '/', '^' :
        uAttribution[character] := operationSymb;
      '1', 'p' :
        uAttribution[character] := constantSymb;
      'q'..'z' :
        uAttribution[character] := varSymb;
      'A'..'Z' :
        uAttribution[character] := functionSymb;
    else
      uAttribution[character] := badThing;
    end; {of case}

  {Translate these messages below to your language. Their meaning}
 {in English is given in the IMPLEMENTATION CONST part.         }

  uStdMessages[1] := 'There is a miswritten number.';
  uStdMessages[2] := 'You did use a character not allowed.';
  uStdMessages[3] := 'Some function or variable was miswritten.';
  uStdMessages[4] := 'Check, please, the parentheses.';
  uStdMessages[5] := 'I cannot cope with so much constants.';
  uStdMessages[6] := 'There is a sintax defect in the formula.';
  uStdMessages[7] := 'Sorry: I cannot understand that formula.';
  uStdMessages[8] := 'This is an expression too long.';
  uStdMessages[9] := 'Perhaps you forgot to write something.';
  uStdMessages[10] := 'The independent variables must be: ';
  uStdMessages[11] :=
    'The allowed variables must be in the range ''p'' to ''z''';
  uStdMessages[12] := 'No parameter may be named as any variable';

  uSpecialValues[0] := Pi/2;
  uSpecialValues[1] := 2*Pi;
  uSpecialValues[2] := eMaxNum;
  uSpecialValues[3] := -eMaxNum;
  uSpecialValues[4] := eMinNum;
  uSpecialValues[5] := -eMinNum;
  uSpecialValues[6] := eMaxExp;
  uSpecialValues[7] := 1/2;
  InitializeCodeVars;
end; {InitializeParser}

function TOCFunctDDDoubleSafe.FirstD(const XV: array of Double): firstDeriv;
begin
  result.fa0:= FoA_prime(XV);
  result.dfa0:= u[1];
end;

function TOCFunctDDDoubleSafe.SecondD(const XV: array of Double): SecondDeriv;
begin
  result.fa0:= FoA_second(XV);
  result.dfa0:= u[1];
  result.ddfa0:= w[1];
end;
procedure TOCFunctDDDoubleSafe.UpdateParameters(parameters:
          array of Double);
  var
    h,i: Integer;
begin
  h:= high(parameters);
  if (h >-1) and (h < parNum) then for i:=0 to high(parameters) do
    data[i+1]:= parameters[i];
end;


procedure ReadConstantField(theStr: AnsiString; var val: Double);
var
  forConstFunct: TOCFunctDDDoubleSafe;
  mc: array of Double;
begin
  val := 0;
  mc:= nil;
  forConstFunct:= TOCFunctDDDoubleSafe.create;
  try
    forConstFunct.ParseNewText(theStr,'','');
    val := forConstFunct.V(mc);
  finally
    forConstFunct.Free;
  end;
end;


initialization
  InitializeParser;
finalization
  ;


end.

