
{-----------------------------------------------------------------------------

 Unit Name: Printing
 Project Name : ImmersionR3R6
 Author:    Angel Montesinos-Amilibia
 Date: 2009
 Purpose: Routines for printing documents
-----------------------------------------------------------------------------}
{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}


unit Printing;

interface

uses
  Windows, Graphics;


function PreparePrint: Byte;
procedure PrintAll;

implementation

uses
  Math, Types, SysUtils, Classes, Forms, Controls, Printers,
  FImmersionR3R6, FMapR3R6, ResourceStrings, FDrawParamSurf, VectorsAndUtil,
  FRenderImplicitSurfaceWithThreads;

const
  cMinPageWidthPixels = 1000;
  cLastSubstr = 7;

type
  StrBlock= record
    maxWidth: Integer;{max width in pixels of the strings in strs}
    strs: TStringList;
  end;

var
  titleStrs: StrBlock;
  subStrs: array[1..cLastSubstr] of StrBlock;
  titleHeight, subStrHeight, lociMargin: Integer;
  r, rOr: TRect;

procedure ComputeLongFormula(var strs: StrBlock;
     can: TCanvas; theText: AnsiString;
     maxwidth, indentFirst, indentOthers: Integer);
{Adds to 'strs' the lines on which 'theText' is wrapped. 'indentFirst' is the number of
characters in which the first line is indented. 'indentOthers', the number of
chars for the indentation of the remaining lines. 'firstDesc' is the height in pixels
that must descend the first line. 'lineHeight' is the descent from line to line.}

const
  kReturn = char(13);
  kLineFeed = char(10);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';

  kLeftArrow = 37;
  kUpArrow = 38;
  kDownArrow = 40;
  digits =['0','1','2','3','4','5','6','7','8','9'];
  specialchars = ['+','-','*','/','=',')',']','}','.',',',':',';', ' '];
var
  theLine, theBareText, firstHead, heads: AnsiString;
  theChar, auxChar: char;
  index, i, textLength, b, previousLineEnd, charW, lineWidth: Integer;
begin
  textLength := Length(theText);
  theBareText:= '';
  strs.maxWidth:= 0;
  if textLength > 0 then begin
    for index := 1 to textLength do begin
      theChar := theText[index];
      if not (theChar in [kReturn, kTab, kLineFeed]) then
        theBareText := theBareText + theChar
      else if length(theBareText) > 0 then begin
        auxChar:=theBareText[length(theBareText)];
        if (auxchar<>' ') then
          theBareText := theBareText + ' ';
      end;
    end;
    textLength := Length(theBareText);
  end;
  if textLength = 0 then
    Exit;

  firstHead:= '';
  if indentFirst > 0 then begin
    for i:= 1 to indentFirst do
      firstHead:= firstHead + ' ';
    theBareText:= firstHead + theBareText;
  end;
  lineWidth:= can.TextWidth(theBareText);
  if (lineWidth <= maxWidth) then begin
    strs.strs.Add(theBareText);
    strs.maxWidth:= lineWidth;
    Exit;
  end;

  textLength:= Length(theBareText);
  heads:= '';
  if indentOthers > 0 then
    for i:= 1 to indentOthers do
      heads:= heads+' ';

  b := 0;
  previousLineEnd:= 0;
  theLine:= firstHead;
  lineWidth:= can.TextWidth(theLine);
  index:= Max(0,indentFirst);
  while index < textLength do begin
    inc(index);
    theChar := theBareText[index];
    charW:= can.TextWidth(theChar);
    if (lineWidth + charW <= maxWidth) and (theChar in specialchars) and
           not((theChar= '.') and (index > 1) and (theBareText[index-1] in digits) and
              (index < textLength) and (theBareText[index+1] in digits)) then
      b:= index;
    lineWidth := lineWidth + charW;
    if index = textLength then begin
      theLine:= theLine + Copy(theBareText, previousLineEnd + 1, index - PreviousLineEnd);
      strs.strs.Add(theLine);

    end
    else  if (lineWidth > maxWidth) then begin
      if b <= previousLineEnd then begin
        theLine:= theLine + Copy(theBareText, previousLineEnd + 1, index -1 - PreviousLineEnd);
        strs.maxWidth:= max(strs.maxWidth, can.TextWidth(theLine));
        strs.strs.Add(theLine);
        theLine:= heads;
        previousLineEnd:= index - 1;
      end
      else begin
        theChar:= theBareText[b];
        if theChar in ['+', '-', '*', '/', '=', ' '] then begin
          theLine:= theLine + Copy(theBareText, previousLineEnd + 1, b -1 - PreviousLineEnd);
          strs.strs.Add(theLine);
          strs.maxWidth:= max(strs.maxWidth, can.TextWidth(theLine));
          if theChar <> ' ' then
            theLine:= heads + theChar
          else
            theLine:= heads;
          previousLineEnd:= b;
        end
        else begin
          theLine:= theLine + Copy(theBareText, previousLineEnd + 1, b - PreviousLineEnd);
          strs.strs.Add(theLine);
          strs.maxWidth:= max(strs.maxWidth, can.TextWidth(theLine));
          theLine:= heads;
          previousLineEnd:= b;
        end;
      end;
      b:= 0;
      lineWidth:= can.TextWidth(theLine);
      index:= previousLineEnd;
    end
  end;
end;


procedure GetTitleStrings(tx: AnsiString; maxWidth: Integer);
{Divide if necessary the title into several lines and store them
in 'titleStrs'}
const
  a1= 15.38 * 1.2;
  b1= 1.27 * 1.2;
  c= 1.3;
var
  can: TCanvas;
  fh: Single;
  i: Integer;
begin
  can:= Printer.Canvas;
  can.Font.Name := 'Helvetica';
  can.Font.Style := [fsBold];

  fh:= a1 + b1*sqrt(maxwidth - cMinPageWidthPixels); {set a size for the text
                       font in terms of 'maxWidth' and 'cMinPageWidthPixels'}
  can.Font.Height := round(fh);

  fh:= c*fh; {space between lines}

  titleStrs.strs.Clear;
  if (tx <> '') then begin
    ComputeLongFormula(titleStrs, can, tx, maxwidth,0,0);
    if ImmersionR3R6.showVeronese then
      ComputeLongFormula(titleStrs, can, eVeronese, maxwidth,0,3)
    else
      ComputeLongFormula(titleStrs, can, eFocalSet, maxwidth,0,3)
  end;
  titleStrs.maxWidth:= 0;
  if titleStrs.strs.count > 0 then begin {compute title height and max width}
    titleHeight:= round((0.5 + titleStrs.strs.count)*fh);
    for i:= 0 to titleStrs.strs.count-1 do
      titleStrs.maxWidth:= Max(titleStrs.maxWidth, can.TextWidth(titleStrs.strs.Strings[i]));
  end
  else
    titleHeight:= 0;
end;



procedure GetSubTextStrings(tx: AnsiString; maxwidth: Integer);
{Divide if necessary each string of the caption into several lines and
 store them in 'subStrs'}
const
  a= 15.38;
  b= 1.27;
  c= 1.3;
var
  theText: AnsiString;
  can: TCanvas;
  shy: Single;
  i, n: Integer;
begin
  for i:= 1 to cLastSubstr do
    if Assigned(subStrs[i].strs) then
       subStrs[i].strs.Clear;
  can:= Printer.Canvas;
  can.Font.Name := 'Arial';
  can.Font.Style := [];

  {set a size for the text font in terms of 'maxWidth' and 'cMinPageWidthPixels'}
  shy:= a + b*sqrt(maxWidth - cMinPageWidthPixels);
  can.Font.Height:= round(shy);

  shy:= c*shy;

  if not ImmersionR3R6.printEquations then begin
    if tx <> '' then   {prepare caption desired by user}
      ComputeLongFormula(subStrs[1], can, tx, maxwidth,3,0);
  end
  else begin {caption consists of the formulas defining the surface}
    theText := 'x1(u,v,w) = ' + ImmersionR3R6.inputStrs[5];
    ComputeLongFormula(subStrs[1], can, theText, maxwidth,0,3);

    theText := 'x2(u,v,w) = ' + ImmersionR3R6.inputStrs[6];
    ComputeLongFormula(subStrs[2], can, theText, maxwidth,0,3);

    theText := 'x3(u,v,w) = ' + ImmersionR3R6.inputStrs[7];
    ComputeLongFormula(subStrs[3], can, theText, maxwidth,0,3);

    theText := 'x4(u,v,w) = ' + ImmersionR3R6.inputStrs[8];
    ComputeLongFormula(subStrs[4], can, theText, maxwidth,0,3);

    theText := 'x5(u,v,w) = ' + ImmersionR3R6.inputStrs[9];
    ComputeLongFormula(subStrs[5], can, theText, maxwidth,0,3);

    theText := 'x6(u,v,w) = ' + ImmersionR3R6.inputStrs[10];
    ComputeLongFormula(subStrs[6], can, theText, maxwidth,0,3);

    theText:= 'u = ' + ImmersionR3R6.inputStrs[17]+ ' ,  v = ' + ImmersionR3R6.inputStrs[18]+
              ' ,  w = ' + ImmersionR3R6.inputStrs[19];
    subStrs[7].strs.Add(theText);
    subStrs[7].maxWidth:= can.TextWidth(theText);

    {compute total height of caption}
    n:= 0;
    for i:= 1 to cLastSubstr do
      n:= n+ subStrs[i].strs.Count;
    if n > 0 then
      subStrHeight:= round((n + 1.5)*shy);
  end;
end;

procedure PrintFrameAndTexts;
const
  a1= 15.38 * 1.2;
  b1= 1.27 * 1.2;

  a2= 15.38;
  b2= 1.27;

  c = 1.3;
var
  can: TCanvas;
  fh: Single;
  x, y, shy, i, j, w, pageW: Integer;
begin
  pageW:= Printer.pageWidth;

  {print title lines}
  if titleStrs.strs.count > 0 then begin
    can:= Printer.Canvas;
    can.Font.Name := 'Helvetica';
    can.Font.Style := [fsBold];
    fh:= a1 + b1*sqrt(r.Right-r.left - cMinPageWidthPixels);
    can.Font.Height := round(fh);
    shy:= round(c*fh); {shift of 'y' from a line to the next}
    y:= rOr.top - titleHeight;

    {center title}
    w:= titleStrs.maxWidth;
    x:= (pageW - w) div 2;
    for i:= 0 to titleStrs.strs.count-1 do begin
      can.TextOut(x,y,titleStrs.strs.Strings[i]);
      y:= y + shy;
    end;
  end;

  {frame the square where drawing will apear}
  Printer.Canvas.Pen.Width := 1;
  Printer.canvas.Pen.Color := clBlack;
  Printer.canvas.moveto(rOr.Left, rOr.Top);
  Printer.canvas.lineto(rOr.Left, rOr.Bottom);
  Printer.canvas.lineto(rOr.Right, rOr.Bottom);
  Printer.canvas.lineto(rOr.Right, rOr.Top);
  Printer.canvas.lineto(rOr.Left, rOr.Top);

  {print caption lines}
  can:= Printer.Canvas;
  can.Font.Name := 'Arial';
  can.Font.Style := [];
  fh:= a2 + b2*sqrt(r.right-r.left - cMinPageWidthPixels);
  can.Font.Height := round(fh);
  shy:= round(c*fh);
  y:= rOr.bottom + shy div 2 ;
  for j:=1 to cLastSubstr do
    if subStrs[j].strs.count > 0 then begin
      {center all lines}
      w:= subStrs[j].maxWidth;
      x:= (pageW - w) div 2;
      for i:= 0 to subStrs[j].strs.count-1 do begin
        can.TextOut(x,y,subStrs[j].strs.Strings[i]);
        y:= y + shy;
      end;
    end;
end;

procedure BitmapToPrinterCanvas(destRect: TRect; source: TBitmap);
{A routine that I have copied from somebody in the web that substitutes
 the crippled 'StrechtDraw' provided by Delphi}
var
  BitmapHeader: pBitmapInfo;
  BitmapImage: POINTER;
  HeaderSize: DWORD;
  ImageSize: DWORD;
  sh: HBitmap;
  ch: Hdc;
  w, h, srcw, srch, n: Integer;
  can: TCanvas;
begin
  can := Printer.Canvas;
  ch:= Printer.canvas.Handle;
  sh:= source.Handle;
  w := destRect.right - destRect.left;
  h := destRect.bottom - destRect.top;
  srcw:= source.Width;
  srch:= source.Height;
  GetDIBSizes(sh, HeaderSize, ImageSize);
  GetMem(BitmapHeader, HeaderSize);
  GetMem(BitmapImage, ImageSize);
  n:= 0;
  try
    GetDIB(sh, source.Palette, BitmapHeader^, BitmapImage^);
    n:= StretchDIBits(ch,
        destRect.left, destRect.top, // Destination Origin
        w, h, // Destination Width and Height
        0, 0, // Source Origin
        srcw, srch, // source Width and Height
        BitmapImage,
        TBitmapInfo(BitmapHeader^),
        DIB_RGB_COLORS,
        SRCCOPY);
  finally
    FreeMem(BitmapImage);
    FreeMem(BitmapHeader);
  end;
  if n=0 then
    can.StretchDraw(destRect,source);
end;

procedure GetPrinterParameters;
{establish margins and square for drawing}
var
  w: Integer;
  textsHeight: Integer;
begin
  textsHeight:= titleHeight + subStrHeight + lociMargin;
  w:= min( Printer.PageHeight*9 div 10 - textsHeight, Printer.PageWidth*9 div 10);
  r.left := (Printer.PageWidth -w) div 2;
  r.top := (Printer.PageHeight - w - textsHeight) div 2 + titleHeight;
  r.right := r.left + w;
  r.bottom := r.top + w;
  rOr := r;
  w:= w div 512;
  InflateRect(rOr, w, w);
end;

procedure ClipPrinter(r: TRect);
{limit the drawing in the printer to the rectangle 'r'}
  var
    rg: HRGN;
begin
  rg:= CreateRectRgnIndirect(r);
  ExtSelectClipRgn(Printer.Canvas.Handle,rg, RGN_AND);
  DeleteObject(rg);
end;

procedure RemoveClipPrinter;
{Make the clipping rectangle be the whole page}
    var
    rg: HRGN;
begin
  rg:= CreateRectRgn(0,0,Printer.PageWidth,Printer.PageHeight);
  SelectClipRgn(Printer.Canvas.Handle,rg);
  DeleteObject(rg);
end;

function PreparePrint: Byte;
var
  sc: Double;
  i, num, fineRes, prWidth: Integer;
begin
  result:= mrCancel;
  num := Printer.copies;
  if num < 1 then
    Exit;
  prWidth:= (Printer.PageWidth * 9) div 10;

  {the width of the bitmap}
  fineRes:= Min(prWidth, ImmersionR3R6.prResolution);

  printer.canvas.Brush.Color:= clWhite;
  printer.canvas.Brush.Style:= bsSolid;

  {create string lists for title and caption}
  if not Assigned(titleStrs.strs) then begin
    titleStrs.strs:= TStringList.Create;
    for i:= 1 to cLastSubstr do
      subStrs[i].strs:= TStringList.Create;
  end;
  try
    titleStrs.strs.Sorted:= False;
    for i:= 1 to cLastSubstr do
      subStrs[i].strs.sorted:= False;
  except on exception do begin
      ImmersionR3R6.InfoBar.SimpleText:= eErrorInPrinting;
      Exit;
    end;
  end;

  GetTitleStrings(ImmersionR3R6.prTitle, prWidth);
  GetSubTextStrings(ImmersionR3R6.prSubText, prWidth);
  GetPrinterParameters;
  sc:= ImmersionR3R6.prResolution/kWidthMainBox;
  if ImmersionR3R6.showVeronese then begin
    if MapR3R6.ComputeVeroneseData(ImmersionR3R6.xDom, ImmersionR3R6.grid) then begin
      dps.ReceiveBasicData(mapR3R6.min3DNorm, mapR3R6.max3DNorm, kPrintGrid,
              ImmersionR3R6.prResolution);
      dps.ScaleCoefficients(sc);
      dps.DrawSurface; {draw to the bitmap}
      dps.ScaleCoefficients(1/sc);
      PrintAll; {continue printing in what it has in common with the printing of the
        Focal locus}
      result:= mrOK;
    end
  end
  else begin
    ImmersionR3R6.printingFocalLocus:= True;
    rist.ScaleCoefficients(sc);
    rist.ComputeAndDraw(ImmersionR3R6.prResolution, ImmersionR3R6.prResolution, False);
    rist.ScaleCoefficients(1/sc);
    {after completing the threads task, the program will call "PrintAll"}
  end;
end;

procedure PrintAll;
{we write and draw everything to the printer}
var
  num, i: Integer;
  prCan: TCanvas;
begin
  num := Printer.copies;
  if num < 1 then
    Exit;
  prCan:= Printer.canvas;
  try
    for i := 1 to num do begin
      prCan.FillRect(rOr);
      PrintFrameAndTexts;
      BitmapToPrinterCanvas(r, bmp);
      if (i < num) then
        Printer.NewPage;
    end;
  finally
    Printer.EndDoc;
    with ImmersionR3R6 do
      DrawScene;
  end;
end;

procedure Finalize;
var
  i: Integer;
begin
  titleStrs.strs.Free;
  for i:= 1 to cLastSubstr do
    subStrs[i].strs.Free;
end;


initialization
  ;
finalization
  Finalize;


end.
 